// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024


#include "batterystatusworker.hpp"
#include "SDL2/SDL_power.h"
#include <QThread>
#include <QDebug>
BatteryStatusWorker::Status status2status(SDL_PowerState state);

BatteryStatusWorker::Status status2status(SDL_PowerState state) {
    BatteryStatusWorker::Status status = BatteryStatusWorker::Status::UNKNOWN;
    switch (state) {
    case SDL_POWERSTATE_ON_BATTERY:
        status = BatteryStatusWorker::Status::ON_BATTERY;
        break;
    case SDL_POWERSTATE_NO_BATTERY:
        status = BatteryStatusWorker::Status::NO_BATTERY;
        break;
    case SDL_POWERSTATE_CHARGING:
        status = BatteryStatusWorker::Status::CHARGING;
        break;
    case SDL_POWERSTATE_CHARGED:
        status = BatteryStatusWorker::Status::CHARGED;
        break;
    default:
        break;
    }
    return status;
}

/// @brief Constructor.
BatteryStatusWorker::BatteryStatusWorker(QObject *parent) : QObject{parent} {
    mPause = false;
    mCancel = false;
    mRunning = false;
    mStatus = status2status(SDL_GetPowerInfo(&mLifetime, &mPercent));
}

/// @brief Get the last registered battery status.
void BatteryStatusWorker::lastStatus(BatteryStatusWorker::Status *status, int *percent, int *lifetime) {
    int s = -1, p = -1;
    Status t = Status::UNKNOWN;
    mDataMutex.lock();
    t = mStatus;
    p = mPercent;
    s = mLifetime;
    mDataMutex.unlock();
    if (status)
        *status = t;
    if (percent)
        *percent = p;
    if (lifetime)
        *lifetime = s;
}

/// @brief Stop processing (soon or later finished will be emitted).
void BatteryStatusWorker::stop() {
    qDebug() << QString("[0x%1] Stop the process").arg(reinterpret_cast<intptr_t>(QThread::currentThreadId()), 0, 16);
    QMutexLocker locker(&mPauseMutex);
    mCancel = true;
    mPauseCond.wakeAll();
}

/// @brief Pause the process.
void BatteryStatusWorker::pause() {
    QMutexLocker locker(&mPauseMutex);
    // already in pause :)
    if (mPause)
        return;
    mPause = true;
    qDebug() << QString("[0x%1] Pause process").arg(reinterpret_cast<intptr_t>(QThread::currentThreadId()), 0, 16);
}

/// @brief Resume the process.
void BatteryStatusWorker::resume() {
    QMutexLocker locker(&mPauseMutex);
    if (!mPause)
        return;
    mPause = false;
    mPauseCond.wakeAll();
    qDebug() << QString("[0x%1] Resume process").arg(reinterpret_cast<intptr_t>(QThread::currentThreadId()), 0, 16);
}

/// @brief Check whether or not the process is paused.
bool BatteryStatusWorker::isPaused() {
    QMutexLocker locker(&mPauseMutex);
    return mPause;
}

/// @brief Check whether or not the process is running (i.e. has not emitted finished signal !).
bool BatteryStatusWorker::isRunning() {
    QMutexLocker locker(&mPauseMutex);
    return mRunning;
}

/// @brief Query regularly for battery status.
void BatteryStatusWorker::process() {
    mPauseMutex.lock();
    mCancel = false;
    mPause = false;
    mRunning = true;
    mPauseMutex.unlock();
    int s, p, ic = 0;
    Status status = Status::UNKNOWN;
    while (true) {
        // Check pause state
        mPauseMutex.lock();
        bool pause = mPause;
        mPauseMutex.unlock();
        mSyncMutex.lock();
        if (pause) {
            mPauseCond.wait(&mSyncMutex);
            ic = 0; // when process will resumed it will be time to check for battery status again!
        }
        mSyncMutex.unlock();

        // query battery status 5 times (~0.5 secs) every 95 iterations (~9.5secs).
        if (ic == 0) {
            // run acq
            bool stop = false;
            int ms = 0, mp = 0, mpc = 0, msc = 0;
            // performs 5 queries: keep last status and average (valid) percent and lifetime
            for (int i = 0; i < 5; i++) {
                status = status2status(SDL_GetPowerInfo(&s, &p));
                if (s > -1) {
                    ms += s;
                    msc++;
                }
                if (p > -1) {
                    mp += p;
                    mpc++;
                }
                // check for cancelling process.
                mPauseMutex.lock();
                stop = mCancel;
                mPauseMutex.unlock();
                if (stop)
                    break;
                QThread::msleep(100);
            }
            if (stop)
                break;
            mp = mpc ? mp / mpc : -1;
            ms = msc ? ms / msc : -1;
            // update internal members
            mDataMutex.lock();
            mStatus = status;
            mPercent = mp;
            mLifetime = ms;
            mDataMutex.unlock();
            emit statusUpdated(status, mp, ms);
        } else if (ic > 94) // time to reset the counter
            ic = -1;
        ic++;
        mPauseMutex.lock();
        bool stop = mCancel;
        mPauseMutex.unlock();
        if (stop)
            break;
        QThread::msleep(100);
    }
    mPauseMutex.lock();
    mPause = false;
    mRunning = false;
    mPauseMutex.unlock();
    emit finished();
}
