// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef FITWORKER_HPP
#define FITWORKER_HPP

#include "isp/isp.hpp"
#include <QtCore>

using namespace ISP;

struct ScaleInfos {
    /**
     * @brief Constructor
     * @param nb Original number of points in the block.
     * @param nc Base number of points in fit configuration.
     * @param d Decimation factor.
     */
    //    ScaleInfos(int nb, int nc, int d = 1)
    //        : nb(nb), nc(nc), d(d), nf(static_cast<int>(std::ceil(static_cast<double>(nb) / d))) {}

    double c2f() const { return static_cast<double>(nc) / nf; } // config -> fit
    double f2c() const { return static_cast<double>(nf) / nc; } // fit -> config

    double c2b() const { return static_cast<double>(nc) / nb; } // config -> block
    double f2b() const { return static_cast<double>(nf) / nb; } // fit -> block

    int nb; // number of points of the ramps in the original block.
    int nc; // number of points of the ramps in the fit configuration.
    int nf; // number of points of the ramps in the actual fit process (may be lower than nb if decimation occured).
    int d;  // decimation factor: nf = ceil(nb/d)
};

class FitWorker : public QObject {
    typedef QMap<QPair<int, int>, FitParams> FitParamsMap;

    Q_OBJECT
  public:
    explicit FitWorker(int instrOffset = 1000, QObject *parent = nullptr);
    bool isRunning() const;

  public slots:
    void fitBlock(const ABlock &inputBlock, double refTime);
    void setConfiguration(const FitConfig &config, bool fastCalculation);

  signals:
    void fitComplete(const ABlock &block, const QVector<FitReport> &reports, const ExecTimes &times);

  private:
    mutable QMutex mRunMutex;
    mutable QMutex mDataMutex;
    FitConfig mConfig;
    bool mFastCalculation;
    bool mReuse;
    bool mRunning;
    FitParamsMap mPreviousFitParams;
    int mInstrOffset;

    DoubleVector setupFitParameters(const SParam &spc, FitParams &fitParams,
                                    const QMap<int, FitParams> &fitParamsMap, const ScaleInfos &scaleInfos,
                                    const double &pres, const double &temp, const double &cH2O0,
                                    const QMap<int, DoubleVector> &spectras, QStringList *errors = nullptr);

    bool setupBlockData(ABlock &block, const FitConfig &fitConfig, const QMap<int, ScaleInfos> &scaleInfos,
                        double &pres, double &temp, double &cH2O0, QMap<int, DoubleVector> &spectras,
                        QMap<int, DoubleVector> &ramps, QMap<int, double> &meanOffsets,
                        QMap<int, QPair<double, double>> &eSignals, QStringList *errors = nullptr);

    void updateDataBlock(ABlock *block, const BasicFitReport &fitReport, const ScaleInfos &infos);
};

#endif // FITWORKER_HPP
