// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "fitplot.hpp"
#include "menusection.hpp"

//#include <numeric>

FitPlot::FitPlot(QWidget *parent) : BasePlot(parent) {
    setNoAntialiasingOnDrag(true);
    setNotAntialiasedElements(QCP::AntialiasedElement::aePlottables);

    emitPosition(false);
    setZoomMode(false);

    mRawSpectraGraph = nullptr;
    mFitSpectraGraph = nullptr;
    mResiduGraph = nullptr;
    mRampGraph = nullptr;
    mBracket1 = nullptr;
    mBracket2 = nullptr;
    mBracket3 = nullptr;
    mBracket4 = nullptr;

    mMarginGroup = new QCPMarginGroup(this);

    // setup axis rects
    clearPlottables();
    plotLayout()->clear(); // WARNING remove everything.
    mMarginGroup->clear();

    // setup axis rects and legends.
    for (int i = 0; i < 3; i++)
        addAxisRect();
    plotLayout()->setRowStretchFactors(QList<double>{1.75, 1.75, 0.95});

    setInteractions(QCP::iRangeZoom | QCP::iRangeDrag);
    setupGraphs();
    synchronizeXaxes();
    setContextMenuPolicy(Qt::CustomContextMenu);
    QObject::connect(this, &FitPlot::customContextMenuRequested, this, &FitPlot::onContextMenuRequested);

    QObject::connect(this, &QCustomPlot::beforeReplot, this, &FitPlot::updateLegends, Qt::UniqueConnection);
}

/**
 * @brief Add new axis rect.
 * @details
 * The method adds a new axis rect in a new row of the plot layout.
 */
void FitPlot::addAxisRect() {
    QCP::MarginSides msides;
    QCPAxisRect *axRect = new QCPAxisRect(this);
    int idx = plotLayout()->rowCount();
    msides = QCP::msLeft | QCP::msRight;
    plotLayout()->addElement(idx, 0, axRect);

    // drag & zoom orientations
    Qt::Orientations orientations(Qt::Horizontal | Qt::Vertical);
    axRect->setRangeZoom(orientations);
    axRect->setRangeDrag(orientations);

    // synchronize margins
    axRect->setAutoMargins(QCP::msAll);
    axRect->setMarginGroup(msides, mMarginGroup);
    axRect->setMinimumMargins(QMargins(50, 5, 5, 5));

    // axes setup.
    for (QCPAxis *axis : axRect->axes()) {
        axis->setLayer("axes");
        axis->grid()->setLayer("grid");
        axis->setRange(0.0, 1.0);
        setAxisTicker(axis, LINEAR);
    }

    // legend
    Legend *legend = new Legend(axRect);
    legend->setVisible(true);
    mLegends.push_back(legend);
}

/// @brief Destructor.
FitPlot::~FitPlot() {}

/// @brief Setup the graphs.
void FitPlot::setupGraphs() {
    QCPAxisRect *axRect = axisRect(0);
    mRampGraph = addGraph(axRect->axis(QCPAxis::atBottom), axRect->axis(QCPAxis::atLeft));

    mBracket1 = new LinePosition(axRect);
    mBracket1->boundToData(true);
    mBracket1->setVisible(false);
    mBracket1->setFixed(true);

    mBracket2 = new LinePosition(axRect);
    mBracket2->boundToData(true);
    mBracket2->setVisible(false);
    mBracket2->setFixed(true);

    mBracket3 = new LinePosition(axRect);
    mBracket3->boundToData(true);
    mBracket3->setVisible(false);
    mBracket3->setFixed(true);
    QPen pen(mBracket3->pen());
    pen.setColor(Qt::red);
    mBracket3->setPen(pen);

    mBracket4 = new LinePosition(axRect);
    mBracket4->boundToData(true);
    mBracket4->setVisible(false);
    mBracket4->setFixed(true);
    pen = mBracket4->pen();
    pen.setColor(Qt::red);
    mBracket4->setPen(pen);

    axRect = axisRect(1);
    legend = mLegends.at(1);
    mRawSpectraGraph = addGraph(axRect->axis(QCPAxis::atBottom), axRect->axis(QCPAxis::atLeft));
    mFitSpectraGraph = addGraph(axRect->axis(QCPAxis::atBottom), axRect->axis(QCPAxis::atLeft));
    axRect = axisRect(2);
    legend = mLegends.at(2);
    mResiduGraph = addGraph(axRect->axis(QCPAxis::atBottom), axRect->axis(QCPAxis::atLeft));

    // graphs setup:
    // 1. ramp
    mRampGraph->setSelectable(QCP::stDataRange);
    mRampGraph->setName("Original ramp");
    mRampGraph->addToLegend(mLegends.at(0));
    {
        QPen pen;
        QPen selectedPen;
        pen.setWidthF(1.0);
        pen.setColor(Qt::black);
        selectedPen = QPen(pen);
        mRampGraph->setLineStyle(QCPGraph::lsLine);
        mRampGraph->setPen(pen);
        mRampGraph->selectionDecorator()->setPen(selectedPen);
    }

    // 2. raw spectra
    mRawSpectraGraph->setSelectable(QCP::stDataRange);
    mRawSpectraGraph->setName("Raw spectra");
    mRawSpectraGraph->addToLegend(mLegends.at(1));
    {
        QPen pen;
        QPen selectedPen;
        pen.setWidthF(1.0);
        pen.setColor(Qt::black);
        selectedPen = QPen(pen);
        mRawSpectraGraph->setLineStyle(QCPGraph::lsLine);
        mRawSpectraGraph->setPen(pen);
        mRawSpectraGraph->selectionDecorator()->setPen(selectedPen);
    }
    // add fill between the 2 spectras graphs.
    mRawSpectraGraph->setChannelFillGraph(mFitSpectraGraph);
    mRawSpectraGraph->setBrush(QColor(255, 161, 0, 50));

    // 3. fit spectra
    mFitSpectraGraph->setSelectable(QCP::stDataRange);
    mFitSpectraGraph->setName("Fit spectra");
    mFitSpectraGraph->addToLegend(mLegends.at(1));
    {
        QPen pen;
        QPen selectedPen;
        pen.setWidthF(1.0);
        pen.setColor(Qt::blue);
        selectedPen = QPen(pen);
        mFitSpectraGraph->setLineStyle(QCPGraph::lsLine);
        mFitSpectraGraph->setPen(pen);
        mFitSpectraGraph->selectionDecorator()->setPen(selectedPen);
    }
    // 2. residual
    mResiduGraph->setSelectable(QCP::stDataRange);
    mResiduGraph->setName("Residual");
    mResiduGraph->addToLegend(mLegends.at(2));
    {
        QPen pen;
        QPen selectedPen;
        pen.setWidthF(1.0);
        pen.setColor(Qt::blue);
        selectedPen = QPen(pen);
        mResiduGraph->setLineStyle(QCPGraph::lsLine);
        mResiduGraph->setPen(pen);
        mResiduGraph->selectionDecorator()->setPen(selectedPen);
    }
}

/**
 * @brief Synchronize the X-axes of the fit and reisuald subplots.
 */
void FitPlot::synchronizeXaxes() {
    QList<QCPAxis *> axes = {mResiduGraph->keyAxis(), mFitSpectraGraph->keyAxis()};
    for (QCPAxis *axis : axes)
        QObject::connect(axis, static_cast<void (QCPAxis::*)(const QCPRange &)>(&QCPAxis::rangeChanged), this,
                         &FitPlot::autoXAxisRange, Qt::UniqueConnection);
}

/**
 * @brief Synchronize X-axis range of subplots.
 * @param newRange New range to set.
 * @details
 * The method is derived from QPlotlib BasePlot::autoAxisRange. It only synchronize the X-axis range
 * of the fit and residuals subplots.
 *
 * The synchronization occurs only if the X-axis destination is not fixed.
 */
void FitPlot::autoXAxisRange(const QCPRange &newRange) {
    QCPAxis *origAxis = static_cast<QCPAxis *>(sender());
    if (!origAxis)
        return;
    QList<QCPAxis *> axes = {mResiduGraph->keyAxis(), mFitSpectraGraph->keyAxis()};
    for (QCPAxis *axe : axes) {
        if (axe == origAxis)
            continue;
        bool old = axe->blockSignals(true);
        axe->setRange(newRange);
        axe->blockSignals(old);
    }
    replot();
}

/// @brief Clear plot data.
void FitPlot::clearData() {

    mBracket1->setVisible(false);
    mBracket2->setVisible(false);
    mBracket3->setVisible(false);
    mBracket4->setVisible(false);

    mRampGraph->data()->clear();
    mRawSpectraGraph->data()->clear();
    mFitSpectraGraph->data()->clear();
    mResiduGraph->data()->clear();

    replot();
}

/**
 * @brief Show/hide the ramp interval brackets.
 * @param visible True to set visible, false otherwise.
 * @note
 * The method checks for the brackets position before switching the visibility state.
 * Brackets positions must be greater than -1.0 to be set visible.
 */
void FitPlot::setRampIntervalVisible(bool visible) {
    mBracket1->setVisible(visible && mBracket1->position() > -1.0);
    mBracket2->setVisible(visible && mBracket2->position() > -1.0);
}

/// @brief Set the ramp interval.
void FitPlot::setRampInterval(const QPair<int, int> &range) {
    double i1 = static_cast<double>(range.first), i2 = static_cast<double>(range.second);
    mBracket1->setPosition(i1);
    mBracket2->setPosition(i2);
    mBracket1->setVisible(i1 > -1.0);
    mBracket2->setVisible(i2 > -1.0);
}

/**
 * @brief Show/hide the ramp interval brackets.
 * @param visible True to set visible, false otherwise.
 * @note
 * The method checks for the brackets position before switching the visibility state.
 * Brackets positions must be greater than -1.0 to be set visible.
 */
void FitPlot::setSpectraIntervalVisible(bool visible) {
    mBracket3->setVisible(visible && mBracket3->position() > -1.0);
    mBracket4->setVisible(visible && mBracket4->position() > -1.0);
}

/// @brief Set the ramp interval.
void FitPlot::setSpectraInterval(const QPair<int, int> &range) {
    double i1 = static_cast<double>(range.first), i2 = static_cast<double>(range.second);
    mBracket3->setPosition(i1);
    mBracket4->setPosition(i2);
    mBracket3->setVisible(i1 > -1.0);
    mBracket4->setVisible(i2 > -1.0);
}

/**
 * @brief Open a context menu with specific actions to perform on the axis-rect.
 * @param pos Position of the mouse event.
 * @details
 * The method generates a context menu with three actions that acts for the AxisRect pointed by the mouse:
 *  - X and Y axes auto-scale.
 *  - X auto scale.
 *  - Y auto scale.
 */
void FitPlot::onContextMenuRequested(const QPoint &pos) {
    QCPAxisRect *axRect = axisRectAt(pos);
    if (!axRect)
        return;
    QAction *action = nullptr;
    QMenu *menu = new QMenu(this);
    menu->setAttribute(Qt::WA_DeleteOnClose);
    /* List of available actions:
     *   * Section: auto-scale
     *   - plot autoscale (X && Y)
     *   - auto scale X
     *   - auto scale Y
     */
    MenuSection *section = new MenuSection(tr("Scaling"), menu);
    menu->addAction(section);
    action = menu->addAction(QIcon(":/plotting/auto-both.png"), tr("Auto-scale"));
    QObject::connect(action, &QAction::triggered, this, [=]() { autoRescale(); });
    action = menu->addAction(QIcon(":/plotting/auto-x.png"), tr("X-axis auto-scale"));
    QObject::connect(action, &QAction::triggered, this, [=]() { autoRescaleX(); });
    action = menu->addAction(QIcon(":/plotting/auto-y.png"), tr("Y-axis auto-scale"));
    QObject::connect(action, &QAction::triggered, this, [=]() { autoRescaleY(); });
    menu->popup(mapToGlobal(pos));
}

void FitPlot::updateLegends() {
    for (Legend *legend : mLegends) {
        if (legend->visible()) {
            legend->setMaximumSize(legend->minimumOuterSizeHint());
        }
    }
}
