// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file legend.cpp
 * @brief Legend class definitions file.
 * @author J. Burgalat
 * @date 2016
 */

#include "legend.hpp"

/// @brief Constructor.
Legend::Legend(QCPAxisRect *clipToAxisRect) : QCPLegend() {
    Q_ASSERT_X(clipToAxisRect != nullptr, "Legend::Legend", "clipAxisRect must be a valid pointer.");
    mAxisRect = clipToAxisRect;
    mOrigin = QPointF(0.05, 0.05);
    clipToAxisRect->insetLayout()->addElement(this, QRectF(0.05, 0.05, 0.33, 0.33));
    mLayoutIndex = clipToAxisRect->insetLayout()->elementCount() - 1;
    setLayer("legend");
    setSelectableParts(QCPLegend::spItems);
}

/**
 * @brief Manage interactions on mouse press.
 * @details
 * The method initializes the drag process of the Legend.
 */
void Legend::mousePressEvent(QMouseEvent *event, const QVariant &details) {
    Q_UNUSED(details)
    QCPLegend::mousePressEvent(event, details);
    if (visible() && event->button() == Qt::LeftButton) {
        // since insetRect is in axisRect coordinates (0..1), we transform the mouse position:
        QPointF mousePoint((event->pos().x() - mAxisRect->left()) / (double)mAxisRect->width(),
                           (event->pos().y() - mAxisRect->top()) / (double)mAxisRect->height());
        mOrigin = mousePoint - mAxisRect->insetLayout()->insetRect(0).topLeft();
        event->accept();
    }
}

/**
 * @brief Manage interactions on mouse move.
 * @details
 * The method updates the position of the Legend.
 */
void Legend::mouseMoveEvent(QMouseEvent *event, const QPointF &startPos) {
    Q_UNUSED(startPos)
    QCPLegend::mouseMoveEvent(event, startPos);
    if (visible() && event->buttons() == Qt::LeftButton) {
        QRectF rect = mAxisRect->insetLayout()->insetRect(mLayoutIndex);
        // since insetRect is in axisRect coordinates (0..1), we transform the mouse position:
        QPointF mousePoint((event->pos().x() - mAxisRect->left()) / (double)mAxisRect->width(),
                           (event->pos().y() - mAxisRect->top()) / (double)mAxisRect->height());
        rect.moveTopLeft(mousePoint - mOrigin);
        mAxisRect->insetLayout()->setInsetRect(mLayoutIndex, rect);
        parentPlot()->replot(QCustomPlot::rpQueuedReplot);
        event->accept();
    }
}

/**
 * @brief Manage interactions on mouse release.
 * @details
 * The method computes the final position of the Legend.
 */
void Legend::mouseReleaseEvent(QMouseEvent *event, const QPointF &startPos) {
    Q_UNUSED(startPos)
    QCPLegend::mouseReleaseEvent(event, startPos);
    if (visible()) {
        QRectF rect = mAxisRect->insetLayout()->insetRect(mLayoutIndex);
        // since insetRect is in axisRect coordinates (0..1), we transform the mouse position:
        QPointF mousePoint((event->pos().x() - mAxisRect->left()) / (double)mAxisRect->width(),
                           (event->pos().y() - mAxisRect->top()) / (double)mAxisRect->height());
        rect.moveTopLeft(mousePoint - mOrigin);
        rect.topLeft().rx() = qBound(0.05, rect.topLeft().rx(), 0.95);
        rect.topLeft().ry() = qBound(0.05, rect.topLeft().ry(), 0.95);
        mAxisRect->insetLayout()->setInsetRect(mLayoutIndex, rect);
        parentPlot()->replot();
    }
    event->accept();
}
