// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "qcp_helpers.hpp"

/// @brief Length of a short dash
#define DASH 4.0
/// @brief Length of a long dash
#define LDASH 8.0
/// @brief Length of a dot
#define DOT 1.0
/// @brief Length of a blank space
#define SPACE 2.0

QVector<QCPScatterStyle::ScatterShape> scatterShapes;
QVector<QString> scatterNames;
QVector<QIcon> scatterIcons;
QVector<QIcon> dashIcons;

/// @brief Set default scatter shapes.
static QVector<QCPScatterStyle::ScatterShape> setDefaultScatterShape() {
    QVector<QCPScatterStyle::ScatterShape> shapes;
    shapes.append(QCPScatterStyle::ssNone);     // None
    shapes.append(QCPScatterStyle::ssDisc);     // Disc
    shapes.append(QCPScatterStyle::ssCross);    // Cross
    shapes.append(QCPScatterStyle::ssPlus);     // Plus
    shapes.append(QCPScatterStyle::ssSquare);   // Square
    shapes.append(QCPScatterStyle::ssDiamond);  // Diamond
    shapes.append(QCPScatterStyle::ssTriangle); // Triangle
    return shapes;
}

/// @brief Set default scatter shape names.
static QVector<QString> setDefaultScatterNames() {
    QVector<QString> names;
    names.append(QCoreApplication::tr("None"));     // None
    names.append(QCoreApplication::tr("Disc"));     // Disc
    names.append(QCoreApplication::tr("Cross"));    // Cross
    names.append(QCoreApplication::tr("Plus"));     // Plus
    names.append(QCoreApplication::tr("Square"));   // Square
    names.append(QCoreApplication::tr("Diamond"));  // Diamond
    names.append(QCoreApplication::tr("Triangle")); // Triangle
    return names;
}

/// @brief Generate icons for default scatter shapes.
static QVector<QIcon> genScatterIcons() {
    if (scatterShapes.isEmpty())
        scatterShapes = setDefaultScatterShape();
    if (scatterNames.isEmpty())
        scatterNames = setDefaultScatterNames();
    QVector<QIcon> icons;
    QCustomPlot *customPlot = new QCustomPlot();
    customPlot->axisRect()->setAutoMargins(QCP::msNone);
    customPlot->axisRect()->setMargins(QMargins(0, 0, 0, 0));
    customPlot->xAxis->setVisible(false);
    customPlot->yAxis->setVisible(false);
    customPlot->xAxis->setRange(-1, 1);
    customPlot->yAxis->setRange(-1, 1);
    customPlot->addGraph();
    customPlot->graph()->addData(0, 0);
    customPlot->graph()->setPen(QPen(QColor(Qt::black)));
    customPlot->setBackground(Qt::transparent);
    for (QVector<QCPScatterStyle::ScatterShape>::const_iterator it = scatterShapes.cbegin(); it != scatterShapes.cend(); it++) {
        customPlot->graph()->setScatterStyle(QCPScatterStyle(*it, 8.5));
        QPixmap pm = customPlot->toPixmap(16, 16);
        icons.append(QIcon(pm));
    }
    delete customPlot;
    return icons;
}

namespace Plotting {
/**
 * @brief Initialize global variables related to plots
 * @details
 * The method initializes the following vectors:
 *   - scatterShapes, a vector of predefined scatter styles.
 *   - scatterNames, a vector of predefined scatter styles names (shares same index as scatterShapes).
 *   - dashPatterns, a vector of predefined dash patterns.
 *   - scatterIcons, a vector of icons with scatters style symbols.
 *   - dashIcons, a vector of icons with dash patterns symbols.
 */
void initHelpers() {
    if (scatterShapes.isEmpty())
        scatterShapes = setDefaultScatterShape();
    if (scatterNames.isEmpty())
        scatterNames = setDefaultScatterNames();
    if (scatterIcons.isEmpty())
        scatterIcons = genScatterIcons();
}

/**
 * @brief Get the scatter shape at given index.
 * @param index Index to process. It generally comes from a combobox that has been populated using ::fillMarkerStyleComboBox.
 * @return Scatter shape associated to the given index. QCPScatterStyle::ssNone is returned if index out of range.
 */
QCPScatterStyle::ScatterShape getScatterShape(int index) {
    if (scatterShapes.isEmpty())
        scatterShapes = setDefaultScatterShape();
    if (index < 0 || index >= scatterShapes.size())
        return QCPScatterStyle::ssNone;
    return scatterShapes[index];
}

/**
 * @brief Get the scatter shape name at given index.
 * @param index Index to process. It generally comes from a combobox that has been populated using ::fillMarkerStyleComboBox.
 * @return Scatter shape name associated to the given index. "None" is returned if index out of range.
 */
QString getScatterName(int index) {
    if (scatterNames.isEmpty())
        scatterNames = setDefaultScatterNames();
    if (index < 0 || index >= scatterNames.size())
        return QCoreApplication::tr("None");
    return scatterNames[index];
}
/**
 * @brief Get the scatter shape icon at given index.
 * @param index Index to process. It generally comes from a combobox that has been populated using ::fillMarkerStyleComboBox.
 * @return Scatter shape icon associated to the given index. QIcon() is returned if index out of range.
 */
QIcon getScatterIcon(int index) {
    if (scatterIcons.isEmpty())
        scatterIcons = genScatterIcons();
    if (index < 0 || index >= scatterIcons.size())
        return QIcon();
    return scatterIcons[index];
};

/// @brief Populate a combobox with default marker styles.
void fillMarkerStyleComboBox(QComboBox *cbx, bool noNone) {
    if (scatterShapes.isEmpty())
        scatterShapes = setDefaultScatterShape();
    if (scatterNames.isEmpty())
        scatterNames = setDefaultScatterNames();
    if (scatterIcons.isEmpty())
        dashIcons = genScatterIcons();
    QString name = cbx->currentText();
    cbx->clear();
    for (int i = (noNone ? 1 : 0); i < scatterIcons.size(); i++)
        cbx->addItem(scatterIcons[i], scatterNames[i], i);
    cbx->setMaxVisibleItems(6);
    cbx->setCurrentText(name);
}

/// @brief Convert AxisRect ratio to plot coordinate.
QPointF axisRectRatioToPlotCoord(QCPAxisRect *axisRect, QPointF pos) {
    double xabs = axisRect->left() + axisRect->width() * pos.x();
    double yabs = axisRect->top() + axisRect->height() * pos.y();
    double x = axisRect->axis(QCPAxis::atBottom)->pixelToCoord(xabs);
    double y = axisRect->axis(QCPAxis::atLeft)->pixelToCoord(yabs);
    return QPointF(x, y);
}

/// @brief Convert plot coordinate to AxisRect ratio.
QPointF plotCoordToAxisRectRatio(QCPAxisRect *axisRect, QPointF pos) {
    double xabs = axisRect->axis(QCPAxis::atBottom)->coordToPixel(pos.x());
    double yabs = axisRect->axis(QCPAxis::atLeft)->coordToPixel(pos.y());
    double x = (xabs - axisRect->left()) / (double)axisRect->width();
    double y = (yabs - axisRect->top()) / (double)axisRect->height();
    return QPointF(x, y);
}

/// @brief Convert the absolute pixel position to axis rect ratio.
QPointF pixelToAxisRectRatio(QCPAxisRect *axisRect, QPointF pos) {
    double x = (pos.x() - axisRect->left()) / (double)axisRect->width();
    double y = (pos.y() - axisRect->top()) / (double)axisRect->height();
    return QPointF(x, y);
}

/// @brief Convert absolute pixel position to plot coordinates
QPointF pixelToPlotCoord(QCPAxisRect *axisRect, QPointF pos) {
    return axisRectRatioToPlotCoord(axisRect, pixelToAxisRectRatio(axisRect, pos));
}

/**
 * @brief Reload list of names.
 * @details
 * The method can be called to reload the list of name. It is wise to call such method when
 * translation files has changed in order to update the names.
 */
void reloadNames() { scatterNames = setDefaultScatterNames(); }
}
