// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "admconfig.hpp"
#include "messagehandler.hpp"

#include <QXmlSchema>
#include <QXmlSchemaValidator>
#include <QtXmlPatterns>

#include "cfg_versions.hpp"

AdmConfig::AdmConfig() {
    QFile file(":/adm2/xsd/adm2.xsd");
    file.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream in(&file);
    mXsdSchema = in.readAll().toLatin1();
    file.close();
    mValid = false;
    mCfgVersion = -1;
    mNumInstrument = -1;
    mNumVersion = -1;
    mLaserWarmup = -1;
}

/**
 * @brief Copy constructor.
 * @details
 * The copy constructor does not perform a deep copy of the key mapper.
 */
AdmConfig::AdmConfig(const AdmConfig &other) {
    mXsdSchema = other.mXsdSchema;
    mValid = other.mValid;
    mDirectory = other.mDirectory;
    mName = other.mName;
    mCfgVersion = other.mCfgVersion;

    mDescriptorPath = other.mDescriptorPath;
    mKeymapPath = other.mKeymapPath;
    mFitPath = other.mFitPath;
    mScenarioPath = other.mScenarioPath;
    mIndicatorPath = other.mIndicatorPath;

    mNetworkConfig = other.mNetworkConfig;
    mNumInstrument = other.mNumInstrument;
    mNumVersion = other.mNumVersion;
    mLaserWarmup = other.mLaserWarmup;
    mRawDescriptor = other.mRawDescriptor;
}
/**
 * @brief Destructor.
 */
AdmConfig::~AdmConfig() {}

/**
 * @brief Copy assignment operator.
 * @details
 * The copy assignment does not perform a deep copy of the spectra and line parameters (shared) pointers.
 */
AdmConfig &AdmConfig::operator=(const AdmConfig &other) {
    if (this != &other) {
        mXsdSchema = other.mXsdSchema;
        mValid = other.mValid;
        mDirectory = other.mDirectory;
        mName = other.mName;
        mCfgVersion = other.mCfgVersion;

        mDescriptorPath = other.mDescriptorPath;
        mKeymapPath = other.mKeymapPath;
        mFitPath = other.mFitPath;
        mScenarioPath = other.mScenarioPath;
        mIndicatorPath = other.mIndicatorPath;

        mNetworkConfig = other.mNetworkConfig;
        mNumInstrument = other.mNumInstrument;
        mNumVersion = other.mNumVersion;
        mLaserWarmup = other.mLaserWarmup;
        mRawDescriptor = other.mRawDescriptor;
    }
    return *this;
}

/// @brief Check whether not the content of the parser is valid.
bool AdmConfig::isValid() const { return mValid; }

/// @brief Checked whether or not the content of the path given by xmlPath is valid.
bool AdmConfig::validate(const QString &xmlPath, QStringList *errors) {
    QFile xmlFile(xmlPath);
    // canot open files.
    if (!xmlFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QCoreApplication::tr("Cannot open %1").arg(xmlPath));
        return false;
    }
    QString instance = xmlFile.readAll();
    xmlFile.close();
    return validateContent(instance, errors);
}

/// @brief Check whether or not the content defined by xmlContent is valid.
bool AdmConfig::validateContent(const QString &xmlContent, QStringList *errors) {
    const QByteArray instanceData = xmlContent.toLatin1();
    MessageHandler messageHandler;
    QXmlSchema schema;
    schema.setMessageHandler(&messageHandler);
    bool errorOccurred = false;
    if (!mXsdSchema.isEmpty())
        schema.load(mXsdSchema);
    if (!schema.isValid()) {
        errorOccurred = true;
        if (errors)
            errors->append(messageHandler.statusMessage());
    } else {
        QXmlSchemaValidator validator(schema);
        if (!validator.validate(instanceData)) {
            errorOccurred = true;
            if (errors)
                errors->append(messageHandler.statusMessage());
        }
    }
    return !errorOccurred;
}

/// @brief Parse the xml file pointed by xmlPath and setup the parameters.
/// @sa Configuration::parseContent
bool AdmConfig::parse(const QString &xmlPath, QStringList *errors) {
    QFile xmlFile(xmlPath);
    // canot open files.
    if (!xmlFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QCoreApplication::tr("Cannot open %1").arg(xmlPath));
        return false;
    }
    QString instance = xmlFile.readAll();
    QFileInfo infos(xmlPath);
    mName = infos.baseName();
    xmlFile.close();
    return parseContent(instance, infos.absolutePath(), errors);
}

/**
 * @brief Validate and parse the input content.
 * @param content XML formatted content to parse.
 * @return True on success, false otherwise.
 */
bool AdmConfig::parseContent(const QString &content, const QString directory, QStringList *errors) {
    clear();
    bool ret = true;
    mValid = validateContent(content, errors);

    if (!mValid)
        return mValid;
    mDirectory = directory;
    QDomDocument document;
    document.setContent(content);
    QDomNode admNode = document.firstChildElement("adm");

    int cfgver = admNode.toElement().attribute("version", "0").toInt();
    // check version: must be:
    // - 2 (before)
    // - 3 (25/01/2021)
    if (cfgver < ADM_FILE_MIN_VER) {
        if (errors)
            errors->append(
                QCoreApplication::tr("ADM configuration version too old (must be >=%1).").arg(ADM_FILE_MIN_VER));
        clear();
        mValid = false;
        return false;
    }
    mCfgVersion = cfgver;

    QString name = admNode.firstChildElement("name").text().trimmed();
    if (!name.isEmpty())
        mName = name;

    // get paths
    mDescriptorPath = processPath(admNode.firstChildElement("descriptor_file").text().trimmed());

    mKeymapPath = processPath(admNode.firstChildElement("keymap_file").text().trimmed());
    mFitPath = processPath(admNode.firstChildElement("inversion_file").text().trimmed());
    mScenarioPath = processPath(admNode.firstChildElement("scenario_file").text().trimmed());
    mIndicatorPath = processPath(admNode.firstChildElement("indicator_file").text().trimmed());

    // setup descriptor
    if (mDescriptorPath.isEmpty()) {
        if (errors)
            errors->append(QCoreApplication::tr("No path defined for <descriptor_file> element."));
        mValid = false;
    }
    StringList serrors;
    if (!mRawDescriptor.setupFromDescriptor(mDescriptorPath.toStdString(), &serrors)) {
        if (errors) {
            for (const auto &s : serrors) {
                errors->append(QString::fromStdString(s));
            }
            errors->append(QCoreApplication::tr("%1 is not a valid descriptor file.").arg(mDescriptorPath));
        }
        mValid = false;
    } else {
        mNumInstrument = mRawDescriptor.instrument();
        mNumVersion = mRawDescriptor.version();
    }

    // network node
    ret = readNetwork(admNode, errors);
    if (errors && !ret)
        errors->append(QCoreApplication::tr("Cannot parse <network> node."));
    mValid = ret && mValid;

    // laser warmup
    mLaserWarmup = -1;
    QDomElement node = admNode.firstChildElement("warmup");
    if (!node.isNull())
        mLaserWarmup = node.toElement().text().trimmed().toInt();

    if (!mValid)
        clear();
    return mValid;
}

/**
 * @brief Process file path.
 * @param filePath Input path to process.
 * @return Output path.
 * @details
 * The method checks whether filePath is empty or if it is an absolute path. In such case, it simply returns filePath.
 * Otherwise it builds a path from the registered directory (AdmConfig::directory) and filePath.
 */
QString AdmConfig::processPath(QString filePath) {
    if (filePath.isEmpty() || QFileInfo(filePath).isAbsolute())
        return filePath;
    return QString("%1/%2").arg(mDirectory, filePath);
}

/// @brief Clear the parser.
void AdmConfig::clear() {
    mValid = false;
    mDirectory = QString();
    mName = QString();
    mCfgVersion = -1;

    mDescriptorPath = QString();
    mKeymapPath = QString();
    mFitPath = QString();
    mScenarioPath = QString();
    mIndicatorPath = QString();

    mNetworkConfig = NetworkConfig();

    mNumInstrument = -1;
    mNumVersion = -1;
    mLaserWarmup = -1;
    mRawDescriptor = ABlock();
}

/**
 * @brief Get the name of this configuration.
 * @details
 * The name of the configuration is either set during configuration parsing (value fo the <name> tag) or
 * by calling directly Configuration::setName.
 */
const QString &AdmConfig::name() const { return mName; }

/**
 * @brief Get the configuration file version.
 * @return Configuration file version if the configuration is valid, 0 otherwise.
 */
int AdmConfig::cfgVersion() const { return isValid() ? mCfgVersion : 0; }

/// @brief Get the path of the associated directory.
const QString &AdmConfig::directory() const { return mDirectory; }

/// @brief Get the (full) path of the registered scenario configuration file.
QString AdmConfig::scenarioFilePath() const { return mScenarioPath; }

/// @brief Get the (full) path of the registered indicators configuration file.
QString AdmConfig::indicatorFilePath() const { return mIndicatorPath; }

/// @brief Get the (full) path of the registered fit configuration file.
QString AdmConfig::fitFilePath() const { return mFitPath; }

/// @brief Get the (full) path of the registered keymap file.
QString AdmConfig::keymapFilePath() const { return mKeymapPath; }

/// @brief Get the instrument number.
int AdmConfig::instrumentNumber() const { return mNumInstrument; }

/// @brief Get the firmware version number.
int AdmConfig::firmwareVersion() const { return mNumVersion; }

/**
 * @brief Get the warmup time of the lasers in seconds.
 * @note
 * The value can be negative (-1) if no warmup has been set in the configuration.
 */
int AdmConfig::laserWarmup() const { return mLaserWarmup; }

/// @brief Get the descriptor that matches AdmConfig::instrumentNumber and AdmConfig::firmwareVersion.
const ABlock &AdmConfig::descriptor() const { return mRawDescriptor; }

/// @brief Get the network configuration.
const NetworkConfig &AdmConfig::networkConfig() const { return mNetworkConfig; }

/**
 * @brief Setup descriptor block from descriptor file path.
 * @param errors List of errors.
 * @return True on success, false otherwise.
 */
bool AdmConfig::setupDescriptor(QStringList *errors) {
    if (mDescriptorPath.isEmpty()) {
        if (errors)
            errors->append(QCoreApplication::tr("No path defined for <descriptor_file> element."));
        return false;
    }
    StringList serrors;
    if (!mRawDescriptor.setupFromDescriptor(mDescriptorPath.toStdString(), &serrors)) {
        if (errors) {
            for (const auto &s : serrors) {
                errors->append(QString::fromStdString(s));
            }
            errors->append(QCoreApplication::tr("%1 is not a valid descriptor file.").arg(mDescriptorPath));
        }
        return false;
    }
    mNumInstrument = mRawDescriptor.instrument();
    mNumVersion = mRawDescriptor.version();
    return true;
}

/**
 * @brief Parse network tag from XML content.
 * @param parentNode Parent node to parse.
 * @param errors Optional pointer to a list of strings with in output the errors.
 * @return True on success, false otherwise.
 * @details
 * The method parses the \<network\> node from parentNode, retrieves the IP address and port for UDP communications
 * as well as the SSID to monitor if any.
 *
 * @internal
 * The method never appends messages in errors even if it fails. The method is intended to be called after XML
 * validation has been performed and then it cannot fail
 */
bool AdmConfig::readNetwork(QDomNode parentNode, QStringList *errors) {
    Q_UNUSED(errors)
    mNetworkConfig = NetworkConfig();
    QDomElement node = parentNode.firstChildElement("network");
    QDomNode n = node.firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            if (tagName.compare("ip") == 0)
                mNetworkConfig.ipAddr = e.text().trimmed();
            else if (tagName.compare("port") == 0)
                mNetworkConfig.port = e.text().trimmed().toInt();
            else if (tagName.compare("ssid") == 0)
                mNetworkConfig.ssidName = e.text().trimmed();
        }
        n = n.nextSibling();
    }
    return mNetworkConfig.isValid();
}

/// @brief Compute a XML formatted string representation of the instance.
QByteArray AdmConfig::toXML() const {
    QString out;
    out = QString("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>"
                  "<adm version=\"%1\">"
                  "<name>%2</name>"
                  "%3"
                  "<descriptor_file>%4</descriptor_file>"
                  "<keymap_file>%5</keymap_file>"
                  "<inversion_file>%6</inversion_file>"
                  "<indicator_file>%7</indicator_file>"
                  "<scenario_file>%8</scenario_file>"
                  "<warmup>%9</warmup>"
                  "</adm>")
              .arg(mCfgVersion)
              .arg(mName)
              .arg(networkToXML())
              .arg(mDescriptorPath)
              .arg(mKeymapPath)
              .arg(mFitPath)
              .arg(mIndicatorPath)
              .arg(mScenarioPath)
              .arg(mLaserWarmup);
    return out.toLatin1();
}

/// @brief Compute the XML representation of the network configuration.
QString AdmConfig::networkToXML() const {
    return QString("<network>"
                   "<ip>%1</ip>"
                   "<port>%2</port>"
                   "<ssid>%3</ssid>"
                   "</network>")
        .arg(mNetworkConfig.ipAddr)
        .arg(mNetworkConfig.port)
        .arg(mNetworkConfig.ssidName);
}
