// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef BASERTPLOTWIDGET_HPP
#define BASERTPLOTWIDGET_HPP

#include <QtWidgets>

#include "datarange.hpp"
#include "globals.hpp"
#include "itemtracer.hpp"
#include "textbox.hpp"

class BasePlot;
class LinePosition;
class QCPGraph;

class BaseRtPlotWidget : public QWidget {
    Q_OBJECT
  public:
    explicit BaseRtPlotWidget(QWidget *parent = nullptr);
    void translateUI();

    void scrollXRange(double reference);
    bool freeMode() const;
    void setFreeMode(bool enabled);
    void setTimeWindow(int seconds);
    int timeWindow() const;
    void setTimeOffset(int milliseconds);
    int timeOffset() const;

    QString titlePrefix() const;
    void setTitlePrefix(const QString &titlePrefix);

    QString name() const;
    void setName(const QString &name);

    void setToolbarVisible(bool visible);
    bool toolbarVisible() const;

    void setXLabelVisible(bool visible);
    bool xLabelVisible() const;

    void setYLabel(const QString &label);
    void setTextBoxText(const QString &text);
    void setTextBoxPosition(const TextBox::Position &position);

    virtual void refreshPlot();

    QLabel *mousePositionLabel() const;
    BasePlot *plot() const;

    QCPGraph *buildGraph();

    void addGraph(GraphContainerPointer data);
    void removeGraph(GraphContainerPointer data);
    void setGraphs(GraphContainerVector data, bool replot = true);

    void addInteractionsActions(QMenu *menu);
    virtual void updateFromSettings();

    void forceTimeRange(const QCPRange &range);

    bool replayMode() const;
    void setReplayMode(bool replayMode);

    void setReplayKey(double key);
    void setReplayTracer(bool enable, double *key = nullptr);

    bool rgLinesVisible() const;
    bool rgLinesEditable() const;
    void setRgLinesPen(QPen pen);

    bool thLineVisible() const;
    bool thLineEditable() const;

    void setThLinePen(QPen pen);

    void fillRgLines(bool enabled);
    void setRgLinesFillBrush(QBrush brush);

  signals:
    void thresholdChanged(double value);
    void colorRangeChanged(const DataRange &range);

  public slots:
    void clearPlot();
    void rescaleYAxis();

  public slots:
    void setRgLinesValues(const DataRange &range);
    void setThLineValue(double value);
    void setRgLinesVisible(bool visible);
    void setRgLinesEditable(bool enable);
    void setThLineVisible(bool visible);
    void setThLineEditable(bool enable);

  private slots:
    void onRgLineChanged(double);
    void emitRangeChange();
    void onThLineChanged(double);
    void emitThresholdChange();

  protected:
    void setupActions();
    virtual void leaveEvent(QEvent *event) override;

  protected:
    bool mReplayMode;   ///< Replay mode control flag.
    bool mFreeMode;     ///< Free mode control flag.
    double mTimeWindow; ///< Size of the time window (seconds).
    double mTimeOffset; ///< Time offset to add to the right edge of the X-axis (seconds).

    bool mWithXLabel;

    QToolBar *mToolBar;
    BasePlot *mPlot;
    TextBox *mLastValueBox;

    QAction *mActionFreeMode;    ///< Free mode control action.
    QAction *mActionYRescale;    ///< Y-axis Rescale action.
    QAction *mActionXRescale;    ///< X-axis Rescale action.
    QAction *mActionRescale;     ///< all axes Rescale action.
    QLabel *mMousePositionLabel; ///< Mouse position label.

    QString mTitlePrefix;
    QString mName;

    ItemTracer *mReplayTracer;

    LinePosition *mUpperLine;
    LinePosition *mLowerLine;
    LinePosition *mThresholdLine;
    bool mThLineVisible;
    QTimer mRgTimer, mThTimer;

  private:
    void buildLines();
};

#endif // BASERTPLOTWIDGET_HPP
