// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "clickablelabel.hpp"

/// @brief Constructor.
ClickableLabel::ClickableLabel(QWidget *parent) : QLabel(parent), mClickable(true), mHighlightable(false) {}

/**
 * @brief Enable/Disable clickable mode.
 * @param enable True to enable clickable mode, false otherwise.
 * @details
 * The method enables/disables the clickable mode.
 *
 * If clickable mode is enabled (default value at initialization), the widget does emit
 * ClickableLabel::clicked signal when the user clicks on the label.
 *
 * @note
 * By default, the label is clickable !
 */
void ClickableLabel::setClickable(bool enable) { mClickable = enable; }

/**
 * @brief Enable/disable label highlighting on mouse over.
 * @param enable True to enable, false otherwise.
 * @details
 * If highlighting is enabled, the label font is changed while the mouse is over the label, it is set to bold and  underline.
 * Additionally, the cursor changes from arrow to pointing hand.
 * Such font features are removed when the mouse cursor leaves the widget.
 * @note
 * By default, highlight is not enabled.
 */
void ClickableLabel::setHighlight(bool enable) { mHighlightable = enable; }

/// @brief Highlight the label's text on enter.
void ClickableLabel::enterEvent(QEvent *event) {
    QLabel::enterEvent(event);
    if (mHighlightable) {
        setCursor(Qt::PointingHandCursor);
        QFont f = font();
        f.setBold(true);
        f.setUnderline(true);
        setFont(f);
    }
    emit mouseEnter();
}
/// @brief Disable Highlight of the label's text on leave.
void ClickableLabel::leaveEvent(QEvent *event) {
    if (mHighlightable) {
        setCursor(Qt::ArrowCursor);
        QFont f = font();
        f.setBold(false);
        f.setUnderline(false);
        setFont(f);
    }
    emit mouseLeave();
    QLabel::leaveEvent(event);
}

/**
 * @brief Overrided mouseReleaseEvent.
 * @param event Mouse event.
 * @details
 * Emit ClickableLabel::clicked signal if the mouse position upon release is still in the label rectangle.
 */
void ClickableLabel::mouseReleaseEvent(QMouseEvent *event) {
    if (mClickable && rect().contains(event->pos(), true)) {
        emit clicked();
    }
    QLabel::mouseReleaseEvent(event);
}
