// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "datamanager.hpp"

#include "baseplot.hpp"
#include "dataitem.hpp"
#include "flightwidget.hpp"
#include "indicatorswidget.hpp"
#include "itemmapper.hpp"
#include "mapwidget.hpp"
#include "mediacontrolwidget.hpp"
#include "rtplotwidget.hpp"
#include "scenario.hpp"
#include "scenarioselector.hpp"
#include "scenariowidget.hpp"
#include "xmlutils.hpp"

// arcgis

#include "GeometryEngine.h"
#include "Graphic.h"
#include "GraphicListModel.h"
#include "GraphicsOverlay.h"
#include "Point.h"
#include "SimpleLineSymbol.h"
#include "SimpleMarkerSymbol.h"
#include "SpatialReference.h"

using namespace Esri::ArcGISRuntime;

static void clearOverlays(QVector<GraphicsOverlay *> &overlays);

static QPair<int, int> getNearestPoint(double timestamp, const QDoubleMatrix &times, int *global = nullptr);

/**
 * @brief Constructor.
 * @param parent Optional QObject parent.
 */
DataManager::DataManager(QObject *parent) : QObject(parent) {
    // plot scrolling
    mAutoScrollInterval = 200; // scrolled every 200 msecs.
    mNowIsReference = true;
    mLastTime = qQNaN();
    mMinSat = 4;
    mReplayData = nullptr;
    mReplayMode = false;
    mAutoRepeat = false;
    mReplayState = ReplayState::STOPPED;
    mReplayFrameRate = 50; // 100;
    mReplayTimer.setInterval(mReplayFrameRate);
    mReplayTimer.setTimerType(Qt::PreciseTimer);
    mFrameIndex = 0;
    mFrameStep = 1;
    mReplayOverlay = new GraphicsOverlay(this);
    mReplayMapGraphData = GraphContainerPointer::create();
    for (int i = 0; i < N_PLOTS; i++) {
        mReplayPlotsGraphData[i] = GraphContainerPointer::create();
    }

    mMapSpatialReference = gWGS84;
    mItemMapper = new ItemMapper(this);
    mColorGradient = QCPColorGradient(QCPColorGradient::gpThermal);
    mBlockCount = 0;
    mTotalCount = 0;
    mSelectedScenario = nullptr;
    mFlightModel = new FlightModel(this);
    mSensorTimes.push_back(QDoubleVector());
    mPcMonTimes.push_back(QDoubleVector());
    mPositions.push_back(PointVector());
    mValidGPS.push_back(BoolVector());

    GraphicsOverlay *go = new GraphicsOverlay(this);
    mOverlays.push_back(go);
    mMapGraphData.push_back(GraphContainerPointer::create());
    for (int i = 0; i < N_PLOTS; i++) {
        mPlotsGraphData[i].push_back(GraphContainerPointer::create());
    }
    mOutline = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("white"), 0.5, this);

    // stuff that can (could) be initialized from settings:

    mBypassGPS = true;

    mVisUseDuration = true;
    mNVisible = 30;
    mSecVisible = 30;

    mNewFlightDelay = 300;    // 5 minutes inactivity delay before creating a new flight !
    mNewFlightOnStrap = true; // always new flight if the physical new-file strap status has changed.
    mUsePCTime = true;
    mAutoCenter = true;

    updateFromSettings(false);

    // nullify widgets pointers
    mMediaControlWidget = nullptr;
    mFlightWidget = nullptr;
    mScenarioWidget = nullptr;
    mIndicatorsWidget = nullptr;
    mMapWidget = nullptr;
    for (int i = 0; i < N_PLOTS; i++)
        mPlotWidgets[i] = nullptr;

    QObject::connect(mFlightModel, &FlightModel::stateChanged, this, &DataManager::onFlightStateChanged);
    QObject::connect(&mReplayTimer, &QTimer::timeout, this, &DataManager::nextFrame);

    // set last time.
    mLastStartTime = QDateTime::currentDateTimeUtc();
}

/// @brief Destructor.
DataManager::~DataManager() { qDeleteAll(mScenarios); }

/**
 * @brief Set the timer interval for auto scrolling.
 * @param interval Interval in milliseconds.
 * @details
 * The interval cannot be less than 40 milliseconds.
 *
 * If auto scrolling is currently enabled, the timer is restarted with the new interval.
 */
void DataManager::setAutoScrollingInterval(int interval) {
    mAutoScrollInterval = qMax(40, interval);
    if (mAutoScrollTimer.isActive()) {
        mAutoScrollTimer.start(mAutoScrollInterval, this);
    }
}

/// @brief Get the timer interval for auto scrolling (msecs).
int DataManager::autoScrollingInterval() const { return mAutoScrollInterval; }

/**
 * @brief Enable/Disable now as time reference for X-axis auto-scrolling.
 * @param enable True to enable, false otherwise.
 * @details
 * When now is set as time reference (default behaviour), auto-scrolling uses the current datetime (i.e. now)
 * to compute the range to display. Otherwise, the last graph data is used (with now as fallback value).
 */
void DataManager::useNowAsTimeReference(bool enable) { mNowIsReference = enable; }

/**
 * @brief Enable time scrolling on registered plot widgets.
 * @param enable True to start scrolling, false to stop.
 */
void DataManager::enableTimeScrolling(bool enable) {
    // start to scroll.
    if (enable)
        mAutoScrollTimer.start(mAutoScrollInterval, this);
    else
        mAutoScrollTimer.stop();
}

/// @brief Check whether or not now is used as time reference.
bool DataManager::nowIsTimeReference() const { return mNowIsReference; }

/**
 * @brief Manage timer events.
 * @param event Time event.
 * @details
 * The method manages time events generated by the internal QBasicTimer. On timeout, it computes the time
 * reference (either the last time a block has been registered or now id MainWindow::mNowIsReference is true).
 * Then it scrolls the x-axis range of each plot widgets (only plot widget not in free mode are actually scrolled, see
 * RTPlotWidget::scrollXRange)
 */
void DataManager::timerEvent(QTimerEvent *event) {
    // auto-scrolling.
    if (event->timerId() == mAutoScrollTimer.timerId()) {
        double timeRef;
        if (!(mNowIsReference || qIsNaN(mLastTime))) {
            timeRef = mLastTime;
        } else {
            timeRef = QDateTime::currentMSecsSinceEpoch() / 1000.0;
        }
        for (BaseRtPlotWidget *pw : mPlotWidgets) {
            if (pw)
                pw->scrollXRange(timeRef);
        }
        if (mMapWidget)
            mMapWidget->scrollXRange(timeRef);
    }
}

/**
 * @brief Get the item mapper of this data manager.
 * @attention
 * The item mapper is owned by the data manager and must not be deleted outside the class.
 */
ItemMapper *DataManager::mapper() { return mItemMapper; }

void DataManager::translateModelHeaders() { mFlightModel->translateHeaders(); }

/**
 * @brief Reset the item mapper.
 * @param mapper Mapper to copy.
 * @details
 * The method resets the internal mapper. In order to maintain consistency with the new mapper, internal data is
 * cleared, including scenarios (see DataManager::cleanup).
 */
void DataManager::resetMapper(const ItemMapper &mapper) {
    cleanup();
    mItemMapper->copyFrom(&mapper);

    if (mReplayData) {
        delete mReplayData;
        mReplayData = nullptr;
    }

    // setup internal data map.
    const QVector<AParam> prms = mItemMapper->parameters();
    for (const AParam &prm : prms) {
        mLiveData[prm] = QDoubleMatrix(1, QDoubleVector());
        mExtremas[prm] = DoublePairVector(1, DoublePair(qQNaN(), qQNaN()));
    }

    // reset last start time
    mLastStartTime = QDateTime::currentDateTimeUtc();
}

/// @brief Register new media control widget.
void DataManager::registerMediaControlWidget(MediaControlWidget *widget) {
    if (mMediaControlWidget) {
        QObject::disconnect(mMediaControlWidget, &MediaControlWidget::play, this, &DataManager::playReplay);
        QObject::disconnect(mMediaControlWidget, &MediaControlWidget::pause, this, &DataManager::pauseReplay);
        QObject::disconnect(mMediaControlWidget, &MediaControlWidget::stop, this, &DataManager::stopReplay);
        QObject::disconnect(mMediaControlWidget, &MediaControlWidget::goToStart, this, &DataManager::goToStart);
        QObject::disconnect(mMediaControlWidget, &MediaControlWidget::goToEnd, this, &DataManager::goToEnd);
        QObject::disconnect(mMediaControlWidget, &MediaControlWidget::autoRepeat, this, &DataManager::setAutoRepeat);
        QObject::disconnect(
            mMediaControlWidget, &MediaControlWidget::frameStepChanged, this, &DataManager::changeFrameStep);
        QObject::disconnect(mMediaControlWidget, &MediaControlWidget::frameIndexChanged, this, &DataManager::goToFrame);
        QObject::disconnect(this, &DataManager::ReplayStateChanged, mMediaControlWidget, &MediaControlWidget::setState);
    }
    mMediaControlWidget = widget;
    if (mMediaControlWidget) {
        QObject::connect(mMediaControlWidget, &MediaControlWidget::play, this, &DataManager::playReplay);
        QObject::connect(mMediaControlWidget, &MediaControlWidget::pause, this, &DataManager::pauseReplay);
        QObject::connect(mMediaControlWidget, &MediaControlWidget::stop, this, &DataManager::stopReplay);
        QObject::connect(mMediaControlWidget, &MediaControlWidget::goToStart, this, &DataManager::goToStart);
        QObject::connect(mMediaControlWidget, &MediaControlWidget::goToEnd, this, &DataManager::goToEnd);
        QObject::connect(mMediaControlWidget, &MediaControlWidget::autoRepeat, this, &DataManager::setAutoRepeat);
        QObject::connect(
            mMediaControlWidget, &MediaControlWidget::frameStepChanged, this, &DataManager::changeFrameStep);
        QObject::connect(mMediaControlWidget, &MediaControlWidget::frameIndexChanged, this, &DataManager::goToFrame);
        QObject::connect(this, &DataManager::ReplayStateChanged, mMediaControlWidget, &MediaControlWidget::setState);
    }
}

/// @brief Register new flight widget.
void DataManager::registerFlightWidget(FlightWidget *widget) {
    if (mFlightWidget) {
        QObject::disconnect(mFlightWidget->tableView->selectionModel(),
            &QItemSelectionModel::selectionChanged,
            mFlightWidget,
            &FlightWidget::onSelectionChanged);
        QItemSelectionModel *m = mFlightWidget->tableView->selectionModel();
        mFlightWidget->tableView->setModel(nullptr);
        if (m)
            m->deleteLater();
        QObject::disconnect(mFlightWidget->addFlightAction, &QAction::triggered, this, &DataManager::createNewFlight);
        QObject::disconnect(
            mFlightWidget->delFlightAction, &QAction::triggered, this, &DataManager::deleteFlightFromAction);
        QObject::disconnect(mFlightWidget->checkAllAction, &QAction::triggered, this, &DataManager::checkAllFlights);
        QObject::disconnect(
            mFlightWidget->uncheckAllAction, &QAction::triggered, this, &DataManager::uncheckAllFlights);
    }
    mFlightWidget = widget;
    if (mFlightWidget) {
        QItemSelectionModel *m = mFlightWidget->tableView->selectionModel();
        mFlightWidget->tableView->setModel(mFlightModel);
        if (m)
            m->deleteLater();
        QObject::connect(mFlightWidget->tableView->selectionModel(),
            &QItemSelectionModel::selectionChanged,
            mFlightWidget,
            &FlightWidget::onSelectionChanged);

        QObject::connect(mFlightWidget->addFlightAction, &QAction::triggered, this, &DataManager::createNewFlight);
        QObject::connect(
            mFlightWidget->delFlightAction, &QAction::triggered, this, &DataManager::deleteFlightFromAction);
        QObject::connect(mFlightWidget->checkAllAction, &QAction::triggered, this, &DataManager::checkAllFlights);
        QObject::connect(mFlightWidget->uncheckAllAction, &QAction::triggered, this, &DataManager::uncheckAllFlights);
    }
}

/// @brief Register new scenario widget.
bool DataManager::registerScenarioSelector(ScenarioSelector *widget) {
    if (mScenarioSelectors.contains(widget))
        return false;
    mScenarioSelectors.append(ScenarioSelectorPointer(widget));
    QObject::connect(widget, &ScenarioSelector::currentIndexChanged, this, &DataManager::selectScenario);
    return true;
}

/// @brief Register new scenario widget.
bool DataManager::unregisterScenarioSelector(ScenarioSelector *widget) {
    if (!mScenarioSelectors.contains(widget))
        return false;
    mScenarioSelectors.removeAll(widget);
    QObject::disconnect(widget, &ScenarioSelector::currentIndexChanged, this, &DataManager::selectScenario);
    return true;
}

/// @brief Register new scenario widget.
void DataManager::registerScenarioWidget(ScenarioWidget *widget) {
    if (mScenarioWidget)
        unregisterScenarioSelector(mScenarioWidget->selector());

    mScenarioWidget = widget;
    if (mScenarioWidget)
        registerScenarioSelector(mScenarioWidget->selector());
}

/// @brief Register new map widget.
void DataManager::registerMapWidget(MapWidget *widget) {
    if (mMapWidget) {
#ifdef SCN_SEL_ON_MAP
        unregisterScenarioSelector(mMapWidget->selector());
#endif
        // perform disconnections if needed.
        QObject::disconnect(mMapWidget->autoCenterAction(), &QAction::toggled, this, &DataManager::setAutoCenter);
        QObject::disconnect(mMapWidget, &MapWidget::colorRangeChanged, this, &DataManager::updateColorRange);
        QObject::disconnect(mMapWidget, &MapWidget::thresholdChanged, this, &DataManager::updateRemanenceThreshold);
        QObject::disconnect(
            mMapWidget, &MapWidget::spatialReferenceChanged, this, &DataManager::updateMapSpatialReference);
        QObject::disconnect(
            mMapWidget, &MapWidget::contextMenuRequested, this, &DataManager::onMapContextMenuRequested);
    }
    mMapWidget = widget;
    if (mMapWidget) {
#ifdef SCN_SEL_ON_MAP
        registerScenarioSelector(mMapWidget->selector());
#endif
        mMapSpatialReference = mMapWidget->spatialReference();
        mMapWidget->setColorGradient(mColorGradient);
        setAutoCenter(mMapWidget->autoCenterAction()->isChecked());
        // perform connections if needed.
        QObject::connect(mMapWidget->autoCenterAction(), &QAction::toggled, this, &DataManager::setAutoCenter);
        QObject::connect(mMapWidget, &MapWidget::colorRangeChanged, this, &DataManager::updateColorRange);
        QObject::connect(mMapWidget, &MapWidget::thresholdChanged, this, &DataManager::updateRemanenceThreshold);
        QObject::connect(
            mMapWidget, &MapWidget::spatialReferenceChanged, this, &DataManager::updateMapSpatialReference);
        QObject::connect(mMapWidget, &MapWidget::contextMenuRequested, this, &DataManager::onMapContextMenuRequested);
    }
}

/// @brief Register new plot widget.
bool DataManager::registerPlotWidget(int index, RtPlotWidget *widget) {
    if (index < 0 || index >= N_PLOTS)
        return false;
    if (mPlotWidgets[index]) {
        // perform disconnections if needed.
        auto p = mPlotWidgets[index]->plot();

        QObject::disconnect(p, nullptr, this, nullptr);
    }
    mPlotWidgets[index] = widget;
    if (mPlotWidgets[index]) {
        // perform connections if needed.
        QObject::connect(widget->plot(),
            &BasePlot::customContextMenuRequested,
            this,
            &DataManager::onPlotWidgetContextMenuRequested);
    }
    return true;
}

/// @brief Register new indicators widget.
void DataManager::registerIndicatorsWidget(IndicatorsWidget *widget) {
    if (mIndicatorsWidget) {
        // perform disconnections if needed.
    }
    mIndicatorsWidget = widget;
    if (mIndicatorsWidget) {
        // perform connections if needed.
    }
}

/**
 * @brief Check all flights.
 * @details
 * The method checks all flights in the internal flight model and sets all the overlays visible.
 */
void DataManager::checkAllFlights() {
    mFlightModel->checkAll(false);
    for (auto overlay : qAsConst(mOverlays))
        overlay->setVisible(true);
}

/**
 * @brief Uncheck all flights.
 * @details
 * The method unchecks all flights in the internal flight model and sets all the overlays hiddens.
 */
void DataManager::uncheckAllFlights() {
    mFlightModel->uncheckAll(false);
    for (auto overlay : qAsConst(mOverlays))
        overlay->setVisible(false);
}

/**
 * @brief Switch between PC time and Sensor time.
 * @param enable True to use PC time as reference timeline.
 * @details
 * The method switches the time series for plot data between the monitoring PC time and the Sensor time.
 *
 * By default, PC time is used to compute timeline. This is preferred behavior as it will not generate drift on the
 * plots.
 */
void DataManager::setPCTimeAsReference(bool enable) {
    if (enable == mUsePCTime)
        return;
    mUsePCTime = enable;

    // update the current data.
    const QDoubleMatrix &times = mUsePCTime ? mPcMonTimes : mSensorTimes;
    QVector<GraphContainerVector> containers;
    containers.reserve(N_PLOTS + 1);
    containers.push_back(mMapGraphData);
    for (int i = 0; i < N_PLOTS; i++)
        containers.push_back(mPlotsGraphData[i]);

    for (GraphContainerVector c : containers) {
        if (c.size() != times.size()) {
            qDebug() << "DataManager::setPCTimeAsReference : Inconsistent size in outer containers";
            continue;
        }
        for (int i = 0; i < times.size(); i++) {
            const QDoubleVector &values = times.at(i);
            GraphContainerPointer gdc = c[i];
            if (gdc.isNull() || (gdc->size() != values.size())) {
                qDebug() << "DataManager::setPCTimeAsReference : Inconsistent size in inner containers";
                continue;
            }
            int j = 0;
            for (auto it = gdc->begin(); it != gdc->end(); ++it) {
                it->key = values.at(j);
                j++;
            }
        }
    }

    // refresh widgets
    if (mMapWidget)
        mMapWidget->refreshPlot();
    for (int i = 0; i < N_PLOTS; i++) {
        if (mPlotWidgets[i])
            mPlotWidgets[i]->refreshPlot();
    }
    return;
}

/// @brief Check whether or not PC time is used as reference for the time line.
///  @sa DataManager::setPCTimeAsReference
bool DataManager::pcTimeIsReference() const { return mUsePCTime; }

/**
 * @brief Compute the name of the file being written in the sensor.
 * @param noSD Optional pointer to a bool with the status of the <i>SD card detection</i> flag (see below for
 * details).
 * @details
 * The method computes the name of the file that the sensor uses for data save based on the
 * (currently) registered value of the item mapper.
 *
 * The method checks for the SD card detection flag value registered in the item mapper. If the parameter
 * is present and its value is zero then it is assumed that no SD card is present in the sensor. In such case
 * the hard-coded (but i18n'ed) string "NO SD CARD" is returned by the method and the argument @b noSD, i given,
 * is set to true.
 */

QString DataManager::computeSensorFileName(bool *noSD) const {
    bool ba, bb, bc, bd;
    qint64 ia, ib, ic, id;
    if (noSD)
        *noSD = false;
    ia = (qint64)mItemMapper->parameter(ItemMapper::F_DATETIME).floatValue(&ba);
    ib = (qint64)mItemMapper->parameter(ItemMapper::A_STRAP_COUNT).floatValue(&bb);
    ic = (qint64)mItemMapper->parameter(ItemMapper::F_NUMBER).floatValue(&bc);
    id = (qint64)mItemMapper->parameter(ItemMapper::SD_DETECT).floatValue(&bd);
    if (bd && id == 0) {
        if (noSD)
            *noSD = true;
        return tr("NO SD CARD");
    }
    if (!(ba && bb && bc))
        return "N/A";
    return QString("20%1_Apodrone_%2_%3.bin").arg(ia).arg(ib).arg(ic);
}

/**
 * @brief Reset data.
 * @details
 * The method resets (clears) the data. Both the item mapper and scenarios are left untouched.
 * Only data related to flights for the current session are removed.
 *
 * This method can be used to partially clean up the UI, at least the registered widgets.
 */
void DataManager::resetData() {
    if (mReplayMode)
        return;
    setupWidgetsUpdatesEnabled(false);
    DataItem *imap = mapItem();
    DataItem *iplots[N_PLOTS];
    for (int i = 0; i < N_PLOTS; i++)
        iplots[i] = plotItem(i);

    // remove item from map and plots (=> remove refs for shared pointers)
    setItemOnMap(nullptr);
    for (int i = 0; i < N_PLOTS; i++)
        setItemOnPlot(i, nullptr);

    // clear internal data: everything expect for the scenarios !
    mTotalCount = 0;
    mSensorTimes.clear();
    mPcMonTimes.clear();
    mPositions.clear();
    mValidGPS.clear();
    for (QMap<AParam, QDoubleMatrix>::iterator it = mLiveData.begin(); it != mLiveData.end(); ++it)
        it.value().clear();
    for (QMap<AParam, DoublePairVector>::iterator it = mExtremas.begin(); it != mExtremas.end(); ++it)
        it.value().clear();

    qDeleteAll(mOverlays);
    mOverlays.clear();
    mMapGraphData.clear();
    for (int i = 0; i < N_PLOTS; i++) {
        mPlotsGraphData[i].clear();
    }

    // reset mapper current values.
    mItemMapper->resetValues();
    // reset flight model.
    mFlightModel->resetFlights();

    // create new vectors
    mSensorTimes.push_back(QDoubleVector());
    mPcMonTimes.push_back(QDoubleVector());
    mPositions.push_back(PointVector());
    mValidGPS.push_back(BoolVector());

    QMap<AParam, QDoubleMatrix>::iterator it1 = mLiveData.begin();
    while (it1 != mLiveData.end()) {
        it1.value().push_back(QDoubleVector());
        ++it1;
    }
    QMap<AParam, DoublePairVector>::iterator it2 = mExtremas.begin();
    while (it2 != mExtremas.end()) {
        it2.value().push_back(DoublePair(qQNaN(), qQNaN()));
        ++it2;
    }
    mOverlays.push_back(new GraphicsOverlay(this));
    mMapGraphData.push_back(GraphContainerPointer::create());
    for (int i = 0; i < N_PLOTS; i++)
        mPlotsGraphData[i].push_back(GraphContainerPointer::create());

    // restore items.
    setItemOnMap(imap);
    for (int i = 0; i < N_PLOTS; i++)
        setItemOnPlot(i, iplots[i]);
    setupWidgetsUpdatesEnabled(true);
    updateWidgets();
}

/**
 * @brief Update registered widgets.
 * @details
 * The method refreshes the plot widgets and updates the indicators.
 */
void DataManager::updateWidgets() {
    if (mMapWidget)
        mMapWidget->refreshPlot();
    for (int i = 0; i < N_PLOTS; i++) {
        if (mPlotWidgets[i])
            mPlotWidgets[i]->refreshPlot();
    }
    if (mIndicatorsWidget)
        mIndicatorsWidget->refreshValues();
}

/**
 * @brief Enable/Disable registered widget updates
 * @param enable True to enable, false otherwise.
 * @details
 * The method is convinient wrapper to QWidget::setUpdatesEnabled this is applied for each registered widgets.
 */
void DataManager::setupWidgetsUpdatesEnabled(bool enable) {
    if (mMapWidget)
        mMapWidget->setUpdatesEnabled(enable);
    for (int i = 0; i < N_PLOTS; i++) {
        if (mPlotWidgets[i])
            mPlotWidgets[i]->setUpdatesEnabled(enable);
    }
    if (mIndicatorsWidget)
        mIndicatorsWidget->setUpdatesEnabled(enable);
    if (mScenarioWidget)
        mScenarioWidget->setUpdatesEnabled(enable);
    if (mFlightWidget)
        mFlightWidget->setUpdatesEnabled(enable);
}

/**
 * @brief Process Amulse data block.
 * @param block Block to process.
 * @param Optional pointer to a bool with in output true if a new flight has been automatically created, false
 * otherwise.
 * @return True if the block has been added to the data manager, false otherwise.
 * @details
 * The method first updates its item mapper (always done whatever the return status).
 * Then it checks for the number of satellites acquired during the acquisition of the block.
 * At least four satellites are expected, otherwise the method stops further processing and returns false.
 * Then it checks for geographic positions and time, all of them must be non-NaN values, otherwise the method
 * returns false.
 *
 * If number of satellites, positions and time are successfully tested then, the method:
 *
 * 1. Checks for new flight and create a new one if needed.
 * 2. Records the block on filesystem if data save is enabled.
 * 3. Adds new data point to each data item of the manager.
 * 4. Emit DataManager::newFlightAdded signal if a new flight has been created.
 *
 * @note
 * After this method has been successfully called, plot widgets should be replot and optionally updated for
 * new graph if a new flight has been generated.
 * @todo
 * The method
 */
bool DataManager::processBlock(const ABlock &block, bool *newFlightCreated) {
    mItemMapper->registerValues(block);

    if (newFlightCreated)
        (*newFlightCreated) = false;

    bool isOk = true, nf = false;
    double lon = qQNaN(), lat = qQNaN(), alt = qQNaN(), stime, ptime;

    // Check for the number of satellite: first degree of confidence !
    int nbSat = mItemMapper->parameter(ItemMapper::NB_SATS).intValue();
    stime = mItemMapper->value(ItemMapper::SENSOR_TIME) + EPOCH1904_OFFSET;
    ptime = mItemMapper->value(ItemMapper::PCMON_TIME) + EPOCH1904_OFFSET;
    mLastTime = ptime;

    bool gpsValid = nbSat >= mMinSat;

    // validation algorithm:
    // 1. Gps is valid (i.e. enough sats.):
    //    1. Time can not be computed: ERROR -> skip block.
    //    2. Position can not be computed: if mBypassGPS -> OK, else -> skip block.
    // 2. Gps is invalid: if mBypassGPS -> OK, else -> skip block.
    if (gpsValid) {
        // Extract geographic position and time.
        lon = mItemMapper->value(ItemMapper::LONGITUDE);
        lat = mItemMapper->value(ItemMapper::LATITUDE);
        alt = mItemMapper->value(ItemMapper::ALTITUDE);
        bool noGPS = (qIsNaN(lon) || qIsNaN(lat) || qIsNaN(alt));
        // uncoverable error.
        if (qIsNaN(stime) || qIsNaN(ptime)) {
            LOG_ERROR(tr("Time is not valid (NaN detected)... Skipping block."));
            isOk = false;
        } else if (noGPS) {
            LOG_WARNING(tr("Geographic position is not valid (NaN detected)...%2")
                            .arg(mBypassGPS ? "" : tr(" Skipping block.")));
            isOk = mBypassGPS;
        }
    } else {
        LOG_WARNING(tr("Insufficient number of satellites detected (%1)...%2")
                        .arg(nbSat)
                        .arg(mBypassGPS ? "" : tr(" Skipping block.")));
        isOk = mBypassGPS;
    }

    // if we allow to bypass GPS, then everything is fine.
    // gpsValid flag which is recorded with the (null) GPS position will prevent anythin nasty to occur
    //  during rendering on map.

    // If something is wrong: stop processing.
    if (mReplayMode || !isOk)
        return false;

    // check for new flight and create if needed:
    // the method will append new vectors to the data containers and add overlay and graph to the registered widgets
    // if an item is selected.
    if (checkForNewFlight(block)) {
        nf = true;
        createNewFlightPvt();
    }

    mLastBlock = block;
    mBlockCount++;
    mTotalCount++;

    mSensorTimes.last().push_back(stime);
    mPcMonTimes.last().push_back(ptime);
    mPositions.last().push_back(Point(lon, lat, alt, gWGS84));
    mValidGPS.last().push_back(gpsValid);

    // compute new extremas (unused as of v2.1)
    QMap<AParam, QDoubleMatrix>::iterator it = mLiveData.begin();
    while (it != mLiveData.end()) {
        // double value = block(it.key());
        double value = mItemMapper->value(it.key());
        it.value().last().append(value);
        DoublePair &p = mExtremas[it.key()].last();
        if (qIsNaN(p.first) || p.first > value)
            p.first = value;
        if (qIsNaN(p.second) || p.second < value)
            p.second = value;
        ++it;
    }

    for (Scenario *scenario : qAsConst(mScenarios)) {
        scenario->registerValues(mItemMapper);
    }

    // Build the Graphics for the map.
    double value = qQNaN();
    double time = mUsePCTime ? ptime : stime;
    bool vis = true; // false;

    // default style for the graphics
    QColor color(Qt::black);
    SimpleMarkerSymbolStyle style = mCFStyle;
    qreal msize = mCFSize;

    DataItem *item = mapItem();
    if (item) {
        vis = true;
        value = item->value();
        color = mColorGradient.color(value, item->colorRange().toQCPRange());
    }

    // manage missing value style
    if (qIsNaN(value)) {
        color = mMVColor;
        style = mMVStyle;
        msize = mMVSize;
        vis = mMVShow;
    }

    bool autoCenter = mAutoCenter && mMapWidget && item;

    QVariantMap attrs = {{"value", value}, {"gps_valid", gpsValid}};
    Point point(lon, lat, alt, gWGS84);
    if (mMapSpatialReference != gWGS84)
        point = GeometryEngine::project(point, mMapSpatialReference);
    Graphic *graphic = new Graphic(point, attrs, mOverlays.last());

    SimpleMarkerSymbol *sms = new SimpleMarkerSymbol(style, color, msize, graphic);
    sms->setOutline(mOutline);
    graphic->setSymbol(sms);
    graphic->setVisible(vis && gpsValid);

    GraphicListModel *model = mOverlays.last()->graphics();
    model->append(graphic);

    // manage points visibility:
    AttributeListModel *attrModel;
    if (mVisUseDuration) {
        int iglobal;
        double t, v, timestamp = ptime - mSecVisible;
        QPair<int, int> tmp = getNearestPoint(timestamp, mPcMonTimes, &iglobal);
        QVector<QPair<int, int>> pairs;
        for (int i = -4; i < 0; i++)
            pairs.push_back(globalToLocals(iglobal + i));
        pairs.push_back(tmp);
        for (const QPair<int, int> &idxs : pairs) {
            graphic = (idxs.first == -1) ? nullptr : mOverlays.at(idxs.first)->graphics()->at(idxs.second);
            if (graphic) {
                t = mPcMonTimes.at(idxs.first).at(idxs.second);
                attrModel = graphic->attributes();
                v = attrModel->attributeValue("value").toDouble();
                bool gpsv = attrModel->attributeValue("gps_valid").toBool();
                if (item) {
                    graphic->setVisible(gpsv && (item->isVisible(v) || t >= timestamp));
                } else {
                    graphic->setVisible(false);
                }
            }
        }
    } else {
        // check whether or not we should hide the point that was added mNVisible points ago ^^
        graphic = getGraphicFromEnd(mNVisible);
        if (graphic) {
            attrModel = graphic->attributes();
            double v = attrModel->attributeValue("value").toDouble();
            bool gpsv = attrModel->attributeValue("gps_valid").toBool();
            if (item) {
                graphic->setVisible(gpsv && item->isVisible(v));
            } else {
                graphic->setVisible(false);
            }
        }
    }

    mMapGraphData.last()->add(QCPGraphData(time, value));

    for (int i = 0; i < N_PLOTS; i++) {
        item = plotItem(i);
        value = item ? item->value() : qQNaN();
        mPlotsGraphData[i].last()->add(QCPGraphData(time, value));
    }

    if (newFlightCreated)
        (*newFlightCreated) = nf;

    if (nf) {
        emit newFlightAdded();
    }

    // center the map on new position if needed
    if (autoCenter)
        mMapWidget->centerMapOn(lat, lon);
    updateWidgets();
    emit mapperUpdated();
    return true;
}

/**
 * @brief Setup new flight (internal method).
 * @details
 * The method creates a new flight to manage:
 *   - The flight model is updated and new flight is added.
 *   - The block counter is re-initialized.
 *   - All internal data containers are updated.
 *   - Add data container (overlay and graph container) to registered map and plots widgets if an item is
 * associated.s
 */
void DataManager::createNewFlightPvt() {
    QDateTime now = QDateTime::currentDateTimeUtc();
    // Create new flight in the flight model ...
    mFlightModel->addFlight(mLastStartTime, now, mBlockCount);

    // ... Create new vectors.
    mSensorTimes.push_back(QDoubleVector());
    mPcMonTimes.push_back(QDoubleVector());
    mPositions.push_back(PointVector());
    mValidGPS.push_back(BoolVector());
    QMap<AParam, QDoubleMatrix>::iterator it1 = mLiveData.begin();
    while (it1 != mLiveData.end()) {
        it1.value().push_back(QDoubleVector());
        ++it1;
    }
    QMap<AParam, DoublePairVector>::iterator it2 = mExtremas.begin();
    while (it2 != mExtremas.end()) {
        it2.value().push_back(DoublePair(qQNaN(), qQNaN()));
        ++it2;
    }

    DataItem *item = mapItem();
    // create new overlays and graph data containers
    if (mOverlays.size()) {
        // update style/size of the current flight (will become a previous flight ^^)
        GraphicListModel *graphics = mOverlays.last()->graphics();
        if (item) {
            for (auto it = graphics->begin(); it != graphics->end(); it++) {
                SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
                symbol->setStyle(mPFStyle);
                symbol->setSize(mPFSize);
            }
        } else {
            // no item: set missing value style  !
            for (auto it = graphics->begin(); it != graphics->end(); it++) {
                SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
                if (symbol) {
                    symbol->setStyle(mMVStyle);
                    symbol->setSize(mMVSize);
                    symbol->setColor(mMVColor);
                }
            }
        }
    }
    // create new containers:
    mOverlays.push_back(new GraphicsOverlay(this));
    mMapGraphData.push_back(GraphContainerPointer::create());
    if (item && mMapWidget) { // add to map if we have a selected item.
        mMapWidget->addOverlay(mOverlays.last());
        mMapWidget->addGraph(mMapGraphData.last());
    }
    for (int i = 0; i < N_PLOTS; i++) {
        mPlotsGraphData[i].push_back(GraphContainerPointer::create());
        // add container to plot widget if needed.
        RtPlotWidget *widget = mPlotWidgets[i];
        item = plotItem(i);
        if (item && widget)
            widget->addGraph(mPlotsGraphData[i].last());
    }
    // reset the block counter.
    mBlockCount = 0;
    // update current start time.
    mLastStartTime = now;
}

/**
 * @brief Setup new flight.
 * @details
 * The method creates a new flight to manage:
 *   - If the data are saved to disk, the internal file handler is freed.
 *   - The flight model is updated and new flight is added.
 *   - The block counter is re-initialized.
 *   - Each data item is updated to add data in a new flight container.
 *   - If the data are saved to disk, the internal file handler is initialized.
 *
 * Finally the method emits the signal DataManager::newFlightAdded.
 *
 * @internal
 * Actually, the method is a simple wrapper to DataManager::createNewFlightPvt that does emit
 * DataManager::newFlightAdded signal.
 */
void DataManager::createNewFlight() {
    createNewFlightPvt();
    emit newFlightAdded();
}

/**
 * @brief Check whether or not a new flight should be created upon block registration.
 * @param block Block to test.
 * @return True if a new flight should be generated, false otherwise.
 *
 * @details
 * Base on the current settings and latest available data block content, the method checks whether or not
 * the new flight should be created before registering the given block.
 *
 * A new flight can be created automatically by DataManager::processBlock if:
 *
 * - The duration derived from h_system parameter of the given block and the last registered one
 *   exceed the delay as returned by DataManager::newFlightDelay.
 * - DataManager::newFlightOnStrap is true and the new-file strap counters of the given block and the last
 * registered one are different.
 *
 * @note
 * [2020/06/09] Apodrone v2 may adjust its internal system time to GPS time in the waiting mode (i.e. Power on /
 * Acquisition strap off). Thus two consecutive blocks may have a large gap for h_system due to the time drift of
 * the NI board. This may generate an extra new flight.
 * @note
 * Using the new-file strap is more reliable as it is a human action.
 */
bool DataManager::checkForNewFlight(const ABlock &block) const {
    // no automatic new-flight is asked.
    if (mNewFlightDelay <= 0 && !mNewFlightOnStrap)
        return false;
    // if one the block is not valid we cannot compare them.
    if (!(mLastBlock.isValid() && block.isValid()))
        return false;

    int ls = 0, cs = 0;
    QString name = mItemMapper->keyName(ItemMapper::F_STRAP_COUNT);
    if (name.size()) {
        ls = mLastBlock.parameter(name.toStdString()).intValue();
        cs = block.parameter(name.toStdString()).intValue();
    }
    double lt = mLastBlock.time(), ct = block.time();
    return (mNewFlightOnStrap && ls != cs) || (mNewFlightDelay > 0 && (ct - lt) > mNewFlightDelay);
}

/// @brief Get the number of scenario registered in the data manager.
int DataManager::scenarioCount() const { return mScenarios.size(); }

/// @brief Check whether or not the manager is allowed to create new flight using the new-flight strap status.
bool DataManager::newFlightOnStrap() const { return mNewFlightOnStrap; }

/// @brief Enable/Disable the automatic creation of flight using the new-flight strap status.
void DataManager::setNewFlightOnStrap(bool enable) { mNewFlightOnStrap = enable; }

/// @brief Get the current delay (in seconds) before generating a new flight.
int DataManager::newFlightDelay() const { return mNewFlightDelay; }

/**
 * @brief Set the delay for the new-flight creation algorithm.
 * @param newFlightDelay Delay to set in seconds.
 * @sa DataManager::checkForNewFlight
 * @details
 * Unless it is set to zero or less (which means that the feature is disabled), the delay allows the manager to
 * generate a new flight if the duration between two consecutive registered blocks are greater than the delay.
 *
 * @note
 * For non-zero (or negative) delay, the method forces the delay to be at least 60 seconds.
 */
void DataManager::setNewFlightDelay(int delay) { mNewFlightDelay = (delay <= 0) ? 0 : qMax(60, delay); }

/**
 * @brief Clear the data manager.
 * @details
 * The method clears all the data stored in the manager including flights and scenarios.
 *
 * The item mapper is not reset by the method.
 */
void DataManager::cleanup() {
    // manage widgets
    // reset the current scenario to cleanup widgets.
    selectScenario(-1);
    qDeleteAll(mScenarios);
    mScenarios.clear();
    resetData();
}

/**
 * @brief Set new items to manage.
 * @param scenarios Vector of scenarios to manage.
 * @details
 * The method replaces the internal scenarios by the given ones.
 *
 * @internal
 * The method clears the current data of the manager and resets the internal raw data map.
 */
void DataManager::setScenarios(const QVector<Scenario *> &scenarios, int *selected) {
    // reset the current scenario to cleanup widgets.
    selectScenario(-1);
    int selIndex = selected ? (*selected) : -1;
    if (selIndex < 0 || selIndex >= scenarios.size())
        selIndex = -1;
    qDeleteAll(mScenarios);
    mScenarios.clear();
    mScenarios = scenarios;

    for (auto p : mScenarioSelectors) {
        p->setup(mScenarios, selIndex);
    }
    mScenarioWidget->setupSelector(mScenarios, selIndex);
    if (selIndex != -1)
        selectScenario(selIndex);
}

/**
 * @brief Add new parameter to the internal map of data.
 * @param names Names of the parameters to add.
 */
void DataManager::addParameters(QStringList names) {
    // build template vectors to assign to each new parameter.
    static const double nan = qQNaN();
    int fc = qMax(1, flightCount()), pc = mLiveData.count();
    DoublePairVector tmp2(fc, DoublePair(qQNaN(), qQNaN()));
    QDoubleMatrix tmp1(fc, QDoubleVector());
    if (pc) {
        const QDoubleMatrix &a = mLiveData.first();
        for (int i = 0; i < fc; ++i) {
            tmp1[i].fill(nan, a[i].size());
        }
    }
    // add new parameters to the internal maps.
    names.removeDuplicates();
    for (const QString &n : names) {
        AParam prm = mItemMapper->parameter(n);
        if (prm.isValid() && !mLiveData.contains(prm)) {
            mLiveData[prm] = tmp1;
            mExtremas[prm] = tmp2;
        }
    }
}

/**
 * @brief Save scenarios to disk.
 */
bool DataManager::saveScenarios(const QString &xmlPath) {
    int selected = -1;
    QStringList errors;
    int itot = 0, stot = mScenarios.size();

    if (!stot) {
        LOG_WARNING("No scenarios to save.");
        LOG_WARNING("Failed to save scenarios to disk.");
        return false;
    }
    for (int i = 0; i < stot; i++) {
        const Scenario *scn = mScenarios.at(i);
        itot += scn->itemCount();
        if (mSelectedScenario && scn == mSelectedScenario) {
            selected = i;
        }
    }
    if (!itot) {
        LOG_WARNING("All scenarios have no items.");
        LOG_WARNING("Failed to save scenarios to disk.");
        return false;
    }
    if (mSelectedScenario) {
        auto it = std::find(mScenarios.begin(), mScenarios.end(), mSelectedScenario);
        if (it != mScenarios.end())
            selected = std::distance(mScenarios.begin(), it);
    }

    bool ret = writeScenariosToXML(xmlPath, mScenarios, &selected, &errors);
    if (ret) {
        LOG_INFO("Scenarios successfully saved to disk.");
    } else {
        LOG_WARNING("Failed to save scenarios to disk.");
        for (const QString &e : errors)
            LOG_DEBUG(e);
    }
    return ret;
}

/**
 * @brief Register new spatial reference of the map and (re)project all data if needed.
 * @param reference New reference of the afferent map.
 */
void DataManager::updateMapSpatialReference(const SpatialReference &reference) {
    if (reference.isEmpty())
        return;
    if (reference != mMapSpatialReference) {
        mMapSpatialReference = reference;
    }
    int no = mOverlays.size();
    for (int i = 0; i < no; i++) {
        auto *overlay = mOverlays.at(i);
        GraphicListModel *graphics = overlay->graphics();
        for (auto it = graphics->begin(); it != graphics->end(); it++) {
            (*it)->setGeometry(GeometryEngine::project((*it)->geometry(), mMapSpatialReference));
        }
    }

    GraphicListModel *graphics = mReplayOverlay->graphics();
    for (auto it = graphics->begin(); it != graphics->end(); it++) {
        (*it)->setGeometry(GeometryEngine::project((*it)->geometry(), mMapSpatialReference));
    }
}

/**
 * @brief Update color range of the current map item (if any).
 * @param range Data range to set.
 * @details
 * The method updates the color range of the current map item and updates the color of each graphics.
 * If no item is currently set on the map, the method does nothing !
 */
void DataManager::updateColorRange(const DataRange &range) {
    DataItem *item = mapItem();
    if (!item || (item->colorRange() == range))
        return;
    item->setColorRange(range);
    setupMarkerColors();
}

/**
 * @brief Update visibility range of the current map item (if any).
 * @param range Data range to set.
 * @details
 * The method updates the visibility range of the current map item and updates the visibility of each graphics.
 * If no item is currently set on the map, the method does nothing !
 */
void DataManager::updateRemanenceThreshold(double value) {
    DataItem *item = mapItem();
    if (!item)
        return;
    item->setThreshold(value);
    setupMarkerVisibility();

    // update threshold lines of all other plots.
    for (int i = 0; i < N_PLOTS; i++) {
        RtPlotWidget *w = mPlotWidgets[i];
        if (w && w->currentItem() == item)
            w->setThLineValue(value);
    }
}

/**
 * @brief Set a new color gradient in the data manager.
 * @param gradient Gradient to set.
 * @details
 * If an item is currently set in the map widget, the method also updates internally the color of
 * all graphics in the overlays.
 */
void DataManager::setColorGradient(const QCPColorGradient &gradient) {
    if (gradient == mColorGradient)
        return;
    mColorGradient = gradient;
    if (mMapWidget)
        mMapWidget->setColorGradient(mColorGradient);
    setupMarkerColors();
}

/**
 * @brief Setup the color of visible marker on the map.
 * @details
 * The method updates the colors of the visible marker in the internal overlays.
 * If no item is currently set on the map (that is the overlays are hidden), the method does nothing.
 */
void DataManager::setupMarkerColors() {
    DataItem *item = mapItem();
    if (!item)
        return;
    QCPRange cRange(item->colorRange().toQCPRange());
    QColor color;
    if (mReplayMode) {
        GraphicListModel *graphics = mReplayOverlay->graphics();
        // NOTE:
        // we should account for mFrameIndex to limit the for-loop.
        // In Replay mode all items indexes greater than mFrameIndex should be invisible...
        // No need to update them.
        for (auto it = graphics->begin(); it != graphics->end(); it++) {
            if ((*it)->isVisible()) {
                double mapValue = (*it)->attributes()->attributeValue("value").toDouble();
                color = qIsNaN(mapValue) ? mMVColor : QColor(mColorGradient.color(mapValue, cRange));
                static_cast<SimpleMarkerSymbol *>((*it)->symbol())->setColor(color);
            }
        }
        return;
    }
    int no = mOverlays.size();
    for (int i = 0; i < no; i++) {
        auto *overlay = mOverlays.at(i);
        GraphicListModel *graphics = overlay->graphics();
        for (auto it = graphics->begin(); it != graphics->end(); it++) {
            if ((*it)->isVisible()) {
                double mapValue = (*it)->attributes()->attributeValue("value").toDouble();
                color = qIsNaN(mapValue) ? mMVColor : QColor(mColorGradient.color(mapValue, cRange));
                static_cast<SimpleMarkerSymbol *>((*it)->symbol())->setColor(color);
            }
        }
    }
}

/**
 * @brief Setup the visibility of the markers on the map.
 * @details
 * The method updates the visibility of the markers in the internal overlays.
 * If no item is currently set on the map, the method does nothing. Otherwise, the visibility is calculated using
 * the item's value and index: an item is visible if its value is higher than the remanence threshold (and not NAN)
 * or its index is within the last N visible indexes (see DataManager::updateFromSettings).
 */
void DataManager::setupMarkerVisibility() {
    DataItem *item = mapItem();
    if (!item)
        return;
    int j = 0, jvis = visibleIndex();
    bool visible;
    AttributeListModel *attrs = nullptr;
    if (mReplayMode) {
        GraphicListModel *graphics = mReplayOverlay->graphics();
        for (auto it = graphics->begin(); it != graphics->end(); it++, j++) {
            if (j > mFrameIndex) {
                (*it)->setVisible(false);
                continue;
            }
            attrs = (*it)->attributes();
            double mapValue = attrs->attributeValue("value").toDouble();
            bool vgps = attrs->attributeValue("gps_valid").toBool();
            if (!vgps) {
                visible = false;
            } else if (qIsNaN(mapValue)) {
                visible = mMVShow && (j >= jvis);
            } else {
                visible = (j >= jvis) || item->isVisible(mapValue);
            }
            (*it)->setVisible(visible);
        }
        return;
    }

    int no = mOverlays.size();
    for (int i = 0; i < no; i++) {
        GraphicListModel *graphics = mOverlays.at(i)->graphics();
        for (auto it = graphics->begin(); it != graphics->end(); it++, j++) {
            attrs = (*it)->attributes();
            double mapValue = attrs->attributeValue("value").toDouble();
            bool vgps = attrs->attributeValue("gps_valid").toBool();
            if (!vgps) {
                visible = false;
            } else if (qIsNaN(mapValue)) {
                visible = mMVShow && (j >= jvis);
            } else {
                visible = (j >= jvis) || item->isVisible(mapValue);
            }
            (*it)->setVisible(visible);
        }
    }
}

/**
 * @brief Compute the index from which data should always be visible.
 * @return Global index.
 * @details
 * The method compute the global index. All indexes
 */
int DataManager::visibleIndex() {
    int jvis = 0;
    if (mReplayMode) {
        if (!mReplayData)
            return jvis;
        if (mVisUseDuration) {
            const QDoubleVector &tmp = mReplayData->times;
            jvis = mFrameIndex;
            double timestamp = tmp.at(jvis) - mSecVisible;
            if (timestamp < 0) {
                jvis = 0;
            } else {
                while (jvis >= 0 && tmp.at(jvis) >= timestamp)
                    jvis--;
            }
        } else {
            jvis = (mFrameIndex + 1) - mNVisible;
        }
    } else {
        if (mVisUseDuration) {
            jvis = 0;
            if (!mPcMonTimes.last().isEmpty()) {
                double lastTime = mPcMonTimes.last().last();
                getNearestPoint(lastTime - mSecVisible, mPcMonTimes, &jvis);
            }
        } else {
            jvis = size() - mNVisible;
        }
    }
    return jvis;
}

/**
 * @brief Compute the global index associated to the given local indexes.
 * @param fi Flight index to process.
 * @param li Local indexes to process.
 * @return Global index.
 * @details
 * The method converts a couple of local indexes (i.e. indexes of the outer and innner vectors of the data matrix),
 * into a global index that ranges from zero to the total number of blocks.
 *
 * If \a fi or \a li are out of range, the method returns -1.
 */
int DataManager::localsToGlobal(int fi, int li) {
    if (fi == -1 || mSensorTimes.size() <= fi || li == -1 || li >= mSensorTimes.at(fi).size())
        return -1;
    int g = 0;
    for (int i = 0; i < fi; i++)
        g += mSensorTimes.at(i).size();
    g += li;
    return g;
}

/**
 * @brief Compute the local indexes associated to the given global index.
 * @param gi Global index to process.
 * @return Pair of local indexes (outer and inner vectors indexes resp.).
 * @sa DataManager::localsToGlobal
 * @details
 * If \a gi is out of range, the method returns {-1,-1}.
 */
QPair<int, int> DataManager::globalToLocals(int gi) {
    if (gi < 0)
        return {-1, -1};
    const QDoubleMatrix &m = mSensorTimes;
    QPair<int, int> out{0, 0};
    int count = 0;
    for (int i = 0; i < m.size(); i++) {
        count += m.at(i).size();
        if (gi < count) {
            out.second = (i == 0) ? gi : gi - count + m.at(i).size();
            out.first = i;
            return out;
        }
    }
    return {-1, -1};
}

/**
 * @brief Get the lower boundary index of the visible interval in replay mode.
 * @param index Upper boundary index to process (generally the current frame !).
 * @return Lower boundary index of the visible interval.
 */
int DataManager::getReplayVisibleMinIndex(int index) {
    if (mVisUseDuration) {
        const QDoubleVector &times = mReplayData->times;
        double timestamp = times.at(index) - mSecVisible;
        while (index > 0 && mReplayData->times.at(--index) >= timestamp)
            ;
        return index;
    } else {
        return index - mNVisible;
    }
}

/**
 * @brief Update the style of all markers.
 * @details
 * The method updates the style of the markers currently displayed on the map (which can be either Replay or live
 * data). Both symbol size,color style and marker visibility are updated.
 */
void DataManager::setupMarkerStyles() {
    DataItem *item = mapItem();
    SimpleMarkerSymbolStyle style;
    QColor color;
    qreal msize;
    double mapValue;
    bool visible, vgps;
    int j = 0, jvis = visibleIndex();
    AttributeListModel *attrs = nullptr;

    if (mReplayMode) {
        GraphicListModel *graphics = mReplayOverlay->graphics();
        if (!item) {
            for (auto it = graphics->begin(); it != graphics->end(); it++, j++) {
                vgps = (*it)->attributes()->attributeValue("gps_valid").toBool();
                (*it)->setVisible(vgps && mMVShow && (j >= jvis) && j <= mFrameIndex);
                SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
                symbol->setColor(mMVColor);
                symbol->setSize(mMVSize);
                symbol->setStyle(mMVStyle);
            }
        } else {

            QCPRange cRange(item->colorRange().toQCPRange());
            for (auto it = graphics->begin(); it != graphics->end(); it++, j++) {
                attrs = (*it)->attributes();
                mapValue = attrs->attributeValue("value").toDouble();
                vgps = attrs->attributeValue("gps_valid").toBool();
                if (qIsNaN(mapValue)) {
                    visible = vgps && mMVShow && (j >= jvis) && (j <= mFrameIndex);
                    color = mMVColor;
                    msize = mMVSize;
                    style = mMVStyle;
                } else {
                    visible = vgps && ((j >= jvis) || item->isVisible(mapValue)) && (j <= mFrameIndex);
                    color = QColor(mColorGradient.color(mapValue, cRange));
                    msize = mCFSize;
                    style = mCFStyle;
                }
                (*it)->setVisible(visible);
                SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
                symbol->setColor(color);
                symbol->setSize(msize);
                symbol->setStyle(style);
            }
        }
        return;
    }
    // RT mode

    int no = mOverlays.size();
    if (!item) { // no item -> apply NaN style ?
        for (int i = 0; i < no; i++) {
            auto *overlay = mOverlays.at(i);
            GraphicListModel *graphics = overlay->graphics();
            for (auto it = graphics->begin(); it != graphics->end(); it++, j++) {
                vgps = (*it)->attributes()->attributeValue("gps_valid").toBool();
                (*it)->setVisible(vgps && mMVShow && (j >= jvis));
                SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
                symbol->setColor(mMVColor);
                symbol->setSize(mMVSize);
                symbol->setStyle(mMVStyle);
            }
        }
        return;
    }

    QCPRange cRange(item->colorRange().toQCPRange());
    for (int i = 0; i < no; i++) {
        auto *overlay = mOverlays.at(i);
        GraphicListModel *graphics = overlay->graphics();
        bool isLast = (i == (no - 1));
        for (auto it = graphics->begin(); it != graphics->end(); it++, j++) {
            mapValue = (*it)->attributes()->attributeValue("value").toDouble();
            vgps = (*it)->attributes()->attributeValue("gps_valid").toBool();
            // missing value
            if (qIsNaN(mapValue)) {
                color = mMVColor;
                msize = mMVSize;
                style = mMVStyle;
                visible = vgps && mMVShow && (j >= jvis);
            } else {
                visible = vgps && ((j >= jvis) || item->isVisible(mapValue));
                color = QColor(mColorGradient.color(mapValue, cRange));
                if (isLast) {
                    msize = mCFSize;
                    style = mCFStyle;
                } else {
                    msize = mPFSize;
                    style = mPFStyle;
                }
            }
            (*it)->setVisible(visible);
            SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
            symbol->setColor(color);
            symbol->setSize(msize);
            symbol->setStyle(style);
        }
    }
    return;
}

/**
 * @brief Select a scenario.
 * @param index Index of the scenario to select.
 * @details
 * The method selects the scenario at given index.
 * It updates the item to display on the registered map and plots widgets accordingly to the selected data item of
 * the scenario (see Scenario::mapItem and Scenario::plotItem). Additionally it updates the indicators to show in
 * the registered indicators widget.
 *
 * The method is a slot connected to the combobox of the registered scenario widget (if any).
 */
void DataManager::selectScenario(int index) {

    ScenarioSelector *selector = qobject_cast<ScenarioSelector *>(QObject::sender());
    if (selector) {
        for (auto p : mScenarioSelectors) {
            if (p == selector)
                continue;
            p->setCurrentIndex(index, true);
        }
    }
    Scenario *newSelection = nullptr;
    if (index >= 0 && index < mScenarios.size()) {
        newSelection = mScenarios[index];
    }

    if (mSelectedScenario) {
        ItemModel *model = mSelectedScenario->itemsModel();
        QObject::disconnect(model, &ItemModel::itemOnMapRequested, this, &DataManager::setItemOnMap);
        QObject::disconnect(model, &ItemModel::itemOnGraphRequested, this, &DataManager::setItemOnPlot);
    }

    mSelectedScenario = newSelection;

    ItemModel *model = nullptr;
    if (mSelectedScenario) {
        model = mSelectedScenario->itemsModel();
        QObject::connect(model, &ItemModel::itemOnMapRequested, this, &DataManager::setItemOnMap);
        QObject::connect(model, &ItemModel::itemOnGraphRequested, this, &DataManager::setItemOnPlot);
    }

    // set the scenario selected item on their afferent widgets.
    setItemOnMap(mapItem());
    for (int i = 0; i < N_PLOTS; i++)
        setItemOnPlot(i, plotItem(i));
    updateIndicators(mSelectedScenario);
    if (mScenarioWidget)
        mScenarioWidget->setItemModel(model);
}

/**
 * @brief Hide/Show flight items on the map.
 * @param index Index of the flight to process.
 * @param checked True to display data, False otherwise.
 * @note
 * Even if checked is true, some items may be hidden, if they lie outside the visiblity range of the current map
 * item.
 */
void DataManager::onFlightStateChanged(int index, bool checked) {
    if (index < 0 || index >= mOverlays.size())
        return;
    mOverlays[index]->setVisible(checked);
}

/**
 * @brief Enable/Disable map auto (re)center.
 * @details
 * If enabled, each time a new (valid) data point is processed, the registered map is recenter to the point (leaving
 * its zoom unchanged).
 */
void DataManager::setAutoCenter(bool enable) { mAutoCenter = enable; }

/**
 * @brief Update the registered indicators widget from the given scenario.
 * @param scenario Scenario to process.
 */
void DataManager::updateIndicators(Scenario *scenario) {
    if (mIndicatorsWidget) {
        mIndicatorsWidget->setupIndicators(
            scenario ? scenario->indicatorsModel()->indicators() : QVector<BaseIndicatorPointer>());
    }
}

/**
 * @brief Display the data associated to the given item on the map widget.
 * @param item DataItem to process.
 *
 * The method immediately returns if no map widget is registered.
 *
 * If item is null or if the manager does not store raw data for its parameter then the method assumes the map
 * widget should be cleared.
 *
 * Otherwise, the method updates the internal containers to display the data item on the map.
 */
void DataManager::setItemOnMap(DataItem *item) {
    if (!mMapWidget)
        return;

    SimpleMarkerSymbolStyle style;
    QColor color;
    qreal msize;
    bool visible, vgps;
    QCPRange cRange;

    int j = 0, jvis = visibleIndex();

    if (mReplayMode) {
        if (!mReplayData)
            return;

        GraphicListModel *graphics = mReplayOverlay->graphics();
        GraphContainerPointer gContainer = mReplayMapGraphData;
        const QMap<AParam, QDoubleVector> &data = mReplayData->data;
        // if no item are to be set on the map OR it cannot be found: remove overlay and graph.
        QMap<AParam, QDoubleVector>::const_iterator itemIterator = item ? data.find(item->parameter()) : data.cend();
        if (itemIterator == data.cend()) {
            mMapWidget->updateFromItem(nullptr, QVector<GraphicsOverlay *>(), GraphContainerVector());
            return;
        }
        // Modify map overlay and graphic container.
        Units::Unit raw_unit = itemIterator.key().units;
        Units::Unit out_unit = item->parameter().units;
        cRange = item->colorRange().toQCPRange();
        const QDoubleVector &values = itemIterator.value();
        int nv = values.size();
        int ng = graphics->size();

        Q_ASSERT_X(ng == nv, "DataManager::setItemOnMap", "Overlays and replay data sizes are not consistent");

        for (auto it = graphics->begin(); it != graphics->end(); it++, j++) {

            double value = Units::convert(values.at(j), raw_unit, out_unit);
            (gContainer->begin() + j)->value = value;              // update graph value.
            (*it)->attributes()->replaceAttribute("value", value); // update map value.
            vgps = (*it)->attributes()->attributeValue("gps_valid").toBool();
            if (qIsNaN(value)) {
                visible = vgps && mMVShow && (j >= jvis) && (j <= mFrameIndex);
                color = mMVColor;
                msize = mMVSize;
                style = mMVStyle;
            } else {
                visible = vgps && ((j >= jvis) || item->isVisible(value)) && (j <= mFrameIndex);
                color = QColor(mColorGradient.color(value, cRange));
                msize = mCFSize;
                style = mCFStyle;
            }
            (*it)->setVisible(visible);
            SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
            symbol->setColor(color);
            symbol->setSize(msize);
            symbol->setStyle(style);
        }

        mMapWidget->updateFromItem(item, {mReplayOverlay}, {gContainer});
        bool tvis = false;
        double key = 0.0;
        if (mReplayData->times.size()) {
            tvis = true;
            key = mReplayData->times.at(mFrameIndex);
        }
        mMapWidget->plotWidget()->setReplayTracer(tvis, &key);
        updateValueRange();
        return;
    }

    // RT mode

    // if no item are to be set on the map OR it cannot be found: simply remove the overlays.
    QMap<AParam, QDoubleMatrix>::iterator itemIterator = item ? mLiveData.find(item->parameter()) : mLiveData.end();
    if (itemIterator == mLiveData.end()) {
        // no item found: assume the map widget should be cleared: clear the map widget !
        mMapWidget->updateFromItem(nullptr, mOverlays, mMapGraphData);
        return;
    }
    // Modify map overlays and graphic container.
    Units::Unit raw_unit = itemIterator.key().units;
    Units::Unit out_unit = item->parameter().units;
    const QDoubleMatrix &itemValues = itemIterator.value();
    cRange = item->colorRange().toQCPRange();
    int no = mOverlays.size(), nvf = itemValues.size();
    Q_ASSERT_X(no == nvf, "DataManager::setItemOnMap", "Overlays and flight sizes are not consistent");

    for (int i = 0; i < no; i++) {
        bool isLast = (i == no - 1);
        auto *overlay = mOverlays.at(i);
        const GraphContainerPointer &gContainer = mMapGraphData.at(i);
        const QDoubleVector &values = itemValues.at(i);
        GraphicListModel *graphics = overlay->graphics();
        int nv = values.size();
        int ng = graphics->size();
        Q_ASSERT_X(nv == ng, "DataManager::setItemOnMap", "graphics and flight data sizes are not consistent");
        int cj = 0;
        for (auto it = graphics->begin(); it != graphics->end(); it++, cj++, j++) {
            double value = Units::convert(values.at(j), raw_unit, out_unit);
            (gContainer->begin() + cj)->value = value;             // update graph value.
            (*it)->attributes()->replaceAttribute("value", value); // update map value.
            vgps = (*it)->attributes()->attributeValue("gps_valid").toBool();
            // missing val
            if (qIsNaN(value)) {
                color = mMVColor;
                msize = mMVSize;
                style = mMVStyle;
                visible = vgps && mMVShow && (j >= jvis);
            } else {
                visible = vgps && ((j >= jvis) || item->isVisible(value));
                color = QColor(mColorGradient.color(value, cRange));
                if (isLast) {
                    msize = mCFSize;
                    style = mCFStyle;
                } else {
                    msize = mPFSize;
                    style = mPFStyle;
                }
            }
            (*it)->setVisible(visible);
            SimpleMarkerSymbol *symbol = static_cast<SimpleMarkerSymbol *>((*it)->symbol());
            symbol->setColor(color);
            symbol->setSize(msize);
            symbol->setStyle(style);
        }
    }
    // update map widget
    mMapWidget->updateFromItem(item, mOverlays, mMapGraphData);
}

/**
 * @brief Display the data associated to the given item on the plot widget.
 * @param index Index of the plot widget to process.
 * @param item DataItem to process.
 *
 * The method immediately returns if no plot widget is registered for the given index.
 *
 * If item is null or if the manager does not store raw data for its parameter then the method assumes the plot
 * widget should be cleared.
 *
 * Otherwise, the method updates the internal containers to display the data item on the plot widget.
 */
void DataManager::setItemOnPlot(int index, DataItem *item) {
    if (index < 0 || index >= N_PLOTS)
        return;
    RtPlotWidget *plotWidget = mPlotWidgets[index];
    if (!plotWidget)
        return;

    if (mReplayMode) {
        if (!mReplayData)
            return;
        const QMap<AParam, QDoubleVector> &data = mReplayData->data;
        // if no item are to be set on the map OR it cannot be found: simply remove the overlays.
        QMap<AParam, QDoubleVector>::const_iterator itemIterator = item ? data.find(item->parameter()) : data.cend();
        if (itemIterator == data.cend()) {
            // no item found: assume the map widget should be cleared:
            // then we call updateFromItem with a null pointer.
            plotWidget->updateFromItem(nullptr, GraphContainerVector());
            plotWidget->setReplayTracer(false);
            return;
        }

        Units::Unit raw_unit = itemIterator.key().units;
        Units::Unit out_unit = item->parameter().units;
        const QDoubleVector &values = itemIterator.value();
        GraphContainerPointer gContainer = mReplayPlotsGraphData[index];
        int ng = gContainer->size(), nv = values.size();
        Q_ASSERT_X(ng == nv, "DataManager::setItemOnPlot", "graphics and flight data sizes are not consistent");
        int j = 0;
        for (auto it = gContainer->begin(); it != gContainer->end(); it++, j++) {
            it->value = Units::convert(values.at(j), raw_unit, out_unit);
        }
        plotWidget->updateFromItem(item, {gContainer});
        bool tvis = false;
        double key = 0.0;
        if (mReplayData->times.size()) {
            tvis = true;
            key = mReplayData->times.at(mFrameIndex);
        }
        plotWidget->setReplayTracer(tvis, &key);
        updateValueRange();
        return;
    }

    // if no item are to be set on the map OR it cannot be found: simply remove the overlays.
    // auto itemIterator = item ? mLiveData.find(item->parameter()) : mLiveData.cend();
    const QMap<AParam, QDoubleMatrix>::const_iterator itemIterator =
        item ? mLiveData.find(item->parameter()) : mLiveData.cend();
    if (itemIterator == mLiveData.cend()) {
        // no item found: assume the map widget should be cleared:
        // then we call updateFromItem with a null pointer.
        plotWidget->updateFromItem(nullptr, GraphContainerVector());
        return;
    }
    Units::Unit raw_unit = itemIterator.key().units;
    Units::Unit out_unit = item->parameter().units;
    const QDoubleMatrix &itemValues = itemIterator.value();
    const GraphContainerVector &gContainers = mPlotsGraphData[index];
    int nc = gContainers.size();
    // rebuild the container.
    for (int i = 0; i < nc; i++) {
        GraphContainerPointer gContainer = gContainers[i];
        const QDoubleVector &values = itemValues[i];
        int ng = gContainer->size(), nv = values.size();
        Q_ASSERT_X(ng == nv, "DataManager::setItemOnPlot", "graphics and flight data sizes are not consistent");
        int j = 0;
        for (auto it = gContainer->begin(); it != gContainer->end(); it++, j++) {
            it->value = Units::convert(values.at(j), raw_unit, out_unit);
        }
    }
    plotWidget->updateFromItem(item, gContainers);
}

/**
 * @brief Build and display the context menu for the plot widget at given index.
 * @param pos Position of the mouse within the plot widget.
 * @details
 * The method builds a menu with the possible user interaction on the plot.
 * If the plot widget allow parameter selection then an extra menu is prepended to the menu with
 * the list of available parameters of the selected scenario.
 *
 * @sa BaseRtPlotWidget::addInteractionsActions
 */
void DataManager::onPlotWidgetContextMenuRequested(const QPoint &pos) {
    BasePlot *plot = qobject_cast<BasePlot *>(QObject::sender());
    RtPlotWidget *plotWidget = qobject_cast<RtPlotWidget *>(plot->parentWidget());
    if (!(plot && plotWidget))
        return;
    int index = -1;
    for (int i = 0; i < N_PLOTS; i++) {
        if (mPlotWidgets[i] == plotWidget) {
            index = i;
            break;
        }
    }
    QMenu *menu = new QMenu(plotWidget);
    menu->setAttribute(Qt::WA_DeleteOnClose);
    if (plotWidget->parameterSelectionEnabled()) {
        QMenu *tmp = nullptr;
        if (mSelectedScenario)
            tmp = mSelectedScenario->itemsModel()->plotContextMenu(index, menu);
        if (tmp)
            menu->addMenu(tmp);
    }
    plotWidget->addInteractionsActions(menu);
    menu->popup(plotWidget->plot()->mapToGlobal(pos));
}

/**
 * @brief Build and display the context menu for the registered map widget.
 * @param pos Position of the mouse within the plot widget.
 * @param isMap True if the signal was emited from the internal mapview of the registered mapwidget.
 * @details
 * The method builds a menu with the possible user interactions on the map widget. The menu is populated with
 * the actions of the plot widget or the map view based on the value isMap argument.
 */
void DataManager::onMapContextMenuRequested(const QPoint &pos, bool isMap) {
    QMenu *menu = new QMenu(mMapWidget);
    menu->setAttribute(Qt::WA_DeleteOnClose);
    if (mMapWidget->parameterSelectionEnabled()) {
        QMenu *tmp = mSelectedScenario ? mSelectedScenario->itemsModel()->mapContextMenu(menu) : nullptr;
        if (tmp)
            menu->addMenu(tmp);
    }
    if (isMap) {
        mMapWidget->addMapInteractionsActions(menu);
    } else {
        mMapWidget->plotWidget()->addInteractionsActions(menu);
    }
    menu->popup(mMapWidget->mapToGlobal(pos));
}

/// @brief Get the currently selected scenario.
Scenario *DataManager::selectedScenario() const { return mSelectedScenario; }

/// @brief Get the list of registered scenarios.
QVector<Scenario *> DataManager::scenarios() const { return mScenarios; }

/// @brief Get the data item associated to the map from the selected scenario.
DataItem *DataManager::mapItem() const {
    return (mSelectedScenario) ? mSelectedScenario->itemsModel()->mapItem() : nullptr;
}

/// @brief Get the data item associated to the plot at the given index from the selected scenario.
DataItem *DataManager::plotItem(int index) const {
    return (mSelectedScenario) ? mSelectedScenario->itemsModel()->plotItem(index) : nullptr;
}

/// @brief Get the flight model.
FlightModel *DataManager::flightModel() const { return mFlightModel; }

/// @brief Delete flight from the action of the registered flight widget.
void DataManager::deleteFlightFromAction() {
    if (mFlightWidget) {
        int index = mFlightWidget->addFlightAction->data().toInt();
        if (index != -1)
            deleteFlight(index);
    }
}

/**
 * @brief Delete the flight at given index.
 * @param index Index of the flight to delete.
 * @return True on success, false otherwise.
 * @details
 * The method attempts to remove the flight at given index (i.e. free all resources of such flight).
 * The current flight (at index DataManager::flightCount-1) cannot be removed using this method.
 * @note
 * The method fails if index is out of range (including index at DataManager::flightCount-1) or if the manager is
 * currently in Replay mode.
 */
bool DataManager::deleteFlight(int index) {

    if (index < 0 || index >= (flightCount() - 1) || mReplayMode)
        return false;

    // 1. delete raw data
    mSensorTimes.remove(index);
    mPcMonTimes.remove(index);
    mPositions.remove(index);
    mValidGPS.remove(index);

    for (QDoubleMatrix &data : mLiveData)
        data.remove(index);
    for (DoublePairVector &data : mExtremas)
        data.remove(index);

    // remove overlay and graph containers.
    GraphicsOverlay *overlay = mOverlays[index];
    mOverlays.remove(index);
    overlay->deleteLater();
    mMapGraphData.remove(index);
    for (int i = 0; i < N_PLOTS; i++)
        mPlotsGraphData[i].remove(index);

    // 2. delete flight in model.
    mFlightModel->removeFlight(index);

    // must reset plot and map
    DataItem *item = mapItem();
    if (item)
        setItemOnMap(item);
    for (int i = 0; i < N_PLOTS; i++) {
        item = plotItem(i);
        if (item)
            setItemOnPlot(i, item);
    }
    return true;
}

/// @brief Get the number of flight stored in the manager.
int DataManager::flightCount() const { return mSensorTimes.size(); }

/// @brief Get the length of the last flight.
int DataManager::flightLength() const { return flightLength(mSensorTimes.size() - 1); }

/**
 * @brief Get the length of the flight at given index.
 * @param index Index of the flight to process.
 * @return Number of points stored in the manager for the given flight index.
 * @details
 * Flight indexes varies between zero (inclusive) and DataManager::flightCount (exclusive). If index is out of
 * range, 0 is returned.
 */
int DataManager::flightLength(int index) const {
    return (index < 0 || index >= mSensorTimes.size()) ? 0 : mSensorTimes[index].size();
}

/// @brief Get the datetime of the last flight creation.
QDateTime DataManager::lastStartTime() const { return mLastStartTime; }

/// @brief Get the parameter data extrema over all flights of the given item.
QPair<double, double> DataManager::dataExtrema(DataItem *item) const {
    double omin = qQNaN(), omax = qQNaN();
    auto it = mExtremas.find(item->parameter());
    if (it == mExtremas.cend())
        return qMakePair(omin, omax);
    const AParam dataParam = it.key();
    const QVector<QPair<double, double>> &extremas = it.value();
    double v;
    for (auto p : extremas) {
        v = p.first;
        if (!qIsNaN(v) && (qIsNaN(omin) || omin > v))
            omin = v;
        v = p.second;
        if (!qIsNaN(v) && (qIsNaN(omax) || omax < v))
            omax = v;
    }
    return qMakePair(omin, omax);
}

/**
 * @brief Get the extrema of the parameter value for the given flight index.
 * @param flightIndex Flight index to process.
 * @return Pair of double value with min/max values. NaN are set if no extrema have been computed.
 * @details
 * If flightIndex is not given (default) or set to -1, the method returns the extrema for the last, i.e. current,
 * flight.
 *
 * If the index is out of range (with the exception of the special value -1), a pair of NaN is returned.
 */
QPair<double, double> DataManager::flightExtrema(DataItem *item, int flightIndex) const {
    int fc = flightCount();
    if (!item || !fc || flightIndex >= fc || flightIndex < -1)
        return qMakePair(qQNaN(), qQNaN());

    auto it = mExtremas.find(item->parameter());
    if (it == mExtremas.cend())
        return qMakePair(qQNaN(), qQNaN());

    const AParam dataParam = it.key();
    const QVector<QPair<double, double>> &extremas = it.value();
    QPair<double, double> origPair(flightIndex == -1 ? extremas.last() : extremas[flightIndex]);
    double mi = Units::convert(origPair.first, dataParam.units, item->parameter().units);
    double ma = Units::convert(origPair.second, dataParam.units, item->parameter().units);
    return qMakePair(mi, ma);
}

/**
 * @brief Get the extrema of the data item for the given flight indexes.
 * @param item DataItem to process.
 * @param flightIndexes Flight indexes to process.
 * @return Vector of pair with the extrema values of the data item for the given flights.
 * @details
 * If item is given as null pointer or if its parameter is not found in the data manager, an empty vector is
 * returned.
 *
 * If flightIndexes is not given, the method returns the extrema for all registered flights.
 * @note
 * All out of range indexes are quietly ignored.
 */
QVector<QPair<double, double>> DataManager::flightsExtremas(DataItem *item, QVector<int> flightIndexes) const {
    QVector<QPair<double, double>> out;
    if (!item) {
        return out;
    }
    auto it = mExtremas.find(item->parameter());
    if (it == mExtremas.cend()) {
        return out;
    }
    const AParam dataParam = it.key();
    const QVector<QPair<double, double>> &extremas = it.value();
    int s = extremas.size();
    if (flightIndexes.isEmpty()) {
        out.reserve(s);
        for (int i = 0; i < s; i++) {
            double mi = Units::convert(extremas[i].first, dataParam.units, item->parameter().units);
            double ma = Units::convert(extremas[i].second, dataParam.units, item->parameter().units);
            out.push_back(qMakePair(mi, ma));
        }
    } else {
        for (int i : flightIndexes) {
            if (i >= 0 && i < s) {
                double mi = Units::convert(extremas[i].first, dataParam.units, item->parameter().units);
                double ma = Units::convert(extremas[i].second, dataParam.units, item->parameter().units);
                out.push_back(qMakePair(mi, ma));
            }
        }
    }
    return out;
}

/// @brief Update map items style from application settings.
void DataManager::updateFromSettings(bool applyMarkerStyle) {
    QSettings settings;
    settings.beginGroup("monitoring");
    setNewFlightDelay(settings.value("nf_delay", 0).toInt() * 60); // delay settings is in minutes, api expects seconds.
    setNewFlightOnStrap(settings.value("nf_onstrap", true).toBool());
    mBypassGPS = settings.value("proc_nogps", true).toBool();
    mMinSat = settings.value("min_nsat", 4).toInt();
    mSecVisible = settings.value("min_sec_vis", 30).toInt();
    mNVisible = settings.value("min_pts_vis", 30).toInt();
    settings.endGroup();

    settings.beginGroup("graph");
    useNowAsTimeReference(settings.value("now_as_ref", true).toBool());
    setAutoScrollingInterval(settings.value("asinterval", 200).toInt());
    settings.endGroup();

    settings.beginGroup("map");
    mMVShow = !settings.value("m_hidemv", false).toBool();
    mMVColor = QColor(settings.value("m_mvcolor", "#000000").toString());
    mMVStyle = static_cast<SimpleMarkerSymbolStyle>(settings.value("m_mvmstyle", 0).toInt());
    mMVSize = qMax(4.0, settings.value("m_mvsize", 10.0).toDouble());
    mCFStyle = static_cast<SimpleMarkerSymbolStyle>(settings.value("m_cmstyle", 0).toInt());
    mCFSize = qMax(4.0, settings.value("m_cmsize", 10.0).toDouble());
    mPFStyle = static_cast<SimpleMarkerSymbolStyle>(settings.value("m_omstyle", 3).toInt());
    mPFSize = qMax(4.0, settings.value("m_omsize", 10.0).toDouble());
    settings.endGroup();

    // update settings for registered widgets.
    for (BaseRtPlotWidget *pw : mPlotWidgets) {
        if (pw)
            pw->updateFromSettings();
    }
    if (mMapWidget)
        mMapWidget->updateFromSettings();

    if (applyMarkerStyle)
        setupMarkerStyles();
}

/**
 * @brief Get the total number of points stored in the manager.
 * @param flightSizes Optional pointer to a vector of int with in output the size of each flight.
 */
int DataManager::size(QVector<int> *flightSizes) const {
    QVector<int> tmp;
    int count = 0;
    for (const QDoubleVector &v : mSensorTimes) {
        int vs = v.size();
        tmp.push_back(vs);
        count += vs;
    }
    if (flightSizes)
        *flightSizes = tmp;
    return count;
}

/// @brief Get the graphic that is stored at interval points from the end of the manager.
Graphic *DataManager::getGraphicFromEnd(int npts) {
    int no = mOverlays.size();
    npts += 1;
    while (no > 0) {
        no--;
        auto *model = mOverlays.at(no)->graphics();
        int on = model->size();
        if (npts > on) {
            npts -= on;
            continue;
        } else {
            return model->at(on - npts);
        }
    }
    return nullptr;
}

void DataManager::setReplayMode(bool enable) {
    for (auto p : mScenarioSelectors)
        p->setInteractionMode(!enable);

    if (mReplayMode == enable)
        return;

    // stop processing updates while we are modifying the data.
    setupWidgetsUpdatesEnabled(false);

    stopReplay();
    mAutoScrollTimer.stop();

    // save the map item and plot item.
    DataItem *imap = mapItem();
    DataItem *iplots[N_PLOTS];
    for (int i = 0; i < N_PLOTS; i++)
        iplots[i] = plotItem(i);

    // reset map and plots
    setItemOnMap(nullptr);
    for (int i = 0; i < N_PLOTS; i++)
        setItemOnPlot(i, nullptr);

    // cleanup ALL containers (live and Replay !)
    {
        mMapGraphData.clear();
        for (int i = 0; i < N_PLOTS; i++)
            mPlotsGraphData[i].clear();
        clearOverlays(mOverlays);

        QVector<GraphicsOverlay *> tmp{mReplayOverlay};
        clearOverlays(tmp);
        mReplayOverlay = nullptr;
        mReplayMapGraphData.clear();
        for (int i = 0; i < N_PLOTS; i++)
            mReplayPlotsGraphData[i].clear();

        if (mReplayData) {
            delete mReplayData;
            mReplayData = nullptr;
        }
    }

    if (enable) {
        // Setup replay containers: EMPTY !
        mReplayOverlay = new GraphicsOverlay(this);
        mReplayMapGraphData = GraphContainerPointer::create();
        for (int i = 0; i < N_PLOTS; i++) {
            mReplayPlotsGraphData[i] = GraphContainerPointer::create();
        }
        if (mMapWidget) {
            mMapWidget->addOverlay(mReplayOverlay);
            mMapWidget->addGraph(mReplayMapGraphData);
        }
        for (int i = 0; i < N_PLOTS; i++) {
            auto w = mPlotWidgets[i];
            if (w)
                w->setGraphs({mReplayPlotsGraphData[i]}, true);
        }
    } else {
        // setup live data containers
        // QVector<int> fsizes;
        int nf = mPositions.size();
        QVariantMap attrs = {{"value", qQNaN()}, {"gps_valid", true}};
        bool isWGS84 = mMapSpatialReference == gWGS84;
        // for each flight: rebuild overlay and QCP container.
        for (int i = 0; i < nf; i++) {
            const QDoubleVector &tt = mUsePCTime ? mPcMonTimes.at(i) : mSensorTimes.at(i);
            int fs = tt.size();
            // Build vector of nan for the plots.
            QVector<QCPGraphData> nanData;
            nanData.reserve(fs);
            for (double t : tt)
                nanData.push_back(QCPGraphData(t, qQNaN()));
            GraphContainerPointer ptr = GraphContainerPointer::create();
            ptr->set(nanData, true);
            mMapGraphData.push_back(ptr);
            for (int j = 0; j < N_PLOTS; j++) {
                GraphContainerPointer ptr = GraphContainerPointer::create();
                ptr.data()->set(nanData, true);
                mPlotsGraphData[j].push_back(ptr);
            }
            // setup overlay data
            const PointVector &fpos = mPositions.at(i);
            const BoolVector &vgps = mValidGPS.at(i);
            GraphicsOverlay *overlay = new GraphicsOverlay(this);
            auto *graphics = overlay->graphics();
            for (int j = 0; j < fpos.size(); ++j) {
                const Point &p = fpos.at(j);
                attrs["gps_valid"] = vgps.at(j);
                Graphic *graphic =
                    isWGS84 ? (new Graphic(p, attrs, overlay))
                            : (new Graphic(GeometryEngine::project(p, mMapSpatialReference), attrs, overlay));
                SimpleMarkerSymbol *sms = new SimpleMarkerSymbol(mMVStyle, mMVColor, mMVSize, graphic);
                sms->setOutline(mOutline);
                graphic->setSymbol(sms);
                graphic->setVisible(false);
                graphics->append(graphic);
            }
            mOverlays.push_back(overlay);
        }
    }
    mReplayMode = enable;

    if (mMapWidget)
        mMapWidget->plotWidget()->setReplayMode(mReplayMode);
    for (int i = 0; i < N_PLOTS; i++) {
        auto pw = mPlotWidgets[i];
        if (pw)
            pw->setReplayMode(mReplayMode);
    }

    // stop/start scrolling.
    bool isLive = !mReplayMode;
    enableTimeScrolling(isLive);
    if (mFlightWidget)
        mFlightWidget->setEnabled(isLive);

    // update mapper (with NaN values)
    mItemMapper->registerValues(QVector<AParam>{});
    for (Scenario *scenario : mScenarios) {
        scenario->registerValues(mItemMapper);
    }

    setItemOnMap(imap);
    for (int i = 0; i < N_PLOTS; i++)
        setItemOnPlot(i, iplots[i]);
    setupWidgetsUpdatesEnabled(true);
    updateWidgets();
}

/**
 * @brief Set new replay data.
 * @param data Data to set.
 * @details
 * The method updates the current replay data. It takes ownership of the pointer, that should not be deleted.
 */
void DataManager::setReplayData(ReplayData *data) {
    QElapsedTimer timer;
    setupWidgetsUpdatesEnabled(false);
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    // save the map item and plot item.
    DataItem *imap = mapItem();
    DataItem *iplots[N_PLOTS];
    for (int i = 0; i < N_PLOTS; i++)
        iplots[i] = plotItem(i);

    stopReplay();
    // reset map and plots
    setItemOnMap(nullptr);
    for (int i = 0; i < N_PLOTS; i++)
        setItemOnPlot(i, nullptr);

    if (mReplayData)
        delete mReplayData;
    mReplayData = data;

    // resize the graphics containers.
    timer.start();
    resizeReplayContainers();
    qDebug() << "resizeReplayContainers:" << timer.restart() << "msecs";
    if (mMediaControlWidget)
        mMediaControlWidget->setupControls(mReplayData->frameCount);
    qDebug() << "setup controls:" << timer.restart() << "msecs";
    setItemOnMap(imap);
    for (int i = 0; i < N_PLOTS; i++) {
        setItemOnPlot(i, iplots[i]);
    }
    qDebug() << "set items:" << timer.restart() << "msecs";
    forceTimeRange();
    updateValueRange();
    goToStart();
    QApplication::restoreOverrideCursor();
    setupWidgetsUpdatesEnabled(true);
    updateWidgets();
}

/// @brief Update Y-axis of each registered plots to fit all data.
void DataManager::updateValueRange() {
    if (mMapWidget)
        mMapWidget->plotWidget()->rescaleYAxis();
    for (int i = 0; i < N_PLOTS; i++) {
        RtPlotWidget *pw = mPlotWidgets[i];
        if (pw)
            pw->rescaleYAxis();
    }
}

/// @brief Update W-axis of each registered plots to fit the replay mode time range.
void DataManager::forceTimeRange() {
    if (!mReplayData)
        return;
    const QDoubleVector &times = mReplayData->times;
    int ns = times.size();
    if (!ns)
        return;
    QCPRange range = ns > 1 ? QCPRange(times.first(), times.last()) : QCPRange(times.first() - 10, times.first() + 10);
    if (mMapWidget)
        mMapWidget->plotWidget()->forceTimeRange(range);
    for (int i = 0; i < N_PLOTS; i++) {
        RtPlotWidget *pw = mPlotWidgets[i];
        if (pw)
            pw->forceTimeRange(range);
    }
}

/**
 * @brief Resize the overlays and QCP graph data containers.
 * @attention
 * The method assumes DataManager::mReplayPositions and DataManager::mReplayTime
 */
void DataManager::resizeReplayContainers() {

    if (!mReplayData)
        return;
    const PointVector &positions = mReplayData->positions;
    const BoolVector &vgps = mReplayData->gpsValidity;
    const QDoubleVector &times = mReplayData->times;
    int npts = positions.size();
    Q_ASSERT_X(npts == times.size(),
        "DataManager::resizeReplayContainers",
        "Inconsistent sizes of Replay positions and time vectors");
    // setup the overlay graphics !
    auto *graphics = mReplayOverlay->graphics();
    int ng = graphics->size();

    if (ng > npts) {
        QList<Graphic *> keep;
        QVector<Graphic *> trash;
        trash.reserve(ng - npts);
        keep.reserve(npts);
        int i = 0;
        for (Graphic *g : *graphics) {
            if (i < npts)
                keep.push_back(g);
            else
                trash.push_back(g);
            i++;
        }
        graphics->clear();
        qDeleteAll(trash);
        graphics->append(keep);
    }

    QVariantMap attrs{{"value", qQNaN()}, {"gps_valid", false}};
    for (int i = 0; i < npts; i++) {
        Graphic *g;
        SimpleMarkerSymbol *sms;
        if (i >= ng) {
            g = new Graphic(attrs, mReplayOverlay);
            sms = new SimpleMarkerSymbol(mMVStyle, mMVColor, mMVSize, g);
            sms->setOutline(mOutline);
            g->setSymbol(sms);
            graphics->append(g);
        }
        g = graphics->at(i);
        attrs = g->attributes()->attributesMap();
        attrs["gps_valid"] = vgps.at(i);
        g->attributes()->setAttributesMap(attrs);
        g->setGeometry(positions.at(i));
        g->setVisible(false);
        sms = static_cast<SimpleMarkerSymbol *>(g->symbol());
        sms->setColor(mMVColor);
        sms->setSize(mMVSize);
        sms->setStyle(mMVStyle);
    }

    // .. and QCP data container.
    QVector<QCPGraphData> nanData;
    nanData.reserve(npts);
    static const double nan = qQNaN();
    for (double t : times)
        nanData.push_back(QCPGraphData(t, nan));
    mReplayMapGraphData->set(nanData, true);
    for (int i = 0; i < N_PLOTS; i++) {
        mReplayPlotsGraphData[i]->set(nanData, true);
    }
}

/**
 * @brief Set the framerate for the replay mode.
 * @param frameRate New framerate to set. It is internally bounded between 1 and 50.
 */
void DataManager::setReplayFrameRate(int frameRate) {
    frameRate = 1000 / qBound(1, frameRate, 50);
    if (mReplayFrameRate != frameRate) {
        mReplayFrameRate = frameRate;
        if (mReplayTimer.isActive())
            mReplayTimer.setInterval(mReplayFrameRate);
    }
}

/// @brief Go to next Replay frame.
void DataManager::nextFrame() {
    if (!mReplayData)
        return;
    // compute new index.
    int index = mFrameIndex + mFrameStep;

    // not yet at the end: go to new frame.
    if (index < mReplayData->frameCount) {
        goToFrame(index);
        return;
    }

    if (mFrameIndex != mReplayData->frameCount - 1) {
        // not exactly at the end: go to the end.
        goToFrame(mReplayData->frameCount - 1);
    } else {
        // exactly at the end: check for auto replay.
        if (mAutoRepeat) {
            goToStart();
        } else {
            stopReplay();
        }
    }
    return;
}

/// @brief Go to previous Replay frame.
void DataManager::prevFrame() {
    int index = mFrameIndex - mFrameStep;
    if (index < 0) {
        index = 0;
        // we have reached the beginning: what should we do ???
        qDebug() << "we have reached the beginning";
    };
    goToFrame(index);
    return;
}

/// @brief Go to the frame at given index.
void DataManager::goToFrame(int index) {
    if (index == mFrameIndex || !mReplayData)
        return;
    // update visibility of the Replay overlay markers !
    auto *m = mReplayOverlay->graphics();
    DataItem *item = mapItem();
    if (item) {
        QCPRange cRange(item->colorRange().toQCPRange());
        QColor color;
        Esri::ArcGISRuntime::Graphic *gr = nullptr;
        double mapValue = qQNaN();
        bool gpsValid = false;
        if (index < mFrameIndex) {
            // we're going back: all point ABOVE new index should be hidden.
            for (int i = mFrameIndex; i > index; --i)
                m->at(i)->setVisible(false);
            // compute the number of visible point from index using visDuration.

            int ivis = getReplayVisibleMinIndex(index);
            // ensure the new current frame and its last visible are visible and updates the color !
            for (int i = index; i > qMax(0, ivis); --i) {
                gr = m->at(i);
                mapValue = gr->attributes()->attributeValue("value").toDouble();
                gpsValid = gr->attributes()->attributeValue("gps_valid").toBool();
                gr->setVisible(gpsValid); // && true);
                color = qIsNaN(mapValue) ? mMVColor : QColor(mColorGradient.color(mapValue, cRange));
                static_cast<SimpleMarkerSymbol *>(gr->symbol())->setColor(color);
            }
        } else {
            int ivis = getReplayVisibleMinIndex(qMin(mReplayData->frameCount - 1, index + 1));
            int imin = qMax(0, getReplayVisibleMinIndex(mFrameIndex)), imax = index + 1;
            bool visible, isnan;
            for (int i = imin; i < imax; ++i) {
                gr = m->at(i);
                mapValue = gr->attributes()->attributeValue("value").toDouble();
                visible = gr->attributes()->attributeValue("gps_valid").toBool();
                if (visible) {
                    isnan = qIsNaN(mapValue);
                    visible = isnan ? (mMVShow && (i >= ivis)) : ((i >= ivis) || item->isVisible(mapValue));
                }
                gr->setVisible(visible);
                if (visible) {
                    color = isnan ? mMVColor : QColor(mColorGradient.color(mapValue, cRange));
                    static_cast<SimpleMarkerSymbol *>(gr->symbol())->setColor(color);
                }
            }
        }
    }

    double timeKey = mReplayData->times.at(index);
    if (mMapWidget && mapItem())
        mMapWidget->plotWidget()->setReplayKey(timeKey);
    for (int i = 0; i < N_PLOTS; i++) {
        auto pw = mPlotWidgets[i];
        if (pw)
            pw->setReplayKey(timeKey);
    }

    QVector<AParam> prms = replayDataAt(*mReplayData, index); // mReplayData->dataAt(index);
    mItemMapper->registerValues(prms);
    for (Scenario *scenario : mScenarios) {
        scenario->registerValues(mItemMapper);
    }

    updateWidgets();
    emit mapperUpdated();

    // update frame index
    mFrameIndex = qBound(0, index, mReplayData->frameCount - 1);
    // mFrameIndex = index;
    if (mMediaControlWidget)
        mMediaControlWidget->setFrameIndex(mFrameIndex);

    return;
}
/**
 * @brief Stop the Replay.
 * @details
 * The method stops the Replay timer and reset the frame index to the first frame of the Replay.
 */
void DataManager::stopReplay() {
    mReplayTimer.stop();
    mReplayState = ReplayState::STOPPED;
    goToStart();
    emit ReplayStateChanged(mReplayState);
}

/**
 * @brief Pause the replay.
 */
void DataManager::pauseReplay() {
    mReplayTimer.stop();
    mReplayState = ReplayState::PAUSED;
    emit ReplayStateChanged(mReplayState);
}
/**
 * @brief Resume the Replay.
 * @details
 * The method starts the timer, only if it is not active.
 */
void DataManager::playReplay() {
    if (!mReplayTimer.isActive()) {
        mReplayTimer.start();
        mReplayState = ReplayState::PLAYING;
        emit ReplayStateChanged(mReplayState);
    }
}

/// @brief Go to the first frame of the Replay (data).
void DataManager::goToStart() { goToFrame(0); }

/// @brief Go to the last frame of the Replay (data).
void DataManager::goToEnd() {
    if (mReplayData)
        goToFrame(mReplayData->frameCount - 1);
}

/// @brief Enable/Disable auto repeat Replay mode.
void DataManager::setAutoRepeat(bool enable) { mAutoRepeat = enable; }

/// @brief Update Replay playing rate.
void DataManager::changeFrameStep(int step) { mFrameStep = qMax(1, step); }

/**
 * @brief Delete given overlays.
 * @param overlays Vector of overlays to delete.
 * @details
 * The method deletes all the \a overlays. For each overlay, the graphics stored in the its GraphicsModel are also
 * deleted.
 */
static void clearOverlays(QVector<GraphicsOverlay *> &overlays) {
    for (auto *o : overlays) {
        if (!o)
            continue;
        auto *graphics = o->graphics();
        int ng = graphics->size();
        QVector<Graphic *> items(ng);
        std::copy(graphics->begin(), graphics->end(), items.begin());
        graphics->clear();
        qDeleteAll(items);
        items.clear();
        delete o;
    }
    overlays.clear();
}

/**
 * @brief Get indexes of the given timestamp in the given time line.
 * @param timestamp Absolute timestamp to search.
 * @param times Time line to process.
 * @param global Optional pointer to an int with in output the global index corresponding to the couple of indexes.
 * @attention
 * The method computes the flight and local indexes of the point that is almost at \a timestamp.
 * The search is performed backward from the end of the time line to the beginning.
 */
static QPair<int, int> getNearestPoint(double timestamp, const QDoubleMatrix &times, int *global) {
    if (global)
        *global = -1;
    QPair<int, int> indexes{-1, -1};
    if (times.isEmpty())
        return indexes;
    auto fit = std::find_if(times.rbegin(), times.rend(), [timestamp](const QDoubleVector &v) {
        int vs = v.size();
        if (!vs)
            return false;
        else if (vs == 1)
            return timestamp > v.first();
        else
            return timestamp > v.first() && timestamp <= v.last();
    });
    if (fit != times.rend())
        indexes.first = std::distance(times.begin(), fit.base()) - 1;
    if (indexes.first != -1) {
        const QDoubleVector &tmp = times.at(indexes.first);
        auto bit = std::find_if(tmp.rbegin(), tmp.rend(), [timestamp](const double &v) { return v < timestamp; });
        if (bit != tmp.rend())
            indexes.second = std::distance(tmp.begin(), bit.base()) - 1;
    }
    if (global) {
        int g = 0;
        for (int i = 0; i < indexes.first; i++)
            g += times.at(i).size();
        *global = g + indexes.second;
    }
    return indexes;
}
