// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "nameeditdelegate.hpp"

#include <QApplication>

#include <QEvent>
#include <QFocusEvent>
#include <QKeyEvent>

#include <QLineEdit>
#include <QMessageBox>

NameEditDelegate::NameEditDelegate(QObject *parent) : QStyledItemDelegate(parent) {}

QWidget *NameEditDelegate::createEditor(QWidget *parent, const QStyleOptionViewItem &option,
                                        const QModelIndex &index) const {
    Q_UNUSED(option)
    if (!index.isValid())
        return QStyledItemDelegate::createEditor(parent, option, index);
    QString value = index.model()->data(index, Qt::EditRole).toString();
    QLineEdit *editor = new QLineEdit(parent);
    editor->setText(value);
    return editor;
}

void NameEditDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const {
    if (!index.isValid())
        QStyledItemDelegate::setEditorData(editor, index);
    QString value = index.data(Qt::DisplayRole).toString();
    QLineEdit *myEditor = qobject_cast<QLineEdit *>(editor);
    if (myEditor) {
        myEditor->setText(value);
        myEditor->selectAll();
    }
}

void NameEditDelegate::setModelData(QWidget *editor, QAbstractItemModel *model, const QModelIndex &index) const {
    if (!index.isValid())
        return;
    QLineEdit *myEditor = qobject_cast<QLineEdit *>(editor);
    if (myEditor) {
        QString name = myEditor->text().simplified();
        if (name.isEmpty())
            return;
        auto list =
            model->match(model->index(0, 0), Qt::DisplayRole, name, -1, Qt::MatchFixedString | Qt::MatchCaseSensitive);
        list.removeAll(index);
        if (list.size()) {
            QMessageBox::warning(
                nullptr, tr("Error"),
                tr("<p>The name %1 is already used for another item. Choose a different name.</p>").arg(name));
            return;
        }
        model->setData(index, name);
    }
}

void NameEditDelegate::updateEditorGeometry(QWidget *editor, const QStyleOptionViewItem &option,
                                            const QModelIndex &index) const {
    QStyledItemDelegate::updateEditorGeometry(editor, option, index);
}

 bool NameEditDelegate::eventFilter(QObject *object, QEvent *event) {
     QWidget *editor = qobject_cast<QWidget *>(object);
     if (!editor)
         return false;
     // for focus out event see qabastractitemdelegate source:
     // we only want to prevent the delegate to commit data ^^
     if (event->type() == QEvent::FocusOut) {
         if (!editor->isActiveWindow() || (QApplication::focusWidget() != editor)) {
             QWidget *w = QApplication::focusWidget();
             while (w) { // don't worry about focus changes internally in the editor
                 if (w == editor)
                     return false;
                 w = w->parentWidget();
             }
             const bool manuallyFixFocus = (event->type() == QEvent::FocusOut) && !editor->hasFocus() &&
                                           editor->parentWidget() &&
                                           (static_cast<QFocusEvent *>(event)->reason() ==
                                           Qt::ActiveWindowFocusReason);
             emit closeEditor(editor, QAbstractItemDelegate::NoHint);
             if (manuallyFixFocus)
                 editor->parentWidget()->setFocus();
         }
     } else if (event->type() != QEvent::KeyPress) {
         // we do not want tab or backtab to allow edition of the next/previous item
         // AND to "validate the input text !
         int key = static_cast<QKeyEvent *>(event)->key();
         if (key == Qt::Key_Backtab || key == Qt::Key_Tab) {
             // emit commitData(editor);
             emit closeEditor(editor, QAbstractItemDelegate::NoHint);
             return true;
         }
         return QStyledItemDelegate::eventFilter(editor, event);
     }
     return QStyledItemDelegate::eventFilter(editor, event);
 }
