// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file globals.cpp
 * @brief Global variables and methods definitions file.
 * @author J. Burgalat
 * @date 2016-2020
 */

#include "globals.hpp"

#include <QSslSocket>
#include <QTcpSocket>

#include "batterystatusworker.hpp"
#include "qcp_helpers.hpp"
#include "wlanwatcher.hpp"

#ifdef DEBUG_BUILD
#include "consoleappender.h"
#endif

I18N *gI18N = nullptr;
FileAppender *gLogAppender = nullptr;
QString gDataDir;
QString gLogFile;
QString gResDir;
QString gLayoutsFile;
QIcon gEmptyIcon;

// arcgis
#include "ArcGISRuntimeEnvironment.h"
using namespace Esri::ArcGISRuntime;
using namespace ISP;

/// @brief Get a string representation of the license status.
QString licenseStatusToString(Esri::ArcGISRuntime::LicenseStatus status) {
    switch (status) {
    case LicenseStatus::Expired:
        return QCoreApplication::tr("expired");
    case LicenseStatus::LoginRequired:
        return QCoreApplication::tr("expired (login required)");
    case LicenseStatus::Valid:
        return QCoreApplication::tr("valid");
    case LicenseStatus::Invalid:
    default:
        return QCoreApplication::tr("invalid");
    }
}

/// @brief Get a string representation of the license level.
QString licenseLevelToString(Esri::ArcGISRuntime::LicenseLevel level) {
    switch (level) {
    case LicenseLevel::Advanced:
        return QCoreApplication::tr("advanced");
    case LicenseLevel::Basic:
        return QCoreApplication::tr("basic");
    case LicenseLevel::Developer:
        return QCoreApplication::tr("developer");
    case LicenseLevel::Lite:
        return QCoreApplication::tr("lite");
    case LicenseLevel::Standard:
        return QCoreApplication::tr("standard");
    }
    return QCoreApplication::tr("unknown");
}

/// @brief Get a string representation of the license type.
QString licenseTypeToString(Esri::ArcGISRuntime::LicenseType type) {

    switch (type) {
    case LicenseType::Developer:
        return QCoreApplication::tr("developer");
    case LicenseType::LicenseKey:
        return QCoreApplication::tr("license-key");
    case LicenseType::NamedUser:
        return QCoreApplication::tr("named-user");
    }
    return QCoreApplication::tr("unknown");
}

/// @brief Initialize global variables of the application.
/// @param dir Application resources directory.
Esri::ArcGISRuntime::LicenseStatus initializeGlobals() {
    QThreadPool::globalInstance()->setMaxThreadCount(QThread::idealThreadCount() - 1);
    qRegisterMetaType<AParam>("AParam");
    qRegisterMetaType<ABlock>("ABlock");
    qRegisterMetaType<BasicFitReport>("BasicFitReport");
    qRegisterMetaType<FitReport>("FitReport");
    qRegisterMetaType<QVector<FitReport>>("FitReportVector");
    qRegisterMetaType<ExecTimes>("ExecTimes");
    qRegisterMetaType<Logger::LogLevel>("LogLevel");

    qRegisterMetaType<WlanWatcher::Notification>("Notification");
    qRegisterMetaType<WlanInterface>("WlanInterface");
    qRegisterMetaType<WlanInfos>("WlanInfos");
    qRegisterMetaType<BatteryStatusWorker::Status>("BatteryStatus");

    QPixmap empty(24, 24);
    empty.fill(Qt::transparent);
    gEmptyIcon = QIcon(empty);

    // initialize qcp data
    Plotting::initHelpers();

    // Initialize translations
    QSettings settings;

    QDir ald = QDir(QStandardPaths::standardLocations(QStandardPaths::AppLocalDataLocation).first());
    ald.mkpath(".");
    gLayoutsFile = ald.absoluteFilePath("layouts.conf");

#ifdef Q_OS_WIN
    gResDir = QString("%1/data/admin2").arg(QApplication::applicationDirPath());
#elif defined(Q_OS_LINUX)
    gResDir = QString("%1/share/ADMin2").arg(QFileInfo(QApplication::applicationDirPath()).absolutePath());
#else
    gResDir = QApplication::applicationDirPath();
#endif
    QString lang = settings.value("global/language", QLocale::system().name()).toString();
    gI18N = new I18N(QString("%1/%2").arg(gResDir, "translations"), lang);
    QString docDir = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).first();
    QString defDir = QString("%1/A-ToolSuite/ADMin2").arg(docDir);
    gDataDir = settings.value("storage/datadir", defDir).toString();
    QDir(gDataDir).mkpath(".");

    gLogFile = QString("%1/logs/ADMin2-%2.log").arg(gDataDir, QDateTime::currentDateTime().toString("yyyyMMdd"));
    QDir(QFileInfo(gLogFile).absolutePath()).mkpath(".");
    bool logToDisk = settings.value("storage/log", true).toBool();
    Logger::LogLevel level =
        static_cast<Logger::LogLevel>(settings.value("storage/log_lvl", static_cast<int>(Logger::Info)).toInt());
    gLogAppender = new FileAppender;
    gLogAppender->setFormat("%{time}{yyyy-MM-dd HH:mm:ss.zzz} %{type:-7}: %{message}\n");
    gLogAppender->setFileName(gLogFile);
    gLogAppender->setDetailsLevel(level);
    if (logToDisk)
        cuteLogger->registerAppender(gLogAppender);

#ifdef DEBUG_BUILD
    ConsoleAppender *consoleAppender = new ConsoleAppender;
    consoleAppender->setFormat("%{time}{yyyy-MM-dd HH:mm:ss.zzz} %{type:-7}: %{message}\n");
    consoleAppender->setDetailsLevel(Logger::Debug);
    cuteLogger->registerAppender(consoleAppender);
#endif

    bool funny = true;
    if (funny) {
        LOG_INFO("");
        LOG_INFO("          _____  __  __ _       ");
        LOG_INFO("    /\\   |  __ \\|  \\/  (_)      ");
        LOG_INFO("   /  \\  | |  | | \\  / |_ _ __  ");
        LOG_INFO("  / /\\ \\ | |  | | |\\/| | | '_ \\ ");
        LOG_INFO(" / ____ \\| |__| | |  | | | | | |");
        LOG_INFO("/_/    \\_\\_____/|_|  |_|_|_| |_|");
    }
    LOG_INFO(QCoreApplication::tr("ADMIN starting..."));

    LOG_INFO(QCoreApplication::tr("Trying to load ArcGIS runtime license key..."));
    Esri::ArcGISRuntime::LicenseStatus result = loadLicence();

    return result;
}

/// @brief Delete global variable pointers.
void finalizeGlobals() {
    LOG_INFO(QCoreApplication::tr("ADMIN closing... Bye pal !"));
    LOG_INFO("---------------------------------------------------------");
    if (!cuteLogger->isRegistered(gLogAppender))
        delete gLogAppender;
    if (gI18N)
        delete gI18N;
    gI18N = nullptr;
}

/**
 * @brief Get icon from current theme.
 * @param name Name of the icon in the theme.
 * @important
 * The method should be only used after globalInit has been called as it uses a global default icon as fallback
 * (Transparent icon of size 24px)
 */
QIcon getIcon(const QString &name) { return QIcon::fromTheme(name, gEmptyIcon); }

/// @brief Set factory default values for application settings.
void setFactoryDefaults(const QString &language) {
    QString docDir = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).first();
    QString dataDir = QString("%1/A-ToolSuite/ADMin2").arg(docDir);
    QString cfgDir = QString("%1/A-ToolSuite/configurations").arg(docDir);
    QSettings settings;

    if (QFileInfo::exists(settings.fileName()))
        QFile::remove(settings.fileName());
    // General default settings
    settings.beginGroup("global");
    settings.setValue("language", language.isEmpty() ? QLocale::system().name() : language);
    settings.setValue("font", QVariant::fromValue(QFont("MS Shell Dlg 2", 8)));
    settings.setValue("save_on_fly", true);

    settings.setValue("show_battery", true);
    settings.setValue("show_battery_pct", true);
    settings.setValue("show_battery_time", true);

    settings.setValue("show_clock", true);
    settings.setValue("show_clock_utc", true);
    settings.setValue("show_clock_date", false);
    settings.setValue("show_clock_sec", true);
    settings.endGroup();

    // Replay mode settings
    settings.beginGroup("replay");
    settings.setValue("max_points", 100000);
    settings.setValue("frame_rate", 10);
    settings.endGroup();

    // warnings
    settings.beginGroup("warnings");
    settings.setValue("conf_change", true);      // configuration changes while another is running.
    settings.setValue("bad_fit_cfg", true);      // invalid fit config.
    settings.setValue("empty_scn", true);        // No (more) scenarios or empty scenarios upon edition complete.
    settings.setValue("map_load_fail", true);    // local map file load failure.
    settings.setValue("no_sd_detected", true);   // NO SD card detected.
    settings.setValue("replay_switch", true);    // switching from RT to Replay mode.
    settings.setValue("no_arcgis_signin", true); // do not ask for arcgis sign-in at startup.
    settings.endGroup();

    // monitoring  settings
    settings.beginGroup("monitoring");
    settings.setValue("multithread", false);
    settings.setValue("proc_internal", false);
    settings.setValue("proc_nogps", true);
    settings.setValue("min_nsat", 4);
    settings.setValue("nf_delay", 0); // 0s -> No automatic flight on delay.
    settings.setValue("nf_onstrap", true);
    settings.setValue("min_sec_vis", 30);
    settings.setValue("min_pts_vis", 30);
    settings.setValue("follow_loc", false);
    settings.setValue("reset_delay", 5000);
    settings.setValue("monitor_wifi", true);
    settings.setValue("max_input_freq", 10); // Maximum input frequency
    settings.setValue("max_inv_freq", 5);    // Maximum input frequency
    settings.setValue("fast_calculations", true);
    settings.setValue("show_fit", false);

    settings.endGroup();

    // graph styles
    settings.beginGroup("graph");
    settings.setValue("hide_toolbar", false);
    settings.setValue("hide_xlabel", false);
    settings.setValue("now_as_ref", true); // not editable yet !
    settings.setValue("asinterval", 200);
    settings.setValue("wtime", 300);   // 300s -> 5 min
    settings.setValue("woffset", 100); // 100 ms
    settings.setValue("box_pos", 2);   // Top-right position.
    settings.setValue("wline", true);
    settings.setValue("lwidth", 1.0);
    settings.setValue("mstyle", 0);
    settings.setValue("msize", 6.0);

    settings.setValue("show_thl", true);
    settings.setValue("thl_color", "#ff0000"); // red
    settings.setValue("thl_width", 1.0);

    settings.setValue("show_rgl", true);
    settings.setValue("rgl_color", "#0000ff"); // blue.
    settings.setValue("rgl_width", 1.0);
    settings.setValue("rgl_fill", true);
    settings.setValue("rgl_fill_color", "#0000ff"); // blue.
    settings.setValue("rgl_fill_alpha", 0.25);

    settings.endGroup();

    // map styles
    settings.beginGroup("map");
    settings.setValue("cs_horiz", false);
    settings.setValue("m_showgrid", false);
    settings.setValue("m_hidemv", false);
    settings.setValue("m_mvmstyle", 5);        // diag. cross
    settings.setValue("m_mvcolor", "#000000"); // black
    settings.setValue("m_mvmsize", 10.0);
    settings.setValue("m_cmstyle", 0); // circle
    settings.setValue("m_cmsize", 10.0);
    settings.setValue("m_omstyle", 3); // square
    settings.setValue("m_omsize", 10.0);
    settings.endGroup();

    // storage
    settings.beginGroup("storage");
    settings.setValue("cfgdir", cfgDir);
    settings.setValue("datadir", dataDir);
    settings.setValue("log", true);
    settings.setValue("log_lvl", static_cast<int>(Logger::Info));
    settings.setValue("save", true);
    settings.endGroup();

    // fit
    settings.beginGroup("fit");
    settings.setValue("maxIter", 20);
    settings.setValue("fTol", 1e-7);
    settings.setValue("pTol", 1e-7);
    settings.setValue("gTol", 1e-9);
    settings.endGroup();
}

/**
 * @brief Build a simple separator widget.
 * @param parent Parent widget.
 * @param orientation Orientation of the separator. Default to Qt::Horizontal.
 */
QFrame *getSeparator(QWidget *parent, Qt::Orientation orientation) {
    QFrame *line = new QFrame(parent);
    line->setFrameShadow(QFrame::Sunken);
    if (orientation == Qt::Horizontal) {
        line->setFrameShape(QFrame::HLine);
        line->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
    } else {
        line->setFrameShape(QFrame::VLine);
        line->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Expanding);
    }
    return line;
}

/**
 * @brief Get the screen of the given widget.
 * @param pWidget Widget to process.
 * @return A pointer to a QScreen. The pointer may be null.
 * @details
 * If the given widget is not a top-level window, the method loops on its parent widget until
 * it finds a top-level widget.
 *
 * As last resort, if no screen has been found, the method returns the primary screen of the
 * application.
 */
QScreen *getActiveScreen(const QWidget *pWidget) {
    QScreen *pActive = nullptr;
    while (pWidget) {
        if (QWindow *w = pWidget->windowHandle()) {
            pActive = w->screen();
            break;
        } else
            pWidget = pWidget->parentWidget();
    }
    if (!pActive)
        pActive = QGuiApplication::primaryScreen();
    return pActive;
}

/**
 * @brief Build a QCPColorGradient from file.
 * @param path Path of the gradient file.
 * @return QCPColorGradient.
 * @details
 * The method reads the file given by path. It expects the file content to be formatted as 4 columns separated by a
 * comma:
 *
 * 1. The position of the color within the gradient (between 0 and 1).
 * 2. The red component of the color (between 0 and 255).
 * 3. The green component of the color (between 0 and 255).
 * 4. The blue component of the color (between 0 and 255).
 *
 * The method stops further processing on the first error and returns an invalid QCPColorGradient in such case.
 */
QCPColorGradient gradientFromFile(const QString &path) {
    QFile file(path);
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
        return QCPColorGradient();
    QTextStream in(&file);
    static const QRegularExpression reSep(" *, *");
    QCPColorGradient out;
    bool ok;
    while (!in.atEnd()) {
        QStringList values = in.readLine().split(reSep);
        if (values.size() < 4) {
            file.close();
            return QCPColorGradient();
        }
        double pos = values.at(0).toDouble(&ok);
        if (!ok) {
            file.close();
            return QCPColorGradient();
        }
        int rc = values.at(1).toInt(&ok);
        if (!ok) {
            file.close();
            return QCPColorGradient();
        }
        int gc = values.at(2).toInt(&ok);
        if (!ok) {
            file.close();
            return QCPColorGradient();
        }
        int bc = values.at(3).toInt(&ok);
        if (!ok) {
            file.close();
            return QCPColorGradient();
        }
        out.setColorStopAt(pos, QColor(rc, gc, bc));
    }
    file.close();
    return out;
}

/// @brief Get the predefined color associated to the given status.
QColor colorFromStatus(StatusColor status) { return QColor(colorStatusToString(status)); }

/// @brief Get the name of color that matches the color status.
QString colorStatusToString(StatusColor status) {
    switch (status) {
    case StatusColor::Green:
        return "green";
    case StatusColor::Orange:
        return "orange";
    case StatusColor::Red:
        return "red";
    case StatusColor::Blue:
        return "blue";
    default:
        return "grey";
        break;
    }
}

/**
 * @brief Get a string representation of the given longitude value.
 * @details
 * The method formats the given value (assumed to be decimal degrees toward East) into a string. The digits precision is
 * fixed to 6.
 */
QString formatLongitude(double value) { return QString("%1°E").arg(value, 11, 'f', 6); }

/**
 * @brief Get a string representation of the given latitude value.
 * @details
 * The method formats the given value (assumed to be decimal degrees toward North) into a string. The digits precision
 * is fixed to 6.
 */
QString formatLatitude(double value) { return QString("%1°N").arg(value, 10, 'f', 6); }

/**
 * @brief Check whether or not internet is reachable.
 * @param urlString URL to check.
 * @param timeout Maximum timeout in milliseconds (default to 2000).
 * @return True if the connection to the host has been done within timeout milliseconds.
 */
bool isInternetReachable(QString urlString, int timeout) {
    bool isSecure = urlString.startsWith("https");
    QUrl url(urlString);
    if (isSecure) {
        QSslSocket socket;
        socket.connectToHostEncrypted(url.host(), 443);
        return socket.waitForConnected(timeout);
    } else {
        QTcpSocket socket;
        socket.connectToHost(url.host(), 80);
        return socket.waitForConnected(timeout);
    }
}

/**
 * @overload
 * @brief Check whether or not internet is reachable.
 * @details
 * A convinient wrapper to isInternetReachable("https://www.arcgis.com", 2000).
 */
bool isInternetReachable() { return isInternetReachable("https://www.arcgis.com", 2000); }

void updateLoggerFromSettings() {
    QSettings settings;
    bool logToDisk = settings.value("log", true).toBool();
    Logger::LogLevel level =
        static_cast<Logger::LogLevel>(settings.value("storage/log_lvl", static_cast<int>(Logger::Info)).toInt());
    gLogFile = QString("%1/logs/ADMin2-%2.log").arg(gDataDir, QDateTime::currentDateTime().toString("yyyyMMdd"));
    QDir(QFileInfo(gLogFile).absolutePath()).mkpath(".");
    gLogAppender->setDetailsLevel(level);
    gLogAppender->setFileName(gLogFile);
    bool isreg = cuteLogger->isRegistered(gLogAppender);
    if (logToDisk) {
        if (!isreg)
            cuteLogger->registerAppender(gLogAppender);
    } else if (isreg) {
        cuteLogger->removeAppender(gLogAppender);
    }
}

/**
 * @brief Save json string in license file.
 * @param json JSON representation of the license to save.
 */
void saveLicence(const QString &json) {
    QFile file(QString("%1/license.txt").arg(gDataDir));
    if (!file.open(QIODevice::WriteOnly)) {
        LOG_DEBUG(QCoreApplication::tr("Cannot save ArcGIS license file (insufficient permissions ?)."));
        return;
    }
    file.write(json.toUtf8());
    file.close();
}

/**
 * @brief load licence file to ArcGIS runtime environment.
 * @return License result.
 * @details
 * The method tries to load a license key from a file named license.txt in the root data directory of the app.
 */
LicenseStatus loadLicence() {
    // add the following to the main.cpp before the application is loaded
    LicenseResult result = ArcGISRuntimeEnvironment::setLicense(ARCGIS_RT_LIC);
    return result.licenseStatus();
}

/// @brief Get the parameters data at given frame.
QVector<AParam> replayDataAt(const ReplayData &data, int frame) {
    QVector<AParam> prms;
    QMap<AParam, QDoubleVector>::const_iterator it = data.data.cbegin();
    prms.reserve(data.data.size());
    while (it != data.data.cend()) {
        prms.push_back(it.key());
        prms.last().value = it.value().at(frame);
        ++it;
    }
    return prms;
}

#include <QBrush>

/**
 * @brief Build a brush from color code and alpha value.buildBrush
 * @param hexColor Hexadecimal color code.
 * @param alpha Color transparency from 0 to 1.
 * @return Qt brush.
 */
QBrush buildBrush(QString hexColor, double alpha) {
    QColor c(hexColor);
    c.setAlphaF(qBound(0.0, alpha, 1.0));
    return QBrush(c);
}

/// @brief Build a pen from color code and width.
QPen buildPen(QString hexColor, double width) { return QPen(buildBrush(hexColor, 1.0), qMax(1.0, width)); }
