// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/**
 * @file globals.hpp
 * @brief Global variables and methods declarations file.
 * @author J. Burgalat
 * @date 2016-2018
 */

#ifndef ADM_GLOBALS_HPP
#define ADM_GLOBALS_HPP

#define NO_RATE

#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#define N_PLOTS 4

#define TB_ICO_SIZE QSize(20, 20)

#include "helpers.hpp"
#include "i18n.hpp"

#include "isp/isp.hpp"

#include "fileappender.h"
#include "logger.h"
#include "qcustomplot.h"
#include <QtCore>
#include <QtWidgets>

// arcgis
#include "License.h"
#include "Point.h"
#include "SpatialReference.h"

using namespace ISP;

#define LOG_MESSAGE(level) CuteMessageLogger(cuteLoggerInstance(), level, __FILE__, __LINE__, Q_FUNC_INFO).write

Q_DECLARE_METATYPE(Qt::CheckState)
Q_DECLARE_METATYPE(QVector<FitReport>)
Q_DECLARE_METATYPE(Units::Unit)

typedef QSharedPointer<QCPGraphDataContainer> GraphContainerPointer;
typedef QVector<QSharedPointer<QCPGraphDataContainer>> GraphContainerVector;
typedef QVector<Esri::ArcGISRuntime::Point> PointVector;
typedef QVector<bool> BoolVector;
typedef QVector<double> QDoubleVector;

/// @brief Global I18N instance.
extern I18N *gI18N;

extern FileAppender *gLogAppender;

extern QString gDataDir;
extern QString gLogFile;
extern QString gResDir;
extern QString gLayoutsFile;

static const Esri::ArcGISRuntime::SpatialReference gWGS84 = Esri::ArcGISRuntime::SpatialReference::wgs84();

QString licenseStatusToString(Esri::ArcGISRuntime::LicenseStatus status);
QString licenseLevelToString(Esri::ArcGISRuntime::LicenseLevel level);
QString licenseTypeToString(Esri::ArcGISRuntime::LicenseType type);

Esri::ArcGISRuntime::LicenseStatus initializeGlobals();
void finalizeGlobals();
void setFactoryDefaults(const QString &language = QString());

void saveLicence(const QString &json);
Esri::ArcGISRuntime::LicenseStatus loadLicence();

void updateLoggerFromSettings();

QFrame *getSeparator(QWidget *parent, Qt::Orientation orientation = Qt::Horizontal);

QIcon getIcon(const QString &name);

QString formatLatitude(double value);

QString formatLongitude(double value);

QScreen *getActiveScreen(const QWidget *pWidget);
QCPColorGradient gradientFromFile(const QString &path);

enum class StatusColor {
    Blue,   ///< Unknown status (not used yet).
    Green,  ///< Normal status.
    Orange, ///< Warning status
    Red,    ///< Critical status.
    Grey    ///< Disabled/Not available status.
};

QColor colorFromStatus(StatusColor status);
QString colorStatusToString(StatusColor status);

bool isInternetReachable(QString urlString, int timeout = 2000);
bool isInternetReachable();

/**
 * @brief A simple validator that prevent empty string.
 */
class NotEmptyValidator : public QValidator {
  public:
    explicit NotEmptyValidator(QObject *parent = nullptr) : QValidator(parent) {}
    virtual QValidator::State validate(QString &input, int &) const override {
        // using Intermediate allows user to temporarily clear the line edit.
        return input.trimmed().isEmpty() ? QValidator::Intermediate : QValidator::Acceptable;
    }
};

/// Enumerate the possible state of the replay mode.
enum class ReplayState {
    STOPPED, ///< The timer is stopped and the frame index is reset to 0 (i.e. start of the movie).
    PLAYING, ///< The timer is active and the frame index changes on timeout.
    PAUSED   ///< The timer is stopped but the frame index is not reset.
};

/**
 * @brief Basic structure for replay mode data.
 */
struct ReplayData {
    int frameCount;
    PointVector positions;
    BoolVector gpsValidity;
    QDoubleVector times;
    QMap<ISP::AParam, QDoubleVector> data;
};

/// @brief Get the parameters data at given frame.
QVector<ISP::AParam> replayDataAt(const ReplayData &data, int frame);

QBrush buildBrush(QString hexColor, double alpha);
QPen buildPen(QString hexColor, double width);

#endif // ADM_GLOBALS_HPP
