// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

/*
 * Copyright (C) 2016-2018  Jeremie Burgalat (jeremie.burgalat at univ-reims dot fr)
 *
 * This file is part of ATLAS.
 *
 * ATLAS is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with ATLAS. If not, see <http://www.gnu.org/licenses/>.
 */
/**
 * @file helpers.cpp
 * @brief Help strings methods definitions file.
 * @author J. Burgalat
 * @date 2016
 */

#include "helpers.hpp"
#include <QApplication>
#include <QWidget>

/// @brief Get help about parameters (for WhatsThis feature).
QString getHelpMessage(HelpFlag flag, bool forSettings) {
    Q_UNUSED(forSettings)
    QString descr;
    switch (flag) {
    case HelpFlag::HIDE_PTB: {
        descr = QApplication::tr("<h4 style='white-space:pre'>Hide plot widgets toolbar.</h4>"
                                 "<p>If enabled, the option instructs to application to hide the toolbar of the plot "
                                 "widgets. This option can be used to save some space on the interface.</p>"
                                 "<p>User interactions provided by the toolbar are still accessible from the context "
                                 "menu of the widget.</p>");
        return descr;
    }
    case HelpFlag::HIDE_PXL: {
        descr = QApplication::tr("<h4 style='white-space:pre'>Hide plot widgets X-axis label.</h4>"
                                 "<p>If enabled, the option instructs to application to hide the X-axis label of the "
                                 "plot widgets. This option can be used to save some space on the interface.</p>"
                                 "<p>X-axis always refers to a time-line.</p>");
        return descr;
    }
    case HelpFlag::ASPERIOD: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Auto-scrolling</h4>"
            "<p>The setting represents the amount of time the application waits between two updates of "
            "the charts in RT mode.</p>");
        return descr;
    }
    case HelpFlag::WTIME: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Time window</h4>"
            "<p>The time window is used during auto-scroll updates to set new x-axis range of the charts.</p>"
            "<p>Upon auto-scroll update, charts will show the last X minutes as defined by <i>time window</i>(plus "
            "the <i>time offset</i>).</p>");
        return descr;
    }
    case HelpFlag::WOFFSET: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Time offset</h4>"
            "<p>The time offset is a small duration added to the time-window during auto-scrolling updates "
            "of the charts in RT mode.</p>"
            "<p>This offset is added to prevent data points to be drawn too close to the chart's right edge.</p>");
        return descr;
    }
    case HelpFlag::LANG: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Application language</h4><p>Change the language of the application.</p>");
        return descr;
    }
    case HelpFlag::FONT: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Application font</h4><p>Change the font of the application.</p>");
        return descr;
    }
    case HelpFlag::SAVE_ON_FLY: {
        descr =
            QApplication::tr("<h4 style='white-space:pre'>Save configuration on the fly</h4>"
                             "<p>If enabled, the option instructs the application to save changes in the current "
                             "configuration <i>on the fly</i>. It occurs in the following situations:</p>"
                             "<ul>"
                             "<li>When a new configuration is loaded. The previous configuration is saved first.</li>"
                             "<li>When the application exits.</li>"
                             "<li>Upon scenario edition using the scenario editor.</li>"
                             "</ul>");
        return descr;
    }
    case HelpFlag::SHOW_BATTERY: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Show computer battery status</h4>"
            "<p>If enabled, the option instructs the application to display the battery status (if any) "
            "in application toolbar.</p>");
        return descr;
    }
    case HelpFlag::SHOW_CLOCK: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Show system clock</h4>"
            "<p>If enabled, the option instructs the application to display the system clock "
            "in application toolbar.</p>"
            "<ul>"
            "<li>Check the option <i>Show date</i> to display/hide the date on the clock.</li>"
            "<li>Check the option <i>Show seconds</i> to display/hide seconds on the clock.</li>"
            "<li>Check the option <i>Show UTC time</i> to display UTC time instead of the local time.</li>"
            "</ul>");
        return descr;
    }
    case HelpFlag::ART_SIGNIN: {
        descr = QApplication::tr("<h4 style='white-space:pre'>ArcGIS Online status</h4>"
                                 "<p>Get informations about ArcGIS Online authentication status.</p>"
                                 "<p>In order to use online basemaps, user must sign-in to ArcGIS Online. It requires "
                                 "an internet connection.</p>"
                                 "<p>Click on the label to sign-in to ArcGIS Online.</p>");
        return descr;
    }
    case HelpFlag::DDIR: {
        descr = QApplication::tr("<h4 style='white-space:pre'>Data files root directory</h4>"
                                 "<p>Set the parent directory of the data files written by the application.</p>"
                                 "<ul>"
                                 "<li>Application log files are stored under logs sub-directory.</li>"
                                 "<li>For each configuration, a directory is created and data files are stored under "
                                 "sub-directories (sorted by date and time) of the former.</li>"
                                 "</ul>"
                                 "<p><b>This option takes effect on the next startup.</b></p>");
        return descr;
    }
    case HelpFlag::DSAVE: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Save data blocks to disk</h4>"
            "<p>When enabled, the application saves all the data processed by the inversion algorithm "
            "during acquisition. Files are written in a sub-directory of the application's <i>root "
            "directory</i>.</p><p>/b>This option takes effect on the next configuration loading.</b></p>");
        return descr;
    }
    case HelpFlag::NFONF: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>New file on new flight</h4>"
            "<p>When enabled, the application creates a new output file when a new flight is created. Otherwise "
            "the application creates a single output file for a given session (i.e. configuration loading).</p>"
            "<p><b>This option takes effect on the next configuration loading.</b></p>");
        return descr;
    }
    case HelpFlag::LOGGING: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Enable application logs</h4>"
            "<p>This option enables/disables log save to the file system. <b>The option takes effect on "
            "the next startup.</b></p>");
        return descr;
    }
    case HelpFlag::LOG_LVL: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Logging severity level</h4>"
            "<p>This option sets the logging verbosity level.</p>"
            "<p>Note that it is not recommended to set the <i>DEBUG</i> level unless a full report of the "
            "application behaviour is required. In <i>DEBUG</i> mode, log file can take a lot of disk space.</p>");
        return descr;
    }
    case HelpFlag::NF_DELAY: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>New flight delay</h4>"
            "<p>This setting allows the application to generate new flights based on the time interval between two "
            "consecutive "
            "blocks. If the latter is greater than the <i>delay</i> indicated by the option, then a new flight is "
            "automatically generated.</p>"
            "<p>If the value is lower or equal to zero, no flight is generated automatically using this "
            "mechanism.</p>");
        return descr;
    }
    case HelpFlag::NF_STRAP: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Enable flight creation on sensor switch status</h4>"
            "<p>If enabled, this setting instructs the application to generate a new flight each time the "
            "<i>new file</i> strap on the sensor is activated.</p>");
        return descr;
    }
    case HelpFlag::CENTER_LOC: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Center map on new sensor location</h4>"
            "<p>This setting indicates if the application should, by default, (re)center the map when the "
            "sensor location has changed. This behaviour can be modified on the fly using the action on map's "
            "toolbar.</p>"
            "<p><b>It takes effect on the next startup.</b></p>");
        return descr;
    }
    case HelpFlag::SHOW_AT_LEAST: {
        descr =
            QApplication::tr("<h4 style='white-space:pre'>Minimum duration to show on map</h4>"
                             "<p>This setting indicates to the application the minimum duration (in seconds) to always "
                             "display whatever the remanence threshold value is.</p>");
        return descr;
    }
    case HelpFlag::CS_ORIENT: {
        descr =
            QApplication::tr("<h4 style='white-space:pre'>Color scale position</h4>"
                             "<p>If checked, this setting instructs the application to place the color scale below the "
                             "map. By default, the color scale is positioned at the right side of the map.</p>"
                             "<p><b>This option takes effect on the next startup.</b></p>");
        return descr;
    }

    case HelpFlag::MT_FIT: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Allow fit calculations on multiple threads</h4>"
            "<p>If enabled, incoming blocks are sent to a thread pool that performs fit calculations on multiple "
            "threads.</p>"
            "<p>The total number of blocks that can be processed in parallel correspond to the number of logical cores "
            "minus "
            "one.</p>"
            "<p>Note that, due to parallel calculations, output block order cannot be guaranteed.</p>");
        return descr;
    }
    case HelpFlag::PROC_INTERNALS: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Process internal parameters of the sensor</h4>"
            "<p>If enabled, all the parameters of the blocks are saved for monitoring. In common "
            "situations, internal parameters are not useful and this option can be disabled (default behaviour).</p>"
            "<p><b>This option takes effect upon the next configuration loading.</b></p>");
        return descr;
    }
    case HelpFlag::PROC_NOGPS: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Process data block with invalid GPS data</h4>"
            "<p>If enabled (default), all incoming blocks are processed, even if they do not store valid GPS data. In "
            "this case, the parameters are not shown on the map but their value are still displayed on the other "
            "widgets of the application (including the real-time plots).</p>");
        return descr;
    }
    case HelpFlag::MON_WIFI: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Monitor WLAN connection</h4>"
            "<p>If enabled, the application checks for the connection to the SSID given in the configuration file. "
            "Additionally, it will regurlarly check for WIFI signal strength.</p>");
        return descr;
    }
    case HelpFlag::MAX_INPUT_FREQ: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Maximum incoming block frequency</h4>"
            "<p>This option indicates the maximum frequency of incoming block to process. It allows the user to limit "
            "the rate of processed block for slow hardware systems.</p>");
        return descr;
    }
    case HelpFlag::MAX_INV_FREQ: {
        descr = QApplication::tr("<h4 style='white-space:pre'>Maximum inversion process frequency</h4>"
                                 "<p>This option indicates the maximum allowed frequency for the inversion process. It "
                                 "allows the user to limit the rate of processed block for slow hardware systems.</p>");
        return descr;
    }
    case HelpFlag::FAST_CALCULATIONS: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Allow fast calculation</h4>"
            "<p>If enabled, the application tries to speed up computations by reducing the number of points when it is "
            "possible. The option has no effect when inversion process is not enabled.</p>"
            "<p><b>This option takes effect upon the next configuration loading.</b></p>");
        return descr;
    }
    case HelpFlag::DISPLAY_FITS: {
        descr = QApplication::tr("<h4 style='white-space:pre'>Display inversion process plots</h4>"
                                 "<p>If enabled, the application generates additional tabs next to the map widget that "
                                 "display the outputs of the inversion process.</p>");
        return descr;
    }
    case HelpFlag::CF_STYLE: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Current flight style</h4>"
            "<p>This group of settings instructs the application how to draw data points of the current "
            "flight on the map.</p>");
        return descr;
    }
    case HelpFlag::PF_STYLE: {
        descr = QApplication::tr("<h4 style='white-space:pre'>Previous flights style</h4>"
                                 "<p>This group of settings instructs the application how to draw data points of the "
                                 "previous flights on the "
                                 "map.</p>");
        return descr;
    }
    case HelpFlag::MV_STYLE: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Missing value point style</h4>"
            "<p>This group of settings instructs the application how to draw data points with missing value on the map."
            "Such style is applied if <i>Hide missing value data point</i> option is not enabled !</p>");
        return descr;
    }
    case HelpFlag::HIDE_MV: {
        descr =
            QApplication::tr("<h4 style='white-space:pre'>Hide missing value point on the map</h4>"
                             "<p>If enabled, the position of the sensor is not visible on the map when the afferent "
                             "parameter value is missing.</p>");
        return descr;
    }
    case HelpFlag::WARN_ARCGIS_SIGNIN: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Prompt user for ArGIS sign-in on startup</h4>"
            "<p>If enabled, the application will prompt the user for ArcGIS Online sign-in at startup.</p>"
            "<p>ArcGIS authentication is only required to access to online basemaps.</p>");
        return descr;
    }
    case HelpFlag::WARN_CONF: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Warn user on configuration changes</h4>"
            "<p>If enabled, the application will ask the user for confirmation before loading a new configuration if "
            "another is already running.</p>"
            "<p>Loading a new configuration always cleans up the previous registered data.</p>");
        return descr;
    }
    case HelpFlag::WARN_INV_CONF: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Warn user about invalid fit configuration</h4>"
            "<p>If enabled, the application warns the user when it has failed to load the <i>inversion "
            "process configuration</i> shipped with the configuration file.</p>"
            "<p>In such situations, inversion process is not available but the sensor can still be "
            "monitored. however molecule concentrations cannot be derived from sensor's incoming data</p>");
        return descr;
    }
    case HelpFlag::WARN_EMPTY_SCN: {
        descr =
            QApplication::tr("<h4 style='white-space:pre'>Warn user about empty scenarios on edition finished</h4>"
                             "<p>If enabled, the application will ask the user for confirmation before "
                             "applying changes in the scenarios editor if:</p>"
                             "<ul>"
                             "<li>The configuration does not store scenarios anymore.</li>"
                             "<li>Some of the configuration's scenarios are empty.</li>"
                             "</ul>"
                             "<p>In such cases, the affected scenarios or the configuration become useless as no more "
                             "parameters can be displayed on the interface.</p>");
        return descr;
    }
    case HelpFlag::WARN_LMFAIL: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Warn user on basemap loading failures</h4>"
            "<p>If enabled, the application warns the user when it has failed to load a basemap, either "
            "from ArcGIS remote servers or from a local tile package file.</p>"
            "<p>In such case, the application reverts to the no-basemap mode.</p>");
        return descr;
    }
    case HelpFlag::WARN_REPLAY: {
        descr = QApplication::tr(
            "<h4 style='white-space:pre'>Warn user before switching to replay mode</h4>"
            "<p>If enabled, the application will ask for confirmation before switching to replay mode.</p>"
            "<p>When the application enters in replay mode, it stops listening to the sensor. Sensor "
            "monitoring is reactivated when the application leaves the replay mode.</p>");
        return descr;
    }
    case HelpFlag::WARN_NOSD: {
        descr = QApplication::tr("<h4 style='white-space:pre'>Warn user about missing SD card in sensor</h4>"
                                 "<p>If enabled, the application will raise a message box that warns about missing SD "
                                 "card in the sensor.</p>"
                                 "<p>When no SD card is present, the sensor still performs acquisitions and sends "
                                 "blocks on the network but <b>no data is saved</b>.</p>");
        return descr;
    }
    default:
        return QString();
    }
}

/// @brief Set tooltip and whatsthis message for the given widget.
void setHelper(QWidget *widget, const QString &message) {
    if (widget) {
        widget->setToolTip(message);
        widget->setWhatsThis(message);
    }
}

/// @brief Set tooltip and whatsthis message for the given widget.
void setHelper(QWidget *widget, HelpFlag flag) {
    if (widget) {
        QString message = getHelpMessage(flag);
        widget->setToolTip(message);
        widget->setWhatsThis(message);
    }
}
