// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "indicatormodel.hpp"
#include "indicators.hpp"
#include "itemmapper.hpp"

#include "cfg_versions.hpp"

#define indRange(x) (qobject_cast<RangeIndicator *>(x))
#define indSwitch(x) (qobject_cast<SwitchIndicator *>(x))
#define indBasic(x) (qobject_cast<BasicIndicator *>(x))

/// @brief Constructor.
IndicatorModel::IndicatorModel(QObject *parent) : QAbstractTableModel(parent) { translateHeaders(); }

/**
 * @brief Full constructor.
 * @param items List of data item to manage. The model takes ownership of the items.
 * @param parent Optional parent object.
 */
IndicatorModel::IndicatorModel(const QVector<BaseIndicator *> &items, QObject *parent)
    : QAbstractTableModel(parent), mItems(items) {
    mItems.removeAll(nullptr);
    for (BaseIndicator *item : mItems)
        item->setParent(this);
    translateHeaders();
}

/// @brief Destructor.
IndicatorModel::~IndicatorModel() {}

/**
 * @brief Get the item flags for the given index.
 */
Qt::ItemFlags IndicatorModel::flags(const QModelIndex &index) const {
    if (!index.isValid())
        return QAbstractTableModel::flags(index) | Qt::ItemIsDropEnabled;

    Qt::ItemFlags f = Qt::ItemIsSelectable | Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled;
    BaseIndicator *item = mItems[index.row()];
    BasicIndicator::Type type = item->type();
    int column = index.column();
    switch (type) {
    case BaseIndicator::BASIC: {
        if (column <= 3)
            f |= Qt::ItemIsEnabled;
        if (column >= 2 && column <= 3)
            f |= Qt::ItemIsEditable;
        break;
    }
    case BaseIndicator::SWITCH: {
        if (column <= 4)
            f |= Qt::ItemIsEnabled;
        if (column >= 2 && column <= 4)
            f |= Qt::ItemIsEditable;
        break;
    }
    case BaseIndicator::RANGE: {
        f |= Qt::ItemIsEnabled;
        if (column >= 2)
            f |= Qt::ItemIsEditable;
        break;
    }
    }
    return f;
}

/// @brief Get the header displayed data.
QVariant IndicatorModel::headerData(int section, Qt::Orientation orientation, int role) const {
    if (role != Qt::DisplayRole)
        return QAbstractTableModel::headerData(section, orientation, role);
    if (orientation == Qt::Horizontal && section >= 0 && section < mHeaders.size())
        return QVariant(mHeaders.at(section));
    else if (orientation == Qt::Vertical)
        return section + 1;
    else
        return QVariant();
}

/// @brief Translate model's headers.
void IndicatorModel::translateHeaders() {
    mHeaders = QStringList{QCoreApplication::tr("Type"),      // 0 RO
                           QCoreApplication::tr("Parameter"), // 1 RO
                           QCoreApplication::tr("Name"),      // 2 RW
                           QCoreApplication::tr("Units"),     // 3 RW
                           QCoreApplication::tr("Reverse"),   // 4 RW
                           QCoreApplication::tr("Warning"),   // 5 RW
                           QCoreApplication::tr("Critical")}; // 6 RW
    emit headerDataChanged(Qt::Horizontal, 0, mHeaders.size() - 1);
}

/**
 * @brief Get the model data under given index and role.
 * @param index Index to process.
 * @param role Role to process.
 * @return Data for the given role at the given index.
 * @details
 * Only Qt::DisplayRole and Qt::TooltipRole are handled by the method on the sole purpose to display the model items in
 * a tableview.
 */
QVariant IndicatorModel::data(const QModelIndex &index, int role) const {
    int ir = index.row(), jc = index.column();

    if (!index.isValid() || ir >= mItems.size())
        return QVariant();

    BaseIndicator *item = mItems[ir];
    BaseIndicator::Type type = item->type();

    if (role == Qt::DisplayRole) {
        switch (jc) {
        case 0: // type
            return indicatorTypeToString(type);
        case 1: // prm name
            return QString::fromStdString(item->parameter().name());
        case 2: // alias
            return item->alias();
        case 3: // units
            return QString::fromStdString(Units::unitToString(item->units(), true));
        case 4: { // reverse (using booleandelegate we should return a null variant).
            return QVariant();
        }
        case 5: { // warn value
            if (type == BasicIndicator::RANGE) {
                double v = indRange(item)->warning();
                return qIsNaN(v) ? QString() : QString::number(v, 'g', 3);
            }
            return QVariant();
        }
        case 6: { // crit value
            if (type == BasicIndicator::RANGE) {
                double v = indRange(item)->critical();
                return qIsNaN(v) ? QString() : QString::number(v, 'g', 3);
            }
            return QVariant();
        }
        default:
            return QVariant();
        }
    } else if (role == Qt::EditRole) {
        switch (jc) {
        case 0:                // indicator type
        case 1:                // prm name
            return QVariant(); // such data is read-only.
        case 2:                // alias
            return item->alias();
        case 3: // units
            return QVariant(item->units());
        case 4: { // reverse (using booleandelegate we should return a null variant).
            if (type == BasicIndicator::BASIC)
                return false;
            else if (type == BasicIndicator::SWITCH)
                return indSwitch(item)->reverse();
            else if (type == BasicIndicator::RANGE)
                return indRange(item)->down();
        }
        case 5: { // warn value
            if (type == BasicIndicator::RANGE)
                return indRange(item)->warning();
            return QVariant();
        }
        case 6: { // crit value
            if (type == BasicIndicator::RANGE)
                return indRange(item)->critical();
            return QVariant();
        }
        default:
            return QVariant();
        }
    } else if (jc == 1 && (role == Qt::ToolTipRole || role == Qt::StatusTipRole)) {
        const AParam &p = item->parameter();
        if (role == Qt::StatusTipRole)
            return QString("%1 (units: %2)").arg(QString::fromStdString(p.name()), QString::fromStdString(Units::unitToString(p.units)));
        else if (p.description.empty()) {
            return QString("<h4 style='white-space:pre'>%1 (units: %2)</h4>")
                .arg(QString::fromStdString(p.name()), QString::fromStdString(Units::unitToString(p.units)));
        } else {
            return QString("<h4 style='white-space:pre'>%1 (units: %2)</h4><p>%3</p>")
                .arg(QString::fromStdString(p.name()), QString::fromStdString(Units::unitToString(p.units)), QString::fromStdString(p.description));
        }
    }
    return QVariant();
}

/**
 * @brief Set data in the model at the given index.
 * @param index Model index to process.
 * @param value Value to set.
 * @param role Model Role.
 * @return True on success, false otherwise.
 * @note
 * The method only deals with Qt::EditRole.
 */
bool IndicatorModel::setData(const QModelIndex &index, const QVariant &value, int role) {
    int row = index.row(), col = index.column();
    if (row == -1 || role != Qt::EditRole)
        return false;
    BaseIndicator *item = mItems[row];
    BaseIndicator::Type type = item->type();
    QVector<int> editableColumns{0, 1, 2, 3};
    if (type == BaseIndicator::SWITCH) {
        editableColumns << 4;
    } else if (type == BaseIndicator::RANGE) {
        editableColumns << 4 << 5 << 6;
    }
    if (!editableColumns.contains(col))
        return false;
    bool updateRange = false;
    switch (col) {
    case 0: // type:not editable yet.
    case 1: // parameter: not editable yet.
        return false;
    case 2: { // alias.
        item->setAlias(value.toString());
        break;
    }
    case 3: { // units
        Units::Unit u = value.value<Units::Unit>();
        // setUnits will scale internal values of the indicator if possible !
        item->setUnits(u, true);
        updateRange = true;
        break;
    }
    case 4: {
        if (type == BaseIndicator::RANGE) {
            RangeIndicator *ind = indRange(item);
            ind->setDown(value.toBool());
            updateRange = true;

        } else {
            SwitchIndicator *ind = indSwitch(item);
            ind->setReverse(value.toBool());
        }
        break;
    }
    case 5: {
        indRange(item)->setWarning(value.toDouble());
        break;
    }
    case 6: {
        indRange(item)->setCritical(value.toDouble());
        break;
    }
    default:
        return false;
    }
    emit dataChanged(index, index);
    if (updateRange && type == BaseIndicator::RANGE)
        emit dataChanged(index.siblingAtColumn(5), index.siblingAtColumn(6));
    return true;
}

/**
 * @brief Add new items to manage.
 * @param items Vector of items to add.
 * @param rejected Reference to a vector with in output the items that have not been added.
 * @details
 * The method adds the given items to the manager. Only items that are not already managed are added, rejected items are
 * returned in rejected reference. It is user responsibility to delete such item if no more used.
 *
 * Unlike IndicatorModel::setItems, the method does not reset the model.
 */
void IndicatorModel::addIndicators(QVector<BaseIndicator *> items, QVector<BaseIndicator *> &rejected) {

    rejected.clear();

    // filter items:
    // - remove null pointers
    items.removeAll(nullptr);
    // - remove duplicates in the vector
    QVector<BaseIndicator *> wrk;
    for (BaseIndicator *item : items) {
        if (wrk.contains(item))
            rejected.push_back(item);
        else
            wrk.push_back(item);
    }
    items = wrk;
    // - remove items already set in mItems:
    wrk.clear();
    for (BaseIndicator *item : items) {
        if (mItems.contains(item))
            rejected.push_back(item);
        else
            wrk.push_back(item);
    }
    // working vector is empty: no items to add !
    if (wrk.isEmpty())
        return;

    int ns = mItems.size();
    beginInsertRows(QModelIndex(), ns, ns + wrk.size() - 1);
    for (BaseIndicator *item : wrk) {
        mItems.push_back(item);
    }
    endInsertRows();
}

/**
 * @brief Add indicator to the model.
 * @param indicator Indicator to add.
 * @return Model index where the indicator has been added. Only the row is relevant, the column of the index is always
 * set to zero.
 * @details
 * The method may fail if the indicator is already found in the model or if it is null, in such case an invalid index is
 * returned. On success, the indicator is appended in the model which takes its ownership.
 */
QModelIndex IndicatorModel::addIndicator(BaseIndicator *indicator) {
    if (!indicator || mItems.contains(indicator))
        return QModelIndex();
    int ns = mItems.size();
    beginInsertRows(QModelIndex(), ns, ns);
    // add the items to the model.
    indicator->setParent(this);
    mItems.push_back(indicator);
    endInsertRows();
    emit itemAdded(indicator);
    return index(rowCount() - 1, 0);
}

/**
 * @brief Duplicate the given indicator.
 * @param indicator Indicator to duplicate.
 * @param uniqueName True to generate a unique alias for the indicator. Default to false.
 * @return Model index where the indicator has been added. Only the row is relevant, the column of the index is always
 *set to zero.
 * @details
 * The indicator is inserted right after the input \a indicator if it belongs to the model, otherwise it is appended to
 *the model.
 *
 * If \a uniqueName is true, then the method appends the suffix " - copy" to the indicator's alias while the latter is
 *found in the aliases of the model.
 */
QModelIndex IndicatorModel::duplicateIndicator(BaseIndicator *indicator, bool uniqueName) {
    BaseIndicator *newItem = BaseIndicator::copyFrom(indicator, this);
    if (!newItem)
        return QModelIndex();
    if (uniqueName) {
        QStringList names;
        for (auto *it : mItems)
            names.append(it->alias());
        QString alias = newItem->alias();
        bool exists = names.contains(alias);
        while (exists) {
            alias = QString("%1 - copy").arg(alias);
            exists = names.contains(alias);
        }
        newItem->setAlias(alias);
    }
    int ridx = mItems.indexOf(indicator);
    if (ridx == -1)
        ridx = mItems.size();
    ridx++;
    beginInsertRows(QModelIndex(), ridx, ridx);
    // add the items to the model.
    mItems.insert(ridx, newItem);
    endInsertRows();
    return index(ridx, 0);
}

/**
 * @overload
 * @brief Duplicate the indicator at given index.
 */
QModelIndex IndicatorModel::duplicateIndicator(QModelIndex index, bool uniqueName) {
    return index.isValid() ? duplicateIndicator(mItems[index.row()], uniqueName) : QModelIndex();
}

/**
 * @brief Remove the indicator at given model index.
 * @param index Model index of the indicator to remove.
 * @return True on success, false otherwise.
 * @details
 * The method removes and deletes the indicator at given index. If \a index is not valid the method immediately returns
 * false.
 *
 * The method emits the signal IndicatorModel::itemAboutToBeRemoved with the pointer of the indicator to be deleted
 * before the indicator is actually removed.
 */
bool IndicatorModel::removeIndicator(const QModelIndex &index) {
    int ir = index.row();
    BaseIndicator *indicator = (ir < 0 || ir >= mItems.size()) ? nullptr : mItems[ir];
    if (!indicator)
        return false;
    emit itemAboutToBeRemoved(BaseIndicatorPointer(indicator));
    beginRemoveRows(QModelIndex(), ir, ir);
    mItems.remove(ir);
    endRemoveRows();
    indicator->deleteLater();
    return true;
}

/// @brief Get the total number of rows in the model.
int IndicatorModel::rowCount(const QModelIndex & /*parent*/) const { return mItems.size(); }

/// @brief Get the number of columns in the model.
int IndicatorModel::columnCount(const QModelIndex & /*parent*/) const { return mHeaders.size(); }

/**
 * @brief Reset the values of the model (set to NaN).
 * @details
 * The method sets all item's value to NaN and refreshes the model afterward.
 */
void IndicatorModel::resetValues() {
    static const double nan = qQNaN();
    for (BaseIndicator *item : mItems) {
        item->setValue(nan);
    }
}

/**
 * @brief Register new values in scenario items and indicators from the given block of data.
 * @param block Block of data to process.
 * @internal
 * The parameter lookup is not very performant:
 */
void IndicatorModel::registerValues(const ABlock &block) {
    for (BaseIndicator *item : mItems) {
        item->setValue(block(item->parameter(), true));
    }
}

/**
 * @brief Register new values in scenario items and indicators from the given item mapper.
 * @param ItemMapper Item mapper to search.
 */
void IndicatorModel::registerValues(const ItemMapper *mapper) {
    if (mapper) {
        for (BaseIndicator *item : mItems) {
            mapper->updateParameter(item->parameter());
        }
    } else {
        qDebug() << "input mapper is null !";
    }
}

/// @brief Get all indicators stored in the model.
QVector<BaseIndicatorPointer> IndicatorModel::indicators() const {
    QVector<BaseIndicatorPointer> out;
    out.reserve(mItems.size());
    for (auto ptr : mItems)
        out.push_back(BaseIndicatorPointer(ptr));
    return out;
}

/// @brief Get the indicator at given index.
BaseIndicatorPointer IndicatorModel::indicator(QModelIndex index) const {
    return index.isValid() ? mItems.at(index.row()) : nullptr;
}

/// @brief Get a XML representation of the model.
QByteArray IndicatorModel::toXML() const {
    QString content;
    QStringList inds;
    for (BaseIndicator *item : mItems) {
        inds.append(QString(item->toXML()));
    }
    content = QString("<indicators version=\"%1\">%2</indicators>").arg(IND_XML_CUR_VER).arg(inds.join(""));
    return content.toLatin1();
}

// DnD support

/// @brief Define the supported drop actions.
Qt::DropActions IndicatorModel::supportedDropActions() const { return Qt::MoveAction; }

/// @brief Define the supported drag actions.
Qt::DropActions IndicatorModel::supportedDragActions() const { return Qt::MoveAction; }

/// @brief Get the list of allowed MIME types.
QStringList IndicatorModel::mimeTypes() const { return QStringList{"IndicatorModel/rows.list"}; }

/// @brief Compute Mime data from a list of indexes.
QMimeData *IndicatorModel::mimeData(const QModelIndexList &indexes) const {
    if (indexes.isEmpty())
        return nullptr;
    QMimeData *mimeData = new QMimeData();
    QByteArray encodedData;
    QDataStream stream(&encodedData, QIODevice::WriteOnly);
    QStringList rows;
    for (const QModelIndex &index : indexes) {
        QString srow = QString::number(index.row());
        if (index.isValid() && !rows.contains(srow))
            rows << srow;
    }
    stream << rows;
    mimeData->setData("IndicatorModel/rows.list", encodedData);
    return mimeData;
}

/// @brief Check whether or not the model can drop data.
bool IndicatorModel::canDropMimeData(const QMimeData *data, Qt::DropAction action, int, int,
                                     const QModelIndex &) const {
    return data->hasFormat("IndicatorModel/rows.list") && action == Qt::MoveAction;
}

/**
 * @brief Drop mime data.
 * @param data Data to drop in the model.
 * @param action Drop action.
 * @param row Destination row.
 * @param column Destination column.
 * @param parent Destination index.
 * @return True on success, false otherwise.
 * @details
 * The method drops mime-data that correspond to a drag and drop operations between the items in the model.
 * It allows to move a \b single item (the first of the list of indexes built in DataIndicatorModel::mimeData) from an
 * index to another.
 */
bool IndicatorModel::dropMimeData(const QMimeData *data, Qt::DropAction, int row, int, const QModelIndex &parent) {
    if (mItems.isEmpty())
        return false;
    QByteArray encodedData = data->data("IndicatorModel/rows.list");
    QDataStream stream(&encodedData, QIODevice::ReadOnly);
    QStringList iRows;
    stream >> iRows;
    if (iRows.isEmpty())
        return false;
    // currently only the first item.
    int iRow = iRows.first().toInt();
    int rCount = rowCount();
    if (row == -1)
        row = parent.row();
    int oRow = (row == -1) ? rCount : row;
    int ovRow = oRow;
    if (oRow > iRow)
        oRow++;
    qDebug() << "IndicatorModel::dropMimeData: FROM" << iRow << "TO" << oRow;

    if (iRow == oRow) // result in no-op for Qt models.
        return false;

    if (beginMoveRows(QModelIndex(), iRow, iRow, QModelIndex(), qMin(oRow, rCount))) {
        mItems.move(iRow, qBound(0, ovRow, rCount - 1)); // from QList::move doc : move(from,to) -> both from and to
                                                         // must be within the index range of the vector.
        endMoveRows();
    }
    return true;
}

#undef indRange
#undef indSwitch
#undef indBasic
