// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "indicators.hpp"
#include <QDomDocument>

BaseIndicator::BaseIndicator(QObject *parent) : QObject(parent) {}

BaseIndicator::~BaseIndicator() {}

/// @brief Constructor.
BaseIndicator::BaseIndicator(const AParam &param, const QString &alias, BaseIndicator::Type type, QObject *parent)
    : QObject(parent), mParam(param), mAlias(alias), mType(type) {}

/// @brief Get the associated parameter (const ref).
const AParam &BaseIndicator::parameter() const { return mParam; }

/// @brief Get the associated parameter.
AParam &BaseIndicator::parameter() { return mParam; }

/**
 * @brief Get the actual type of indicator.
 * @details
 * The type of the BaseIndicator can be used to safely cast a BaseIndicator to a SwitchIndicator or RangeIndicator.
 */
BaseIndicator::Type BaseIndicator::type() const { return mType; }

/// @brief Get the name of this data item.
QString BaseIndicator::alias() const { return mAlias; }

/// @brief Set a name for this indicator.
void BaseIndicator::setAlias(const QString &alias) { mAlias = alias; }

/**
 * @brief Associate a value to this indicator.
 * @param value Value to associate. The method assumes the value is consistent with the indicator parameter's unit.
 */
void BaseIndicator::setValue(double value) { mParam.value = value; }

/// @brief Get the last registered value.
double BaseIndicator::value() const { return mParam.floatValue(); }

/**
 * @brief Set new units.
 * @param unit New unit to set.
 * @param update True (default) to update internal value.
 * @details
 * If \a update is set to true and units are consistent, the method updates the internal value of the indicator.
 */
void BaseIndicator::setUnits(Units::Unit unit, bool update) {
    double v = mParam.floatValue();
    if (update && !Units::inconsistent(unit, mParam.units) && !qIsNaN(v)) {
        v = Units::convert(v, mParam.units, unit);
        mParam.value = v;
    }
    mParam.units = unit;
}

/// @brief Get the parameter units.
Units::Unit BaseIndicator::units() const { return mParam.units; }

/// @brief Get the color of the last registered value.
QColor BaseIndicator::valueColor() { return color(mParam.floatValue()); }

/**
 * @brief Get the color status associated to the last registered value.
 * @return Associated color status (always Grey in base class).
 */
StatusColor BaseIndicator::valueStatusColor() const { return statusColor(mParam.floatValue()); }

/**
 * @brief Get a string representation of the last registered value including units if any.
 * @param prec Optional decimal precision (default to 3).
 */
QString BaseIndicator::valueAndUnits(int prec) const {
    double val = mParam.floatValue();
    if (qIsNaN(val))
        return "N/A";
    QString sym(QString::fromStdString(Units::unitToString(mParam.units)));
    return (sym.compare("N/A") == 0) ? QString::number(val, 'g', prec)
                                     : QString("%1 %2").arg(val, 0, 'g', prec).arg(sym);
}

/// @brief Get the name and units of this indicator.
QString BaseIndicator::aliasAndUnits(bool split) const {
    QString units = QString::fromStdString(Units::unitToString(mParam.units));
    if (units.compare("N/A") == 0) {
        return mAlias;
    } else {
        return QString("%1%2(%3)").arg(mAlias).arg((mAlias.size() > 16 && split) ? "\n" : " ").arg(units);
    }
}

/**
 * @brief Get the color associated to the given value.
 * @param value Value to process (unused in base class).
 * @return Associated color (always Grey in base class).
 */
QColor BaseIndicator::color(double value) const { return colorFromStatus(statusColor(value)); }

/**
 * @brief Get the color status associated to the given value.
 * @param value Value to process (unused in base class).
 * @return Associated color status (always Grey in base class).
 */
StatusColor BaseIndicator::statusColor(double /*value*/) const { return StatusColor::Grey; }

/**
 * @brief Convert the given value to string.
 * @param value Value to convert;
 * @return A String with the value and the parameter's units if relevant.
 */
QString BaseIndicator::valueToString(double value, int prec) const {
    QString text = "N/A";
    if (!qIsNaN(value)) {
        QString sym(QString::fromStdString(Units::unitToString(mParam.units)));
        text = (sym.compare("N/A", Qt::CaseInsensitive) == 0) ? QString::number(value, 'g', prec)
                                                              : QString("%1 %2").arg(value, 0, 'g', prec).arg(sym);
    }
    return text;
}

/**
 * @brief Check whether or not this indicator and the other are comparable.
 * @return True if both instances are comparable, false otherwise.
 * @details
 * The base class compares the type of indicator, the parameter and the alias (case-insensitive) in this
 * order.
 */
bool BaseIndicator::comparable(const BaseIndicator &other) const {
    return mType == other.mType && mParam == other.mParam && mAlias.compare(other.mAlias, Qt::CaseInsensitive) == 0;
}

// --------------------------------------------------

/// @brief (Invalid) constructor.
BasicIndicator::BasicIndicator(QObject *parent) : BaseIndicator(AParam(), "", BASIC, parent) {}

/// @brief Constructor.
BasicIndicator::BasicIndicator(const AParam &param, const QString &alias, QObject *parent)
    : BaseIndicator(param, alias, BASIC, parent) {}

BasicIndicator::~BasicIndicator() {}

/**
 * @brief Check whether or not this indicator and the other are comparable.
 * @return True if both instances are comparable, false otherwise.
 * @details
 * The base class compares the type of indicator, the parameter and the alias (case-insensitive) in this
 * order.
 */
bool BasicIndicator::comparable(const BasicIndicator &other) const {
    return mType == other.mType && mParam == other.mParam && mAlias.compare(other.mAlias, Qt::CaseInsensitive) == 0;
}

/// @brief Compute a XML formatted string representation of the instance.
QByteArray BasicIndicator::toXML() const {
    QString out = QString("<basic>"
                          "<name>%1</name>"
                          "<alias>%2</alias>"
                          "<units>%3</units>"
                          "</basic>")
                      .arg(QString::fromStdString(mParam.name()))
                      .arg(mAlias)
                      .arg(QString::fromStdString(Units::unitToString(mParam.units, true)));
    return out.toLatin1();
}

// --------------------------------------------------

/// @brief (Invalid) constructor.
SwitchIndicator::SwitchIndicator(QObject *parent)
    : BaseIndicator(AParam(), "", SWITCH, parent), mReverse(false), mThreshold(0.5) {}

/// @brief Constructor.
SwitchIndicator::SwitchIndicator(const AParam &param, const QString &alias, bool reverse, QObject *parent)
    : BaseIndicator(param, alias, SWITCH, parent), mReverse(reverse), mThreshold(0.5) {}

SwitchIndicator::~SwitchIndicator() {}

/**
 * @brief Get the color status associated to the given value.
 * @param value Value to process.
 * @return Associated color status.
 * @details
 * The status color first depends on value. If a NaN is given Grey is returned.
 * Otherwise, the method checks if the value is greater or equal to SwitchIndicator::threshold (default to 0.5) and sets
 * the color to Green (SwitchIndicator::reverse is false) or Red (SwitchIndicator::reverse is true).
 */
StatusColor SwitchIndicator::statusColor(double value) const {
    StatusColor color = StatusColor::Grey;
    if (qIsNaN(value))
        return color;
    if (mReverse)
        color = value < mThreshold ? StatusColor::Green : StatusColor::Red;
    else
        color = value > mThreshold ? StatusColor::Green : StatusColor::Red;
    return color;
}

/**
 * @brief Convert the given value to string.
 * @param value Value to convert;
 * @param prec Optional precision. The argument is not used in the overload.
 * @return Either "N/A", "ON" or "OFF" based on the given value.
 * @details
 * The method simply returns:
 * - "N/A" if value is NaN.
 * - "ON" if the value corresponds to a Green status color.
 * - "OFF" if the value corresponds to a Red status color.
 *
 * See SwitchIndicator::statusColor for color status details.
 */
QString SwitchIndicator::valueToString(double value, int /*prec*/) const {
    if (qIsNaN(value))
        return "N/A";
    if (mReverse)
        return value < mThreshold ? QCoreApplication::tr("ON") : QCoreApplication::tr("OFF");
    else
        return value > mThreshold ? QCoreApplication::tr("ON") : QCoreApplication::tr("OFF");
}
/**
 * @brief Check whether or not this indicator and the other are comparable.
 * @return True if both instances are comparable, false otherwise.
 * @details
 * In addition to the base class comparison, the method also compares the reverse flag.
 */
bool SwitchIndicator::comparable(const SwitchIndicator &other) const {
    bool ret = BaseIndicator::comparable(other);
    return ret && mReverse == other.mReverse;
}

/// @brief Compute a XML formatted string representation of the instance.
QByteArray SwitchIndicator::toXML() const {
    QString out;
    out = QString("<switch reverse=\"%3\">"
                  "<name>%1</name>"
                  "<alias>%2</alias>"
                  "</switch>")
              .arg(QString::fromStdString(mParam.name()))
              .arg(mAlias)
              .arg(mReverse ? "yes" : "no");
    return out.toLatin1();
}

/**
 * @brief Set new units.
 * @param unit New unit to set.
 * @param update True (default) to update internal values.
 * @details
 * If \a update is set to true and units are consistent, the method updates the internal value and threshold of the
 * indicator.
 */
void SwitchIndicator::setUnits(Units::Unit unit, bool update) {
    if (update && !Units::inconsistent(unit, mParam.units)) {
        double v = mParam.floatValue();
        if (!qIsNaN(v))
            mParam.value = Units::convert(v, mParam.units, unit);
        v = mThreshold;
        if (!qIsNaN(v))
            mThreshold = Units::convert(v, mParam.units, unit);
    }
    mParam.units = unit;
}

/**
 * @brief Indicate how values should be compared during tests.
 * @details
 * If SwitchIndicator::reverse is true, then the switch indicator is assumed to be \a green if the valuer is \b greater
 * than SwitchIndicator::threshold. In the other case, the indicator is \a green if the value is \b lower than
 * SwitchIndicator::threshold.
 */
bool SwitchIndicator::reverse() const { return mReverse; }

/// @brief Set the reverse mode.
/// @sa SwitchIndicator::reverse.
void SwitchIndicator::setReverse(bool reverse) { mReverse = reverse; }

/// @brief Get the threshold value used for status color calculations.
double SwitchIndicator::threshold() const { return mThreshold; }

/// @brief Set the threshold value used for status color calculations.
/// @note
/// If \a threshold is NaN, the method does nothing.
void SwitchIndicator::setThreshold(double threshold) {
    if (!qIsNaN(threshold))
        mThreshold = threshold;
}

// --------------------------------------------------

/// @brief (Invalid) constructor.
RangeIndicator::RangeIndicator(QObject *parent)
    : BaseIndicator(AParam(), "", RANGE, parent), mWarning(qQNaN()), mCritical(qQNaN()), mDown(false) {}

/// @brief Constructor.
RangeIndicator::RangeIndicator(
    const AParam &param, const QString &alias, double warnValue, double critValue, bool down, QObject *parent)
    : BaseIndicator(param, alias, RANGE, parent), mWarning(warnValue), mCritical(critValue), mDown(down) {
    swapValues();
}

RangeIndicator::~RangeIndicator() {}

/**
 * @brief Swap warning and critical values if needed.
 * @details
 * The method swaps the warning and critical values if needed, based on the ascending flag.
 * - If ::mDown is false: warning and critical values are swapped if the latter is lower than the former.
 * - If ::mDown is true: warning and critical values are swapped if the former is lower than the latter.
 */
void RangeIndicator::swapValues() {
    if (qIsNaN(mWarning) || qIsNaN(mCritical))
        return;
    if ((!mDown && mCritical < mWarning) || (mDown && mCritical > mWarning))
        qSwap(mCritical, mWarning);
}

/**
 * @brief Compute a status color from the given value.
 * @param value Value to process.
 * @return Status color enum value.
 * @details
 * The method computes a status color from the given values and the parameters of this indicator:
 *
 * If value is NaN or both warning and critical (internal) values are NaN then Grey is returned.
 *
 * Otherwise the color is set to:
 *
 * - Red if the value is above (::mDown is false) or below (::mDown is true) the critical value.
 * - Orange if the value is between critical and warning values (both values must not be NaN).
 * - Green if the value is below/above the warning (if any) or the critical value (otherwise).
 */
StatusColor RangeIndicator::statusColor(double value) const {
    StatusColor color = StatusColor::Grey;
    if (qIsNaN(value))
        return color;
    bool hasWarn = !qIsNaN(mWarning), hasCrit = !qIsNaN(mCritical);
    if (!(hasWarn || hasCrit))
        return color;

    double warn = mWarning, crit = mCritical;

    // if swapping is needed every time color is computed.
    // note that we do not update the class members here !!
    //    if ((!mDown && crit < warn) || (mDown && crit > warn))
    //        qSwap(crit, warn);

    if (mDown) {
        if (hasCrit && value < crit) {
            color = StatusColor::Red;
        } else if (hasWarn && value < warn) {
            color = StatusColor::Orange;
        } else {
            color = StatusColor::Green;
        }
    } else {
        if (hasCrit && value > crit) {
            color = StatusColor::Red;
        } else if (hasWarn && value > warn) {
            color = StatusColor::Orange;
        } else {
            color = StatusColor::Green;
        }
    }
    return color;
}

/**
 * @brief Check whether or not this indicator and the other are comparable.
 * @return True if both instances are comparable, false otherwise.
 * @details
 * In addition to the base class comparison, the method also compares the down flag and the warning and critical values
 * of each instances.
 */
bool RangeIndicator::comparable(const RangeIndicator &other) const {
    bool ret = BaseIndicator::comparable(other);
    return ret && mDown == other.mDown && dblCompare(mWarning, other.mWarning) &&
           dblCompare(mCritical, other.mCritical);
}

/// @brief Compute a XML formatted string representation of the instance.
QByteArray RangeIndicator::toXML() const {
    QString out = QString("<range down=\"%6\">"
                          "<name>%1</name>"
                          "<alias>%2</alias>"
                          "<units>%3</units>"
                          "<warning>%4</warning>"
                          "<critical>%5</critical>"
                          "</range>")
                      .arg(QString::fromStdString(mParam.name()))
                      .arg(mAlias)
                      .arg(QString::fromStdString(Units::unitToString(mParam.units, true)))
                      .arg(qIsNaN(mWarning) ? "" : QString::number(mWarning))
                      .arg(qIsNaN(mCritical) ? "" : QString::number(mCritical))
                      .arg(mDown ? "yes" : "no");

    return out.toLatin1();
}

/**
 * @brief Set new units.
 * @param unit New unit to set.
 * @param update True (default) to update internal values.
 * @details
 * If \a update is set to true and units are consistent, the method updates the internal value and thresholds of the
 * indicator.
 */
void RangeIndicator::setUnits(Units::Unit unit, bool update) {
    if (update && !Units::inconsistent(unit, mParam.units)) {
        double v = mParam.floatValue();
        if (!qIsNaN(v))
            mParam.value = Units::convert(v, mParam.units, unit);
        v = mWarning;
        if (!qIsNaN(v))
            mWarning = Units::convert(v, mParam.units, unit);
        v = mCritical;
        if (!qIsNaN(v))
            mCritical = Units::convert(v, mParam.units, unit);
    }
    mParam.units = unit;
}

/// @brief Get the warning state value.
double RangeIndicator::warning() const { return mWarning; }

/// @brief Set the warning state value.
void RangeIndicator::setWarning(double warning) { mWarning = warning; }

/// @brief Get the critical state value.
double RangeIndicator::critical() const { return mCritical; }

/// @brief Set the critical state value.
void RangeIndicator::setCritical(double critical) { mCritical = critical; }

/**
 * @brief Indicate the order of the states values for RangeIndicator::colorStatus.
 * @details
 * The method indicates the reading order of the states value for ::colorStatus method.
 */
bool RangeIndicator::down() const { return mDown; }

/**
 * @brief Set the order of the sates values.
 * @param down True to \a read warning and critical in descreasing order.
 * @note
 * The method may swap warning and critical values (if not nan) to reflect the new \a reading order.
 */
void RangeIndicator::setDown(bool down) {
    mDown = down;
    swapValues();
}

QString indicatorTypeToString(BaseIndicator::Type type) {
    switch (type) {
    case BaseIndicator::BASIC:
        return QCoreApplication::tr("BASIC");
    case BaseIndicator::SWITCH:
        return QCoreApplication::tr("SWITCH");
    case BaseIndicator::RANGE:
        return QCoreApplication::tr("RANGE");
    default:
        return "";
    }
}

/// @brief Get a rich-text description of the indicator type.
QString indicatorTypeDescription(BaseIndicator::Type type) {
    QString descr;
    switch (type) {
    case BaseIndicator::BASIC:
        descr =
            QCoreApplication::tr("<p>The BASIC indicator is represented in the UI by a label with the parameter name "
                                 "followed by its current value.</p>");
        return descr;
    case BaseIndicator::SWITCH:
        descr = QCoreApplication::tr(
            "<p>The SWITCH indicator is represented in the UI by a LED-like image followed by the parameter name.</p>"
            "<p>The <i>LED</i> color, either GREEN or RED, is updated based on the current parameter value.</p>"
            "<p>Such indicator is suitable to represent boolean parameter.</p>");
        return descr;
    case BaseIndicator::RANGE:
        descr = QCoreApplication::tr("<p>The RANGE indicator is represented in the UI by a LED-like image followed by "
                                     "the parameter name and value.</p>"
                                     "<p>The <i>LED</i> color, either GREEN, ORANGE or RED, is updated based on the "
                                     "current parameter value and the "
                                     "different theshold(s) defined in the indicator.</p>");
        return descr;
    default:
        return QCoreApplication::tr("No description available");
    }
}

/**
 * @brief Get the indicator Type matching the input string.
 * @param string String representation of the type to process.
 * @param ok Optional pointer to a bool with in output the result of the process.
 * @return Output BaseIndicator::Type
 * @attention
 * If the input string does not match one of the enum value, the output type is set to BaseIndicator::BASIC.
 * It is wise to check the conversion process using \a ok argument.
 */
BaseIndicator::Type stringToIndicatorType(const QString &string, bool *ok) {
    bool bok = true;
    BaseIndicator::Type type;
    if (string.compare(QCoreApplication::tr("BASIC")) == 0) {
        type = BaseIndicator::BASIC;
    } else if (string.compare(QCoreApplication::tr("SWITCH")) == 0) {
        type = BaseIndicator::SWITCH;
    } else if (string.compare(QCoreApplication::tr("RANGE")) == 0) {
        type = BaseIndicator::RANGE;
    } else {
        type = BaseIndicator::BASIC;
        bok = false;
    }
    if (ok)
        *ok = bok;
    return type;
}

/**
 * @brief Duplicate the given indicator.
 * @return New indicator.
 */
BaseIndicator *BaseIndicator::copyFrom(BaseIndicator *indicator, QObject *parent) {
    if (!indicator)
        return nullptr;
    BaseIndicator::Type type = indicator->type();
    AParam param = indicator->parameter();
    QString alias = indicator->alias();

    switch (type) {
    case BaseIndicator::BASIC: {
        return new BasicIndicator(param, alias, parent);
    }
    case BaseIndicator::SWITCH: {
        bool reverse = qobject_cast<SwitchIndicator *>(indicator)->reverse();
        return new SwitchIndicator(param, alias, reverse, parent);
    }
    case BaseIndicator::RANGE: {
        RangeIndicator *tmp = qobject_cast<RangeIndicator *>(indicator);
        return new RangeIndicator(param, alias, tmp->warning(), tmp->critical(), tmp->down(), parent);
    }
    default:
        return nullptr;
    }
    return nullptr;
}
