// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef INDICATORS_HPP
#define INDICATORS_HPP

#include "globals.hpp"

class BaseIndicator : public QObject {
    Q_OBJECT
  public:
    enum Type { BASIC, SWITCH, RANGE };
    Q_ENUM(Type)
    explicit BaseIndicator(QObject *parent = nullptr);
    virtual ~BaseIndicator();
    Type type() const;
    QString alias() const;
    void setAlias(const QString &alias);
    const AParam &parameter() const;
    AParam &parameter();
    void setValue(double value);
    double value() const;

    virtual void setUnits(Units::Unit unit, bool update = true);
    Units::Unit units() const;

    bool comparable(const BaseIndicator &other) const;

    virtual QColor valueColor();
    virtual StatusColor valueStatusColor() const;
    virtual QString valueAndUnits(int prec = 3) const;
    virtual QString aliasAndUnits(bool split = false) const;

    virtual QColor color(double value) const;
    virtual StatusColor statusColor(double value) const;
    virtual QString valueToString(double value, int prec = 3) const;
    virtual QByteArray toXML() const = 0;

    static BaseIndicator *copyFrom(BaseIndicator *item, QObject *parent = nullptr);

  protected:
    BaseIndicator(const AParam &param, const QString &alias, Type type, QObject *parent = nullptr);
    AParam mParam;
    QString mAlias;
    Type mType;

  private:
    Q_DISABLE_COPY(BaseIndicator)
};

typedef QPointer<BaseIndicator> BaseIndicatorPointer;

class BasicIndicator : public BaseIndicator {
    Q_OBJECT
  public:
    explicit BasicIndicator(QObject *parent);
    BasicIndicator(const AParam &param, const QString &alias, QObject *parent = nullptr);
    virtual ~BasicIndicator();
    virtual QByteArray toXML() const override;
    bool comparable(const BasicIndicator &other) const;

  private:
    Q_DISABLE_COPY(BasicIndicator)
};

class SwitchIndicator : public BaseIndicator {
    Q_OBJECT
  public:
    explicit SwitchIndicator(QObject *parent = nullptr);
    explicit SwitchIndicator(const AParam &param, const QString &alias, bool reverse = false, QObject *parent = nullptr);
    virtual ~SwitchIndicator();

    bool comparable(const SwitchIndicator &other) const;

    virtual StatusColor statusColor(double value) const override;
    virtual QString valueToString(double value, int prec = 3) const override;
    virtual QByteArray toXML() const override;

    virtual void setUnits(Units::Unit unit, bool update = true) override;

    bool reverse() const;
    void setReverse(bool reverse);

    double threshold() const;
    void setThreshold(double threshold);

  private:
    bool mReverse;
    double mThreshold;
    Q_DISABLE_COPY(SwitchIndicator)
};

class RangeIndicator : public BaseIndicator {
    Q_OBJECT
  public:
    explicit RangeIndicator(QObject *parent = nullptr);
    RangeIndicator(const AParam &param, const QString &alias, double warnValue = qQNaN(), double critValue = qQNaN(),
                   bool down = false, QObject *parent = nullptr);
    virtual ~RangeIndicator();

    bool comparable(const RangeIndicator &other) const;

    virtual StatusColor statusColor(double value) const override;
    virtual QByteArray toXML() const override;

    virtual void setUnits(Units::Unit unit, bool update = true) override;

    double warning() const;
    void setWarning(double warning);

    double critical() const;
    void setCritical(double critical);

    bool down() const;
    void setDown(bool down);

  private:
    double mWarning;
    double mCritical;
    bool mDown;

    void swapValues();

    Q_DISABLE_COPY(RangeIndicator)
};

Q_DECLARE_METATYPE(QVector<BaseIndicator::Type>)

QString indicatorTypeToString(BaseIndicator::Type type);
QString indicatorTypeDescription(BaseIndicator::Type type);
BaseIndicator::Type stringToIndicatorType(const QString &string, bool *ok = nullptr);

#endif // INDICATORS_HPP
