// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "itemchoicedialog.hpp"
#include "itemmapper.hpp"

ItemChoiceDialog::ItemChoiceDialog(const QVector<AParam> &parameters, int selected, QWidget *parent, Qt::WindowFlags flags)
    : QDialog(parent, flags) {
    mParameters = parameters;
    mLblParameter = new QLabel(this);
    mCbxParameter = new QComboBox(this);
    mLblParamDescr = new QLabel(this);
    mLblParamDescrValue = new QLabel(this);
    mLblParamDescrValue->setWordWrap(true);
    mButtonBox = new QDialogButtonBox(this);
    mButtonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel);
    QVBoxLayout *layout = new QVBoxLayout(this);
    QFormLayout *flayout = new QFormLayout;
    flayout->setContentsMargins(0, 0, 0, 0);
    flayout->addRow(mLblParameter, mCbxParameter);
    flayout->addRow(mLblParamDescr, mLblParamDescrValue);
    layout->addLayout(flayout);
    layout->addStretch(1);
    layout->addWidget(mButtonBox);

    static const QString noDescr = QCoreApplication::tr("No description available");
    int i = 0;
    for (const AParam &p : mParameters) {
        mCbxParameter->addItem(QString::fromStdString(p.name()));
        mCbxParameter->setItemData(i++, p.description.empty() ? noDescr : QString::fromStdString(p.description), Qt::ToolTipRole);
    }

    QObject::connect(mCbxParameter, QOverload<int>::of(&QComboBox::currentIndexChanged), this,
                     &ItemChoiceDialog::onParameterIndexChanged);
    QObject::connect(mButtonBox, &QDialogButtonBox::accepted, this, &QDialog::accept);
    QObject::connect(mButtonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);
    mLblParameter->setText(tr("Parameter:"));
    mLblParamDescr->setText(tr("Description:"));
    setWindowTitle(tr("Parameter selection"));
    setParameterIndex(selected);
    setMinimumWidth(200);
    setMaximumWidth(800);
}

/// @brief Set combobox index programmatically.
void ItemChoiceDialog::setParameterIndex(int index) {
    index = qBound(-1, index, mParameters.size() - 1);
    mCbxParameter->setCurrentIndex(index);
}

/// @brief Get the currently selected parameter.
AParam ItemChoiceDialog::parameter() const {
    int index = mCbxParameter->currentIndex();
    return (index == -1) ? AParam() : mParameters.at(index);
}

/// @brief Update description label on parameter selection changed.
void ItemChoiceDialog::onParameterIndexChanged(int index) {
    if (index >= 0 && index < mParameters.size()) {
        mLblParamDescrValue->setText(QString::fromStdString(mParameters.at(index).description));
        mButtonBox->button(QDialogButtonBox::Ok)->setEnabled(true);
    } else {
        mLblParamDescrValue->setText(tr("No parameter selected"));
        mButtonBox->button(QDialogButtonBox::Ok)->setEnabled(false);
    }
}
/// @brief Get the size hint of the widget.
QSize ItemChoiceDialog::sizeHint() const {
    QSize size = minimumSizeHint();
    int &w = size.rwidth();
    w = qBound(200, w, 800);
    return size;
}

/**
 * @brief Static convenience function to let the user select a parameter from a list of parameters.
 * @param parent Optional parent widget.
 * @param parameters List of parameter to set.
 * @param current Optional index of the parameter to show in the combobox.
 * @param ok Option control flag with, in output, true if the OK button was clicked, false otherwise.
 * @param flags Window flag.
 * @return Selected parameter or an invalid AParam if no selection occured.
 * @details
 * An invalid parameter may be returned in the following cases:
 * - the combobox was left untouched.
 * - the dialog was rejected (ESC key has been pressed or the CANCEL button was clicked). In such case \a ok is set to false.
 */
AParam ItemChoiceDialog::getParameter(QWidget *parent, const QVector<AParam> &parameters, int current, bool *ok,
                                      Qt::WindowFlags flags) {
    ItemChoiceDialog *dialog = new ItemChoiceDialog(parameters, current, parent, flags);
    const int ret = dialog->exec();
    if (ok)
        *ok = (ret == QDialog::Accepted);
    AParam result = ret ? dialog->parameter() : AParam();
    dialog->deleteLater();
    return result;
}
