// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "itemmapper.hpp"
#include "globals.hpp"
#include <QDomDocument>

#include "cfg_versions.hpp"

/**
 * @brief Get the default parameter key map.
 * @details
 * The default special parameter key map contains the usual name found in AMULSE descriptor for special parameters to
 * be retrieved from a block.
 *
 * All keys followed by (*) must be found in descriptor at runtime, otherwise the application will not be able to
 * compute geographic positions which are mandatory for tracking !!
 *
 * All other keys are optional. If not found some part of the UI may no be updated.
 *
 * @par version 1
 *
 * | Key                     | Parameter name    |
 * | :---------------------- | :---------------: |
 * | PrmKey::SENSOR_TIME (*) | h_system          |
 * | PrmKey::LONGITUDE (*)   | longitude_gps     |
 * | PrmKey::LATITUDE (*)    | latitude_gps      |
 * | PrmKey::ALTITUDE (*)    | altitude_gps      |
 * | PrmKey::NB_SATS (*)     | nb_sats_gps       |
 * | PrmKey::GPS_TIME        | time_gps          |
 * | PrmKey::GPS_DATE        | date_gps          |
 * | PrmKey::PCMON_TIME      | h_pcmon           |
 * | PrmKey::CLOCK_SYNC      | h_gps_sync        |
 * | PrmKey::A_STRAP_COUNT   | acq_strap_counter |
 * | PrmKey::A_PERIOD        | t_mes_ai          |
 * | PrmKey::A_NUMBER        | num_acq           |
 * | PrmKey::F_STRAP_COUNT   | f_strap_counter   |
 * | PrmKey::F_NUMBER        | f_number          |
 * | PrmKey::F_DATETIME      | f_datetime        |
 * | PrmKey::UDP_FREQUENCY   | udp_freq          |
 * | PrmKey::DIO_ACQ_STRAP   | dio_strap         |
 *
 */
KeyMap defaultKeyMap() {
    return {
        {ItemMapper::SENSOR_TIME, "h_system"},
        {ItemMapper::LONGITUDE, "longitude_gps"},
        {ItemMapper::LATITUDE, "latitude_gps"},
        {ItemMapper::ALTITUDE, "altitude_gps"},
        {ItemMapper::NB_SATS, "nb_sats_gps"},

        {ItemMapper::GPS_TIME, "time_gps"},
        {ItemMapper::GPS_DATE, "date_gps"},
        {ItemMapper::PCMON_TIME, "h_pcmon"},
        {ItemMapper::CLOCK_SYNC, "h_gps_sync"},

        {ItemMapper::A_STRAP_COUNT, "acq_strap_counter"},
        {ItemMapper::A_PERIOD, "t_mes_ai"},
        {ItemMapper::A_NUMBER, "num_acq"},
        {ItemMapper::F_STRAP_COUNT, "f_strap_counter"},
        {ItemMapper::F_NUMBER, "f_number"},
        {ItemMapper::F_DATETIME, "f_datetime"},
        {ItemMapper::UDP_FREQUENCY, "udp_freq"},
        {ItemMapper::DIO_ACQ_STRAP, "dio_strap"},
    };
}

const QVector<ItemMapper::PrmKey> ItemMapper::gMandatoryKeys{SENSOR_TIME, LONGITUDE, LATITUDE, ALTITUDE, NB_SATS};

/**
 * @brief Constructor
 * @details
 * The internal keymap is initialized to the hard-coded default keymap (see ::defaultKeyMap).
 */
ItemMapper::ItemMapper(QObject *parent) : QObject(parent) {
    mKeyMapVersion = KMP_XML_CUR_VER;
    mKeymapFile = "";
    mKeyMap = defaultKeyMap();
    mLastValid = false;
}

/**
 * @brief Copy the content of the given mapper.
 * @param mapper Item mapper to copy.
 */
void ItemMapper::copyFrom(const ItemMapper *mapper) {
    if (!mapper)
        return;
    mLastValid = mapper->mLastValid;
    mKeyMapVersion = mapper->mKeyMapVersion;
    mKeymapFile = mapper->mKeymapFile;
    mKeyMap = mapper->mKeyMap;
    mParams = mapper->mParams;
}

/// @brief Get the version of the key map.
int ItemMapper::keymapVersion() const { return mKeyMapVersion; }

/// @brief Reset the keymap to its default factory value.
void ItemMapper::resetKeymap() { mKeyMap = defaultKeyMap(); }

/// @brief Reset the parameters.
void ItemMapper::resetParameters() { mParams.clear(); }

/// Get the (parameter) name of the given key.
QString ItemMapper::keyName(PrmKey key) const { return mKeyMap.value(key, ""); }

/// @brief Set parameters to map.
void ItemMapper::setParameters(const QVector<AParam> &parameters) {
    mParams.clear();
    for (const AParam &p : parameters) {
        if (!mParams.contains(p))
            mParams.push_back(p);
    }
}

/// @brief set parameters to map from given descriptor.
void ItemMapper::setParameters(const ABlock &descriptor) {
    QVector<AParam> ou;
    auto &&in = descriptor.parameters();
    std::copy(in.begin(), in.end(), std::back_inserter(ou));
    setParameters(ou);
}

/**
 * @brief Get the registered parameters.
 * @param withInternals Optional boolean flag with true to return also internal parameters of the mapper. Default to
 * false.
 * @return Vector of parameters.
 * @details
 * By default, the method does not return the parameters that belongs to the (hard-coded) category \a internal, unless
 * the parameter belongs also to the keymap.
 *
 * This default behaviour allows to reduce the parameters that are stored in a DataManager that uses this item mapper.
 * In common situations, internal parameters are not useful for the front end user.
 */
QVector<AParam> ItemMapper::parameters(bool withInternals) const {
    if (withInternals)
        return mParams;
    QVector<AParam> out;
    QStringList keys = mKeyMap.values();
    for (const AParam &prm : mParams) {
        QStringList ou;
        auto &in = prm.categories;
        for (auto & v : in)
            ou.push_back(QString::fromStdString(v));
        if (ou.contains("internal", Qt::CaseInsensitive) && !keys.contains(QString::fromStdString(prm.name())))
            continue;
        out.push_back(prm);
    }
    return out;
}

/**
 * @brief Get the parameter associated to the given key.
 * @param key Parameter key to process.
 * @return Parameter that matches the key or an invalid parameter if no match is found.
 */
AParam ItemMapper::parameter(PrmKey key) const {
    QString name = mKeyMap.value(key, "");
    auto it = std::find_if(mParams.begin(), mParams.end(), [name](const AParam &p) {
        return QString::fromStdString(p.name()).compare(name) == 0;
    });
    return (it == mParams.end()) ? AParam() : AParam(*it);
}

/**
 * @brief Get the parameter associated to the given name.
 * @param name Name of the parameter to search.
 * @return Parameter that matches the name or an invalid parameter if no match is found.
 */
AParam ItemMapper::parameter(QString name) const {
    name = name.toLower();
    auto it = std::find_if(mParams.begin(), mParams.end(), [name](const AParam &p) {
        return QString::fromStdString(p.name()).compare(name) == 0;
    });
    if (it == mParams.end())
        return AParam();
    else {
        return AParam(*it);
    }
}

/**
 * \overload
 * @brief Get the parameter associated to the given name.
 * @param name Name of the parameter to search.
 * @param units Output parameter units (for value scaling).
 * @return Parameter that matches the name or an invalid parameter if no match is found.
 * @note
 * If units is not consistent with the parameter's units, then NaN is set for the output parameter value.
 */
AParam ItemMapper::parameter(QString name, Units::Unit units) const {
    name = name.toLower();
    auto it = std::find_if(mParams.begin(), mParams.end(), [name](const AParam &p) {
        return QString::fromStdString(p.name()).compare(name) == 0;
    });
    if (it == mParams.end())
        return AParam();
    else {
        AParam out(*it);
        if (Units::inconsistent(units, it->units)) {
            out.value = qQNaN();
        } else {
            out.value = Units::convert(out.floatValue(), out.units, units);
        }
        out.units = units;
        return out;
    }
}

/// @brief Get the current value of the given \a parameter.
double ItemMapper::value(const AParam &parameter) const {
    auto it = std::find(mParams.begin(), mParams.end(), parameter);
    return (it == mParams.end()) ? qQNaN() : it->floatValue();
}

/// @brief Get the current value of the given parameter \a name.
double ItemMapper::value(QString name) const {
    name = name.toLower();
    auto it = std::find_if(mParams.begin(), mParams.end(), [name](const AParam &p) {
        return QString::fromStdString(p.name()).compare(name) == 0;
    });
    return (it == mParams.end()) ? qQNaN() : it->floatValue();
}

/// @brief Get the current value of the given \a key.
double ItemMapper::value(PrmKey key) const { return value(mKeyMap.value(key, "")); }

/**
 * @brief Update the given parameter.
 * @param param Parameter to update.
 * @return Reference to param.
 * @details
 * The method tries to update the given parameter with the value of its counterpart in the mapper.
 * If no parameter is found in the mapper or if the units of the parameters are not consistent, the value is set to NaN.
 * Otherwise, the method updates the parameter value (and possibly scale the units).
 */
AParam &ItemMapper::updateParameter(AParam &param) const {
    auto it = std::find(mParams.begin(), mParams.end(), param);
    if (it == mParams.end() || Units::inconsistent(param.units, it->units))
        param.value = qQNaN();
    else
        param.value = Units::convert(it->floatValue(), it->units, param.units);
    return param;
}

/**
 * @brief Reset the current values of the mapper.
 *@details
 *
 * The method invalidates the mapper and sets all its values to NaN.
 */
void ItemMapper::resetValues() {
    mLastValid = false;
    for (AParam &p : mParams)
        p.value = qQNaN();
}

/**
 * @brief Register new values in the mapper.
 * @param parameters Vector of the parameters to process.
 * @details
 * The method updates the values of all parameters within the mapper from the values of the given ones.
 *
 * If units are inconsistent within the two sources or if some of the mapper's parameters are not found in the input
 * vectors, then the afferent parameters in the mapper are set to NaN.
 *
 * Passing a empty vector will set NaN to all parameters in the mapper and invalidates the registration: next call of
 * ItemMapper::validRegistration will return false.
 */
void ItemMapper::registerValues(const QVector<AParam> &parameters) {
    for (AParam &p : mParams) {
        double value = qQNaN();
        QVector<AParam>::const_iterator it =
            std::find_if(parameters.begin(), parameters.end(), [p](const AParam &other) { return other == p; });
        if (it != parameters.end()) {
            // return nan if units are not consistent or input value is nan.
            value = Units::convert(it->floatValue(), it->units, p.units);
        }
        p.value = value;
    }
    mLastValid = !parameters.isEmpty();
}

/**
 * @overload
 * @brief Register new values in the mapper.
 * @param block Input data block to process.
 * @details
 * The method is a equivalent wrapper to:
 *
 * @code{.cpp}
 * ItemMapper::registerValues(block.parameters())
 * @endcode
 *
 * However the registration is assumed to be valid if the block is valid (see ABlock::isValid).
 */
void ItemMapper::registerValues(const ABlock &block) {
    QVector<AParam> parameters;
    auto && in = block.parameters();
    std::copy(in.begin(),in.end(),std::back_inserter(parameters));
    for (AParam &p : mParams) {
        double value = qQNaN();
        QVector<AParam>::const_iterator it =
            std::find_if(parameters.begin(), parameters.end(), [p](const AParam &other) { return other == p; });
        if (it != parameters.end()) {
            // return nan if units are not consistent or input value is nan.
            if (it->units == Units::UKN || Units::NOUNIT)
                value = it->floatValue();
            else
                value = Units::convert(it->floatValue(), it->units, p.units);
        }
        p.value = value;
    }
    mLastValid = block.isValid();
}

/// @brief Check whether or not the last registration process was valid.
bool ItemMapper::validRegistration() const { return mLastValid; }

/**
 * @brief Get the list of available categories.
 * @param toUpper True to convert name in upper case. Default to false, that is return categories in lower case.
 * @return List of categories names.
 */
QStringList ItemMapper::categories(bool toUpper) const {
    QStringList out;
    for (const AParam &p : mParams) {
        QStringList ou;
        auto &in = p.categories;
        for (auto & v : in)
            ou.push_back(QString::fromStdString(v));
        if (ou.isEmpty())
            continue;
        for (QString c : ou) {
            if (c.trimmed().size() && !out.contains(c, Qt::CaseInsensitive))
                out.push_back(toUpper ? c.toUpper() : c.toLower());
        }
    }
    return out;
}

/**
 * @brief Get the list of parameters that matches the given categories.
 * @param categories Categories to process.
 * @return Vector of Parameter.
 * @details
 * Duplicates are filtered out.
 */
QVector<AParam> ItemMapper::parametersFromCategories(const QStringList &categories) const {
    QVector<AParam> out;
    for (const AParam &p : mParams)
        for (const QString &c : categories) {
            QStringList ou;
            auto &in = p.categories;
            for (auto & v : in)
                ou.push_back(QString::fromStdString(v));
            if (ou.contains(c, Qt::CaseInsensitive) && !out.contains(p))
                out.push_back(p);
        }
    return out;
}

/**
 * @brief Check whether or not the current keymap is consistent with the parameters registered in the mapper.
 * @param mandatoryMissings Optional pointer to a list of string with in output the names of the missing mandatory
 * parameters.
 * @param optionalMissings Optional pointer to a list of string with in output the names of the missing optional
 * parameters.
 * @return Pair of boolean with the results (see method documentation).
 * @details
 * The method checks if the current keymap is consistent with the parameters registered in the mapper. On output two
 * boolean flags are returned:
 *
 * - The first indicates whether or not all the mandatory parameters have been found. These parameters are associated to
 * GPS location of the sensor and are mandatory to display data on a map.
 * - The second indicates whether or not all optional parameters have been found. These parameters are used to get
 * additional informations about the sensor.
 */
QPair<bool, bool> ItemMapper::checkKeymap(QStringList *mandatoryMissings, QStringList *optionalMissings) const {
    QStringList misso, missm;
    QMetaEnum me = QMetaEnum::fromType<PrmKey>();
    bool ook = true, mok = true;

    for (int i = 0; i < me.keyCount(); i++) {
        PrmKey key = static_cast<PrmKey>(me.value(i));
        QString keyName = QString(me.key(i)).toLower();
        QString keyTag = QString("<%1>").arg(QString(me.key(i)).toLower());
        QString name = mKeyMap.value(key);
        bool ismok = ItemMapper::gMandatoryKeys.contains(key);
        if (name.isEmpty()) {
            LOG_DEBUG(QString("ItemMapper::isConsistent(), key \"%1\" does not store a value !").arg(keyName));
            if (ismok)
                mok = false;
            else
                ook = false;
            continue;
        }
        if (!hasParameter(name)) {
            if (ismok) {
                mok = false;
                missm.append(QString("\"%1\" (<%2>)").arg(name, keyName));
            } else {
                ook = false;
                misso.append(QString("\"%1\" (<%2>)").arg(name, keyName));
            }
        }
    }
    if (mandatoryMissings)
        *mandatoryMissings = missm;
    if (optionalMissings)
        *optionalMissings = misso;
    return qMakePair(mok, ook);
}

/**
 * @brief Check whether or not the given parameter is stored in the mapper.
 * @details
 * The method searches for the first parameter in the mapper with the same name as the given one and returns true on
 * success.
 */
bool ItemMapper::hasParameter(const AParam &parameter) const {
    QString name = QString::fromStdString(parameter.name());
    QVector<AParam>::const_iterator it = std::find_if(
        mParams.begin(), mParams.end(), [name](const AParam &param) { return QString::fromStdString(param.name()).compare(name) == 0; });
    return (it != mParams.end());
}

/// @brief Check whether or not the mapper contains a parameter with the given name.
bool ItemMapper::hasParameter(QString name) const {
    name = name.toLower();
    QVector<AParam>::const_iterator it = std::find_if(
        mParams.begin(), mParams.end(), [name](const AParam &param) { return QString::fromStdString(param.name()).compare(name) == 0; });
    return (it != mParams.end());
}

/**
 * @brief Parse the xml file pointed by xmlPath and setup the internal key map.
 * @param xmlPath Path of the file to parse.
 * @param errors Optional pointer to a QStringList with error messages on output (if any).
 * @return True on success, false otherwise.
 * @sa ItemMapper::parseContent
 *
 * 1. The internal keymap is reset to default values.
 * 2. The file located at xmlPath is tested. If no file is found the method immediately returns false.
 * 3. The file is then parsed. For each missing key or missing key's value an error message is
 *    stored in errors (if any) and the method will return false.
 * 4. The internal keymap is updated with the (partial) keymap parsed on step 3.
 * 5. Input file path is registered in the mapper.
 */
bool ItemMapper::loadKeymap(const QString &xmlPath, QStringList *errors) {
    mKeyMap = defaultKeyMap();
    QFile xmlFile(xmlPath);
    // canot open files.
    if (!xmlFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QCoreApplication::tr("Cannot open %1.").arg(xmlPath));
        return false;
    }
    QString instance = xmlFile.readAll();
    xmlFile.close();
    bool ret = parseContent(instance, false, errors);
    mKeymapFile = QFileInfo(xmlPath).absoluteFilePath();
    return ret;
}

/**
 * @brief Validate and parse the input content.
 * @param content XML formatted content to parse.
 * @param strict Instruct the method whether to be strict or not. Default to false.
 * @return True on success, false otherwise.
 * @details
 * The method attempts to parse the given XML content. It may fail if at least one mandatory tag is not found or if it
 * has no value.
 *
 * Once all content is parsed, if an error has been detected, the behaviour of the method depends on strict :
 *
 * - If \strict is false (default), the method updates partially the internal map with valid pair of key/values.
 * - Otherwise, the internal map is not updated at all.
 *
 * In both case, the method still returns false on error.
 */
bool ItemMapper::parseContent(const QString &content, bool strict, QStringList *errors) {
    QDomDocument document;
    document.setContent(content);
    QDomNode root = document.firstChildElement("keymap");
    if (root.isNull()) {
        if (errors)
            errors->append(QCoreApplication::tr("No <keymap> node found."));
        return false;
    }
    // search for version attr.
    mKeyMapVersion = root.toElement().attribute("version", "0").toInt();
    if (mKeyMapVersion < KMP_XML_MIN_VER) {
        if (errors)
            errors->append(QCoreApplication::tr("keymap version too old (must be >=%1).").arg(KMP_XML_MIN_VER));
        return false;
    }
    QMetaEnum me = QMetaEnum::fromType<PrmKey>();
    KeyMap map;
    bool ok = true;
    for (int i = 0; i < me.keyCount(); i++) {
        QString tag = QString(me.key(i)).toLower();
        PrmKey key = static_cast<PrmKey>(me.value(i));
        QDomElement elt = root.firstChildElement(tag);
        bool imk = ItemMapper::gMandatoryKeys.contains(key);
        if (elt.isNull()) {
            ok = false;
            if (errors) {
                if (imk)
                    errors->append(QCoreApplication::tr("<%1> mandatory tag is missing").arg(tag));
                else
                    errors->append(QCoreApplication::tr("<%1> optional tag is missing").arg(tag));
            }
        } else {
            QString value = elt.text().trimmed().toLower();
            if (value.isEmpty()) {
                ok = false;
                if (errors) {
                    if (imk)
                        errors->append(QCoreApplication::tr("<%1> mandatory tag has no value").arg(tag));
                    else
                        errors->append(QCoreApplication::tr("<%1> optional tag has no value").arg(tag));
                }
            } else {
                map[key] = value;
            }
        }
    }
    if (ok || !strict) {
        KeyMap::const_iterator i = map.constBegin();
        while (i != map.constEnd()) {
            mKeyMap[i.key()] = i.value();
            ++i;
        }
    }
    return ok;
}

/**
 * @brief Validate the content of the given XML file path.
 * @param xmlPath Path of the file to process.
 * @param errors Optional pointer to a QStringList with error messages on output (if any).
 * @return True if the file is valid, false otherwise.
 */
bool ItemMapper::validateXML(const QString &xmlPath, QStringList *errors) {
    QFile xmlFile(xmlPath);
    // canot open files.
    if (!xmlFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QCoreApplication::tr("Cannot open %1.").arg(xmlPath));
        return false;
    }
    QString instance = xmlFile.readAll();
    xmlFile.close();
    return validateXMLContent(instance, errors);
}

/**
 * @brief Validate the given content.
 * @param xmlContent XML content to validate.
 * @param errors Optional pointer to a QStringList with error messages on output (if any).
 * @return True if the content is valid, false otherwise.
 * @details
 * The XML content is valid if and only if all the expected tags are found under a \<keymap\> node and all the afferent
 *elements
 * are non-empty.
 *
 * The expected tags are derived from the meta-enum ItemMapper::PrmKey. Their order does not matter in the content.
 *
 * @internal
 * The key mapper does not use XML schema to valid the content of the XML node. We rather use the features of Qt
 *meta-enum to get the name of each enum and check them in the xml content: The validation process is tied to the source
 *code !
 */
bool ItemMapper::validateXMLContent(const QString &xmlContent, QStringList *errors) {
    QDomDocument document;
    document.setContent(xmlContent);
    QDomNode root = document.firstChildElement("keymap");
    if (root.isNull()) {
        if (errors)
            errors->append(QCoreApplication::tr("No <keymap> node found."));
        return false;
    }
    // search for version attr.
    int cfgver = root.toElement().attribute("version", "0").toInt();
    if (cfgver < KMP_XML_MIN_VER) {
        if (errors)
            errors->append(QCoreApplication::tr("keymap version too old (must be >=%1).").arg(KMP_XML_MIN_VER));
        return false;
    }

    QMetaEnum me = QMetaEnum::fromType<PrmKey>();
    bool ok = true;
    for (int i = 0; i < me.keyCount(); i++) {
        QString tag = QString(me.key(i)).toLower();
        QDomElement elt = root.firstChildElement(tag);
        if (elt.isNull()) {
            ok = false;
            if (errors)
                errors->append(QCoreApplication::tr("<%1> tag is missing").arg(tag));
        } else if (elt.text().trimmed().isEmpty()) {
            ok = false;
            if (errors)
                errors->append(QCoreApplication::tr("<%1> element has no value").arg(tag));
        }
    }
    return ok;
}

/**
 * @brief Save the internal key map to XML.
 * @return A string with the xml representation of the key map.
 */
QByteArray ItemMapper::keymapToXML(bool xmlHeader) const {
    QMetaEnum me = QMetaEnum::fromType<PrmKey>();
    QStringList elts;
    for (int i = 0; i < me.keyCount(); i++) {
        QString tag = QString(me.key(i)).toLower(), value = mKeyMap.value(static_cast<PrmKey>(me.value(i)), "");
        elts << QString("<%1>%2</%1>").arg(tag, value);
    }
    QString content;
    if (xmlHeader) {
        content = QString("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>"
                          "<keymap version=\"%1\">%2</keymap>");
    } else {
        content = QString("<keymap version=\"%1\">%2</keymap>");
    }
    return content.arg(mKeyMapVersion).arg(elts.join("")).toLatin1();
}

/// @brief Get the associated keymap file on disk (if any).
QString ItemMapper::keymapFilePath() const { return mKeymapFile; }
