// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef ITEMMAPPER_HPP
#define ITEMMAPPER_HPP

#include "globals.hpp"
#include <QtCore>

class ItemMapper : public QObject {
    Q_OBJECT
  public:
    /**
     * @brief Enumeration for special parameters to extract from data block.
     * @details
     * This enumeration covers all the special parameters the application needs to extract value from incoming data
     * block.
     */
    enum PrmKey {
        SENSOR_TIME = 0, ///< Sensor system time.
        LONGITUDE,       ///< GPS longitude coordinate.
        LATITUDE,        ///< GPS latitude coordinate.
        ALTITUDE,        ///< GPS altitude coordinate.
        NB_SATS,         ///< Number of GPS satellites.

        GPS_TIME,   ///< GPS time.
        GPS_DATE,   ///< GPS date.
        PCMON_TIME, ///< Monitoring PC system time.
        CLOCK_SYNC, ///< Clock synchronization flag.

        A_STRAP_COUNT, ///< Acquisition strap counter.
        A_PERIOD,      ///< Acquisition measurement period.
        A_NUMBER,      ///< Acquisition counter.
        F_STRAP_COUNT, ///< New-file strap counter.
        F_NUMBER,      ///< File number.
        F_DATETIME,    ///< File datetime.
        UDP_FREQUENCY, ///< UDP frequency.
        DIO_ACQ_STRAP, ///< Acquisition strap DIO status.
        SD_DETECT,     ///< SD card detection flag.
    };
    Q_ENUM(PrmKey)

    enum ErrorType {
        NO_ERRORS = 0, ///< No error: the item matches.
        NOT_FOUND,     ///< The item's parameter has no match in the mapper.
        INVALID_UNIT,  ///< The item parameter's units is not consistent with the relevant parameter in the mapper.
    };

    explicit ItemMapper(QObject *parent = nullptr);

    bool operator==(const ItemMapper &other);

    void copyFrom(const ItemMapper *mapper);

    ErrorType validate(const AParam &parameter) const;

    int keymapVersion() const;
    QString keyName(PrmKey key) const;

    AParam parameter(PrmKey key) const;
    AParam parameter(QString name) const;
    AParam parameter(QString name, Units::Unit units) const;
    AParam &updateParameter(AParam &param) const;

    double value(PrmKey key) const;
    double value(QString name) const;
    double value(const AParam &parameter) const;

    void setParameters(const QVector<AParam> &parameters = QVector<AParam>());
    void setParameters(const ABlock &descriptor);

    QVector<AParam> parameters(bool withInternals = false) const;

    QString keymapFilePath() const;
    bool loadKeymap(const QString &xmlPath, QStringList *errors = nullptr);

    QPair<bool, bool> checkKeymap(QStringList *mandatoryMissings = nullptr,
                                  QStringList *optionalMissings = nullptr) const;

    void resetValues();
    void resetKeymap();
    void resetParameters();

    void registerValues(const QVector<AParam> &parameters);
    void registerValues(const ABlock &block);
    bool validRegistration() const;

    QStringList categories(bool toUpper = false) const;
    QVector<AParam> parametersFromCategories(const QStringList &categories) const;

  private:
    bool parseContent(const QString &content, bool strict = false, QStringList *errors = nullptr);
    bool validateXML(const QString &xmlPath, QStringList *errors = nullptr);
    bool validateXMLContent(const QString &xmlContent, QStringList *errors = nullptr);
    QByteArray keymapToXML(bool xmlHeader = true) const;

  private:
    bool mLastValid;
    int mKeyMapVersion;
    QString mKeymapFile;
    QMap<PrmKey, QString> mKeyMap;
    QVector<AParam> mParams;
    static const QVector<PrmKey> gMandatoryKeys;

    bool hasParameter(const AParam &parameter) const;
    bool hasParameter(QString name) const;
    Q_DISABLE_COPY(ItemMapper)
};

/** Type definition for the map of parameters key / names. */
typedef QMap<ItemMapper::PrmKey, QString> KeyMap;

/** Type definition for the map of parameters key / values. */
typedef QMap<ItemMapper::PrmKey, AParam> ValueMap;

#endif // ITEMMAPPER_HPP
