// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "itemmodel.hpp"

#include "dataitem.hpp"
#include "itemmapper.hpp"

#include "cfg_versions.hpp"

/// @brief Constructor.
ItemModel::ItemModel(QObject *parent) : QAbstractTableModel(parent) {
    mMapItem = nullptr;
    for (int i = 0; i < N_PLOTS; i++)
        mPlotItems[i] = nullptr;
    translateHeaders();
}

/**
 * @brief Full constructor.
 * @param items List of data item to manage. The model takes ownership of the items (and reparents them).
 * @param parent Optional parent object.
 */
ItemModel::ItemModel(const QVector<DataItem *> &items, QObject *parent) : QAbstractTableModel(parent), mItems(items) {
    mMapItem = nullptr;
    for (int i = 0; i < N_PLOTS; i++)
        mPlotItems[i] = nullptr;
    mItems.removeAll(nullptr);
    for (DataItem *item : mItems)
        item->setParent(this);
    translateHeaders();
}

/// @brief Destructor.
ItemModel::~ItemModel() {
    // force deletion of items regardles of their actual QObject parent :)
    qDeleteAll(mItems);
}

/**
 * @brief Get the item flags for the given index.
 */
Qt::ItemFlags ItemModel::flags(const QModelIndex &index) const {

    Qt::ItemFlags f = QAbstractTableModel::flags(index);
    if (index.isValid()) {
        f |= Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled;
        if (index.column() >= 1) // ItemModel::mapColumn)
            f |= Qt::ItemIsEditable;
    } else {
        f |= Qt::ItemIsDropEnabled;
    }
    return f;
}

/// @brief Get the header displayed data.
QVariant ItemModel::headerData(int section, Qt::Orientation orientation, int role) const {
    if (role != Qt::DisplayRole)
        return QAbstractTableModel::headerData(section, orientation, role);
    if (orientation == Qt::Horizontal && section >= 0 && section < mHeaders.size())
        return QVariant(mHeaders.at(section));
    else if (orientation == Qt::Vertical)
        return section + 1;
    else
        return QVariant();
}

/**
 * @brief Get the model data under given index and role.
 * @param index Index to process.
 * @param role Role to process.
 * @return Data for the given role at the given index.
 * @details
 * Only Qt::DisplayRole and Qt::TooltipRole are handled by the method on the sole purpose to display the model items in
 * a tableview.
 */
QVariant ItemModel::data(const QModelIndex &index, int role) const {
    int ir = index.row(), jc = index.column();

    if (!index.isValid() || ir >= mItems.size())
        return QVariant();

    DataItem *item = mItems[ir];

    if (role == Qt::DisplayRole) {
        switch (jc) {
        case 0:
            return QString::fromStdString( item->parameter().name());
        case 1:
            return item->alias();
        case 2:
            return QString::fromStdString(Units::unitToString(item->units(), true));
        case 3:
            return QString::number(item->colorRange().lower(), 'g', 3);
        case 4:
            return QString::number(item->colorRange().upper(), 'g', 3);
        case 5: {
            double v = item->threshold();
            return qIsNaN(v) ? QString() : QString::number(v, 'g', 3);
        }
        default:
            return QVariant();
        }
    } else if (role == Qt::EditRole) {
        if (jc < ItemModel::mapColumn) {
            switch (jc) {
            case 0:
                return QString::fromStdString(item->parameter().name());
            case 1:
                return item->alias();
            case 2:
                return QVariant(item->units());
            case 3:
                return item->colorRange().lower();
            case 4:
                return item->colorRange().upper();
            case 5: {
                return item->threshold();
            }
            default:
                return QVariant();
            }
        } else { // map & plots checkstates
            if (jc == ItemModel::mapColumn) {
                return item == mMapItem;
            } else {
                return item == mPlotItems[jc - ItemModel::plotColumnStart];
            }
        }
    } else if ((jc == 0 || jc == 1) && (role == Qt::ToolTipRole || role == Qt::StatusTipRole)) {
        const AParam &p = item->parameter();
        if (role == Qt::StatusTipRole)
            return QString("%1 (%2), units: %3").arg(item->alias(), QString::fromStdString(p.name()), QString::fromStdString(Units::unitToString(p.units)));
        else if (p.description.empty()) {
            return QString("<h4 style='white-space:pre'>%1 (%2), units: %3</h4>")
                .arg(item->alias(), QString::fromStdString(p.name()), QString::fromStdString(Units::unitToString(p.units)));
        } else {
            return QString("<h4 style='white-space:pre'>%1 (%2), units: %3</h4><p>%4</p>")
                .arg(item->alias(), QString::fromStdString(p.name()), QString::fromStdString(Units::unitToString(p.units)), QString::fromStdString(p.description));
        }
    }
    return QVariant();
}

bool ItemModel::setData(const QModelIndex &index, const QVariant &value, int role) {
    int row = index.row(), col = index.column();
    if (row == -1 || role != Qt::EditRole)
        return false;
    DataItem *item = mItems[row];
    if (col < ItemModel::mapColumn) {
        switch (col) {
        case 0: // parameter: not editable yet.
            return false;
        case 1: { // alias.
            item->setAlias(value.toString());
            emit dataChanged(index, index);
            break;
        }
        case 2: { // units
            Units::Unit u = value.value<Units::Unit>();
            // setUnits will scale internal values of the indicator if possible !
            item->setUnits(u, true);
            emit dataChanged(index, index.siblingAtColumn(5));
            break;
        }
        case 3:
            item->colorRange().setLower(value.toDouble());
            item->colorRange().normalize();
            emit dataChanged(index, index.siblingAtColumn(4));
            break;
        case 4:
            item->colorRange().setUpper(value.toDouble());
            item->colorRange().normalize();
            emit dataChanged(index.siblingAtColumn(3), index);
            break;
        case 5:
            item->setThreshold(value.toDouble());
            emit dataChanged(index, index);
            break;
        }
    } else {
        bool state = value.toBool();
        // column #2 is for map:
        if (col == ItemModel::mapColumn) {
            if (state) {
                mMapItem = item;
            } else {
                mMapItem = nullptr;
            }
            emit dataChanged(this->index(0, col), this->index(rowCount() - 1, col));
            emit itemOnMapRequested(mMapItem);
        } else {
            // graphics
            int gCol = col - ItemModel::plotColumnStart;
            if (state) {
                mPlotItems[gCol] = item;
            } else {
                mPlotItems[gCol] = nullptr;
            }
            emit dataChanged(this->index(row, 0), this->index(row, columnCount() - 1));
            emit dataChanged(this->index(0, col), this->index(rowCount() - 1, col));
            emit itemOnGraphRequested(gCol, mPlotItems[gCol]);
        }
    }
    return true;
}

/**
 * @brief Add new items to manage.
 * @param items Vector of items to add.
 * @param rejected Reference to a vector with in output the items that have not been added.
 * @details
 * The method adds the given items to the model (and reparents them). Only items that are not already stored are added,
 * rejected items are returned in rejected reference. It is user responsibility to delete such items if no more used.
 */
void ItemModel::addItems(QVector<DataItem *> items, QVector<DataItem *> &rejected) {
    rejected.clear();
    // filter items:
    // - remove null pointers
    items.removeAll(nullptr);
    // - remove duplicates in the vector
    QVector<DataItem *> wrk;
    for (DataItem *item : items) {
        if (wrk.contains(item))
            rejected.push_back(item);
        else
            wrk.push_back(item);
    }
    items = wrk;
    wrk.clear();
    // - remove items already set in mItems:
    for (DataItem *item : items) {
        if (mItems.contains(item))
            rejected.push_back(item);
        else
            wrk.push_back(item);
    }
    // working vector is empty: no items to add !
    if (wrk.isEmpty())
        return;

    int ns = mItems.size();
    beginInsertRows(QModelIndex(), ns, ns + wrk.size() - 1);
    for (DataItem *item : wrk) {
        item->setParent(this);
        mItems.push_back(item);
    }
    endInsertRows();
}

/**
 * @brief Add data item to the model.
 * @param item Item to add.
 * @return Model index where the item has been added. Only the row is relevant, the column of the index is always set to
 * zero.
 * @details
 * The method may fail if the item is already found in the model or if it is null, in such case an invalid index is
 * returned. On success, the item is appended in the model which takes its ownership.
 */
QModelIndex ItemModel::addItem(DataItem *item) {
    if (!item || mItems.contains(item))
        return QModelIndex();
    int ns = mItems.size();
    beginInsertRows(QModelIndex(), ns, ns);
    // add the items to the model.
    item->setParent(this);
    mItems.push_back(item);
    endInsertRows();
    emit itemAdded(item);
    return index(ns, 0);
}

/**
 * @brief Duplicate the given item.
 * @param item DataItem to duplicate.
 * @param uniqueName True to generate a unique alias for the item.  (append " - copy" suffix). Default to false.
 * @return Model index where the item has been added. Only the row is relevant, the column of the index is always set to
 * zero.
 * @details
 * The item is inserted right after the input \a item if it belongs to the model. Otherwise it is appended to the model.
 *
 * If \a uniqueName is true, then the method appends the suffix " - copy" to the item's alias while the latter is found
 * in the item aliases of the model.
 */
QModelIndex ItemModel::duplicateItem(DataItem *item, bool uniqueName) {
    DataItem *newItem = DataItem::copyFrom(item, this);
    if (!newItem)
        return QModelIndex();
    if (uniqueName) {
        QStringList names;
        for (auto *it : mItems)
            names.append(it->alias());
        QString alias = newItem->alias();
        bool exists = names.contains(alias);
        while (exists) {
            alias = QString("%1 - copy").arg(alias);
            exists = names.contains(alias);
        }
        newItem->setAlias(alias);
    }
    int ridx = mItems.indexOf(item);
    if (ridx == -1)
        ridx = mItems.size();
    ridx++;
    beginInsertRows(QModelIndex(), ridx, ridx);
    // add the items to the model.
    mItems.insert(ridx, newItem);
    endInsertRows();
    return index(ridx, 0);
}

/**
 * @overload
 * @brief Duplicate the item at given index.
 */
QModelIndex ItemModel::duplicateItem(QModelIndex index, bool uniqueName) {
    return index.isValid() ? duplicateItem(mItems[index.row()], uniqueName) : QModelIndex();
}

/**
 * @brief Remove the item at given model index.
 * @param index Model index of the data item to remove.
 * @return True on success, false otherwise.
 * @details
 * The method removes and deletes the item at given index. If \a index is not valid the method immediately returns
 * false.
 *
 * The following signals are emitted:
 * - ItemModel::itemAboutToBeRemoved with the pointer of the item.
 * - ItemModel::itemOnMapRequested with a null pointer if the item to be deleted is currently sets as the map item.
 * - ItemModel::itemOnGraphRequested with a null pointer for each plot where the item to be deleted is currently set.
 *
 * The signals are emitted before the item is actually removed from the model and its internal data vector.
 *
 * The, registered plot widgets and map widget will be reset if they use the item.
 */
bool ItemModel::removeItem(const QModelIndex &index) {
    int ir = index.row();
    DataItem *item = (ir < 0 || ir >= mItems.size()) ? nullptr : mItems[ir];
    if (!item)
        return false;
    // emit some signals before actually remove and delete the item:
    // 1. the "about to be remove" signal.
    emit itemAboutToBeRemoved(DataItemPointer(item));
    // 2. the itemOnMapRequested signal if the item is the current map item
    //    (in such case we nullify the current map item.
    if (item == mMapItem) {
        emit itemOnMapRequested(nullptr);
        mMapItem = nullptr;
    }
    // 3. the itemOnGraphRequested signal for each plot where the item is used.
    //    (in such case we nullify the plot's item).
    for (int i = 0; i < N_PLOTS; ++i) {
        if (mPlotItems[i] == item) {
            emit itemOnGraphRequested(i, nullptr);
            mPlotItems[i] = nullptr;
        }
    }
    // now we remove the item from the model.
    beginRemoveRows(QModelIndex(), ir, ir);
    mItems.remove(ir);
    endRemoveRows();
    // and delete it (later ^^).
    item->deleteLater();
    return true;
}

/// @brief Get the total number of rows in the model.
int ItemModel::rowCount(const QModelIndex & /*parent*/) const { return mItems.size(); }

/// @brief Get the number of columns in the model.
int ItemModel::columnCount(const QModelIndex & /*parent*/) const { return mHeaders.size(); }

/// @brief translate model headers.
void ItemModel::translateHeaders() {
    mHeaders = QStringList{QCoreApplication::tr("Parameter"), // 0
                           QCoreApplication::tr("Name"),      // 1
                           QCoreApplication::tr("Units"),     // 2
                           QCoreApplication::tr("Min."),      // 3
                           QCoreApplication::tr("Max."),      // 4
                           QCoreApplication::tr("Rem."),      // 5
                           QCoreApplication::tr("Map")};      // 6
    for (int i = 1; i < N_PLOTS + 1; i++) {
        mHeaders.append(QCoreApplication::tr("G#%1").arg(i));
    }
    emit headerDataChanged(Qt::Horizontal, 0, mHeaders.size() - 1);
}

/**
 * @brief Reset the values of the model (set to NaN).
 * @details
 * The method sets all item's value to NaN and refreshes the model afterward.
 */
void ItemModel::resetValues() {
    static const double nan = qQNaN();
    for (DataItem *item : mItems) {
        item->setValue(nan);
    }
}

/**
 * @brief Build a context menu for item selection on plot widgets.
 * @param plotIndex Index of the plot widget that has requested a context menu.
 * @param parentMenu Parent menu of the menu to build.
 * @return A pointer to QMenu with the item selection actions.
 */
QMenu *ItemModel::plotContextMenu(int plotIndex, QMenu *parentMenu) {
    DataItem *checkedItem = plotItem(plotIndex);
    int rowIndex = -1, nr = mItems.size();
    QMenu *menu = new QMenu(tr("Parameters"), parentMenu);
    for (int i = 0; i < nr; i++) {
        DataItem *item = mItems[i];
        QModelIndex idx = this->index(i, plotIndex + ItemModel::plotColumnStart);
        if (item != checkedItem) {
            QAction *action = new QAction(item->alias(), menu);
            QObject::connect(action, &QAction::triggered, this,
                             [this, idx]() { this->setData(idx, true, Qt::EditRole); });
            menu->addAction(action);
        } else
            rowIndex = i;
    }
    if (rowIndex > -1) {
        QModelIndex idx = this->index(rowIndex, plotIndex + ItemModel::plotColumnStart);
        menu->addSeparator();
        QAction *action = new QAction(tr("Clear"), menu);
        action->setToolTip(tr("Remove the current parameter from the plot."));
        QObject::connect(action, &QAction::triggered, this, [this, idx]() { this->setData(idx, false, Qt::EditRole); });
        menu->addAction(action);
    }
    return menu;
}

/**
 * @brief Build a context menu for item selection on map widget.
 * @param parentMenu Parent menu of the menu to build.
 * @return A pointer to QMenu with the item selection actions.
 */
QMenu *ItemModel::mapContextMenu(QMenu *parentMenu) {

    DataItem *checkedItem = mapItem();
    int rowIndex = -1, nr = mItems.size();
    QMenu *menu = new QMenu(tr("Parameters"), parentMenu);
    for (int i = 0; i < nr; i++) {
        DataItem *item = mItems[i];
        QModelIndex idx = this->index(i, ItemModel::mapColumn);
        if (item != checkedItem) {
            QAction *action = new QAction(item->alias(), menu);
            QObject::connect(action, &QAction::triggered, this,
                             [this, idx]() { this->setData(idx, true, Qt::EditRole); });
            menu->addAction(action);
        } else
            rowIndex = i;
    }
    if (rowIndex > -1) {
        QModelIndex idx = this->index(rowIndex, ItemModel::mapColumn);
        menu->addSeparator();
        QAction *action = new QAction(tr("Clear"), menu);
        action->setToolTip(tr("Remove the current parameter from the map."));
        QObject::connect(action, &QAction::triggered, this, [this, idx]() { this->setData(idx, false, Qt::EditRole); });
        menu->addAction(action);
    }
    return menu;
}


/**
 * @brief Sort the items in the order given by the input vector.
 * @param order New index order vector.
 * @details
 * \a order vector must be of the same size than the model rows count, and must contains only in-range indexes,
 * otherwise the method returns immediately.
 */
void ItemModel::sortItems(const QVector<int> &order) {
    int nr = rowCount();
    bool ok = order.size() == nr;
    if (!ok)
        return;
    QVector<int> rows(nr);
    std::iota(rows.begin(), rows.end(), 0);
    // check that order vector is consistent.
    for (int ir : order) {
        if (!rows.contains(ir))
            return;
    }
    QVector<DataItem *> items;
    items.reserve(nr);
    for (int ir : order)
        items.push_back(mItems.at(ir));
    mItems = items;
    emit dataChanged(index(0, 0), index(nr - 1, columnCount() - 1));
}

/**
 * @brief Normalize the color ranges of items in the model.
 * @details
 * The method iterates over each item of the model and normalize (swap min and max values if needed) their color range
 * if needed.
 */
void ItemModel::normalizeRanges() {
    for (DataItem *item : mItems) {
        item->colorRange().normalize();
    }
}

void ItemModel::reOrderItem(int oldIndex, int newIndex) {
    int nr = rowCount(), nc = columnCount();
    if (oldIndex < 0 || oldIndex >= nr || newIndex < 0 || newIndex >= nr || oldIndex == newIndex)
        return;
    if (oldIndex < newIndex)
        newIndex++;
    // mItems.insert(qMin(nr - 1, newIndex), mItems.takeAt(oldIndex));
    mItems.move(oldIndex, qMin(nr - 1, newIndex));
    emit dataChanged(index(0, 0), index(nr - 1, nc - 1));
    emit headerDataChanged(Qt::Vertical, 0, nr - 1);
}

/**
 * @brief Register new values in scenario items and indicators from the given block of data.
 * @param block Block of data to process.
 * @internal
 * The parameter lookup is not very performant:
 */
void ItemModel::registerValues(const ABlock &block) {
    for (DataItem *item : mItems) {
        item->setValue(block(item->parameter(), true));
    }
}

/**
 * @brief Register new values in scenario items and indicators from the given item mapper.
 * @param ItemMapper Item mapper to search.
 */
void ItemModel::registerValues(const ItemMapper *mapper) {
    if (mapper) {
        for (DataItem *item : mItems) {
            mapper->updateParameter(item->parameter());
        }
    } else {
        qDebug() << "input mapper is null !";
    }
}

/**
 * @brief Get the data item associated to the map.
 * @return The data item associated to the map, may be null if no data item is associated.
 */
DataItemPointer ItemModel::mapItem() const { return DataItemPointer(mMapItem); }

/**
 * @brief Get the data item associated to the given plot (by index).
 * @param index Index of the plot to process (between 0 and 3, hard-coded).
 * @return The data item associated to the plot, may be null if no data item is associated.
 */
DataItemPointer ItemModel::plotItem(int index) const {
    if (index < 0 || index >= N_PLOTS)
        return nullptr;
    return DataItemPointer(mPlotItems[index]);
}

/**
 * @brief Get the row index of the current map item.
 * @return Row index or -1 if no item is registered for the map.
 */
int ItemModel::mapItemRowIndex() const { return mItems.indexOf(mMapItem); }

/**
 * @brief Get the row index of the current plot item.
 * @param index Index of the plot to process.
 * @return Row index or -1 if no item is registered for the afferent plot.
 */
int ItemModel::plotItemRowIndex(int index) const { return mItems.indexOf(mPlotItems[index]); }

/// @brief Get a XML representation of the scenario.
QByteArray ItemModel::toXML() const {
    QString content;
    QStringList prms, inds;

    for (DataItem *item : mItems) {
        bool onMap = (item == mMapItem);
        int pIdx = -1;
        for (int i = 0; i < N_PLOTS; i++) {
            if (mPlotItems[i] == item) {
                pIdx = i;
                break;
            }
        }
        prms.append(QString(item->toXML(&onMap, &pIdx)));
        // update the "initial" range and threshold to match the current ones.
        item->updateSaveThreshold();
        item->updateSaveRange();
    }
    content = QString("<parameters version=\"%1\">%2</parameters>").arg(PRM_XML_CUR_VER).arg(prms.join(""));
    return content.toLatin1();
}

/// @brief Get all items stored in the scenario.
QVector<DataItemPointer> ItemModel::items() const {
    QVector<DataItemPointer> out;
    out.reserve(mItems.size());
    for (auto ptr : mItems)
        out.push_back(DataItemPointer(ptr));
    return out;
}

/// @brief Get the item at given index.
DataItemPointer ItemModel::item(QModelIndex index) const { return index.isValid() ? mItems.at(index.row()) : nullptr; }

// DnD support

/// @brief Define the supported drop actions.
Qt::DropActions ItemModel::supportedDropActions() const { return Qt::MoveAction; }

/// @brief Define the supported drag actions.
Qt::DropActions ItemModel::supportedDragActions() const { return Qt::MoveAction; }

/// @brief Get the list of allowed MIME types.
QStringList ItemModel::mimeTypes() const { return QStringList{"ItemModel/rows.list"}; }

/// @brief Compute Mime data from a list of indexes.
QMimeData *ItemModel::mimeData(const QModelIndexList &indexes) const {
    if (indexes.isEmpty())
        return nullptr;
    QMimeData *mimeData = new QMimeData();
    QByteArray encodedData;
    QDataStream stream(&encodedData, QIODevice::WriteOnly);
    QStringList rows;
    for (const QModelIndex &index : indexes) {
        QString srow = QString::number(index.row());
        if (index.isValid() && !rows.contains(srow))
            rows << srow;
    }
    stream << rows;
    mimeData->setData("ItemModel/rows.list", encodedData);
    return mimeData;
}

/// @brief Check whether or not the model can drop data.
bool ItemModel::canDropMimeData(const QMimeData *data, Qt::DropAction action, int, int, const QModelIndex &) const {
    return data->hasFormat("ItemModel/rows.list") && action == Qt::MoveAction;
}

/**
 * @brief Drop mime data.
 * @param data Data to drop in the model.
 * @param action Drop action.
 * @param row Destination row.
 * @param column Destination column.
 * @param parent Destination index.
 * @return True on success, false otherwise.
 * @details
 * The method drops mime-data that correspond to a drag and drop operations between the items in the model.
 * It allows to move a \b single item (the first of the list of indexes built in DataItemModel::mimeData) from an index
 * to another.
 */
bool ItemModel::dropMimeData(const QMimeData *data, Qt::DropAction, int row, int, const QModelIndex &parent) {
    if (mItems.isEmpty())
        return false;
    QByteArray encodedData = data->data("ItemModel/rows.list");
    QDataStream stream(&encodedData, QIODevice::ReadOnly);
    QStringList iRows;
    stream >> iRows;
    if (iRows.isEmpty())
        return false;
    // currently only the first item.
    int iRow = iRows.first().toInt();
    int rCount = rowCount();
    if (row == -1)
        row = parent.row();
    int oRow = (row == -1) ? rCount : row;
    int ovRow = oRow;
    if (oRow > iRow)
        oRow++;

    if (iRow == oRow) // result in no-op for Qt models.
        return false;

    if (beginMoveRows(QModelIndex(), iRow, iRow, QModelIndex(), qMin(oRow, rCount))) {
        mItems.move(iRow, qBound(0, ovRow, rCount - 1)); // from QList::move doc : move(from,to) -> both from and to
                                                         // must be within the index range of the vector.
        endMoveRows();
    }
    return true;
}
