// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "layoutmanagerdialog.hpp"
#include "aspectratiopixmaplabel.hpp"
#include "layoutmanager.hpp"
#include "listview.hpp"
#include "nameeditdelegate.hpp"
#include <QDebug>

#include <QDialogButtonBox>
#include <QFileDialog>
#include <QHBoxLayout>
#include <QListView>
#include <QMessageBox>
#include <QPushButton>
#include <QVBoxLayout>

LayoutManagerDialog::LayoutManagerDialog(LayoutManager *manager, const QString &selectedName, QWidget *parent,
                                         Qt::WindowFlags f)
    : QDialog(parent, f), mManager(manager) {

    setWindowTitle(tr("Layouts manager"));

    mListView = new ListView(this);
    mListView->setModel(mManager);
    mListView->setSelectionBehavior(QListView::SelectRows);
    mListView->setSelectionMode(QListView::SingleSelection);
    NameEditDelegate *delegate = new NameEditDelegate(mListView);
    mListView->setItemDelegate(delegate);
    mListView->setSizeAdjustPolicy(QAbstractScrollArea::AdjustToContents);
    mListView->setSizePolicy(QSizePolicy::Preferred, QSizePolicy::MinimumExpanding);
    mListView->setResizeMode(QListView::Adjust);
    mListView->installEventFilter(this);

    mBtnImport = new QPushButton(tr("Import"), this);
    mBtnExport = new QPushButton(tr("Export"), this);
    mBtnClose = new QPushButton(tr("Close"), this);
    mBtnClose->setDefault(true);
    mPixmapLabel = new AspectRatioPixmapLabel(this);
    mPixmapLabel->setStyleSheet("border: 1px solid black;");

    QVBoxLayout *vlayout = new QVBoxLayout;
    vlayout->setContentsMargins(0, 0, 0, 0);
    vlayout->addWidget(mListView, 1);
    vlayout->addWidget(mBtnImport, 0);
    vlayout->addWidget(mBtnExport, 0);
    vlayout->addWidget(mBtnClose, 0);

    QHBoxLayout *hlayout = new QHBoxLayout(this);
    hlayout->addLayout(vlayout, 0);
    hlayout->addWidget(mPixmapLabel, 1);

    QObject::connect(mBtnImport, &QPushButton::clicked, this, &LayoutManagerDialog::importLayouts);
    QObject::connect(mBtnExport, &QPushButton::clicked, this, &LayoutManagerDialog::exportLayouts);
    QObject::connect(mBtnClose, &QPushButton::clicked, this, &LayoutManagerDialog::close);

    QItemSelectionModel *ism = mListView->selectionModel();
    if (ism)
        QObject::connect(ism, &QItemSelectionModel::selectionChanged, this, &LayoutManagerDialog::onSelectionChanged);
    // load UI geometry if any.
    QSettings settings;
    QByteArray array = settings.value("layouts/layout_diag/geometry").toByteArray();
    if (array.size())
        restoreGeometry(array);
    int idx = manager->indexOf(selectedName);
    if (idx != -1 && ism)
        ism->select(mManager->index(idx, 0), QItemSelectionModel::ClearAndSelect);
}

void LayoutManagerDialog::closeEvent(QCloseEvent *event) {
    QSettings settings;
    settings.setValue("layouts/layout_diag/geometry", saveGeometry());
    QDialog::closeEvent(event);
}

/// @brief Get the size hint of the widget.
QSize LayoutManagerDialog::sizeHint() const {
    return minimumSizeHint();
    QSize sz(1024, 768);
    if (parentWidget())
        sz = parentWidget()->size() * 0.85;
    qDebug() << sz;
    return sz;
}

void LayoutManagerDialog::onSelectionChanged(const QItemSelection &, const QItemSelection &) {
    auto *m = mListView->selectionModel();
    if (!(m && m->hasSelection())) {
        mPixmapLabel->resetPixmap();
        return;
    }
    QModelIndex idx = m->selectedIndexes().first();
    mPixmapLabel->setPixmap(idx.data(LayoutManager::ScreenshotRole).value<QPixmap>());
}

/**
 * @brief Import layouts from file to the layout manager.
 */
void LayoutManagerDialog::importLayouts() {
    QString iniDir = QStandardPaths::standardLocations(QStandardPaths::DesktopLocation).first();
    QString f1 = tr("Layout files (*.layout)"), f2 = tr("All files (*.*)"), fs = f1;
    QString fileName =
        QFileDialog::getOpenFileName(this, tr("Select layouts file"), iniDir, QString("%1;;%2").arg(f1, f2), &fs);
    if (fileName.isEmpty())
        return;
    bool ok;
    QVector<Layout> layouts = LayoutManager::readFromFile(fileName, &ok);
    mManager->addLayouts(layouts);
}

/**
 * @brief Export the layout of the layout manager to file.
 */
void LayoutManagerDialog::exportLayouts() {
    QString iniDir = QStandardPaths::standardLocations(QStandardPaths::DesktopLocation).first();
    QString f1 = tr("Layout files (*.layout)"), f2 = tr("All files (*.*)"), fs = f1;
    QString fileName =
        QFileDialog::getSaveFileName(this, tr("Save layouts to file"), iniDir, QString("%1;;%2").arg(f1, f2), &fs);
    if (fileName.isEmpty())
        return;
    if (fs == f1 && !fileName.endsWith(".layout"))
        fileName.append(".layout");
    mManager->saveToFile(fileName);
}

bool LayoutManagerDialog::eventFilter(QObject *object, QEvent *event) {
    if (object == mListView && event->type() == QEvent::KeyPress) {
        QKeyEvent *keyEvent = static_cast<QKeyEvent *>(event);
        bool ok = keyEvent->key() == Qt::Key_Delete;
        auto m = mListView->selectionModel();
        if (ok && m && m->hasSelection()) {
            auto idx = m->selectedIndexes().first();
            mManager->removeRow(idx.row());
            return true;
        }
    }
    return QDialog::eventFilter(object, event);
}
