// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "layoutselector.hpp"
#include "layoutmanager.hpp"
#include "layoutmanagerdialog.hpp"
#include "layoutnamedialog.hpp"
#include "mainwindow.hpp"
#include <QComboBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QPushButton>

LayoutSelector::LayoutSelector(QWidget *parent) : QWidget{parent} {
    mManager = nullptr;
    mLabel = new QLabel(this);
    mBtnSaveLayout = new QPushButton(QIcon::fromTheme("save"), "", this);
    mBtnEditLayout = new QPushButton(QIcon::fromTheme("edit"), "", this);

    mBtnEditLayout->setStyleSheet("border : none");
    mBtnSaveLayout->setStyleSheet("border : none");

    mCbxSelectLayout = new QComboBox(this);
    mCbxSelectLayout->setMinimumWidth(70);
    mCbxSelectLayout->setSizeAdjustPolicy(QComboBox::AdjustToContents);

    QHBoxLayout *layout = new QHBoxLayout(this);
    layout->addWidget(mLabel);
    layout->addWidget(mCbxSelectLayout);
    layout->addWidget(mBtnEditLayout);
    layout->addWidget(mBtnSaveLayout);

    QObject::connect(mBtnSaveLayout, &QPushButton::clicked, this, &LayoutSelector::saveCurrentLayout);
    QObject::connect(mBtnEditLayout, &QPushButton::clicked, this, &LayoutSelector::editLayouts);
    translateUI();
}

/// @brief Get the combobox.
QComboBox *LayoutSelector::combobox() const { return mCbxSelectLayout; }

void LayoutSelector::onRowAboutToBeRemoved(const QModelIndex &parent, int first, int last) {
    int idx = mCbxSelectLayout->currentIndex();
    if (idx >= first && idx <= last)
        mCbxSelectLayout->setCurrentIndex(-1);
}

/// @brief Register the given layout manager as the internal combobox's model.
void LayoutSelector::setLayoutManager(LayoutManager *manager) {
    if (mManager) {
        QObject::disconnect(mManager, &QAbstractListModel::rowsAboutToBeRemoved, this,
                            &LayoutSelector::onRowAboutToBeRemoved);
        QObject::disconnect(mCbxSelectLayout, qOverload<int>(&QComboBox::activated), mManager,
                            qOverload<int>(&LayoutManager::loadLayout));
    }
    mManager = manager;
    if (mManager) {
        QObject::connect(mManager, &QAbstractListModel::rowsAboutToBeRemoved, this,
                         &LayoutSelector::onRowAboutToBeRemoved);
        QObject::connect(mCbxSelectLayout, qOverload<int>(&QComboBox::activated), mManager,
                         qOverload<int>(&LayoutManager::loadLayout));
    }
    mCbxSelectLayout->setModel(mManager);
}

/// @brief Translate UI texts.
void LayoutSelector::translateUI() { mLabel->setText(tr("Layout")); }

/// @brief Save the current layout in the layout manager.
void LayoutSelector::saveCurrentLayout() {
    if (mManager) {
        QString current = mCbxSelectLayout->currentText();
        QString name = LayoutNameDialog::getLayoutName(current, mManager->layoutNames());
        if (name.isEmpty())
            return;
        int idx = mManager->saveLayout(name);
        if (idx != -1) {
            QSignalBlocker blocker(mCbxSelectLayout);
            mCbxSelectLayout->setCurrentIndex(idx);
        }
    }
}

void LayoutSelector::editLayouts() {
    if (mManager) {
        LayoutManagerDialog *dialog =
            new LayoutManagerDialog(mManager, mCbxSelectLayout->currentText(), mManager->window());
        dialog->setAttribute(Qt::WA_DeleteOnClose);
        dialog->exec();
    }
}
