// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "loginview.hpp"

#include "ArcGISRuntimeEnvironment.h"
#include "Credential.h"
#include "Portal.h"
#include "TaskWatcher.h"
#include <AuthenticationChallenge.h>
#include <AuthenticationManager.h>

using namespace Esri::ArcGISRuntime;

void LoginView::onAuthenticationChallenge(AuthenticationChallenge *challenge) {

    if (challenge->error().errorType() != ErrorType::Success)
        qDebug() << challenge->error().message() << challenge->failureCount();
    if (challenge->authenticationChallengeType() == AuthenticationChallengeType::UsernamePassword && challenge->failureCount() > 1) {
        challenge->cancel();
        return;
    }
    if (challenge->authenticationChallengeType() == AuthenticationChallengeType::UsernamePassword)
        challenge->continueWithUsernamePassword(mLdtUserName->text(), mLdtPassword->text());
}
/// @brief Constructor.
LoginView::LoginView(const QString customMessage, bool withCheckBox, QWidget *parent)
    : QDialog(parent, Qt::Dialog | Qt::WindowTitleHint | Qt::WindowCloseButtonHint | Qt::WindowStaysOnTopHint) {
    mAutoClose = true;
    mTimer.setSingleShot(true);
    mCursorOverride = false;
    mIsFetching = false;
    mLblInfos = new QLabel(this);
    mLblResult = new QLabel(this);
    mLdtUserName = new QLineEdit(this);
    mLdtPassword = new QLineEdit(this);
    mLdtPassword->setEchoMode(QLineEdit::Password); // QLineEdit::PasswordEchoOnEdit);
    mBtnConnect = new QPushButton(this);

    mChbWarn = withCheckBox ? new QCheckBox(this) : nullptr;

    mLdtUserName->setFixedWidth(200);
    mLdtPassword->setFixedWidth(200);
    mBtnConnect->setFixedWidth(200);

    for (QLabel *lbl : {mLblInfos, mLblResult}) {
        lbl->setWordWrap(true);
        lbl->setMaximumWidth(600);
    }

    QVBoxLayout *layout = new QVBoxLayout(this);
    layout->addWidget(mLblInfos, 1);
    layout->addWidget(mLdtUserName, 0, Qt::AlignCenter);
    layout->addWidget(mLdtPassword, 0, Qt::AlignCenter);
    layout->addWidget(mBtnConnect, 0, Qt::AlignCenter);
    layout->addWidget(mLblResult);
    layout->addStretch(0);
    if (mChbWarn)
        layout->addWidget(mChbWarn);

    // set texts
    setWindowTitle(tr("Esri sign-in"));
    QString msg = tr("<p>Sign-in on your Esri account to access ARCGIS online resources.</p>");
    if (!customMessage.isEmpty()) {
        msg = QString("%1%2").arg(msg, customMessage);
    }
    mLblInfos->setText(msg);
    mLdtUserName->setPlaceholderText(tr("Esri Account name"));
    mLdtPassword->setPlaceholderText(tr("Esri Account password"));
    mBtnConnect->setText(tr("Sign-in"));
    mLblResult->setText(tr("Info: No operation in progress."));
    if (mChbWarn)
        mChbWarn->setText(tr("Do not show again"));
    QSettings settings;
    QString uname = settings.value("global/last_esri_user", "").toString();
    if (!uname.isEmpty())
        mLdtUserName->setText(uname);

    mCredentials = new Credential(this);
    mPortal = new Portal(true, this);
    QObject::connect(mBtnConnect, &QPushButton::clicked, this, &LoginView::tryLogin);
    QObject::connect(AuthenticationManager::instance(), &AuthenticationManager::authenticationChallenge, this,
                     &LoginView::onAuthenticationChallenge);
    QObject::connect(mPortal, &Portal::loadStatusChanged, this, &LoginView::onPortalLoadStatusChanged);
}

LoginView::~LoginView() { QObject::disconnect(AuthenticationManager::instance(), nullptr, this, nullptr); }

/**
 * @brief Close event.
 * @details
 * Prevents the dialog to close if the license if currently being fetched.
 */
void LoginView::closeEvent(QCloseEvent *event) {
    if (mChbWarn) {
        QSettings settings;
        settings.setValue("warnings/no_arcgis_signin", !mChbWarn->isChecked());
    }
    if (mIsFetching) {
        event->ignore();
    } else {
        QObject::disconnect(&mTimer);
        if (mCursorOverride)
            QApplication::restoreOverrideCursor();
        cancelTask();
        QDialog::closeEvent(event);
    }
}

/// @brief Get the size hint of the dialog.
QSize LoginView::sizeHint() const { return QSize(520, minimumHeight()); }

/**
 * @brief Enable/Disable auto close mode.
 * @param enable True to enable auto close mode. False otherwise.
 * @details
 * The method enables/disables the auto close mode. If enabled, the dialog will automatically close upon successfull
 * sign-in.
 *
 * by default, the mode is enabled.
 */
void LoginView::autoCloseOnSignIn(bool enable) { mAutoClose = enable; }

/**
 * @brief Try to sign in to ArsGIS onlin using current credentials.
 * @details
 * The method attempts to fetch a new license using a Portal with the credentials sets in the line edits.
 *
 * The method requires an internet connection. Then it implements a timeout that will stop the task after
 * 15 seconds which is, I believe, sufficient to retrieve the license.
 */
void LoginView::tryLogin() {
    if (mPortal->loadStatus() == LoadStatus::Loading)
        return;
    mLblResult->setText(tr("Info: Trying to acquire token from ArcGIS server."));
    mCredentials->setUserAccount(mLdtUserName->text(), mLdtPassword->text());
    mPortal->setCredential(mCredentials);
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    mCursorOverride = true;
    mIsFetching = true;
    // generate a timeout: in 5 secs, the task will be canceled.
    QObject::connect(&mTimer, &QTimer::timeout, this, &LoginView::cancelTask);
    mTimer.start(10000);
    mPortal->load();
}

void LoginView::onPortalLoadStatusChanged(LoadStatus loadStatus) {
    bool finished = loadStatus == LoadStatus::Loaded || loadStatus == LoadStatus::FailedToLoad;
    if (!finished)
        return;
    mTimer.stop();
    if (mCursorOverride)
        QApplication::restoreOverrideCursor();
    mCursorOverride = false;
    mIsFetching = false;
    QSettings settings;
    settings.setValue("global/last_esri_user", mLdtUserName->text());
    bool ok = loadStatus == LoadStatus::Loaded;
    if (ok) {
        mLblResult->setText(tr("Info: User successfully signed in."));
    } else {
        auto error = mPortal->loadError();
        qDebug() << error.message() << error.additionalMessage() << (int)error.errorType();
        mLblResult->setText(tr("Info: Cannot sign-in (bad credentials ?)."));
    }
    if (mAutoClose || ok)
        QTimer::singleShot(2000, this, &LoginView::close);
}

/// @brief Cancel the fetch task (if any).
void LoginView::cancelTask() {
    mTimer.stop();
    if (mCursorOverride)
        QApplication::restoreOverrideCursor();
    if (mPortal->loadStatus() == LoadStatus::Loading) {
        mPortal->cancelLoad();
    }
    mCursorOverride = false;
    mIsFetching = false;
    mLblResult->setText(tr("Info: Cannot sign-in (operation timeout)."));
}
