// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef MAINWINDOW_HPP
#define MAINWINDOW_HPP

#include "globals.hpp"

#include "admconfig.hpp"

using namespace ISP;

class AUDPSocket;
class DataManager;
class BaseIndicator;
class ScreenshotManager;
class ItemMapper;
class ScenarioWidget;
class RtPlotWidget;
class FitPlotWidget;
class MapWidget;
class FlightWidget;
class TimestampWidget;
class FilenameWidget;
class IndicatorsWidget;
class GPSWidget;
class CpuMemWidget;
class MediaControlWidget;
class InversionStatusWidget;
class CommWidget;
class ClockWidget;
class BatteryStatusWidget;
class LayoutManager;
class LayoutSelector;
class ScenarioSelector;
class FitController;

class MainWindow : public QMainWindow {
    Q_OBJECT

  public:
    MainWindow(
        const QString &configPath = QString(), QWidget *parent = nullptr, Qt::WindowFlags flags = Qt::WindowFlags());
    virtual ~MainWindow() override;
    virtual QSize sizeHint() const override;
    void translateUI();

    void centerToScreen();
    MapWidget *mapWidget() const;
  public slots:
    void openSettings();

    void lockUI(bool locked);

  protected:
    virtual void closeEvent(QCloseEvent *event) override;
    virtual void changeEvent(QEvent *event) override;

  private slots:
    void editScenarios(int current);
    bool saveScenarios();
    void openConfiguration();
    void updateFromSettings();
    void captureScreen();
    void aboutApp();
    void clearData();
    void incomingBlock(ABlock block); // -> no const ref we need to modify the block.

    void reportInvalidTemplate(int instr, int vers);
    void enableInversion(bool enable);
    void openHelp();

    void onBlockProcessed(const ABlock &block, const QVector<FitReport> &reports, const ExecTimes &times);

    void switchFullscreen();
    void onNewFlightCreated();
    void onMapperUpdated();
    void onHeaderClick(int index);

    void switchModes(bool replay);
    void loadReplayData();

    void refreshTexts();
    void onResetTimeout();

  private:
    AdmConfig mAdmConfig;

    AUDPSocket *mUdpSocket;
    QString mLastDataRootDir;
    QString mCaptureDirectory;
    bool mInversionEnabled;
    bool mInvalidTemplateReported;
    bool mShowFit;
    ScreenshotManager *mScreenshotManager;

    LayoutManager *mLayoutManager;

    FCParameters mFCParams;
    FitConfig mFitConfig;
    FitController *mFitController;
    DataManager *mDataManager;
    QVector<BaseIndicator *> mInstrumentIndicators;

    BinFileHandler *mResultFileHandler;
    bool mSaveData;
    bool mNewFileOnNewFlight;
    bool mMultiThreaded;

    bool mNoSDWarned;
    bool mNoRefresh;
    QTimer mRefreshTimer;
    int mResetInterval;
    QTimer mResetTimer;

    // child widgets.
    ClockWidget *mClockWidget;
    BatteryStatusWidget *mBatteryWidget;

    // Map widget
    MapWidget *mMapWidget;
    QTabWidget *mTabWidget;
    QVector<FitPlotWidget *> mFitWidgets;

    // other widgets are docked !

    // Plot widgets
    RtPlotWidget *mPlotWidgets[N_PLOTS];
    QVector<QDockWidget *> mPlotDockWidgets; // associated dockwidget.
    int mHeaderOffset;

    // Scenario parameter view widget
    FlightWidget *mFlightWidget;
    // Scenario parameter view widget
    ScenarioWidget *mScenarioWidget;
    // Flight table view
    QTableView *mFlightTableView;

    // Resources monitor widget
    CpuMemWidget *mCpuMemWidget;

    // Wifi & udp widget
    CommWidget *mCommWidget;

    // Inversion status
    InversionStatusWidget *mInversionWidget;

    // Movie controls panel.
    MediaControlWidget *mMediaControlWidget;
    QDockWidget *mDockMCW;
    // indicators

    // GPS
    GPSWidget *mGPSWidget;
    // timestamps
    TimestampWidget *mTimestampWidget;
    // filenames
    FilenameWidget *mFilenameWidget;
    // Instrument status
    IndicatorsWidget *mInstrumentWidget;
    // "User-defined" parameter indicators
    IndicatorsWidget *mIndicatorsWidget;

    // selectors
    LayoutSelector *mLayoutSelector;

#ifndef SCN_SEL_ON_MAP
    ScenarioSelector *mScenarioSelector;
    QAction *mActionScenario;
#endif

    // actions
    QAction *mActionLayout;
    QAction *mActionClock;
    QAction *mActionBattery;
    QAction *mActionLockUI;
    QAction *mActionLoad;
    QAction *mActionQuit;
    QAction *mActionSettings;
    QAction *mActionToggleInv;
    QAction *mActionClear;
    QAction *mActionReplay;
#ifdef O_FULLSCREEN
    QAction *mActionFullScreen;
#endif
    QAction *mActionScreenshot;
    QAction *mActionAbout;
    QAction *mActionHelp;

    // Toolbar
    QToolBar *mToolBar;
    // menus
    QMenu *mFileMenu;
    QMenu *mViewMenu;
    QMenu *mViewPlotSubMenu;
    QMenu *mViewIndicatorsSubMenu;
    QMenu *mHelpMenu;

  private:
    // UI

    void setupUI();
    void loadUI();
    void saveUI();
    void createActions();

    void setMultiThreadMode(bool enable);
    // configuration
    void loadConfiguration(const QString &configPath);

    bool initResultFileHandler(QDateTime startTime);
    void freeResultFileHandler();

    void setupFitWidgets();
    QString computeSensorFileName();
    bool saveScenariosOnTheFly();
    void checkSDCard(bool sdDetected);
  signals:
    /**
     * @brief Fit workers cancellation signal.
     * @details
     * The signal is used to instruct current fit processes (in the threadpool) to abort.
     * It is emitted only in MainWindow::abortFit and should not be used anywhere else.
     * @todo
     * In practice, fit workers take less than a second to complete (actually it depends on how many spectras must be
     * processed).
     * I may think:
     *
     * - about removing the cancellation process in the worker. It will remove an extra mutex.
     * - use a more consumer/producer approach that work only on the last available block (i.e. most fresh data).
     */
    void abortFitWorkers();
};

#endif // MAINWINDOW_HPP
