// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef MAPWIDGET_HPP
#define MAPWIDGET_HPP

#include "datarange.hpp"
#include "globals.hpp"

class ScenarioSelector;
class DataManager;
class DataItem;
class ColorScale;
class BaseRtPlotWidget;
struct BasemapInfos;
class BasemapSelector;
class MapItemEditor;

namespace Esri {
namespace ArcGISRuntime {
class Map;

class Basemap;
class GraphicsOverlay;
class LatitudeLongitudeGrid;
enum class LoadStatus;
} // namespace ArcGISRuntime
} // namespace Esri

#include "MapGraphicsView.h"
#include "SpatialReference.h"
#include "Viewpoint.h"

class MapWidget : public QWidget {
    Q_OBJECT
  public:
    explicit MapWidget(QWidget *parent = nullptr);

    void loadTilePackage(const QString &path);
    void scrollXRange(double reference);
    void refreshPlot();

    void setColorGradient(const QCPColorGradient &gradient);

    void saveUI();
    void loadUI();
    void translateUI();

    QByteArray splitterState() const;
    void setSplitterState(const QByteArray &state);

    QPointer<BaseRtPlotWidget> plotWidget() const;
    QPointer<Esri::ArcGISRuntime::MapGraphicsView> mapView() const;

    void enableParameterSelection(bool enable);
    bool parameterSelectionEnabled() const;

#ifdef SCN_SEL_ON_MAP
    ScenarioSelector *selector() const;
    QAction *selectorAction() const;
#endif
    virtual QSize sizeHint() const override;

    void clearGraphsAndOverlays();
    void addOverlay(Esri::ArcGISRuntime::GraphicsOverlay *overlay);
    void removeOverlay(Esri::ArcGISRuntime::GraphicsOverlay *overlay);

    void addGraph(GraphContainerPointer data);
    void removeGraph(GraphContainerPointer data);
    void updateFromItem(DataItem *item, QVector<Esri::ArcGISRuntime::GraphicsOverlay *> overlays,
                        GraphContainerVector graphContainers);

    void centerMapOn(double latitude, double longitude);
    QAction *autoCenterAction();

    Esri::ArcGISRuntime::SpatialReference spatialReference() const;

    double zoomFactor() const;
    void setZoomFactor(double zoomFactor);

    void updateFromSettings();

    void addMapInteractionsActions(QMenu *menu);
  public slots:
    void setGridVisible(bool visible);
  signals:

    void contextMenuRequested(const QPoint &pos, bool isMap);
    void colorRangeChanged(const DataRange &range);
    void thresholdChanged(double value);
    void spatialReferenceChanged(Esri::ArcGISRuntime::SpatialReference);

  private slots:

    void setPositionText(double x, double y);
    void centerOnLastPosition();
    void rescaleMap();

    void onCScaleRangeChanged(const DataRange &range);
    void onIERangeChanged(const DataRange &range);
    void onBPRangeChanged(const DataRange &range);

    void onIEThresholdChanged(double value);
    void onBPThresholdChanged(double value);

    void loadStatusChanged(Esri::ArcGISRuntime::LoadStatus status);
    void onBasemapSelectionChanged(const BasemapInfos &selected);

    void zoomIn();
    void zoomOut();

    void onLockToggled(bool checked);
    void onResetClicked();

  private:
    DataItem *mCurrentItem;

    QSplitter *mSplitter; // main widget

    // map & color bar
    Esri::ArcGISRuntime::Map *mMap = nullptr;
    Esri::ArcGISRuntime::MapGraphicsView *mMapView = nullptr;
    Esri::ArcGISRuntime::LatitudeLongitudeGrid *mMapGrid = nullptr;
    Esri::ArcGISRuntime::SpatialReference mSpatialReference;
    ColorScale *mColorScale;
    MapItemEditor *mItemEditor;
    // plot
    BaseRtPlotWidget *mBaseRtPlotWidget;

    // tool bar & actions
    QToolBar *mToolBar;
    QAction *mMARescale;
    QAction *mMACenter;
    QAction *mMAZoomIn;
    QAction *mMAZoomOut;
    QAction *mMABasemaps;
    QAction *mMALock;
    QAction *mMAReset;

#ifdef SCN_SEL_ON_MAP
    ScenarioSelector *mScenarioSelector;
    QAction *mActionScenario;
#endif

    BasemapSelector *mBasemapsWidget;

    Esri::ArcGISRuntime::Viewpoint mLastViewpoint;
    double mZoomFactor;

    bool mEnableParamSelection;
    bool mShowThreshold;
    bool mShowRange;

  private:
    void setupActions();
    void setupGrid();
    void setBasemap(Esri::ArcGISRuntime::Basemap *basemap);
    void zoom(double factor);
};

#endif // MAPWIDGET_HPP
