// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "networkconfig.hpp"

/// @brief Constructor.
NetworkConfig::NetworkConfig(const QString &ip, quint16 port, const QString &ssid) : ipAddr(ip), port(port), ssidName(ssid) {}

/**
 * @brief Check whether or not the configuration is valid.
 * @return True if valid, false otherwise.
 * @details
 * A network configuration is considered valid if:
 *   - its IP address is not empty.
 *   - Its port is greater than 49151 .
 */
bool NetworkConfig::isValid() const { return port > 49151 && !ipAddr.isEmpty(); }

/**
 * @brief Check whether or not the two configuration are equals.
 * @param other Other configuration to check.
 * @return True if both instances stores the same members values (include ssid name), False otherwise.
 */
bool NetworkConfig::operator==(const NetworkConfig &other) const {
    return ipAddr.compare(other.ipAddr, Qt::CaseInsensitive) == 0 && port == other.port &&
           ssidName.compare(other.ssidName) == 0;
}

/**
 * @brief Check whether or not the two configuration are different.
 * @param other Other configuration to check.
 * @return False if both instances stores the same members values (include ssid name), True otherwise.
 * @details
 * The operator is the strict opposite of NetworkConfig::operator==.
 */
bool NetworkConfig::operator!=(const NetworkConfig &other) const { return !(*this == other); }

/**
 * @brief Check whether or not this configuration and the other are comparable.
 * @return True if both instances are comparable, false otherwise.
 * @details
 * Unlike the equality/inequality operators, the method only checks for equality of the IP address and port.
 */
bool NetworkConfig::comparable(const NetworkConfig &other) const {
    return ipAddr.compare(other.ipAddr, Qt::CaseInsensitive) == 0 && port == other.port;
}

