// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "prm_utils.hpp"

#include <list>

static QVector<QVector<int>> groupedIndexes(const QDoubleVector &input, const QDoubleVector &output);
int diffRspPoints(double timeStep, const DoubleVector &input, double bref, double eref, int maxPoints);

static inline double isNaN(double v) { return qIsNaN(v); };
static inline double isNotNaN(double v) { return !qIsNaN(v); };

double qnanmin(QDoubleVector::iterator begin, QDoubleVector::iterator end);
double qnanmax(QDoubleVector::iterator begin, QDoubleVector::iterator end);

/**
 * @brief Reduce \a values to a scalar.
 * @param prm Parameter associated to the value. It instructs the method which kind of operation to perform on the input
 * vector.
 * @param values Vector of values to reduce.
 * @param ignoreNaN True (default) to ignore NaN in \a values.
 * @return Reduced value.
 * @note
 * If \a ignoreNaN is set to False the results of the operations is systematically NaN for the operations: MIN, MAX, SUM
 * and MEAN.
 */
double reduceData(const AParam &prm, const DoubleVector &values, bool ignoreNaN) {
    double out = qQNaN();
    DoubleVector tmp;
    const DoubleVector *ptr = nullptr;
    // 1. filter nan data if we ignore NaN.
    if (ignoreNaN) {
        tmp.resize(values.size());
        auto it = std::copy_if(values.begin(), values.end(), tmp.begin(), [](double v) { return !qIsNaN(v); });
        tmp.erase(it, tmp.end());
    } else {
        ptr = &values;
    }
    if (ptr->empty()) {
        return out;
    }
    // 2. apply operation
    switch (prm.reduction) {
    case Reduction::FIRST:
        out = ptr->front();
        break;
    case Reduction::LAST:
        out = ptr->back();
        break;
    case Reduction::MEAN:
        out = std::accumulate(ptr->begin(), ptr->end(), 0) / ptr->size();
        break;
    case Reduction::SUM:
        out = std::accumulate(ptr->begin(), ptr->end(), 0);
        break;
    case Reduction::MIN:
        out = *std::min_element(ptr->begin(), ptr->end());
        break;
    case Reduction::MAX:
        out = *std::max_element(ptr->begin(), ptr->end());
        break;
    default:
        break;
    }
    // 3. return
    return out;
}

/**
 * @brief Reduce \a values to a scalar.
 * @param operation Reduction operation to perform.
 * @param values Vector of values to reduce.
 * @param ignoreNaN True (default) to ignore NaN in \a values.
 * @return Reduced value.
 * @note
 * If \a ignoreNaN is set to False the result is systematically NaN for the operations: MIN, MAX, SUM and
 * MEAN.
 */
double reduceData(
    Reduction operation, const QDoubleVector::Iterator &begin, const QDoubleVector::Iterator &end, bool ignoreNaN) {
    double out = qQNaN();
    DoubleVector tmp;
    switch (operation) {
    case Reduction::FIRST: {
        if (ignoreNaN) {
            auto it = std::find_if_not(begin, end, isNaN);
            if (it != end)
                out = (*it);
        } else
            out = (*begin);
        break;
    }
    case Reduction::LAST: {
        if (ignoreNaN) {
            auto it = std::find_if_not(std::make_reverse_iterator(end), std::make_reverse_iterator(begin), isNaN);
            if (it != std::make_reverse_iterator(begin))
                out = (*it);
        } else {
            out = *(end - 1);
        }
        break;
    }
    case Reduction::MEAN: {
        if (ignoreNaN) {
            out = std::accumulate(begin, end, 0.0, [](const double &a, const double &b) {
                return qIsNaN(b) ? a : (a + b);
            }) / std::count_if(begin, end, isNotNaN);
        } else {
            out = std::accumulate(begin, end, 0.0) / std::distance(begin, end);
        }
        break;
    }
    case Reduction::SUM: {
        if (ignoreNaN) {
            out = std::accumulate(
                begin, end, 0.0, [](const double &a, const double &b) { return qIsNaN(b) ? a : (a + b); });
        } else {
            out = std::accumulate(begin, end, 0);
        }
        break;
    }
    case Reduction::MIN: {
        if (ignoreNaN) {
            out = qnanmin(begin, end);
        } else {
            out = *std::min_element(begin, end);
        }
        break;
    }
    case Reduction::MAX: {
        if (ignoreNaN) {
            out = qnanmax(begin, end);
        } else {
            out = *std::max_element(begin, end);
        }
        break;
    }
    default:
        break;
    }
    // 3. return
    return out;
}

/**
 * @brief Resample the input data.
 * @param data Map of data to resample.
 * @param timeParam Time parameter. It must be found in \a data, and its values should be defined in seconds.
 * @param frequency Output frequency in hertz.
 * @param operations Optional reduction operations to perform.
 * @param stripNaN Indicate whether or not remove output NaN produced by the method. Default to false.
 * @return True on success, false otherwise.
 * @details
 * The method resamples the values of \a data to the given \a frequency. On the first error, the method exits with \a
 * false. In such case, \a data is not modified.
 *
 * The algorithm is as follow:
 *
 * 1. Search for input time line using \a timeParam. If not found, the method exits.
 * 2. Check for input sizes (both time line and \a data map) and frequency.
 * 3. Build output time line using the minimum and maximum input time line values (rounded down and up resp.) and the \a
 *    frequency.
 * 4. Build groups of indexes of input timeline over the output time line.
 * 5. Apply reduction operation as defined by \a operations (or as set in the \a data key parameters) on the each group
 * of indexes.
 *
 * @note
 * The method assumes that the time line is defined in seconds: no tests is performed on the AParam units.
 */
// bool resample(QMap<AParam, DoubleVector> &data, const AParam &timeParam, double frequency, bool stripNaN) {

//     bool okTime = data.contains(timeParam);
//     if (!okTime)
//         return false;

//     DoubleVector &time_in = data[timeParam];
//     int ni = time_in.size();

//     if (data.isEmpty() || dblCompare(frequency, 0.0) || !ni)
//         return false;

//     // filter nan values in time (should never happen but who actually knows ?)
//     if (hasNaN(time_in)) {
//         QVector<int> goods;
//         goods.reserve(ni);
//         for (int i = 0; i < ni; i++) {
//             if (qIsNaN(time_in.at(i)))
//                 continue;
//             goods.push_back(i);
//         }
//         goods.squeeze();
//         ni = goods.size();
//         QMap<AParam, DoubleVector>::iterator it = data.begin();
//         while (it != data.end()) {
//             DoubleVector tmp(it.value());
//             DoubleVector &values = it.value();
//             values.resize(ni);
//             for (int i = 0; i < ni; i++)
//                 values[i] = tmp.at(goods.at(i));
//             it++;
//         }
//     }
//     time_in = data[timeParam];
//     DoubleVector time_ou;
//     std::list<QPair<int, int>> groups = timePairIndexes(time_in, frequency, time_ou,stripNaN);

//     int nout = time_ou.size();

//     QMap<AParam, DoubleVector>::iterator itMap = data.begin();
//     DoubleVector tmp(nout);

//     while (itMap != data.end()) {
//         const AParam &p = itMap.key();
//         DoubleVector &values = itMap.value();
//         if (p == timeParam) {
//             values = time_ou;
//         } else {
//             int i = 0;
//             for (auto it = groups.begin(); it != groups.end(); it++, i++) {
//                 int first = it->first, second = it->second;
//                 tmp[i] = reduceData(p.reduction, values.begin() + first, values.begin() + second);
//             }
//             values = tmp;
//         }
//         itMap++;
//     }
//     return true;
// }

/**
 * @brief Decimate the data map.
 * @param data Map of data to process.
 * @param step Decimation step.
 * @return True on success, false otherwise. The method fails only if \a data is empty or if data's values are empty.
 */
bool decimate(QMap<AParam, DoubleVector> &data, int step) {
    QMap<AParam, DoubleVector>::iterator it = data.begin();
    if (it == data.end())
        return false;
    int ni = it.value().size();
    if (!ni)
        return false;
    step = qMax(1, step);
    int no = qMax(1, ni / step);
    DoubleVector tmp(no);
    while (it != data.end()) {
        DoubleVector &values = it.value();
        for (int i = 0, j = 0; i < ni; i += step, j++) {
            tmp[j] = values.at(i);
        }
        values = tmp;
    }
    return true;
}

/**
 * @brief Get the minimum resample time step that genrates at most \a maxPoints on the output time line.
 * @param maxPoints Maximum number of points allowed.
 * @param input Input time line.
 * @param avgTimestep Optional pointer to a double with the input average time step.
 * @return Minimum resample time.
 * @details
 * The method uses the bisection method to find the minimum resample time step that, used in ::timePairIndexes,
 * will generate at most \a maxPoints on the output groups list.
 *
 * If given, \a avgTimestep is used as a hint to compute a proper range for bisection. Otherwise, the step is computed
 * internally from \a input vector.
 *
 * @note
 * The method may return a resample step that produces slightly less points than \a maxPoints .
 */
double getMinRspStep(int maxPoints, const DoubleVector &input, double *avgTimestep) {
    int ni = input.size();
    double avgstep;
    if (avgTimestep)
        avgstep = *avgTimestep;
    else {
        for (int i = 0; i < ni - 1; i++)
            avgstep = input[i + 1] - input[i];
        avgstep /= (ni - 1);
    }
    if (ni <= maxPoints)
        return avgstep;
    double bref, eref, frac;
    frac = std::modf(input.front(), &bref);
    frac = std::modf(input.back(), &eref);
    double a = avgstep, b = (a * ni) / maxPoints, c = a;
    double e = 1e-3 * avgstep;
    int o = -qMax(50, maxPoints / 100);

#define func(x) static_cast<double>(diffRspPoints(x, input, bref, eref, maxPoints))

    while (func(b) > 0)
        b *= 1.5;
    double fa = func(a), fc;
    while ((b - a) >= e) {
        c = (a + b) / 2;
        fc = func(c);
        // Check root: we accept zero but also a little less (up to o)
        if (fc == 0.0 || (fc < 0 && fc > o))
            break;
        if (fc * fa < 0) {
            b = c;
        } else {
            a = c;
            fa = fc;
        }
    }
#undef func
    return c;
}

/**
 * @brief Compute a new time line from given <i>input</i> and <i>timeStep</i>.
 * @param input Input time line (assumed to be in seconds and already sorted in ascending order).
 * @param timeStep New timestep in seconds.
 * @param output Output time line.
 * @param noEmpty Control flag that instructs to remove empty groups. Default to true.
 * @return Vector of indexes from the input timeline on the output timeline.
 * @todo
 * Document the method algorithm :)
 */
std::list<QPair<int, int>> timePairIndexes(
    const QDoubleVector &input, double timeStep, QDoubleVector &output, bool noEmpty) {
    std::list<QPair<int, int>> iout;
    if (timeStep <= 0.0) {
        int ni = input.size();
        for (int i = 0; i < ni; i++)
            iout.push_back({i, i + 1});
        output = input;
        return iout;
    }
    double bref, eref, frac;
    // get the integral part of the starting/ending times of the line.
    frac = std::modf(input.front(), &bref);
    frac = std::modf(input.back(), &eref);
    QVector<double> tmp;
    double v = bref;
    while (v < eref) {
        tmp.push_back(v);
        v += timeStep;
    }
    int nout = tmp.size();

    std::list<double> outlist;
    int i, ilast = 0;
    for (int j = 0; j < nout; j++) {
        QPair<int, int> cp{-1, -1};
        int sz = 0;
        // last item.
        if (j == nout - 1) {
            cp.first = ilast;
            cp.second = input.size();
            if (cp.first == cp.second && noEmpty)
                cp.first = -1;
        } else {
            // other items.
            cp.first = ilast;
            for (i = ilast; i < input.size(); i++) {
                if (input[i] <= tmp[j + 1]) {
                    sz++;
                } else {
                    cp.second = i;
                    break;
                }
            }
            if (sz == 0) {
                cp.first = (noEmpty) ? -1 : (cp.first - 1);
            }
            ilast = i;
        }
        if (cp.first != -1) {
            iout.push_back(cp);
            outlist.push_back(tmp[j] + timeStep / 2.0);
        }
    }
    output.resize(outlist.size());
    std::copy(outlist.begin(), outlist.end(), output.begin());
    return iout;
}

/**
 * @brief Group \a input data indexes on \a output data intervals.
 * @param input Input data values.
 * @param output Output data values.
 * @return Vector of grouped indexes.
 * @details
 * The method generates groups of indexes that belong to \a input for which afferent values lies between two adjacent
 * values of \a output. The method assumes \a output values to be sorted in ascending order.
 *
 * The returned vector has the same size than \a output. Values from \a input lower than the first value of \a output
 * are always ignored. The last item of the returned vector contains all indexes of \a input for which values are
 * greater or equal to the last value of \a output.
 *
 * @warning
 * The method assumes all values of \a output are non-NaN. Nan values may lead to unexpected behaviour.
 */
static QVector<QVector<int>> groupedIndexes(const QDoubleVector &input, const QDoubleVector &output) {
    int nin = input.size();
    int nou = output.size();
    QVector<QVector<int>> indexes;
    std::list<std::pair<int, double>> tmp;
    double omin = output.first();
    for (int i = 0; i < nin; ++i) {
        const double &vin = input.at(i);
        if (qIsNaN(vin) || vin < omin)
            continue;
        tmp.push_back({i, vin});
    }
    indexes.resize(nou);
    for (int j = 0; j < nou - 1; j++) {
        QVector<int> &idxs = indexes[j];
        idxs.reserve(qMin(1, static_cast<int>(tmp.size()) / (nou - (j + 1))));
        const double &vou = output[j], voup1 = output[j + 1];
        auto it = tmp.begin();
        while (it != tmp.end()) {
            const double &vin = it->second;
            if (vin >= vou && vin < voup1) {
                idxs.push_back(it->first);
                it = tmp.erase(it);
            } else {
                it++;
            }
        }
    }
    // last group : with all our assumptions all remainging items in tmp should be copied !
    QVector<int> &idxs = indexes[nou - 1];
    idxs.reserve(tmp.size());
    for (auto it = tmp.begin(); it != tmp.end(); ++it)
        idxs.push_back(it->first);
    for (QVector<int> &idxs : indexes)
        idxs.squeeze();
    return indexes;
}

int diffRspPoints(double timeStep, const DoubleVector &input, double bref, double eref, int maxPoints) {
    int ni = input.size();
    if (timeStep <= 0.0)
        return ni - maxPoints;
    int icount = 0, i = 0, ilast = 0, sz = 0;
    double vtime = bref;
    while (vtime < eref) {
        vtime += timeStep;
        sz = 0;
        for (i = ilast; i < ni; i++) {
            if (input[i] > vtime)
                break;
            sz++;
        }
        if (sz)
            icount++;
        ilast = i;
        if (ilast >= ni)
            break;
    }
    return icount - maxPoints;
}
/**
 * @brief Get the minimum value of the input vector ignoring NaN.
 * @param begin Iterator to the start of the range.
 * @param end Iterator to the end of the range (exclusive).
 * @return Minimum non-NaN value.
 * @note
 * The method may still return NaN, if the vector contains only NaN values !
 */
double qnanmin(QDoubleVector::iterator begin, QDoubleVector::iterator end) {
    double out = qQNaN();
    for (auto it = begin; it != end; ++it) {
        if (qIsNaN(*it))
            continue;
        if (qIsNaN(out) || out > (*it))
            out = (*it);
    }
    return out;
}
/**
 * @brief Get the maximum value of the input vector ignoring NaN.
 * @param begin Iterator to the start of the range.
 * @param end Iterator to the end of the range (exclusive).
 * @return Maximum non-NaN value.
 * @note
 * The method may still return NaN, if the vector contains only NaN values !
 */
double qnanmax(QDoubleVector::iterator begin, QDoubleVector::iterator end) {
    double out = qQNaN();
    for (auto it = begin; it != end; ++it) {
        if (qIsNaN(*it))
            continue;
        if (qIsNaN(out) || out < (*it))
            out = (*it);
    }
    return out;
}
