// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "scenariosmodel.hpp"
#include "scenario.hpp"

/// @brief Constructor.
ScenariosModel::ScenariosModel(QObject *parent) : QAbstractTableModel(parent) { translateHeaders(); }

/// @brief Destructor.
ScenariosModel::~ScenariosModel() { qDeleteAll(mScenarios); }

/// @brief Get all the scenarios stored in the model.
QVector<ScenarioPointer> ScenariosModel::scenarios() const {
    QVector<ScenarioPointer> out;
    out.reserve(mScenarios.size());
    for (auto ptr : mScenarios)
        out.push_back(ScenarioPointer(ptr));
    return out;
}

/**
 * @brief Set new scenarios in the model.
 * @param scenarios Scenarios to set.
 * @details
 * The method replace the current scenarios of the model by the given ones.
 * @note
 * The model takes ownership of the scenarios which are also reparented to the model.
 */
void ScenariosModel::setScenarios(const QVector<Scenario *> &scenarios) {
    beginResetModel();
    mScenarios = scenarios;
    for (auto ptr : mScenarios)
        ptr->setParent(this);
    endResetModel();
}

/**
 * @brief Get the scenario at given index.
 * @param index Index to process.
 * @return Scenario or null pointer if index is not valid.
 */
ScenarioPointer ScenariosModel::scenario(QModelIndex index) const {
    return index.isValid() ? mScenarios[index.row()] : nullptr;
}

/**
 * @brief Add scenario to the model.
 * @param scenario Scenario to add.
 * @return Model index where the scenario has been added. Only the row is relevant, the column of the index is always set to
 * zero.
 * @details
 * The method may fail if the scenario is already found in the model or if it is null, in such case an invalid index is
 * returned.
 * On success, the scenario is appended in the model which takes its ownership.
 */
QModelIndex ScenariosModel::addScenario(Scenario *scenario) {
    if (!scenario || mScenarios.contains(scenario))
        return QModelIndex();
    int nr = rowCount();
    beginInsertRows(QModelIndex(), nr, nr);
    scenario->setParent(this);
    mScenarios.push_back(scenario);
    endInsertRows();
    return index(nr, 0);
}

/**
 * @brief Duplicate the given scenario.
 * @param scenario Scenario to duplicate.
 * @param updateName True to update the scenario name (append "- Copy" suffix). Default to false.
 * @return Model index associated to the new scenario. Only the row is relevant, the column of the index is always set to zero.
 * @details
 * The scenario is inserted right after the input \a item if it belongs to the model. Otherwise it is appended to the model.
 *
 * If \a uniqueName is true, then the method appends the suffix " - copy" to the item's alias while the latter is found in the
 * item aliases of the model.
 *
 * If \a uniqueName is true, then the method appends the suffix " - copy" to the scenario name while the latter is found in the
 * scenario names of the model.
 */
QModelIndex ScenariosModel::duplicateScenario(Scenario *scenario, bool uniqueName) {
    Scenario *scn = Scenario::copyFrom(scenario, this);
    if (!scn)
        return QModelIndex();
    if (uniqueName) {
        QStringList names;
        for (auto *it : mScenarios)
            names.append(it->name());
        QString name = scn->name();
        bool exists = names.contains(name);
        while (exists) {
            name = QString("%1 - copy").arg(name);
            exists = names.contains(name);
        }
        scn->setName(name);
    }
    int ridx = mScenarios.indexOf(scenario);
    if (ridx == -1)
        ridx = mScenarios.size();
    ridx++;
    beginInsertRows(QModelIndex(), ridx, ridx);
    mScenarios.insert(ridx, scn);
    endInsertRows();
    return index(ridx, 0);
}

/**
 * @overload
 * @brief Duplicate the scenario at given index.
 */
QModelIndex ScenariosModel::duplicateScenario(QModelIndex index, bool uniqueName) {
    return index.isValid() ? duplicateScenario(mScenarios[index.row()], uniqueName) : QModelIndex();
}

/**
 * @brief Remove and delete the scenario at given index from the model.
 * @param index Model index of the scenario to remove. Index's column is irrelevant, only row is used.
 * @return True on success, false otherwise.
 * @details
 * The method removes the scenario located at the given index. If \a index is invalid, the method does nothing.
 *
 * The method emits ScenariosModel::scenarioAboutToBeDeleted before actual scenario removal.
 */
bool ScenariosModel::removeScenario(QModelIndex index) {
    int rowIndex = index.row();
    if (rowIndex < 0 || rowIndex >= mScenarios.size())
        return false;
    Scenario *scn = mScenarios[rowIndex];
    emit scenarioAboutToBeDeleted(scn);
    beginRemoveRows(QModelIndex(), rowIndex, rowIndex);
    mScenarios.removeAt(rowIndex);
    endRemoveRows();
    scn->deleteLater();
    return true;
}

/// @brief Get model flags for the given index.
Qt::ItemFlags ScenariosModel::flags(const QModelIndex &index) const {
    Qt::ItemFlags f = QAbstractTableModel::flags(index);
    if (index.isValid()) {
        f |= Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled;
    } else {
        f |= Qt::ItemIsDropEnabled;
    }
    return f;
}

/**
 * @brief Get the data of the model.
 * @param index Index to process.
 * @param role Data role. Currently only Qt::DisplayRole is implemented.
 * @return Data. The variant can be safely converted to string.
 */
QVariant ScenariosModel::data(const QModelIndex &index, int role) const {
    if (!index.isValid())
        return QVariant();
    int row = index.row(), col = index.column();
    Scenario *scn = mScenarios.at(row);
    if (role == Qt::DisplayRole) {
        switch (col) {
        case 0:
            return scn->name();
        case 1:
            return scn->comments();
        case 2:
            return scn->itemsModel()->rowCount();
        case 3:
            return scn->indicatorsModel()->rowCount();
        default:
            return QVariant();
        }
    }
    return QVariant();
}

/**
 * @brief Set data in the model.
 * @param index Index to process.
 * @param value Value to set.
 * @param role Data role. Currently only Qt::EditRole is allowed.
 * @return True on success, false otherwise.
 * @details
 * The method allows to set new names and/or comments for internal scenarios.
 * @note
 * Trying to setData with a column index greater than one does not modify internally the scenario but the method still emit
 * dataChanged signal. Consequently, the number of items/indicators should be updated in views if needed.
 */
bool ScenariosModel::setData(const QModelIndex &index, const QVariant &value, int role) {
    if (!(index.isValid() && role == Qt::EditRole))
        return false;
    int column = index.column(), row = index.row();
    Scenario *scn = mScenarios.at(row);
    if (column == 0)
        scn->setName(value.toString());
    else if (column == 1)
        scn->setComments(value.toString());
    emit dataChanged(index, index);
    return true;
}

/// @brief Get the header data at given section, orientation and role.
QVariant ScenariosModel::headerData(int section, Qt::Orientation orientation, int role) const {
    if (role != Qt::DisplayRole)
        return QAbstractTableModel::headerData(section, orientation, role);
    if (orientation == Qt::Horizontal && section >= 0 && section < mHeaders.size())
        return QVariant(mHeaders.at(section));
    else if (orientation == Qt::Vertical)
        return section + 1;
    else
        return QVariant();

    //    if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
    //        return mHeaders.at(section);
    //    else if (orientation == Qt::Vertical && role == Qt::DisplayRole)
    //        return section;
    //    return QVariant();
}

/// @brief Get the number of rows in the model.
int ScenariosModel::rowCount(const QModelIndex & /*parent*/) const { return mScenarios.size(); }

/// @brief Get the number of columns in the model.
int ScenariosModel::columnCount(const QModelIndex & /*parent*/) const { return mHeaders.size(); }

/// @brief Translate headers.
void ScenariosModel::translateHeaders() { mHeaders = QStringList{tr("Name"), tr("Comments"), tr("Items"), tr("Indicators")}; }

// DnD support

/// @brief Define the supported drop actions.
Qt::DropActions ScenariosModel::supportedDropActions() const { return Qt::MoveAction; }

/// @brief Define the supported drag actions.
Qt::DropActions ScenariosModel::supportedDragActions() const { return Qt::MoveAction; }

/// @brief Get the list of allowed MIME types.
QStringList ScenariosModel::mimeTypes() const { return QStringList{"ScenariosModel/rows.list"}; }

/// @brief Compute Mime data from a list of indexes.
QMimeData *ScenariosModel::mimeData(const QModelIndexList &indexes) const {
    if (indexes.isEmpty())
        return nullptr;
    QMimeData *mimeData = new QMimeData();
    QByteArray encodedData;
    QDataStream stream(&encodedData, QIODevice::WriteOnly);
    QStringList rows;
    for (const QModelIndex &index : indexes) {
        QString srow = QString::number(index.row());
        if (index.isValid() && !rows.contains(srow))
            rows << srow;
    }
    stream << rows;
    mimeData->setData("ScenariosModel/rows.list", encodedData);
    return mimeData;
}

/// @brief Check whether or not the model can drop data.
bool ScenariosModel::canDropMimeData(const QMimeData *data, Qt::DropAction action, int, int, const QModelIndex &) const {
    return data->hasFormat("ScenariosModel/rows.list") && action == Qt::MoveAction;
}

/**
 * @brief Drop mime data.
 * @param data Data to drop in the model.
 * @param action Drop action.
 * @param row Destination row.
 * @param column Destination column.
 * @param parent Destination index.
 * @return True on success, false otherwise.
 * @details
 * The method drops mime-data that correspond to a drag and drop operations between the items in the model.
 * It allows to move a \b single item (the first of the list of indexes built in DataItemModel::mimeData) from an index to
 * another.
 */
bool ScenariosModel::dropMimeData(const QMimeData *data, Qt::DropAction, int row, int, const QModelIndex &parent) {
    if (mScenarios.isEmpty())
        return false;
    QByteArray encodedData = data->data("ScenariosModel/rows.list");
    QDataStream stream(&encodedData, QIODevice::ReadOnly);
    QStringList iRows;
    stream >> iRows;
    if (iRows.isEmpty())
        return false;
    // currently only the first item.
    int iRow = iRows.first().toInt();
    int rCount = rowCount();
    if (row == -1)
        row = parent.row();
    int oRow = (row == -1) ? rCount : row;
    int ovRow = oRow;
    if (oRow > iRow)
        oRow++;
    if (iRow == oRow) // result in no-op for Qt models.
        return false;
    if (beginMoveRows(QModelIndex(), iRow, iRow, QModelIndex(), qMin(oRow, rCount))) {
        mScenarios.move(iRow, qBound(0, ovRow, rCount - 1)); // from QList::move doc : move(from,to) -> both from and to must be
                                                             // within the index range of the vector.
        endMoveRows();
    }
    return true;
}
