// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "scenariowidget.hpp"
#include "booleandelegate.hpp"
#include "scenario.hpp"
#include "scenarioselector.hpp"

ScenarioWidget::ScenarioWidget(QWidget *parent) : QWidget(parent) {
    mEmptyProxyModel = new ItemROModel(this);
    mEmptyProxyModel->setSourceModel(new ItemModel(this));
    mEmptyProxyModel->setDynamicSortFilter(false);

    mScenarioSelector = new ScenarioSelector(this);
    mScenarioSelector->layout()->setContentsMargins(0, 0, 0, 0);
    mScenarioSelector->combobox()->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);
    mTableView = new QTableView(this);
    mTableView->setSortingEnabled(false);

    mTableView->verticalHeader()->hide();
    mTableView->setSelectionBehavior(QAbstractItemView::SelectRows);
    mTableView->setSelectionMode(QAbstractItemView::NoSelection); // QAbstractItemView::SingleSelection);

    mTableView->setFocusPolicy(Qt::NoFocus);

    // add checkbox for the map and plots columns

    int mapIndex = 1; // DataItemModel::mapColumn; // if full model.
    int pltIndex = 2; // DataItemModel::plotColumnStart; // if full model.

    mTableView->setItemDelegateForColumn(mapIndex, new BooleanDelegate(mTableView));
    for (int i = 0; i < N_PLOTS; i++) {
        mTableView->setItemDelegateForColumn(i + pltIndex, new BooleanDelegate(mTableView));
    }
    mTableView->setMouseTracking(true);

    mTableView->setModel(mEmptyProxyModel);
    setupHeaders();

    QVBoxLayout *vlayout = new QVBoxLayout(this);

    vlayout->setContentsMargins(4, 4, 4, 4);
    vlayout->setSpacing(4);

    vlayout->addWidget(mScenarioSelector, 0);
    vlayout->addWidget(mTableView, 2);

    translateUI();
}

ScenarioWidget::~ScenarioWidget() {}

void ScenarioWidget::setupHeaders() {

    // setup column resizing policy: all column except the first (stretched) resize to its content.
    QHeaderView *header = mTableView->horizontalHeader();
    header->setSectionsMovable(false);
    header->setSectionResizeMode(QHeaderView::ResizeToContents);
    header->setSectionResizeMode(0, QHeaderView::Stretch);
}

/// @brief Get the internal parameter table view.
QTableView *ScenarioWidget::tableView() const { return mTableView; }

/// @brief Get the internal scenario selector.
ScenarioSelector *ScenarioWidget::selector() const { return mScenarioSelector; }

/// @brief Translate widgets texts.
void ScenarioWidget::translateUI() {
    mScenarioSelector->translateUI();
    qobject_cast<ItemModel *>(mEmptyProxyModel->sourceModel())->translateHeaders();
}

/**
 * @brief Set new scenarios.
 * @param scenarios Scenarios to process.
 * @param selectedIndex Optional current index to set in the combobox after scenarios have been processed. Default to
 * -1.
 * @details
 * The method sets new combobox items using the given scenarios.
 * Then it blocks combobox signals and sets the current index of the combobox to selectedIndex.
 */
void ScenarioWidget::setupSelector(const QVector<Scenario *> &scenarios, int selectedIndex) {
    mScenarioSelector->setup(scenarios, selectedIndex);
}

/**
 * @brief Set new item model for the internal table view.
 * @param model Item model to set.
 */
void ScenarioWidget::setItemModel(ItemModel *model) {
    QAbstractItemModel *oldModel = mTableView->model();
    QItemSelectionModel *m = mTableView->selectionModel();
    if (!model) {
        mTableView->setModel(mEmptyProxyModel);
    } else {
        ItemROModel *newModel = new ItemROModel(this);
        newModel->setSourceModel(model);
        mTableView->setModel(newModel);
    }
    if (oldModel && oldModel != mEmptyProxyModel)
        oldModel->deleteLater();
    if (m)
        m->deleteLater();
}
