// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "screenshotmanager.hpp"
#include "globals.hpp"

#include <QDir>
#include <QPixmap>
#include <QScreen>

QPixmap grabScreens() {
    QList<QScreen *> screens = QGuiApplication::screens();
    QList<QPixmap> scrs;
    int w = 0, h = 0, p = 0;

    foreach (auto scr, screens) {
        QRect g = scr->geometry();
        QPixmap pix = scr->grabWindow(0, g.x(), g.y(), g.width(), g.height());
        w += pix.width();
        h = qMax(h, pix.height());
        scrs.append(pix);
    }

    QPixmap final(w, h);
    QPainter painter(&final);
    final.fill(Qt::black);
    foreach (auto scr, scrs) {
        painter.drawPixmap(QPoint(p, 0), scr);
        p += scr.width();
    }

    return final;
}

/**
 * @brief Constructor
 * @param widget Widget to process for screen management.
 * @param folder Output image folder.
 * @param fileName Output filename.
 * @param enableCounter Whether or not append an increment to the output filename.
 */
ScreenshotManager::ScreenshotManager(QWidget *widget, QString folder, QString fileName) {
    Q_ASSERT_X(widget, "ScreenshotManager::ScreenshotManager", "Widget cannot be null");
    mWidget = widget;
    mAutoDatetime = true;
    setFolder(folder);
    setFileName(fileName);
}

/**
 * @brief Save a peculiar window as a PNG file.
 * @param winId Optional window ID of the window to save, if not given, the entire screen is saved.
 */
bool ScreenshotManager::saveScreen(WId winId) {
    QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
    if (mWidget->isFullScreen())
        winId = mWidget->winId();

    QScreen *screen = getActiveScreen(mWidget);
    QPixmap pixmap = screen->grabWindow(winId);

    // - Save this picture
    QString filePath = generateFilePath();
    bool ret = filePath.isEmpty() ? false : pixmap.save(filePath);
    QApplication::restoreOverrideCursor();
    return ret;
}

/**
 * @brief Set folder for file save.
 * @param folder Folder to set.
 * @return True on success, false otherwise.
 * @details
 * The method saves the directory path for file save. If the path does not exist, the method attempts to create the directory.
 * @attention
 * If the method fails, all subsequent calls to ScreenshotManager::saveScreen will fail too.
 */
bool ScreenshotManager::setFolder(QString folder) {
    mFolder = folder;
    QDir dir(mFolder);
    if (!dir.exists())
        dir.mkpath(".");
    mFolder = dir.absolutePath();
    return dir.exists();
}

/**
 * @brief Set new file name.
 * @param fileName New file name to set.
 * @details
 * The final name of the file saved to disk is computed from fileName with an additional number as suffix if the counter is
 * enabled (see ScreenshotManager::counterEnabled). The extension png is automatically added to the file.
 */
void ScreenshotManager::setFileName(QString fileName) { mFileName = fileName.replace('/', '_'); }

/// @brief Enable/Disable the datetime suffix in the filename.
void ScreenshotManager::enableDatetimeSuffix(bool enable) { mAutoDatetime = enable; }

/// @brief Check whether or not the datetime is added in the filename.
bool ScreenshotManager::datetimeSuffixEnabled() const { return mAutoDatetime; }

/**
 * @brief Generate a file path based on the current settings of the manager.
 * @return A stirng with the file (full) path or an empty string if the file path could not be generated.
 * @details
 * The method fails (that is returns an empty string) if the folder is not set or does not exist.
 */
QString ScreenshotManager::generateFilePath() {
    static const QString format = "png";
    QDir dir(mFolder);
    if (mFolder.isEmpty() || !dir.exists())
        return "";
    QString timestamp;
    if (mAutoDatetime)
        timestamp = QDateTime::currentDateTimeUtc().toString("_yyyy-MM-dd-HH'h'mm'm'ss's'");
    QString basename = QString("%1%2%3.%4").arg(mFileName, timestamp, "%1", format);
    QString filename = basename.arg("");
    int counter = 0;
    while (dir.exists(filename)) {
        counter++;
        filename = basename.arg(QString("_%1").arg(counter));
    }
    return dir.absoluteFilePath(filename);
}
