// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "graphpage.hpp"
#include "colorchooserlabel.hpp"
#include "qcp_helpers.hpp"

/**
 * @brief Constructor
 * @param parent Parent widget.
 * @details
 * The constructor only builds up the UI.
 * Additionally, GraphStyleWidget::loadSettings() and GraphStyleWidget::translateUI() should be called to
 * populate the widgets with values and set texts.
 */
GraphPage::GraphPage(QWidget *parent) : BasePage(parent) {
    mPageTitle = tr("Graphs");
    mPageIcon = getIcon("graph");
    buildUI();
    translateUI();
}

/// @brief Initialize widgets of the plot group box.
void GraphPage::buildUI() {

    mButtonGroup = new QButtonGroup(this);
    QFormLayout *flayout = new QFormLayout(this);

    mChbHideToolbar = new QCheckBox(this);
    flayout->addRow(mChbHideToolbar);
    mChbHideXLabel = new QCheckBox(this);
    flayout->addRow(mChbHideXLabel);

    flayout->addRow(getSeparator(this));

    mLblASPeriod = new QLabel(this);
    mSpbASPeriod = new QSpinBox(this);
    mSpbASPeriod->setRange(40, 2000);
    mSpbASPeriod->setSingleStep(10);
    mSpbASPeriod->setSuffix(tr(" ms"));
    mSpbASPeriod->setAccelerated(true);
    flayout->addRow(mLblASPeriod, mSpbASPeriod);
    mLblTimeWindow = new QLabel(this);
    mSpbTimeWindow = new QSpinBox(this);
    mSpbTimeWindow->setRange(1, 30);
    mSpbTimeWindow->setSuffix(tr(" min"));
    mSpbTimeWindow->setAccelerated(true);
    flayout->addRow(mLblTimeWindow, mSpbTimeWindow);
    mLblTimeOffset = new QLabel(this);
    mSpbTimeOffset = new QSpinBox(this);
    mSpbTimeOffset->setRange(50, 60000);
    mSpbTimeOffset->setSuffix(tr(" ms"));
    mSpbTimeOffset->setSingleStep(10);
    mSpbTimeOffset->setStepType(QSpinBox::AdaptiveDecimalStepType);
    mSpbTimeOffset->setAccelerated(true);
    flayout->addRow(mLblTimeOffset, mSpbTimeOffset);
    mLblTextBox = new QLabel(this);
    mCbxTextBox = new QComboBox(this);
    mCbxTextBox->addItems({
        tr("Top left"),
        tr("Top center"),
        tr("Top right"),
        tr("Bottom left"),
        tr("Bottom center"),
        tr("Bottom right"),
    });
    flayout->addRow(mLblTextBox, mCbxTextBox);

    flayout->addRow(getSeparator(this));

    mLblPlotType = new QLabel(this);
    mBtnPLine = new QRadioButton(this);
    mBtnPScatter = new QRadioButton(this);
    mButtonGroup->addButton(mBtnPLine, 0);
    mButtonGroup->addButton(mBtnPScatter, 1);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mBtnPLine);
    hlayout->addWidget(mBtnPScatter);
    flayout->addRow(mLblPlotType, hlayout);
    hlayout = new QHBoxLayout;
    mLblPLineSize = new QLabel(this);
    mSpbPLineSize = new QDoubleSpinBox(this);
    flayout->addRow(mLblPLineSize, mSpbPLineSize);
    hlayout = new QHBoxLayout;
    mLblPSymbol = new QLabel(this);
    mCbxPSymbol = new QComboBox(this);
    Plotting::fillMarkerStyleComboBox(mCbxPSymbol);
    mSpbPSymbolSize = new QDoubleSpinBox(this);
    mSpbPSymbolSize->setRange(4.0, 24.0);
    mSpbPSymbolSize->setValue(6.0);
    mSpbPSymbolSize->setSingleStep(1.0);
    mSpbPSymbolSize->setSuffix(" px");
    mSpbPSymbolSize->setAccelerated(true);
    hlayout->addWidget(mCbxPSymbol);
    hlayout->addWidget(mSpbPSymbolSize);
    flayout->addRow(mLblPSymbol, hlayout);

    mChbThreshold = new QCheckBox(this);
    mLblThColor = new QLabel(this);
    mLblThCC = new ColorChooserLabel(this);
    mLblThSize = new QLabel(this);
    mSpbThLineSize = new QDoubleSpinBox(this);

    flayout->addRow(getSeparator(this));
    flayout->addRow(mChbThreshold);
    flayout->addRow(mLblThColor, mLblThCC);
    flayout->addRow(mLblThSize, mSpbThLineSize);

    // setup width spinboxes properties.
    for (QDoubleSpinBox *s : {mSpbPLineSize, mSpbThLineSize}) {
        s->setRange(0.5, 5.0);
        s->setValue(1.0);
        s->setSingleStep(0.1);
        s->setSuffix(" px");
        s->setAccelerated(true);
    }

    QObject::connect(mChbThreshold, &QCheckBox::toggled, [=](bool checked) {
        mLblThCC->setHighlight(checked);
        mLblThColor->setEnabled(checked);
        mLblThSize->setEnabled(checked);
        mSpbThLineSize->setEnabled(checked);
    });

    QObject::connect(mCbxPSymbol, qOverload<int>(&QComboBox::currentIndexChanged), [=](int) {
        if (mBtnPLine->isChecked()) {
            mLastLMS = mCbxPSymbol->currentData().toInt();
        } else {
            mLastSMS = mCbxPSymbol->currentData().toInt();
        }
    });

    QObject::connect(mButtonGroup, &QButtonGroup::idToggled, this, &GraphPage::updatePlotWidgets);
}

/**
 * @brief Update graph line connection style as a function of the checked radio button.
 * @param id Identifier of the button
 * @param checked Check state.
 * @details
 * If the linestyle button checkstate has changed, the method enables/disables the line dash pattern and width widgets.
 *
 * If the scatterstyle button checkstate has changed, the method update the scatter style combobox selected value if
 * it is set to none.
 */
void GraphPage::updatePlotWidgets(int id, bool checked) {

    // connection by line selected.
    if (id == 0) {
        mSpbPLineSize->setEnabled(checked);
    }
    if (id == 1) {
        int firstData = mCbxPSymbol->itemData(0).toInt();
        bool old = mCbxPSymbol->blockSignals(true);
        if (checked) {
            if (firstData == 0)
                mCbxPSymbol->removeItem(0);
        } else {
            if (firstData != 0)
                mCbxPSymbol->insertItem(0, Plotting::getScatterIcon(0), Plotting::getScatterName(0), 0);
        }
        mCbxPSymbol->blockSignals(old);
    }
    if (mBtnPLine->isChecked())
        mCbxPSymbol->setCurrentIndex(mCbxPSymbol->findData(mLastLMS));
    if (mBtnPScatter->isChecked())
        mCbxPSymbol->setCurrentIndex(mCbxPSymbol->findData(mLastSMS));
}

/// @brief Load plot settings from configuration file.
void GraphPage::loadSettings() {
    QSettings settings;

    settings.beginGroup("monitoring");
    mSpbASPeriod->setValue(settings.value("asinterval", 200).toInt());
    mSpbTimeWindow->setValue(settings.value("wtime", 300).toInt() / 60);
    mSpbTimeOffset->setValue(settings.value("woffset", 100).toInt());
    settings.endGroup();

    settings.beginGroup("graph");

    mChbHideToolbar->setChecked(settings.value("hide_toolbar", false).toBool());
    mChbHideXLabel->setChecked(settings.value("hide_xlabel", false).toBool());

    mCbxTextBox->setCurrentIndex(settings.value("box_pos", 2).toInt());
    mBtnPLine->setChecked(settings.value("wline", true).toBool());
    mSpbPLineSize->setValue(settings.value("lwidth", 1.0).toDouble());
    int ms = settings.value("mstyle", 0).toInt();
    mCbxPSymbol->setCurrentIndex(qMax(0, mCbxPSymbol->findData(ms)));
    mSpbPSymbolSize->setValue(settings.value("msize", 6.0).toDouble());

    mChbThreshold->setChecked(settings.value("show_thl", true).toBool());
    mLblThCC->setColor(QColor(settings.value("thl_color", "#ff0000").toString()));
    mSpbThLineSize->setValue(settings.value("thl_width", 1.0).toDouble());
    settings.endGroup();

    if (mBtnPLine->isChecked()) {
        mLastLMS = ms;
        mLastSMS = 1;
    } else {
        mLastLMS = 0;
        mLastSMS = ms;
    }
}

/// @brief Save plot settings in configuration file.
void GraphPage::saveSettings() {
    QSettings settings;

    settings.beginGroup("graph");

    settings.setValue("hide_toolbar", mChbHideToolbar->isChecked());
    settings.setValue("hide_xlabel", mChbHideXLabel->isChecked());

    settings.setValue("now_as_ref", true); // not editable yet !
    settings.setValue("asinterval", qBound(40, mSpbASPeriod->value(), 2000));
    settings.setValue("wtime", qBound(60, mSpbTimeWindow->value() * 60, 1800));
    settings.setValue("woffset", qBound(50, mSpbTimeOffset->value(), 5000));
    settings.setValue("box_pos", mCbxTextBox->currentIndex());

    settings.setValue("wline", mBtnPLine->isChecked());
    settings.setValue("lwidth", mSpbPLineSize->value());
    settings.setValue("mstyle", mCbxPSymbol->currentData().toInt());
    settings.setValue("msize", mSpbPSymbolSize->value());

    settings.setValue("show_thl", mChbThreshold->isChecked());
    settings.setValue("thl_color", mLblThCC->color().name());
    settings.setValue("thl_width", mSpbThLineSize->value());

    settings.endGroup();
}

/// @brief Translates widgets of the plot groupbox.
void GraphPage::translateUI() {

    mChbHideToolbar->setText(tr("Hide plot toolbar."));
    mChbHideXLabel->setText(tr("Hide X-axis label."));

    mLblASPeriod->setText(tr("Auto-scrolling"));
    mLblTimeWindow->setText(tr("Time window"));
    mLblTimeOffset->setText(tr("Time offset"));

    mLblTextBox->setText(tr("Value box position"));
    int idx = mCbxTextBox->currentIndex();
    QStringList list{
        tr("Top left"), tr("Top center"), tr("Top right"), tr("Bottom left"), tr("Bottom center"), tr("Bottom right"),
    };
    mCbxTextBox->clear();
    mCbxTextBox->addItems(list);
    mCbxTextBox->setCurrentIndex(idx);

    setHelper(mChbHideToolbar, HelpFlag::HIDE_PTB);
    setHelper(mChbHideXLabel, HelpFlag::HIDE_PXL);

    setHelper(mLblASPeriod, HelpFlag::ASPERIOD);
    setHelper(mSpbASPeriod, HelpFlag::ASPERIOD);
    setHelper(mLblTimeWindow, HelpFlag::WTIME);
    setHelper(mSpbTimeWindow, HelpFlag::WTIME);
    setHelper(mLblTimeOffset, HelpFlag::WOFFSET);
    setHelper(mSpbTimeOffset, HelpFlag::WOFFSET);

    setHelper(mSpbTimeOffset, HelpFlag::WOFFSET);
    setHelper(mSpbTimeOffset, HelpFlag::WOFFSET);

    // graph
    Plotting::reloadNames();
    Plotting::fillMarkerStyleComboBox(mCbxPSymbol, mBtnPScatter->isChecked());
    mBtnPLine->setText(tr("line"));
    mBtnPScatter->setText(tr("scatter"));
    mLblPlotType->setText(tr("Graph type"));
    mLblPLineSize->setText(tr("Line width"));
    mLblPSymbol->setText(tr("Marker style"));

    // threshold line
    mChbThreshold->setText(tr("Show remanence threshold line"));
    mLblThColor->setText(tr("Line color"));
    mLblThSize->setText(tr("Line width"));
}
