// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "mappage.hpp"
#include "colorchooserlabel.hpp"

static void fillStyleCombobox(QComboBox *cbx);

/**
 * @brief Constructor
 * @param parent Parent widget.
 * @details
 * The constructor only builds up the UI.
 * Additionally, MapStyleWidget::loadSettings() and MapStyleWidget::translateUI() should be called to
 * populate the widgets with values and set texts.
 */
MapPage::MapPage(QWidget *parent) : BasePage(parent) {
    mPageTitle = tr("Map");
    mPageIcon = getIcon("map");
    buildUI();
    translateUI();
}

/// @brief Initialize widgets of the plot group box.
void MapPage::buildUI() {

    mChbMapGrid = new QCheckBox(this);
    mChbHoriCB = new QCheckBox(this);
    mChbHideMV = new QCheckBox(this);

    mChbFollowLocation = new QCheckBox(this);

    mLblSecVisible = new QLabel(this);
    mSpbSecVisible = new QSpinBox(this);
    mSpbSecVisible->setRange(1, 3600);
    mSpbSecVisible->setSuffix(tr(" seconds"));
    mSpbSecVisible->setSingleStep(1);
    mSpbSecVisible->setAccelerated(true);
    mSpbSecVisible->setStepType(QSpinBox::AdaptiveDecimalStepType);

    // missing value style
    mGbxMissingValue = new QGroupBox(this);

    mLblMVColor = new QLabel(mGbxMissingValue);
    mLblMVColorChooser = new ColorChooserLabel(mGbxMissingValue);
    mLblMVColorChooser->setHighlight(true);
    mLblMVStyle = new QLabel(mGbxMissingValue);
    mCbxMVStyle = new QComboBox(mGbxMissingValue);
    fillStyleCombobox(mCbxMVStyle);
    mSpbMVSize = new QDoubleSpinBox(mGbxMissingValue);
    mSpbMVSize->setSuffix(" px");
    mSpbMVSize->setRange(4.0, 20.0);
    mSpbMVSize->setSingleStep(0.5);

    QFormLayout *flayout = new QFormLayout(mGbxMissingValue);
    flayout->addRow(mLblMVColor, mLblMVColorChooser);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxMVStyle);
    hlayout->addWidget(mSpbMVSize);
    flayout->addRow(mLblMVStyle, hlayout);

    // current flight style
    mGbxCurrentFlight = new QGroupBox(this);
    mLblCFStyle = new QLabel(mGbxCurrentFlight);
    mCbxCFStyle = new QComboBox(mGbxCurrentFlight);
    fillStyleCombobox(mCbxCFStyle);
    mSpbCFSize = new QDoubleSpinBox(mGbxCurrentFlight);
    mSpbCFSize->setSuffix(" px");
    mSpbCFSize->setRange(4.0, 20.0);
    mSpbCFSize->setSingleStep(0.5);
    flayout = new QFormLayout(mGbxCurrentFlight);
    hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxCFStyle);
    hlayout->addWidget(mSpbCFSize);
    flayout->addRow(mLblCFStyle, hlayout);

    // previous flights style
    mGbxPreviousFlights = new QGroupBox(this);
    mLblPFStyle = new QLabel(mGbxPreviousFlights);
    mCbxPFStyle = new QComboBox(mGbxPreviousFlights);
    fillStyleCombobox(mCbxPFStyle);
    mSpbPFSize = new QDoubleSpinBox(mGbxPreviousFlights);
    mSpbPFSize->setSuffix(" px");
    mSpbPFSize->setRange(4.0, 20.0);
    mSpbPFSize->setSingleStep(0.5);

    flayout = new QFormLayout(mGbxPreviousFlights);
    hlayout = new QHBoxLayout;
    hlayout->addWidget(mCbxPFStyle);
    hlayout->addWidget(mSpbPFSize);
    flayout->addRow(mLblPFStyle, hlayout);

    mGbxRange = new QGroupBox(this);
    mGbxRange->setCheckable(true);
    mLblRgColor = new QLabel(mGbxRange);
    mLblRgCC = new ColorChooserLabel(mGbxRange);
    mSpbRgLineSize = new QDoubleSpinBox(mGbxRange);
    mSpbRgLineSize->setRange(0.5, 5.0);
    mSpbRgLineSize->setValue(1.0);
    mSpbRgLineSize->setSingleStep(0.1);
    mSpbRgLineSize->setPrefix(tr("line width: "));
    mSpbRgLineSize->setSuffix(" px");
    mSpbRgLineSize->setAccelerated(true);
    mChbRgFill = new QCheckBox(mGbxRange);
    mSpbRgFillAlpha = new QSpinBox(mGbxRange);
    mSpbRgFillAlpha->setPrefix("alpha: ");
    mSpbRgFillAlpha->setRange(0, 100);

    QVBoxLayout *vlayout = new QVBoxLayout(mGbxRange);

    hlayout = new QHBoxLayout;
    hlayout->addWidget(mLblRgColor, 0);
    hlayout->addWidget(mLblRgCC, 1);
    hlayout->addWidget(mSpbRgLineSize, 0, Qt::AlignRight);
    vlayout->addLayout(hlayout);

    hlayout = new QHBoxLayout;
    hlayout->addWidget(mChbRgFill, 1);
    hlayout->addWidget(mSpbRgFillAlpha, 0, Qt::AlignRight);
    vlayout->addLayout(hlayout);

    vlayout = new QVBoxLayout(this);
    vlayout->addWidget(mChbMapGrid);
    vlayout->addWidget(mChbHoriCB);
    vlayout->addWidget(mChbHideMV);
    vlayout->addWidget(mChbFollowLocation);

    hlayout = new QHBoxLayout;
    hlayout->addWidget(mLblSecVisible, 1);
    hlayout->addWidget(mSpbSecVisible, 0, Qt::AlignRight);
    vlayout->addLayout(hlayout);
    vlayout->addWidget(getSeparator(this));

    vlayout->addWidget(mGbxMissingValue);
    vlayout->addWidget(mGbxCurrentFlight);
    vlayout->addWidget(mGbxPreviousFlights);
    vlayout->addWidget(getSeparator(this));

    vlayout->addWidget(mGbxRange);
    vlayout->addStretch(1);

    QObject::connect(mChbHideMV, &QCheckBox::toggled, [=](bool checked) {
        mGbxMissingValue->setDisabled(checked);
        mLblMVColorChooser->setHighlight(!checked);
    });

    QObject::connect(mGbxRange, &QGroupBox::toggled, [=](bool checked) {
        if (checked)
            mSpbRgFillAlpha->setEnabled(mChbRgFill->isChecked());
    });

    QObject::connect(mChbRgFill, &QCheckBox::toggled, [=](bool checked) { mSpbRgFillAlpha->setEnabled(checked); });
}

/// @brief Load plot settings from configuration file.
void MapPage::loadSettings() {
    QSettings settings;

    settings.beginGroup("monitoring");
    mSpbSecVisible->setValue(settings.value("min_sec_vis", 30).toInt());
    mChbFollowLocation->setChecked(settings.value("follow_loc", false).toBool());
    settings.endGroup();

    settings.beginGroup("map");
    mChbMapGrid->setChecked(settings.value("m_showgrid", false).toBool());
    mChbHoriCB->setChecked(settings.value("cs_horiz", false).toBool());
    mChbHideMV->setChecked(settings.value("m_hidemv", false).toBool());
    mLblMVColorChooser->setColor(QColor(settings.value("m_mvcolor", "#000000").toString()));
    mCbxMVStyle->setCurrentIndex(mCbxCFStyle->findData(settings.value("m_mvmstyle", 0).toInt()));
    mSpbMVSize->setValue(settings.value("m_mvsize", 10.0).toDouble());
    mCbxCFStyle->setCurrentIndex(mCbxCFStyle->findData(settings.value("m_cmstyle", 0).toInt()));
    mSpbCFSize->setValue(settings.value("m_cmsize", 10.0).toDouble());
    mCbxPFStyle->setCurrentIndex(mCbxPFStyle->findData(settings.value("m_omstyle", 0).toInt()));
    mSpbPFSize->setValue(settings.value("m_omsize", 10.0).toDouble());
    settings.endGroup();

    settings.beginGroup("graph");
    mGbxRange->setChecked(settings.value("show_rgl", true).toBool());
    mLblRgCC->setColor(QColor(settings.value("rgl_color", "#000000").toString()));
    mSpbRgLineSize->setValue(settings.value("rgl_width", 1.0).toDouble());
    mChbRgFill->setChecked(settings.value("rgl_fill", true).toBool());
    mLblRgCC->setColor(QColor(settings.value("rgl_fill_color", "#0000ff").toString()));
    mSpbRgFillAlpha->setValue(settings.value("rgl_fill_alpha", 0.2).toDouble() * 100);
    settings.endGroup();
}

/// @brief Save plot settings in configuration file.
void MapPage::saveSettings() {
    QSettings settings;

    settings.beginGroup("monitoring");
    settings.setValue("min_sec_vis", qBound(1, mSpbSecVisible->value(), 3600));
    settings.setValue("follow_loc", mChbFollowLocation->isChecked());
    settings.endGroup();

    settings.beginGroup("map");
    settings.setValue("cs_horiz", mChbHoriCB->isChecked());
    settings.setValue("m_showgrid", mChbMapGrid->isChecked());
    settings.setValue("m_hidemv", mChbHideMV->isChecked());
    settings.setValue("m_mvmstyle", mCbxMVStyle->currentData().toInt());
    settings.setValue("m_mvcolor", mLblMVColorChooser->color().name());
    settings.setValue("m_mvmsize", mSpbMVSize->value());
    settings.setValue("m_cmstyle", mCbxCFStyle->currentData().toInt());
    settings.setValue("m_cmsize", mSpbCFSize->value());
    settings.setValue("m_omstyle", mCbxPFStyle->currentData().toInt());
    settings.setValue("m_omsize", mSpbPFSize->value());
    settings.endGroup();

    settings.beginGroup("graph");
    settings.setValue("show_rgl", mGbxRange->isChecked());
    settings.setValue("rgl_color", mLblRgCC->color().name());
    settings.setValue("rgl_width", mSpbRgLineSize->value());
    settings.setValue("rgl_fill", mChbRgFill->isChecked());
    settings.setValue("rgl_fill_color", mLblRgCC->color().name());
    settings.setValue("rgl_fill_alpha", mSpbRgFillAlpha->value() / 100.0);
    settings.endGroup();
}

/// @brief Translates widgets of the plot groupbox.
void MapPage::translateUI() {

    mLblSecVisible->setText(tr("Show at least"));
    mChbFollowLocation->setText(tr("Center map on new sensor location."));

    setHelper(mLblSecVisible, HelpFlag::SHOW_AT_LEAST);
    setHelper(mSpbSecVisible, HelpFlag::SHOW_AT_LEAST);
    setHelper(mChbFollowLocation, HelpFlag::CENTER_LOC);

    // map
    mChbMapGrid->setText(tr("Show map grid."));
    mChbHoriCB->setText(tr("Display the colorbar horizontally."));
    mChbHideMV->setText(tr("Hide missing value point on the map."));
    mGbxMissingValue->setTitle(tr("Missing value point"));
    mLblMVColor->setText(tr("Color"));
    mLblMVStyle->setText(tr("Symbol"));
    mGbxCurrentFlight->setTitle(tr("Current flight"));
    mLblCFStyle->setText(tr("Symbol"));
    mGbxPreviousFlights->setTitle(tr("Previous flights"));
    mLblPFStyle->setText(tr("Symbol"));

    setHelper(mGbxCurrentFlight, HelpFlag::CF_STYLE);
    setHelper(mGbxPreviousFlights, HelpFlag::PF_STYLE);
    setHelper(mGbxMissingValue, HelpFlag::MV_STYLE);
    setHelper(mChbHoriCB, HelpFlag::CS_ORIENT);
    setHelper(mChbHideMV, HelpFlag::HIDE_MV);

    // range lines
    mGbxRange->setTitle(tr("Show color range lines"));
    mLblRgColor->setText(tr("Color"));
    mSpbRgLineSize->setPrefix(tr("line width: "));
    mChbRgFill->setText(tr("Apply fill between lines."));
}

static void fillStyleCombobox(QComboBox *cbx) {
    int idx = cbx->currentIndex();
    cbx->clear();
    cbx->addItem(QCoreApplication::tr("Circle"), 0);
    cbx->addItem(QCoreApplication::tr("Cross (+)"), 1);
    cbx->addItem(QCoreApplication::tr("Diamond"), 2);
    cbx->addItem(QCoreApplication::tr("Square"), 3);
    cbx->addItem(QCoreApplication::tr("Triangle"), 4);
    cbx->addItem(QCoreApplication::tr("Diag. cross (X)"), 5);
    cbx->setCurrentIndex(idx);
}
