// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "monitoringpage.hpp"

MonitoringPage::MonitoringPage(QWidget *parent) : BasePage(parent) {
    mPageTitle = tr("Monitoring");
    mPageIcon = getIcon("monitoring");
    buildUI();
    translateUI();
}

/// @brief Initialize widgets of the Plot group box.
void MonitoringPage::buildUI() {

    mLblNewFlightDelay = new QLabel(this);
    mSpbNewFlightDelay = new QSpinBox(this);
    mSpbNewFlightDelay->setRange(0, 60);
    mSpbNewFlightDelay->setSpecialValueText(tr("disabled"));
    mSpbNewFlightDelay->setSuffix(tr(" min"));
    mSpbNewFlightDelay->setSingleStep(1);
    mSpbNewFlightDelay->setAccelerated(true);

    mChbNewFlightOnStrap = new QCheckBox(this);
    mChbProcInternals = new QCheckBox(this);
    mChbProcNoGPS = new QCheckBox(this);
    mChbMonitorWifi = new QCheckBox(this);

    mLblInputFrequency = new QLabel(this);
    mSpbInputFrequency = new QSpinBox(this);
    mSpbInputFrequency->setSuffix(tr(" Hz"));
    mSpbInputFrequency->setSingleStep(1);
    mSpbInputFrequency->setRange(0, 50);
    mSpbInputFrequency->setSpecialValueText(tr("disabled"));

    mLblInvFrequency = new QLabel(this);
    mSpbInvFrequency = new QSpinBox(this);
    mSpbInvFrequency->setSuffix(tr(" Hz"));
    mSpbInvFrequency->setSingleStep(1);
    mSpbInvFrequency->setRange(0, 50);
    mSpbInvFrequency->setSpecialValueText(tr("disabled"));

    mChbFastCalculation = new QCheckBox(this);
    mChbDisplayFit = new QCheckBox(this);

    QFormLayout *layout = new QFormLayout(this);

    layout->addRow(mLblNewFlightDelay, mSpbNewFlightDelay);
    layout->addRow(mChbNewFlightOnStrap);
    layout->addRow(getSeparator(this));
    layout->addRow(mChbProcNoGPS);
    layout->addRow(mChbProcInternals);
    layout->addRow(getSeparator(this));
    layout->addRow(mChbMonitorWifi);
    layout->addRow(getSeparator(this));
    layout->addRow(mLblInputFrequency, mSpbInputFrequency);
    layout->addRow(getSeparator(this));
    layout->addRow(mLblInvFrequency, mSpbInvFrequency);
    layout->addRow(mChbFastCalculation);
    layout->addRow(mChbDisplayFit);
}

/// @brief Load monitoring settings from configuration file.
void MonitoringPage::loadSettings() {
    QSettings settings;
    settings.beginGroup("monitoring");
    mSpbNewFlightDelay->setValue(settings.value("nf_delay", 0).toInt());
    mChbNewFlightOnStrap->setChecked(settings.value("nf_onstrap", true).toBool());
    mChbProcInternals->setChecked(settings.value("proc_internal", false).toBool());
    mChbProcNoGPS->setChecked(settings.value("proc_nogps", true).toBool());
    mChbMonitorWifi->setChecked(settings.value("monitor_wifi", true).toBool());
    int v = settings.value("max_input_freq", 10).toInt();
    if (v < 0)
        v = 0;
    mSpbInputFrequency->setValue(v);
    v = settings.value("max_inv_freq", 5).toInt();
    if (v < 0)
        v = 0;
    mSpbInvFrequency->setValue(v);

    mChbFastCalculation->setChecked(settings.value("fast_calculations", true).toBool());
    mChbDisplayFit->setChecked(settings.value("show_fit", false).toBool());
    settings.endGroup();
}

/// @brief Save monitoring settings in configuration file.
void MonitoringPage::saveSettings() {
    QSettings settings;
    settings.beginGroup("monitoring");
    settings.setValue("nf_delay", qBound(0, mSpbNewFlightDelay->value(), 60));
    settings.setValue("nf_onstrap", mChbNewFlightOnStrap->isChecked());
    settings.setValue("proc_internal", mChbProcInternals->isChecked());
    settings.setValue("proc_nogps", mChbProcNoGPS->isChecked());
    settings.setValue("monitor_wifi", mChbMonitorWifi->isChecked());
    int v = mSpbInputFrequency->value();
    if (v == 0)
        v = -1;
    settings.setValue("max_input_freq", v);
    v = mSpbInvFrequency->value();
    if (v == 0)
        v = -1;
    settings.setValue("max_inv_freq", v);
    settings.setValue("fast_calculations", mChbFastCalculation->isChecked());
    settings.setValue("show_fit", mChbDisplayFit->isChecked());
    settings.endGroup();
}

/// @brief Translates widgets of the monitoring groupbox.
void MonitoringPage::translateUI() {
    mPageTitle = tr("Monitoring");
    mLblNewFlightDelay->setText("New flight delay");
    mSpbNewFlightDelay->setSpecialValueText(tr("disabled"));
    mChbNewFlightOnStrap->setText(tr("Enable new flight on 'new file' switch activation."));
    mChbProcInternals->setText(tr("Process internal parameters of the sensor."));
    mChbProcNoGPS->setText(tr("Process data block with invalid GPS data."));
    mChbMonitorWifi->setText(tr("Monitor WLAN connection."));
    mLblInputFrequency->setText("Maximum incoming block frequency");
    mLblInvFrequency->setText("Maximum inversion process frequency");
    mChbFastCalculation->setText("Fast calculations mode.");
    mChbDisplayFit->setText("Display inversion process plots.");
    setHelper(mLblNewFlightDelay, HelpFlag::NF_DELAY);
    setHelper(mSpbNewFlightDelay, HelpFlag::NF_DELAY);
    setHelper(mChbNewFlightOnStrap, HelpFlag::NF_STRAP);
    setHelper(mChbMonitorWifi, HelpFlag::MON_WIFI);
    setHelper(mChbProcInternals, HelpFlag::PROC_INTERNALS);
    setHelper(mChbProcNoGPS, HelpFlag::PROC_NOGPS);
    setHelper(mLblInputFrequency, HelpFlag::MAX_INPUT_FREQ);
    setHelper(mSpbInputFrequency, HelpFlag::MAX_INPUT_FREQ);
    setHelper(mLblInvFrequency, HelpFlag::MAX_INV_FREQ);
    setHelper(mSpbInvFrequency, HelpFlag::MAX_INV_FREQ);
    setHelper(mChbFastCalculation, HelpFlag::FAST_CALCULATIONS);
}
