// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "storagepage.hpp"

QVector<QPair<Logger::LogLevel, QString>> getLogLevels() {
    return {
        {Logger::Debug, QCoreApplication::tr("DEBUG")},
        {Logger::Info, QCoreApplication::tr("INFO")},
        {Logger::Warning, QCoreApplication::tr("WARNING")},
        {Logger::Error, QCoreApplication::tr("ERROR")},
    };
}

/// @brief Constructor
StoragePage::StoragePage(QWidget *parent) : BasePage(parent) {
    mPageTitle = tr("Storage");
    mPageIcon = getIcon("storage");
    buildUI();
    translateUI();
}

/// @brief Initialize widgets of the Storage group box.
void StoragePage::buildUI() {

    mLblDataFolder = new QLabel(this);
    mWdtDataFolder = new DirectoryWidget(this);

    mChbLog = new QCheckBox(this);
    mChbData = new QCheckBox(this);
    mChbNFONF = new QCheckBox(this);

    mLblLogLevel = new QLabel(this);
    mCbxLogLevel = new QComboBox(this);

    for (auto p : getLogLevels())
        mCbxLogLevel->addItem(p.second, static_cast<int>(p.first));

    QFormLayout *formLayout = new QFormLayout(this);
    formLayout->addRow(mLblDataFolder, mWdtDataFolder);
    formLayout->addRow(mChbData);
    formLayout->addRow(mChbNFONF);
    formLayout->addRow(mChbLog);
    formLayout->addRow(mLblLogLevel, mCbxLogLevel);

    QObject::connect(mChbData, &QCheckBox::toggled, mChbNFONF, &QCheckBox::setEnabled);
}

/// @brief Load storage settings from configuration file.
void StoragePage::loadSettings() {
    QSettings settings;
    settings.beginGroup("storage");
    // storage
    mWdtDataFolder->setDirectory(settings.value("datadir", gDataDir).toString());
    mChbData->setChecked(settings.value("save", true).toBool());
    mChbNFONF->setChecked(settings.value("nfonf", true).toBool());
    mChbLog->setChecked(settings.value("log", true).toBool());
    mCbxLogLevel->setCurrentIndex(mCbxLogLevel->findData(settings.value("log_lvl", static_cast<int>(Logger::Info)).toInt()));
    settings.endGroup();
}

/// @brief Save storage settings in configuration file.
void StoragePage::saveSettings() {
    QString defDir =
        QString("%1/A-ToolSuite/ADMin").arg(QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).first());
    QString dataFolder = mWdtDataFolder->directory().isEmpty() ? defDir : mWdtDataFolder->directory();
    QSettings settings;
    settings.beginGroup("storage");
    settings.setValue("datadir", dataFolder);
    settings.setValue("save", mChbData->isChecked());
    settings.setValue("nfonf", mChbNFONF->isChecked());
    settings.setValue("log", mChbLog->isChecked());
    settings.setValue("log_lvl", mCbxLogLevel->currentData().toInt());
    settings.endGroup();
    //    gDataDir = dataFolder;
    //    QDir(gDataDir).mkpath(".");
}

/// @brief Translates widgets of the storage groupbox.
void StoragePage::translateUI() {
    mPageTitle = tr("Storage");
    mLblDataFolder->setText(tr("Root data directory"));
    mChbData->setText(tr("Save data blocks to disk."));
    mChbNFONF->setText(tr("Create new file on new flight."));
    mChbLog->setText(tr("Enable application logs."));
    mLblLogLevel->setText(tr("Logging level"));

    // rebuild labels for the combobox
    int idx = mCbxLogLevel->currentIndex();
    mCbxLogLevel->clear();
    for (auto p : getLogLevels())
        mCbxLogLevel->addItem(p.second, static_cast<int>(p.first));
    mCbxLogLevel->setCurrentIndex(idx);

    setHelper(mLblDataFolder, HelpFlag::DDIR);
    setHelper(mWdtDataFolder, HelpFlag::DDIR);
    setHelper(mChbData, HelpFlag::DSAVE);
    setHelper(mChbNFONF, HelpFlag::NFONF);
    setHelper(mChbLog, HelpFlag::LOGGING);
    setHelper(mLblLogLevel, HelpFlag::LOG_LVL);
    setHelper(mCbxLogLevel, HelpFlag::LOG_LVL);
}
