// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "wlanwidget.hpp"

#include "wlanwatcher.hpp"

// const QSize WLANWidget::iconSize = QSize(48, 48);
const QSize WLANWidget::iconSize = QSize(24, 24);

WLANWidget::WLANWidget(QWidget *parent) : QWidget(parent) {
    QSettings settings;

    mNotconnectedPixmap = QPixmap(":/adm2/images/wifi/no-wifi.png").scaled(WLANWidget::iconSize, Qt::KeepAspectRatio);
    mSignalPixmaps = {QPixmap(":/adm2/images/wifi/wifi-ukn.png").scaled(WLANWidget::iconSize, Qt::KeepAspectRatio),
                      QPixmap(":/adm2/images/wifi/wifi-000.png").scaled(WLANWidget::iconSize, Qt::KeepAspectRatio),
                      QPixmap(":/adm2/images/wifi/wifi-025.png").scaled(WLANWidget::iconSize, Qt::KeepAspectRatio),
                      QPixmap(":/adm2/images/wifi/wifi-050.png").scaled(WLANWidget::iconSize, Qt::KeepAspectRatio),
                      QPixmap(":/adm2/images/wifi/wifi-075.png").scaled(WLANWidget::iconSize, Qt::KeepAspectRatio),
                      QPixmap(":/adm2/images/wifi/wifi-100.png").scaled(WLANWidget::iconSize, Qt::KeepAspectRatio)};

    mWlanWatcher = new WlanWatcher(this);

    mLblIcon = new QLabel(this);
    mLblSSID = new QLabel(this);
    mLblStatus = new QLabel(this);

    mLblIcon->setFixedSize(WLANWidget::iconSize);

    QHBoxLayout *layout = new QHBoxLayout(this);
    layout->addWidget(mLblIcon, 0, Qt::AlignLeft);
    layout->addWidget(mLblSSID, 0, Qt::AlignRight);
    layout->addWidget(mLblStatus, 1, Qt::AlignRight);

    reset();

    QObject::connect(mWlanWatcher, &WlanWatcher::connected, this, &WLANWidget::onConnection);
    QObject::connect(mWlanWatcher, &WlanWatcher::signalChanged, this, &WLANWidget::onSignalChanged);
    QObject::connect(mWlanWatcher, &WlanWatcher::disconnected, this, &WLANWidget::onDisconnection);
    QObject::connect(mWlanWatcher, &WlanWatcher::interfaceChange, this, &WLANWidget::onInterfaceChange);
}

/// @brief Destructor.
WLANWidget::~WLANWidget() {}

/// @brief Set SSID to monitor.
void WLANWidget::setSSID(const QString &ssid) {
    mWlanWatcher->monitorSSID(ssid);
    mLblSSID->setText(ssid);
}

/// @brief Enable/Disable monitoring.
void WLANWidget::monitor(bool enabled) {
    if (enabled) {
        mWlanWatcher->start();
        QString ssid = mLblSSID->text();
        if (ssid.isEmpty()) {
            mLblIcon->setPixmap(mSignalPixmaps.at(0));
            mLblStatus->setText(tr("No SSID to watch."));
        } else if (!mWlanWatcher->interfacesCount()) {
            mLblIcon->setPixmap(mSignalPixmaps.at(0));
            mLblStatus->setText(tr("No WLAN interface."));
        } else if (!mWlanWatcher->isConnected()) {
            if (mWlanWatcher->isAvailable()) {
                mLblIcon->setPixmap(mNotconnectedPixmap);
                mLblStatus->setText(tr("Not connected to SSID."));
            } else {
                mLblIcon->setPixmap(mSignalPixmaps.at(0));
                mLblStatus->setText(tr("SSID not found."));
            }
        }
    } else {
        mWlanWatcher->stop();
        mLblIcon->setPixmap(mSignalPixmaps.at(0));
        mLblStatus->setText(tr("Monitoring disabled."));
    }
    setEnabled(enabled);
}


/// @brief Reset widget to display defaults.
void WLANWidget::reset() {
    mLblIcon->setPixmap(mSignalPixmaps.at(0));
    mLblSSID->setText("");
    mLblStatus->setText(tr("No SSID to watch."));
}

void WLANWidget::onInterfaceChange(WlanInterface) {
    QString ssid = mWlanWatcher->monitoredSSID();
    int icount = mWlanWatcher->interfacesCount();
    bool empty = ssid.isEmpty();
    if (empty || !icount) {
        mLblIcon->setPixmap(mSignalPixmaps.at(0));
        mLblStatus->setText(empty ? tr("No SSID to watch.") : tr("No WLAN interface."));
    } else if (!mWlanWatcher->isConnected()) {
        if (mWlanWatcher->isAvailable()) {
            mLblIcon->setPixmap(mNotconnectedPixmap);
            mLblStatus->setText(tr("Not connected to SSID."));
        } else {
            mLblIcon->setPixmap(mSignalPixmaps.at(0));
            mLblStatus->setText(tr("SSID not found."));
        }
    }
}

/// @brief Update widget on monitored ssid disconnection.
void WLANWidget::onConnection(int signal) {
    LOG_INFO("WIFI Connection detected");
    onSignalChanged(signal);
}

/// @brief Update widget on monitored ssid disconnection.
void WLANWidget::onDisconnection() {
    LOG_WARNING("WIFI Connection lost");
    mLblIcon->setPixmap(mNotconnectedPixmap);
    mLblStatus->setText(tr("Not connected to SSID."));
}

/// @brief Update widget on signal strength value change.
void WLANWidget::onSignalChanged(int signal) {
    setWifiPixmap(signal);
    mLblStatus->setText(QString("Signal: %1%").arg(signal));
}

/// @brief Set pixmap from the given signal strength value.
void WLANWidget::setWifiPixmap(int value) {
    if (value < 0.0)
        mLblIcon->setPixmap(mSignalPixmaps.at(0));
    else if (value < 25.0)
        mLblIcon->setPixmap(mSignalPixmaps.at(1));
    else if (value < 50.0)
        mLblIcon->setPixmap(mSignalPixmaps.at(2));
    else if (value < 75.0)
        mLblIcon->setPixmap(mSignalPixmaps.at(3));
    else if (value < 100.0)
        mLblIcon->setPixmap(mSignalPixmaps.at(4));
    else
        mLblIcon->setPixmap(mSignalPixmaps.at(5));
}
