// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#ifndef WLANWATCHER_HPP
#define WLANWATCHER_HPP

#include "wlaninterface.hpp"

#include <wlanapi.h>

class WlanWatcher : public QObject {
    Q_OBJECT

  public:
    enum Notification {
        START = 0,
        AUTOCONF_ENABLED,
        AUTOCONF_DISABLED,
        BACKGOUND_SCAN_ENABLED,
        BACKGOUND_SCAN_DISABLED,
        BSS_TYPE_CHANGE,
        POWER_SETTINGS_CHANGE,
        SCAN_COMPLETE,
        SCAN_FAIL,
        CONNECTION_START,
        CONNECTION_COMPLETE,
        CONNECTION_ATTEMPT_FAIL,
        FILTER_LIST_CHANGE,
        INTERFACE_ARRIVAL,
        INTERFACE_REMOVAL,
        PROFILE_CHANGE,
        PROFILES_EXHAUSTED,
        NETWORK_NOT_AVAILABLE,
        NETWORK_AVAILABLE,
        DISCONNECTING,
        DISCONNECTED,
        ADHOC_NETWORK_STATE_CHANGE,
        PROFILE_UNBLOCKED,
        SCREEN_POWER_CHANGE,
        PROFILE_BLOCKED,
        SCAN_LIST_REFRESH,
        END
    };

    WlanWatcher(QObject *parent = nullptr);
    virtual ~WlanWatcher();
    bool isValid() const;
    void monitorSSID(const QString &ssid);
    void start(int refreshTimeout = 1000);
    void stop();
    QVector<WlanInfos> availableNetworks() const;
    QStringList availableSSIDs() const;
    const QString &monitoredSSID() const;
    bool isAvailable() const;
    bool isConnected() const;
    static QString notificationToString(Notification notification);
    int interfacesCount() const;
signals:
    /// @brief Emitted when at least one WLAN interface is connected to the monitored SSID.
    void connected(int signalStrength);

    void signalChanged(int signalStrength);

    /// @brief Emitted when no more WLAN interface is connected to the monitored SSID.
    void disconnected();

    /**
     * @brief Emitted whenever a change occur in the interfaces.
     * @param iface Updated interface.
     * @details
     * The signal is emitted notably when:
     * - a new interface has been added.
     * - an interface has been removed.
     * - an interface has been connected to an SSID (not necessarily the monitored one).
     */
    void interfaceChange(WlanInterface iface);

  public slots:
    void updateStates();

  private:
    QTimer mTimer;
    bool mAvailable;
    bool mConnected;
    QString mSSID;
    HANDLE mHClient;
    QMap<GUID, WlanInterface> mInterfaces;
    Notification mLastNotification;
    Q_DISABLE_COPY(WlanWatcher)
    static const DWORD dwClientVersion;

  private:
    static VOID WINAPI wlanNotification(PWLAN_NOTIFICATION_DATA wlanNotifData, PVOID context);
    void setLastNotification(Notification notification, GUID guid);
    void resetIfaceInfos();
    bool checkAvailability(const QString &ssid) const;
};

#endif // WLANWATCHER_HPP
