// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "csvfilewriter.hpp"
#include "logger.h"

/**
 * @brief Constructor
 * @param path Path of the output file.
 * @param data Pair of data to write.
 * @param parent Parent object.
 */
CSVFileWriter::CSVFileWriter(const QString &path, const PairOfData &data, int precision, QObject *parent)
    : QObject(parent), mPath(path), mCancel(false), mIsRunning(false), mPrecision(precision) {
    mData = data.second;
    mHeaders = data.first;
}

/// @brief Check whether or not the worker in the running.
bool CSVFileWriter::isRunning() const {
    QMutexLocker locker(&mDataMutex);
    return mIsRunning;
}

/// @brief Get the output path.
QString CSVFileWriter::outputPath() const {
    QMutexLocker locker(&mDataMutex);
    return mPath;
}

/// @brief Process data.
void CSVFileWriter::run() { process(); }

/// @brief Write data to file(s).
void CSVFileWriter::process() {
    mDataMutex.lock();
    mCancel = false;
    mIsRunning = true;
    mDataMutex.unlock();

    int npts = 0;
    for (QVector<double> *c : mData) {
        int ns = c->size();
        if (ns > npts)
            npts = ns;
    }
    int nprms = mData.size();

    // check size.
    if (!(npts && nprms == mHeaders.size())) {
        if (nprms == mHeaders.size())
            LOG_DEBUG("[CSV] Inconsistent header size !");
        else
            LOG_DEBUG("[CSV] No data to write !");
        emit finished(false);
        return;
    }
    emit progressRangeChanged(0, npts);
    // check file.
    QFile file(mPath);
    if (!file.open(QIODevice::Text | QIODevice::WriteOnly) || mData.isEmpty()) {
        emit finished(false);
        return;
    }
    emit progressChanged(0);

    QStringList values;
    values.reserve(nprms);
    for (int i = 0; i < nprms; i++)
        values.append("nan");
    QTextStream out(&file);
    out << mHeaders.join(";") << Qt::endl;
    for (int i = 0; i < npts; i++) {
        for (int j = 0; j < nprms; j++) {
            QVector<double> *c = mData.at(j);
            values[j] = i >= c->size() ? "nan" : QString::number(c->at(i), 'e', mPrecision);
        }
        out << values.join(";") << Qt::endl;
        if (i % 100 == 0) {
            emit progressChanged(i);
            mDataMutex.lock();
            bool cancel = mCancel;
            mDataMutex.unlock();
            if (cancel) {
                file.close();
                file.remove();
                mDataMutex.lock();
                mIsRunning = false;
                mDataMutex.unlock();
                LOG_DEBUG("[CSV] Job canceled: file has been removed.");
                emit finished(true);
                return;
            }
        }
    }
    file.close();
    emit progressChanged(npts);

    mDataMutex.lock();
    mIsRunning = false;
    mDataMutex.unlock();
    emit finished(false);
}

/// @brief Cancel operations.
void CSVFileWriter::cancel() {
    mDataMutex.lock();
    mCancel = true;
    mDataMutex.unlock();
}
