// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "datasetinfowidget.hpp"
#include "labelwidget.hpp"

#include "amdataset.hpp"

/// @brief Set custom font.
void setCustomFont(QWidget *widget) {
    QPalette palette = widget->palette();
    palette.setColor(widget->foregroundRole(), Qt::blue);
    QFont f = widget->font();
    f.setUnderline(true);
    widget->setPalette(palette);
    widget->setFont(f);
}

/// @brief Restore font.
void restoreFont(QWidget *widget) {
    QPalette palette = widget->palette();
    palette.setColor(widget->foregroundRole(), Qt::black);
    QFont f = widget->font();
    f.setUnderline(false);
    widget->setPalette(palette);
    widget->setFont(f);
}

/// @brief Constructor.
DatasetInfoWidget::DatasetInfoWidget(QWidget *parent) : QGroupBox(parent) {
    mLblFilesD = new ClickableLabel(this);
    mLblFiles = new QLabel(this);
    mLblParametersD = new QLabel(this);
    mLblParameters = new QLabel(this);
    mLblInstrumentD = new QLabel(this);
    mLblInstrument = new QLabel(this);
    mLblVersionD = new QLabel(this);
    mLblVersion = new QLabel(this);

    QHBoxLayout *layout = new QHBoxLayout(this);
    QFormLayout *flayout = new QFormLayout;
    flayout->addRow(mLblFilesD, mLblFiles);
    flayout->addRow(mLblParametersD, mLblParameters);
    layout->addLayout(flayout);
    flayout = new QFormLayout;
    flayout->addRow(mLblInstrumentD, mLblInstrument);
    flayout->addRow(mLblVersionD, mLblVersion);
    layout->addLayout(flayout);
    setDataSet(nullptr);
    QObject::connect(mLblFilesD, &ClickableLabel::clicked, this, &DatasetInfoWidget::showFilePaths);
    translateUI();
}

/// @brief Translate UI texts.
void DatasetInfoWidget::translateUI() {
    setTitle(tr("Dataset informations"));
    mLblFilesD->setText(tr("Num. of files"));
    mLblParametersD->setText(tr("Num. of parameters"));
    mLblInstrumentD->setText(tr("Instrument number"));
    mLblVersionD->setText(tr("Firmware version"));
}

/// @brief Update informations from given dataset.
void DatasetInfoWidget::setDataSet(AMDataset *dataset) {
    filePaths.clear();
    if (dataset) {
        setCustomFont(mLblFilesD);
        setCustomFont(mLblParametersD);
        mLblFiles->setText(QString::number(dataset->fileCount()));
        mLblParameters->setText(tr("%1 (%2 blocks)").arg(dataset->parameterCount()).arg(dataset->blockCount()));
        mLblInstrument->setText(QString::number(dataset->instrument()));
        mLblVersion->setText(QString::number(dataset->version()));
        filePaths = dataset->filePaths();
    } else {
        restoreFont(mLblFilesD);
        restoreFont(mLblParametersD);
        mLblFiles->setText("N/A");
        mLblParameters->setText("N/A");
        mLblInstrument->setText("N/A");
        mLblVersion->setText("N/A");
    }
}

/**
 * @brief Show list of files in an external window.
 */
void DatasetInfoWidget::showFilePaths() {

    QListWidget *listWidget = new QListWidget;
    listWidget->setAttribute(Qt::WA_DeleteOnClose, true);
    listWidget->addItems(filePaths);
    listWidget->show();
    listWidget->activateWindow();

//    QString message = QString("<ul><li>%1</li></ul>").arg(filePaths.join("</li><li>"));
//    LabelWidget *label = new LabelWidget(message);
//    label->setAttribute(Qt::WA_DeleteOnClose, true);
//    label->setText(message);
//    label->show();
//    label->activateWindow();
}
