// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "mainwindow.hpp"
#include "amdataset.hpp"
#include "isp/descriptor.hpp"
#include "multifileloader.hpp"

static bool isRaw(const QFileInfo &pathInfo) { return pathInfo.suffix().compare("bin", Qt::CaseInsensitive) == 0; }
static bool isResult(const QFileInfo &pathInfo) {
    return pathInfo.suffix().compare("results", Qt::CaseInsensitive) == 0;
}

/**
 * @brief Constructor.
 * @param parent Parent widget.
 */
MainWindow::MainWindow(QWidget *parent) : QMainWindow(parent) {
    mThreadPool = QThreadPool::globalInstance();
    mThreadPool->setMaxThreadCount(1);

    mDebug = false;
    mLogWidget = new LogWidget(gSignalAppender, "ui/logw");
    mLogWidget->setWindowFlags(Qt::Tool);
    setAcceptDrops(true);

    mXMLDContainer.setupFromDirectory(gDescrDir);
    mXMLDCurrent = nullptr;

    QWidget *widget = new QWidget(this);

    mDatasetInfoWidget = new DatasetInfoWidget(widget);
    mPrmChoiceWidget = new PrmChoiceWidget(widget);
    mChbSaveChanges = new QCheckBox(widget);

    mSpbPrecision = new QSpinBox(widget);
    // note that a precision greater than 16 is non-sense: 1e-16 is the epsilon for double.
    mSpbPrecision->setRange(0, 20);

    mBtnConvert = new QPushButton(widget);
    mBtnConvert->setEnabled(false);
    QHBoxLayout *hlayout = new QHBoxLayout;
    hlayout->addWidget(mChbSaveChanges, 0, Qt::AlignLeft);
    hlayout->addStretch(1);
    hlayout->addWidget(mSpbPrecision, 0, Qt::AlignRight);
    hlayout->addWidget(mBtnConvert, 0, Qt::AlignRight);

    QVBoxLayout *layout = new QVBoxLayout(widget);
    layout->addWidget(mDatasetInfoWidget, 0, Qt::AlignTop);
    layout->addWidget(mPrmChoiceWidget, 2);
    layout->addLayout(hlayout, 0);

    setCentralWidget(widget);

    mStatusPBar = new StatusProgressBar;
    mProgressBar = mStatusPBar->progressBar();
    mProgressBar->progressBar()->setAlignment(Qt::AlignCenter);
    mProgressBar->setVisible(false);
    setStatusBar(mStatusPBar);

    createActions();

    QObject::connect(mProgressBar->button(), &QPushButton::clicked, this, &MainWindow::cancelOperations);

    QObject::connect(mPrmChoiceWidget, &PrmChoiceWidget::itemMoved, this, &MainWindow::onItemMoved);
    QObject::connect(mBtnConvert, &QPushButton::clicked, this, &MainWindow::convertParameters);
    translateUI();
    loadUI();
}

/// @brief Destructor.
MainWindow::~MainWindow() {
    if (mLogWidget) {
        mLogWidget->saveUI();
        mLogWidget->close();
        mLogWidget->deleteLater();
    }
}

/**
 * @brief changeEvent event handler.
 * @details
 * If event is of type QEvent::LanguageChange, the method (re)translates (completely) the application (event ==
 * QEvent::LanguageChange), change application Locale (event == QEvent::LocaleChange), or simply pass the event to the
 * base class changeEvent method (any other case).
 */
void MainWindow::changeEvent(QEvent *event) {
    if (event->type() == QEvent::LanguageChange)
        translateUI();
    else if (event->type() == QEvent::LocaleChange) {
        QString locale = QLocale::system().name();
        locale.truncate(locale.lastIndexOf('_'));
        if (gI18N)
            gI18N->setLanguage(locale);
    }
    QMainWindow::changeEvent(event);
}

/**
 * @brief Perform final actions on close.
 * @param event CloseEvent.
 */
void MainWindow::closeEvent(QCloseEvent *event) {
    saveDescriptor();
    saveUI();
    cancelOperations();
    QMainWindow::closeEvent(event);
}

/// @brief Translate texts.
void MainWindow::translateUI() {
    mDatasetInfoWidget->translateUI();
    mPrmChoiceWidget->translateUI();
    mChbSaveChanges->setText(tr("Always saves changes."));
    mBtnConvert->setText(tr("Convert"));

    mSpbPrecision->setPrefix(tr("Precision: "));
    mLogWidget->translateUI();
    mLogWidget->setWindowTitle(tr("Log activity"));
    mLogWidget->toggleViewAction()->setText(tr("Log activity"));
    mMenuFiles->setTitle(tr("File"));
    mMenuSubFiles->setTitle(tr("Open"));
    mMenuHelp->setTitle(tr("Help"));
    mActionOpen->setText(tr("Files"));
    mActionOpenDirectory->setText(tr("Directory"));
    mActionQuit->setText(tr("Quit"));
#ifndef NO_HELP
    mActionHelp->setText(tr("Manual"));
#endif
    mActionDebug->setData(tr("Degub log"));
    mActionAbout->setText(tr("About AFC"));
    mActionOpen->setToolTip(tr("Import files to process."));
    mActionOpenDirectory->setToolTip(tr("Import files to process from a directory tree."));
    mActionQuit->setToolTip(tr("Stop everything and quit the application."));
    mActionDebug->setToolTip(tr("Enable/Disable debug messages in log activity."));
    mActionAbout->setToolTip(tr("Display the application about box."));
    mSpbPrecision->setToolTip(tr("Set the number of digits after the decimal point."));
#ifndef NO_HELP
    QString msg;
    if (mActionHelp->isEnabled()) {
        msg = tr("Show the manual.");
    } else {
        msg = tr("Manual is not available.");
    }
    mActionHelp->setToolTip(msg);
#endif
}

/// @brief Get the size hint of the window.
QSize MainWindow::sizeHint() const {
    QSize sz = minimumSizeHint();
    sz.rwidth() = qMax(sz.width(), 400);
    sz.rheight() = qMax(sz.height(), 600);
    return sz;
}

/// @brief Load UI layout settings.
void MainWindow::loadUI() {
    QSettings settings;
    restoreGeometry(settings.value("ui/main_win/geometry").toByteArray());
    restoreState(settings.value("ui/main_win/state").toByteArray());
    mChbSaveChanges->setChecked(settings.value("global/saveprms", true).toBool());
    mSpbPrecision->setValue(qBound(0, settings.value("global/precision", 12).toInt(), 20));
}

/// @brief Save UI layout settings.
void MainWindow::saveUI() {
    QSettings settings;
    settings.setValue("ui/main_win/geometry", saveGeometry());
    settings.setValue("ui/main_win/state", saveState());
    settings.setValue("global/saveprms", mChbSaveChanges->isChecked());
    settings.setValue("global/precision", mSpbPrecision->value());
}

/// @brief Create actions.
void MainWindow::createActions() {
    QMenuBar *menubar = menuBar();
    mMenuFiles = menubar->addMenu(tr("File"));
    mMenuSubFiles = mMenuFiles->addMenu(tr("Open"));
    mActionOpen = mMenuSubFiles->addAction(QIcon::fromTheme("browse-file"), tr("Files"));
    QObject::connect(mActionOpen, &QAction::triggered, this, &MainWindow::openFiles);
    mActionOpenDirectory = mMenuSubFiles->addAction(QIcon::fromTheme("browse-folder"), tr("Directory"));
    QObject::connect(mActionOpenDirectory, &QAction::triggered, this, &MainWindow::openDirectory);
    mMenuFiles->addSeparator();
    mActionQuit = mMenuFiles->addAction(QIcon::fromTheme("exit"), tr("Quit"));
    QObject::connect(mActionQuit, &QAction::triggered, this, &MainWindow::close);

    mMenuHelp = menubar->addMenu(tr("Help"));
#ifndef NO_HELP
    mActionHelp = mMenuHelp->addAction(tr("Manual"));
    // setup help button: if manual is not found disable the button.
    QString path = QString("%1/manual/afc.pdf").arg(gResDir);
    mActionHelp->setEnabled(QFileInfo::exists(path));
    QObject::connect(mActionHelp, &QAction::triggered, this, &MainWindow::openHelp);
#endif
    mMenuHelp->addSeparator();
    mActionDebug = mMenuHelp->addAction(tr("Debug log"));
    mActionDebug->setCheckable(true);
    QObject::connect(mActionDebug, &QAction::toggled, this, &MainWindow::enableDebugLog);
    mMenuHelp->addAction(mLogWidget->toggleViewAction());
    mMenuHelp->addSeparator();
    mActionAbout = mMenuHelp->addAction(tr("About AFC"));
    QObject::connect(mActionAbout, &QAction::triggered, this, &MainWindow::aboutApp);
}

/// @brief Open html help file.
void MainWindow::openHelp() {
    QString path = QString("%1/manual/afc.pdf").arg(gResDir);
    if (QFileInfo::exists(path))
        QDesktopServices::openUrl(QUrl::fromLocalFile(path));
}

/// @brief Enable/Disable debug log in file.
void MainWindow::enableDebugLog(bool enable) {
    mDebug = enable;
    gSignalAppender->setDetailsLevel(enable ? Logger::Debug : Logger::Info);
    gFileAppender->setDetailsLevel(enable ? Logger::Debug : Logger::Info);
    LOG_INFO(enable ? "Debug log enabled." : "Debug log disabled.");
}

/// @brief Display about box.
void MainWindow::aboutApp() {
    QString caption = QString("<h3>AFC v%1</h3>").arg(QCoreApplication::applicationVersion());
    QString aboutText =
        QMessageBox::tr("<p>Amulse/Apodrone file conversion program</p>"
                        "<p>A simple application to convert Amulse/Apodrone binary files to CSV formatted file.</p>"
                        "<p>J. Burgalat, URCA.</p>"
                        "<p>Copyright (C) 2019-2021</p>");
    QMessageBox *msgBox = new QMessageBox(this);
    msgBox->setAttribute(Qt::WA_DeleteOnClose);
    msgBox->setWindowTitle(tr("About AFC"));
    msgBox->setText(caption);
    msgBox->setInformativeText(aboutText);
    QIcon icon = QIcon(":/images/app.png");
    QSize size = icon.actualSize(QSize(128, 128));
    msgBox->setIconPixmap(icon.pixmap(size));
    msgBox->exec();
}

/// @brief Open files from dialog.
void MainWindow::openFiles() {
    QSettings settings;
    QString lastDir =
        settings
            .value("global/lastdatadir", QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).first())
            .toString();
    QString filters = "Amulse files (*.bin *results);;Raw files (*.bin);;Results files (*.results)",
            selectedFilter = "Amulse files (*.bin *results)";
    QStringList filePaths = QFileDialog::getOpenFileNames(this, tr("Select files"), lastDir, filters, &selectedFilter);
    if (filePaths.isEmpty())
        return;
    settings.setValue("global/lastdatadir", QFileInfo(filePaths.first()).dir().absolutePath());
    LOG_DEBUG(tr("Processing %1 file path(s).").arg(filePaths.size()));
    filterListOfPaths(filePaths);
}

/// @brief Open files from directory dialog.
void MainWindow::openDirectory() {
    QSettings settings;
    QString lastDir =
        settings
            .value("global/lastdatadir", QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).first())
            .toString();
    QString directory = QFileDialog::getExistingDirectory(this, tr("Select directory"), lastDir);
    if (directory.isEmpty())
        return;
    settings.setValue("global/lastdatadir", directory);
    LOG_DEBUG(tr("Processing %1 directory").arg(directory));
    filterListOfPaths(QStringList() << directory);
}

/**
 * @brief Handle the drop action within the widget.
 * @param de the DropEvent.
 * @details The method uses the Mime Data as input configuration file if the former is an url.
 * Otherwise it simply calls the base dropEvent (i.e. from QWidget).
 */
void MainWindow::dropEvent(QDropEvent *event) {
    if (event->mimeData()->hasUrls() && event->mimeData()->urls().size() > 0) {
        QStringList inputs;
        for (const QUrl &url : event->mimeData()->urls())
            inputs.append(url.toLocalFile());
        LOG_DEBUG(tr("Dropping %1 url(s)").arg(inputs.size()));
        filterListOfPaths(inputs);
    } else
        QWidget::dropEvent(event);
}

/**
 * @brief Accept the dragging process when mouse move in the widget.
 * @param de Event handler of the drag process.
 */
void MainWindow::dragMoveEvent(QDragMoveEvent *event) {
    // We accept everything here.
    event->accept();
}

/**
 * @brief Check if the window accepts the dragged data.
 * @param event Event handler of the drag process.
 */
void MainWindow::dragEnterEvent(QDragEnterEvent *event) {
    if (event->mimeData()->hasUrls() && event->mimeData()->urls().size() > 0)
        event->acceptProposedAction();
}

/**
 * @brief Sort and filter the given list of files.
 * @param inputs Inputn file paths to filter and sort
 * @details
 * The method filters out from inputs all files which extension is not ".results" or ".bin".
 * Then it splits the resulting list in two sub-lists, one for results files and the other for raw files.
 *
 * If the two lists are not empty, user is asked for a choice (the application manage only one kind of file at
 * a time). Otherwise, the non-empty list is passed to MainWindow::checkFiles.
 */
void MainWindow::filterListOfPaths(const QStringList &inputs) {
    QStringList binFiles, resultFiles;

    Qt::CaseSensitivity sens = Qt::CaseSensitive;
#ifdef Q_OS_WIN32
    sens = Qt::CaseInsensitive;
#endif
    for (const QString &path : inputs) {
        QFileInfo info(path);
        if (info.isDir()) {
            QSettings settings;
            QDirIterator it(path,
                QStringList() << "*.bin"
                              << "*.results",
                QDir::Files | QDir::NoDotAndDotDot,
                QDirIterator::Subdirectories);
            while (it.hasNext()) {
                QString fpath = it.next();
                QFileInfo finfo(fpath);
                if (!finfo.size())
                    continue;
                QString fullpath = finfo.absoluteFilePath();
                if (isResult(finfo) && !resultFiles.contains(fullpath, sens))
                    resultFiles << fullpath;
                else if (isRaw(finfo) && !binFiles.contains(fullpath, sens))
                    binFiles << fullpath;
            }
        } else if (isResult(info) && !resultFiles.contains(path, sens))
            resultFiles << path;
        else if (isRaw(info) && !binFiles.contains(path, sens))
            binFiles << path;
    }
    int nb = binFiles.size(), nr = resultFiles.size();
    // User must make a choice.
    if (nb && nr) {
        LOG_DEBUG(tr("Detected two kind of files: What should I do ??"));
        QMessageBox *msgBox = new QMessageBox(this);
        msgBox->setIcon(QMessageBox::Question);
        msgBox->setWindowTitle(tr("Time to make a choice"));
        msgBox->setText(tr("<p>The selected files have different kinds. The application can only "
                           "manage one kind of files at a time.</p>"
                           "<p>Please select the kind of files you wish to process.</p>"));
        msgBox->setStandardButtons(QMessageBox::Cancel);
        msgBox->setDefaultButton(QMessageBox::Cancel);
        msgBox->setEscapeButton(QMessageBox::Cancel);
        QAbstractButton *binBtn = msgBox->addButton("Binary files", QMessageBox::YesRole);
        msgBox->addButton("Results files", QMessageBox::NoRole);
        msgBox->exec();
        QAbstractButton *clkBtn = msgBox->clickedButton();
        if (clkBtn == msgBox->button(QMessageBox::Cancel)) {
            LOG_DEBUG(tr("Operation canceled by the user"));
            return;
        }
        bool isBin = clkBtn == binBtn;
        msgBox->deleteLater();
        if (isBin) {
            LOG_DEBUG(tr("User has made its choice: loading binary files."));
            checkAndLoad(binFiles);
        } else {
            LOG_DEBUG(tr("User has made its choice: loading results files."));
            checkAndLoad(resultFiles);
        }

    } else if (nb)
        checkAndLoad(binFiles);
    else if (nr)
        checkAndLoad(resultFiles);
    else {
        LOG_DEBUG(tr("No valid files to process from user selection."));
    }
}

/// @brief Load file paths in the model.
void MainWindow::checkAndLoad(const QStringList &files) {
    QStringList errors;
    QStringList valids;
    int instr = -1, vers = -1, tmp1, tmp2;
    for (int i = 0; i < files.size(); i++) {
        QString path = files.at(i);
        QFile file(path);
        bool ret = file.open(QIODevice::ReadOnly);
        if (!ret) {
            errors.append(tr("%1: Cannot open file for reading.").arg(path));
            continue;
        }
        if (!file.size()) {
            LOG_DEBUG(QString("%1: file is empty.").arg(path));
            continue;
        }
        ret = readInstrVerFlags(path.toStdString(), false, tmp1, tmp2);
        if (!ret) {
            errors.append(QString("%1: Cannot read instrument number and version flags.").arg(path));
            continue;
        }
        // the first valid.
        if (instr == -1 || vers == -1) {
            instr = tmp1;
            vers = tmp2;
        } else if (instr != tmp1 || vers != tmp2) {
            errors.append(
                QString("%1: Instrument number and/or firmware version inconsistent with previous files.").arg(path));
            continue;
        }
        valids.append(path);
    }

    // WE SHOULD PROBABLY WARN USER ABOUT INVALID FILES !
    if (errors.size()) {
        for (QString e : errors)
            LOG_WARNING(e);
        QMessageBox::information(this,
            tr("Inconsistent files"),
            tr("<p>Some files are not valid or not consistent with each other.</p>"
               "<p>Only files consistent with the first valid file will be taken into account.</p>"
               "<p>See the log activity window for more detail.</p>"));
    }
    if (valids.size()) {
        // Check for a fuckin descriptor !
        StringList dirs;
        for (const QString &p : valids)
            dirs.push_back(QFileInfo(p).absolutePath().toStdString());
        if (getDescriptorPath(instr, vers, dirs).empty()) {
            LOG_ERROR(tr("No descriptor in the database matches the couple (i:%1,v:%2).").arg(instr).arg(vers));
            // no descriptor :(
            QMessageBox::critical(this,
                tr("No descriptor"),
                tr("<p>No descriptor has been found matching the list of valid files.</p>"
                   "<p>There should be a descriptor file (*.csv) in the parent directory of each selected files. "
                   "If such file is not found, you have probably moved the files elsewhere without moving the "
                   "descriptor.</p>"
                   "<p>Unfortunately, I cannot do anything more for you...</p>"));
            mXMLDCurrent = nullptr;
            mPrmChoiceWidget->setupFromXMLDescriptor(mXMLDCurrent);
            return;
        }
        // then load files.
        loadFiles(QVector<QStringList>() << valids);
    }
}

/// @brief Load files in dataset
void MainWindow::loadFiles(const QVector<QStringList> &paths) {
    if (paths.isEmpty())
        return;
    setAcceptDrops(false);
    mBtnConvert->setEnabled(false);
    mPrmChoiceWidget->setEnabled(false);
    mActionOpen->setEnabled(false);
    int nf = 0;
    for (const QStringList &l : paths)
        nf += l.size();
    LOG_INFO(tr("Loading %1 data files...").arg(nf));

    mProgressBar->progressBar()->setRange(0, nf);
    mProgressBar->setVisible(true);
    mFirst = true;
    MultiFileLoader *loader = new MultiFileLoader(QVector<QStringList>() << paths);

    // connect signals/slots
    // progressions
    QObject::connect(loader, &MultiFileLoader::progressRangeChanged, this, &MainWindow::onProgressRangeChanged);
    QObject::connect(loader, &MultiFileLoader::progressChanged, this, &MainWindow::onProgressChanged);
    // on dataset loaded.
    QObject::connect(loader, &MultiFileLoader::datasetLoaded, this, &MainWindow::onDatasetLoaded);
    // informations messages
    QObject::connect(loader, &MultiFileLoader::message, this, &MainWindow::logMessage);
    // clean up before deletion
    QObject::connect(loader, &MultiFileLoader::finished, this, &MainWindow::onLoadingFinished);
    // abort process.
    QObject::connect(this, &MainWindow::abortTasks, loader, &MultiFileLoader::cancel);

    mProgressBar->progressBar()->setValue(0);
    mThreadPool->start(loader);
}

/// brief Send message to global logger instance (DEBUG level).
void MainWindow::logMessage(const QString &message) {
    if (mDebug)
        LOG_DEBUG(message);
}

/// @brief Convert the list of selected parameters.
void MainWindow::convertParameters() {
    if (!(mXMLDCurrent || mPrmChoiceWidget->hasSelection())) {
        if (!mXMLDCurrent)
            LOG_DEBUG("No current descriptor is defined.");
        if (!mPrmChoiceWidget->hasSelection())
            LOG_DEBUG("No parameters are selected for conversion.");
        return;
    }
    QSettings settings;
    QString iDir =
        settings.value("global/lastdir", QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0))
            .toString();
    QString fileName = QFileDialog::getSaveFileName(this, tr("Save selection to file"), iDir, tr("CSV files (*.csv)"));
    if (fileName.isEmpty())
        return;
    LOG_DEBUG(tr("Selected output file: %1").arg(fileName));
    iDir = QFileInfo(fileName).absoluteDir().absolutePath();
    settings.setValue("global/lastdir", iDir);

    PairOfData pairs = dataSelection(mPrmChoiceWidget->selected());

    CSVFileWriter *writer = new CSVFileWriter(fileName, pairs);

    // connect signals/slots
    // progressions
    QObject::connect(writer, &CSVFileWriter::progressRangeChanged, this, &MainWindow::onProgressRangeChanged);
    QObject::connect(writer, &CSVFileWriter::progressChanged, this, &MainWindow::onProgressChanged);
    // clean up before deletion
    QObject::connect(writer, &CSVFileWriter::finished, this, &MainWindow::onDataConverted);
    // abort process.
    QObject::connect(this, &MainWindow::abortTasks, writer, &CSVFileWriter::cancel);

    mPrmChoiceWidget->setEnabled(false);
    mActionOpen->setEnabled(false);
    setAcceptDrops(false);
    LOG_DEBUG(tr("Starting conversion process."));
    mThreadPool->start(writer);
    return;
}

/// @brief Load dataset and setup widgets.
void MainWindow::onDatasetLoaded(AMDataset *dataset) {
    if (!dataset)
        return;
    if (!(mFirst && dataset->isValid())) {
        if (mFirst)
            LOG_ERROR(tr("Failed to load files."));
        dataset->deleteLater();
        dataset = nullptr;
        return;
    }
    if (mFirst)
        mFirst = false;
    LOG_INFO(tr("%1 data files loaded (%2 blocks, i:%3, v:%4).")
                 .arg(dataset->fileCount())
                 .arg(dataset->blockCount())
                 .arg(dataset->instrument())
                 .arg(dataset->version())

    );
    mInstrument = dataset->instrument();
    mVersion = dataset->version();
    mDatasetInfoWidget->setDataSet(dataset);

    mData.clear();
    const ABlock &descriptor = dataset->descriptor();
    for (const AParam &p : descriptor.parameters()) {
        DoubleVector v = dataset->parameter(p);
        QVector<double> tmp;
        for (const auto &vv : v)
            tmp.push_back(vv);
        mData.insert(QString::fromStdString(p.name()), tmp);
    }

    mXMLDCurrent = mXMLDContainer.get(mInstrument, mVersion, dataset->parameters());
    mPrmChoiceWidget->setupFromXMLDescriptor(mXMLDCurrent);
    mBtnConvert->setEnabled(mPrmChoiceWidget->hasSelection());
    dataset->deleteLater();
}

/// @brief Update some widgets on loading finished.
void MainWindow::onLoadingFinished(bool aborted) {
    if (aborted)
        statusBar()->showMessage(tr("Job aborted."), 1000);
    MultiFileLoader *loader = qobject_cast<MultiFileLoader *>(QObject::sender());
    if (loader)
        QObject::disconnect(this, &MainWindow::abortTasks, loader, &MultiFileLoader::cancel);
    setAcceptDrops(true);
    mActionOpen->setEnabled(true);
    mPrmChoiceWidget->setEnabled(true);
    mProgressBar->progressBar()->reset();
    mProgressBar->hide();
}

/**
 * @brief Perform conversion post-actions.
 * @param aborted True if the conversion has been aborted, false otherwise.
 */
void MainWindow::onDataConverted(bool aborted) {
    CSVFileWriter *writer = qobject_cast<CSVFileWriter *>(QObject::sender());
    if (writer)
        QObject::disconnect(this, &MainWindow::abortTasks, writer, &CSVFileWriter::cancel);
    mProgressBar->progressBar()->reset();
    mProgressBar->hide();
    mActionOpen->setEnabled(true);
    mPrmChoiceWidget->setEnabled(true);
    setAcceptDrops(true);
    if (aborted) {
        LOG_WARNING(tr("Conversion aborted."));
        statusBar()->showMessage(tr("Conversion aborted."), 1000);
    } else {
        LOG_INFO(tr("Conversion done."));
        statusBar()->showMessage(tr("Conversion done."), 1000);
        saveDescriptor();
    }
}

/// @brief Save the current descriptor to disk.
void MainWindow::saveDescriptor() {
    if (mChbSaveChanges->isChecked() && mXMLDCurrent) {
        // update current XML descriptor.
        mPrmChoiceWidget->updateOriginal();
        // save the file in gDescrDir
        QString xmlpath = QString("%1/%2").arg(gDescrDir, mXMLDCurrent->fileName());
        QFile file(xmlpath);
        if (!file.open(QIODevice::Text | QIODevice::WriteOnly)) {
            LOG_WARNING(tr("Failed to save the current descriptor."));
            return;
        }
        file.write(mXMLDCurrent->toXML(true));
        file.close();
        LOG_INFO(tr("Current descriptor successfully saved to disk."));
    }
}

/**
 * @brief Get a pair of headers, data pointers corresponding to the given pair of strings.
 * @param pairs Pairs of string with the parameter name and alias to output.
 * @internal
 * Pointers are owned by the model and should not be deleted.
 */
PairOfData MainWindow::dataSelection(QVector<PairOfStrings> pairs) {
    QMutableVectorIterator<PairOfStrings> it(pairs);
    PairOfData out;
    QVector<QVector<double> *> &params = out.second;
    QStringList &headers = out.first;
    while (it.hasNext()) {
        PairOfStrings p = it.next();
        if (!mData.contains(p.first))
            it.remove();
        else {
            headers.append(p.second);
            params.push_back(&mData[p.first]);
        }
    }
    return out;
}

/// @brief Check whether or not the conversion buton can be enabled.
void MainWindow::onItemMoved() { mBtnConvert->setEnabled(mPrmChoiceWidget->hasSelection()); }

/**
 * @brief Update progress bar(s) on range changed.
 * @param minValue Minimum progress value.
 * @param maxValue Maximum progress value.
 */
void MainWindow::onProgressRangeChanged(int minValue, int maxValue) {
    mProgressBar->progressBar()->setRange(minValue, maxValue);
    mProgressBar->setVisible(true);
}

/// @brief update some widget on progress changes.
void MainWindow::onProgressChanged(int value) {
    mProgressBar->progressBar()->setValue(value);
    int numTasks = mProgressBar->progressBar()->maximum();
    if (qobject_cast<MultiFileLoader *>(QObject::sender()))
        statusBar()->showMessage(tr("Loading files (%1/%2)...").arg(value).arg(numTasks));
    else if (qobject_cast<CSVFileWriter *>(QObject::sender()))
        statusBar()->showMessage(tr("Conversion in progress. Please wait..."));
}

void MainWindow::cancelOperations() {
    if (mThreadPool->activeThreadCount() > 0) {
        QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
        emit abortTasks();
        mThreadPool->waitForDone();
        QApplication::restoreOverrideCursor();
    }
}
