// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "xmldescriptors.hpp"
#include "messagehandler.hpp"
#include <QApplication>
#include <QXmlSchema>
#include <QXmlSchemaValidator>
#include <QtXmlPatterns>

/// @brief Constructor.
XMLDescriptor::XMLDescriptor() : instrument(-1), version(-1) {}

/// @brief Copy constructor.
XMLDescriptor::XMLDescriptor(const XMLDescriptor &other) {
    instrument = other.instrument;
    version = other.version;
    parameters = other.parameters;
    selection = other.selection;
    inputPath = other.inputPath;
}

/// @brief Copy assignment operator.
XMLDescriptor &XMLDescriptor::operator=(const XMLDescriptor &other) {
    if (this != &other) {
        instrument = other.instrument;
        version = other.version;
        parameters = other.parameters;
        selection = other.selection;
        inputPath = other.inputPath;
    }
    return *this;
}

/// @brief Get a (xml) string representation of the instance.
QByteArray XMLDescriptor::toXML(bool pretty) const {
    QString tplPrm("<parameter><name>%1</name><alias>%2</alias></parameter>");
    QString tplDes("<descriptor><instrument>%1</instrument><version>%2</version>%3</descriptor>");
    QString tplSel("<selection>%1</selection>");
    QStringList prms;
    for (const PairOfStrings &p : parameters)
        prms.append(p.second.isEmpty() ? tplPrm.arg(p.first, p.first) : tplPrm.arg(p.first, p.second));
    QString sel = (selection.isEmpty()) ? QString() : QString("<name>%1</name>").arg(selection.join("</name><name>"));
    QString xmlIn = QString("<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
                            "<adescr version=\"1\">%1%2</adescr>")
                        .arg(tplDes.arg(instrument).arg(version).arg(prms.join("")), tplSel.arg(sel));

    if (!pretty)
        return xmlIn.toUtf8();

    QString xmlOut;
    QXmlStreamReader reader(xmlIn);
    QXmlStreamWriter writer(&xmlOut);
    writer.setAutoFormatting(true);
    while (!reader.atEnd()) {
        reader.readNext();
        if (!reader.isWhitespace()) {
            writer.writeCurrentToken(reader);
        }
    }
    return xmlOut.toUtf8();
}

/**
 * @brief Get/compute a filename from the container informations.
 * @return File name.
 * @details
 * If the internal inputPath of the descriptor is empty a file name is built
 * using the instrument number, firmware version and number of parameters.
 * Otherwise the method returns the filename of the input path.
 */
QString XMLDescriptor::fileName() const {
    return inputPath.isEmpty() ? QString("i%1_v%2_p%3.adescr").arg(instrument).arg(version).arg(parameters.size())
                               : QFileInfo(inputPath).fileName();
}

/// @brief Check whether or not the given parameter name is stored in the container.
bool XMLDescriptor::hasParameter(const QString &name) {
    for (const PairOfStrings &p : parameters) {
        if (p.first.compare(name, Qt::CaseInsensitive) == 0)
            return true;
    }
    return false;
}

/**
 * @brief Remove and return the pair matching the given parameter name.
 * @param name Name of the parameter to search (case-insensitive).
 * @return Pair of string with the parameter name and its alias if found in the container. Otherwise, an empty pair is
 * returned.
 */
PairOfStrings XMLDescriptor::pop(const QString &name) {
    auto it = std::find_if(parameters.begin(), parameters.end(), [=](const PairOfStrings &p) {
        return p.first.compare(name, Qt::CaseInsensitive) == 0;
    });
    if (it == parameters.end())
        return PairOfStrings();

    PairOfStrings out(*it);
    parameters.erase(it);
    return out;
}

/// @brief Set new alias for the given parameter name.
void XMLDescriptor::setAlias(const QString &name, const QString &alias) {
    auto it = std::find_if(parameters.begin(), parameters.end(), [=](const PairOfStrings &p) {
        return p.first.compare(name, Qt::CaseInsensitive) == 0;
    });
    if (it != parameters.end())
        it->second = alias;
}

/// @brief Constructor.
XMLDContainer::XMLDContainer() {
    QFile file(":/files/adescr.xsd");
    if (file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        QTextStream in(&file);
        mXsdSchema = in.readAll().toUtf8();
    }
}

/// @brief Destructor.
XMLDContainer::~XMLDContainer() {
    qDeleteAll(mDescriptors);
    mDescriptors.clear();
}

/**
 * @brief Load all adescr files from the given directory and stores them in the container.
 * @param path Directory path to process.
 * @return Number of XML descriptor parsed.
 */
int XMLDContainer::setupFromDirectory(const QString &path) {
    QDirIterator it(path, QStringList() << "*.adescr", QDir::Files | QDir::NoDotAndDotDot);
    mDescriptors.clear();
    int count = 0;
    while (it.hasNext()) {
        QStringList errors;
        QString p = it.next();
        if (!parse(p, &errors)) {
            for (const QString &e : errors)
                LOG_ERROR(e);
        } else
            count++;
    }
    return count;
}

/**
 * @brief Get the XML Descriptor that matches instrument and version values.
 * @param instrument Instrument number to check.
 * @param version Firmware version to check.
 * @param nparams Optional expected number of parameters.
 * @return Pointer to an XML descriptor container.
 * If no internal XML descriptors matches instrument and version, the method attempts
 * to build an XML Descriptor from a CSV descriptor. As last resort, a null XMLDescriptor pointer is returned.
 * @attention
 * Unless null, the pointer is owned by the container and should not be deleted.
 *
 * @internal
 * The method is already patched for a bug introduced by ADMin:
 * ADmin sets the same instrument number and version than the input block coming from Amulse for the results file.
 * This is a violation of the convention used:
 *   - results file should have the same firmware version than their raw file counterpart.
 *   - instrument number of results files should be incremented, by default, by 1000 (from file fit workers)
 *     or 2000 (block fit workers, specific to ADMin).
 *
 * Data generated by ADMin before 27/03/2019 contains same instrument number and version. Then we
 * use nparams as a last resort to get a usable database of XML descriptor.
 */
XMLDescriptor *XMLDContainer::get(int instrument, int version, const QVector<AParam> &params) {
    int nparams = params.size();
    auto it = std::find_if(mDescriptors.begin(), mDescriptors.end(), [=](XMLDescriptor *d) {
        bool ret = d->instrument == instrument && d->version == version;
        if (!ret)
            return false;
        else if (nparams > 0)
            return d->parameters.size() == nparams;
        return true;
    });
    if (it != mDescriptors.end())
        return *it;
    // no descriptor found: let's build a descriptor from the list of parameters.
    if (!nparams)
        return nullptr;
    XMLDescriptor *out = new XMLDescriptor;
    out->instrument = instrument;
    out->version = version;
    for (const AParam &p : params)
        out->parameters.append({QString::fromStdString(p.name()), QString::fromStdString(p.name())});
    mDescriptors.append(out);
    return out;
}

/// @brief Parse the xml file pointed by xmlPath and setup the parameters.
/// @sa Configuration::parseContent
bool XMLDContainer::parse(const QString &xmlPath, QStringList *errors) {
    QFile xmlFile(xmlPath);
    // cannot open files.
    if (!xmlFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QApplication::tr("Cannot open %1.").arg(xmlPath));
        return false;
    }
    QString instance = xmlFile.readAll();
    xmlFile.close();
    return parseContent(instance, &xmlPath, errors);
}

/**
 * @brief Validate and parse the input content.
 * @param content XML formatted content to parse.
 * @return True on success, false otherwise.
 * @note
 * On success, a new XMLDescriptor instance is added to the internal vector.
 */
bool XMLDContainer::parseContent(const QString &content, const QString *path, QStringList *errors) {
    if (!validateContent(content, errors))
        return false;
    QDomDocument document;
    document.setContent(content);
    QDomNode rootNode = document.firstChildElement("adescr");
    XMLDescriptor *descr = new XMLDescriptor;
    if (path)
        descr->inputPath = *path;
    // read descriptor node
    QDomNode n = rootNode.firstChildElement("descriptor").firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            bool ok = true;
            if (tagName.compare("instrument") == 0) {
                descr->instrument = e.text().trimmed().toInt(&ok);
                if (!ok)
                    descr->instrument = -1;
            } else if (tagName.compare("version") == 0) {
                descr->version = e.text().trimmed().toInt(&ok);
                if (!ok)
                    descr->version = -1;
            } else if (tagName.compare("parameter") == 0)
                descr->parameters.push_back(readParameter(e));
        }
        n = n.nextSibling();
    }
    n = rootNode.firstChildElement("selection").firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            if (tagName.compare("name") == 0)
                descr->selection.push_back(e.text().trimmed());
        }
        n = n.nextSibling();
    }
    qDebug() << descr->instrument << descr->version << (descr->instrument == -1 || descr->version == -1);
    if (!(descr->instrument == -1 || descr->version == -1))
        mDescriptors.push_back(descr);
    return true;
}

/// @brief Read a parameter node.
PairOfStrings XMLDContainer::readParameter(QDomElement parent) {
    PairOfStrings out("", "");
    QDomNode n = parent.firstChild();
    while (!n.isNull()) {
        if (n.isElement()) {
            QDomElement e = n.toElement();
            QString tagName = e.tagName();
            if (tagName.compare("name") == 0)
                out.first = e.text().trimmed();
            else if (tagName.compare("alias") == 0)
                out.second = e.text().trimmed();
        }
        n = n.nextSibling();
    }
    return out;
}

/// @brief Clear the container.
void XMLDContainer::clear() {
    qDeleteAll(mDescriptors);
    mDescriptors.clear();
}

/// @brief Checked whether or not the content of the path given by xmlPath is valid.
bool XMLDContainer::validate(const QString &xmlPath, QStringList *errors) {
    QFile xmlFile(xmlPath);
    // canot open files.
    if (!xmlFile.open(QIODevice::ReadOnly | QIODevice::Text)) {
        if (errors)
            errors->append(QApplication::tr("Cannot open %1.").arg(xmlPath));
        return false;
    }
    QString instance = xmlFile.readAll();
    xmlFile.close();
    return validateContent(instance, errors);
}

/// @brief Check whether or not the content defined by xmlContent is valid.
bool XMLDContainer::validateContent(const QString &xmlContent, QStringList *errors) {
    const QByteArray instanceData = xmlContent.toUtf8();
    MessageHandler messageHandler;
    QXmlSchema schema;
    schema.setMessageHandler(&messageHandler);
    bool errorOccurred = false;
    if (!mXsdSchema.isEmpty())
        schema.load(mXsdSchema);
    if (!schema.isValid()) {
        errorOccurred = true;
        if (errors)
            errors->append(messageHandler.statusMessage());
    } else {
        QXmlSchemaValidator validator(schema);
        if (!validator.validate(instanceData)) {
            errorOccurred = true;
            if (errors)
                errors->append(messageHandler.statusMessage());
        }
    }
    return !errorOccurred;
}
