// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "fit.hpp"
#include "isp/isp.hpp"
#include "simplefilelogger.hpp"

ExecTimes fitBlock(ABlock &block, const FitConfig &fitConfig, const std::map<int, std::pair<double, double>> &eSignals,
    const std::map<int, double> &ptScales, const double &pres, const double &temp, const double &cH2O0,
    const std::map<int, DoubleVector> &spectras, const FitParamReuse &reuseParam,
    const std::map<std::pair<int, int>, FitParams> &previousFitParams, std::vector<BasicFitReport> *reports = nullptr,
    StringList *errors = nullptr);

#include <chrono>

/// @brief Process input file and set output data in output file.
bool processFile(const QString &inputPath, const QString &outputPath, const FitConfig &fitConfig, int instrOffset,
    const FitParamReuse &reuseParams, int blockStep, Decimation decimTo, bool preLoad, bool logToFile,
    bool stripSpectras, bool asCSV, ExecTimes *times) {

    using std::chrono::duration;
    using std::chrono::duration_cast;
    using std::chrono::high_resolution_clock;
    using std::chrono::milliseconds;
    using std::chrono::nanoseconds;

    ExecTimes ttimes;
    if (times)
        *times = ttimes;
    QFileInfo info(outputPath);
    if (!info.absoluteDir().mkpath("."))
        return false;
    ADataset *dataset = new ADataset(StringList{inputPath.toStdString()});

    std::map<std::pair<int, int>, FitParams> previousFitParams;

    // Log file (if any)
    QString logfile;
    if (logToFile)
        logfile = QString("%1/%2.log").arg(info.absolutePath(), info.completeBaseName());
    SimpleFileLogger *fileLogger = new SimpleFileLogger(logfile, true);
    fileLogger->write(QString("INFO : %1: Starting inversion process.\n").arg(inputPath));
    if (preLoad)
        dataset->loadRamps();

    int nb = dataset->blockCount();
    int npts = dataset->rampSize();

    int decimFactor = 1;
    if (decimTo != Decimation::No) {
        decimFactor = npts / static_cast<int>(decimTo);
        if (decimFactor <= 1)
            fileLogger->write(QString("WARN : I will not decimate %1 points to %2 points, because it is non-sense !\n")
                                  .arg(npts, static_cast<int>(decimTo)));
    }

    FileHandler *fileHandler = nullptr;
    if (asCSV) {
        fileHandler = new CsvFileHandler(outputPath.toStdString(), false);
    } else {
        fileHandler = new BinFileHandler(outputPath.toStdString(), false);
    }

    //    ADataFileHandler *fileHandler = new ADataFileHandler(outputPath, false, 0, 1000);
    StringList errors;
    for (int i = 0; i < nb; i += blockStep) {
        auto t1 = high_resolution_clock::now();
        std::map<int, double> ptScales, mOffsets;
        std::map<int, std::pair<double, double>> eSignals;
        std::map<int, DoubleVector> ramps;
        std::vector<BasicFitReport> reports;
        errors.clear();
        ABlock block = dataset->blockAt(i);
        // decimate if possible.
        block.decimate(decimFactor);
        // update the instrument number flag.
        block.setInstrument(block.instrument() + instrOffset);
        block.updateValue("iversion", fitConfig.cfgVersion());
        double pres, temp, cH2O0;
        std::map<int, DoubleVector> spectras;
        // setup data from block.
        if (!setupDataFromBlock(
                block, fitConfig, ptScales, pres, temp, cH2O0, spectras, ramps, mOffsets, eSignals, &errors)) {
            if (logToFile) {
                for (const auto &e : errors)
                    fileLogger->write(QString("ERROR: %1\n").arg(QString::fromStdString(e)));
                fileLogger->write(QString("ERROR: Skipping invalid block #%1.\n").arg(i + 1));
            }
            if (stripSpectras)
                block.stripSpectras();
            // save the block in file handler.
            if (!fileHandler->registerBlock(block))
                fileLogger->write(
                    QString("ERROR: Block #%1 is not consistent with output file structure.\n").arg(i + 1));
            continue;
        }
        auto t2 = high_resolution_clock::now();

        // fit the block
        ExecTimes btimes = fitBlock(block,
            fitConfig,
            eSignals,
            ptScales,
            pres,
            temp,
            cH2O0,
            spectras,
            reuseParams,
            previousFitParams,
            &reports,
            &errors);

        btimes.preFit += (qint64)duration_cast<nanoseconds>(t2 - t1).count();

        t1 = high_resolution_clock::now();
        // log errors if any
        if (errors.size() && logToFile) {
            for (const auto &e : errors)
                fileLogger->write(QString("ERROR: %1\n").arg(QString::fromStdString(e)));
        }

        // strip spectras from output block before send it to the file handler.
        if (stripSpectras)
            block.stripSpectras();
        // save the block in file handler.
        if (!fileHandler->registerBlock(block))
            fileLogger->write(QString("ERROR: Block #%1 is not consistent with output file structure.\n").arg(i + 1));

        // save previous parameters
        for (const BasicFitReport &report : reports) {
            if (report.spectraId == -1 || report.lineId == -1 || !report.fitSuccess())
                continue;
            std::pair<int, int> slid = std::make_pair(report.spectraId, report.lineId);
            previousFitParams[slid] = report.finalParams;
        }
        t2 = high_resolution_clock::now();
        btimes.postFit += (qint64)duration_cast<nanoseconds>(t2 - t1).count();

        ttimes += btimes;

        // additional logging:
        if (logToFile) {
            QString msg =
                QString("INFO : Block #%1, %2\n").arg(i + 1).arg(QString::fromStdString(btimes.toString(Units::USEC)));
            fileLogger->write(msg);
            // 3.4 Log out summary
            for (const BasicFitReport &report : qAsConst(reports)) {
                if (!(report.spectraId == -1 || report.lineId == -1))
                    fileLogger->write(QString("INFO : Block #%1, Spectra #%2, Line #%3 report:\n%4\n")
                                          .arg(i + 1)
                                          .arg(report.spectraId)
                                          .arg(report.lineId)
                                          .arg(QString::fromStdString(report.summary())));
            }
        }
    }
    fileHandler->close();
    fileLogger->write(QString("INFO: %1: Inversion process finished !\n").arg(inputPath));
    fileLogger->close();
    delete fileLogger;
    delete fileHandler;
    delete dataset;

    if (times)
        *times = ttimes;

    return true;
}

/**
 * @brief Fit the spectras of the given block.
 */
ExecTimes fitBlock(ABlock &block, const FitConfig &fitConfig, const std::map<int, std::pair<double, double>> &eSignals,
    const std::map<int, double> &ptScales, const double &pres, const double &temp, const double &cH2O0,
    const std::map<int, DoubleVector> &spectras, const FitParamReuse &reuseParams,
    const std::map<std::pair<int, int>, FitParams> &previousFitParams, std::vector<BasicFitReport> *reports,
    StringList *errors) {
    ExecTimes ttimes;
    QElapsedTimer timer;
    qint64 tmp = 0, lastElapsed = 0;
    // process fits.
    timer.start();
    // map of all result FitParams.
    std::map<int, FitParams> fitParamsMap;
    if (reports) {
        reports->resize(fitConfig.spectras().size());
        for (BasicFitReport &report : *reports)
            report = BasicFitReport();
    }
    int scounter = 0;
    for (const SParam &sp : fitConfig.spectras()) {
        ExecTimes stimes;
        tmp = timer.nsecsElapsed();
        lastElapsed = tmp;
        LParam lptr = sp.line;
        int sid = sp.id;
        int lid = lptr.id;
        // 1. Setup the parameter for the fit.
        FitParams iFitParams;
        DoubleVector spc =
            setupFitParameters(sp, iFitParams, fitParamsMap, ptScales, pres, temp, cH2O0, spectras, errors);
        int spcsize = spc.size();
        // No spectra data: some errors occured in the setup:
        //  ==> Finalize the process for this spectra and go to the next.
        if (spc.empty()) {
            const StringList names = iFitParams.names();
            // set nan for each parameter value and error !
            for (const std::string &name : names) {
                iFitParams.setValue(name, qQNaN());
                iFitParams.setError(name, qQNaN());
            }
            // We must set a fake fit params set (in case of spectra dependents).
            fitParamsMap[sid] = iFitParams;
            continue;
        }

        // update initial guess from previous valid fit params if required
        std::pair<int, int> slid = std::make_pair(sid, lid);

        if (reuseParams.atLeast() && previousFitParams.find(slid) != previousFitParams.end()) {
            const FitParams ofp = previousFitParams.at(slid);
            if (reuseParams.concentration) {
                for (const MParam &p : ofp.molecules()) {
                    if (p.vary && !qIsNaN(p.value)) // can reuse H2O conc !
                        iFitParams.setValue(p.name, p.value);
                }
            }
            if (reuseParams.baseLine) {
                for (const CParam &p : ofp.baseLine()) {
                    if (p.vary && !qIsNaN(p.value))
                        iFitParams.setValue(p.name, p.value);
                }
            }
            if (reuseParams.sigmaScale) {
                for (const CParam &p : ofp.sigmaScale()) {
                    if (p.vary && !qIsNaN(p.value))
                        iFitParams.setValue(p.name, p.value);
                }
            }
        }

        // 2. Setup MPFIT solver
        // 2.1 build raw parameters data, mapping.
        std::map<std::string, int> mapping;
        double *varParams = new double[static_cast<unsigned int>(iFitParams.size())];
        mp_par_struct *parameters = buildMpFitParameters(iFitParams, mapping, varParams);
        if (parameters == nullptr) {
            delete[] varParams;
            const StringList names = iFitParams.names();
            for (const std::string &name : names) {
                iFitParams.setValue(name, qQNaN());
                iFitParams.setError(name, qQNaN());
            }
            fitParamsMap[sid] = iFitParams;
            continue;
        }

        // 2.2 generate struct of fixed parameters for the residual function.
        FitData fixedVars(mapping, iFitParams, spc);
        // 2.3 Prepare MPFIT configuration.
        const FCParameters &fcParams = fitConfig.fcParameters();
        mp_config config = {fcParams.fTol,
            fcParams.pTol,
            fcParams.gTol,
            MP_MACHEP0,
            100.0,
            1e-12,
            fcParams.maxIter,
            fcParams.maxFEV,
            0,
            0,
            0,
            nullptr};
        // prepare fit result structure
        mp_result result;
        int np = mapping.size();
        double *xerror = new double[np];
        memset(&result, 0, sizeof(result));
        result.xerror = xerror;
        // 3. launch mpfit
        tmp = timer.nsecsElapsed();
        stimes.preFit += tmp - lastElapsed;
        lastElapsed = tmp;
        int info =
            mpfit(&mpResFunc, spcsize, np, varParams, parameters, &config, static_cast<void *>(&fixedVars), &result);
        result.status = info; // force status: it is not updated if the process fails before actual computations start.
        tmp = timer.nsecsElapsed();
        stimes.fit += tmp - lastElapsed;
        lastElapsed = tmp;
        // 4. Compute the fit report.
        BasicFitReport report;
        mpFillReportFromFit(
            &report, result, iFitParams, varParams, mapping, fcParams.nullLower, fcParams.nullUpper, sid, lid);
        fitParamsMap[sid] = report.finalParams;
        if (report.status != 0 && errors)
            errors->push_back(QString("Failed to fit spectra #%1 of line #%2.").arg(sid).arg(lid).toStdString());
        // clean up.
        delete[] xerror;
        delete[] varParams;
        delete[] parameters;
        // 5. update current data block.
        updateDataBlock(&block, report, spc);
        // setup exec times.

        tmp = timer.nsecsElapsed();
        stimes.postFit += tmp - lastElapsed;
        stimes.global = stimes.preFit + stimes.fit + stimes.postFit;
        report.times = stimes;
        ttimes += stimes;
        if (reports)
            (*reports)[scounter] = report;
        scounter++;
    }
    ttimes.global = timer.nsecsElapsed();
    return ttimes;
}

#include <iostream>
void printString(const QString &string) { std::cerr << string.toStdString(); }
