// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2018

/**
 * @file logwidget.cpp
 * @brief Log widget definitions file
 * @author J. Burgalat
 * @date 2016-2018
 */

#include <QtCore>
#include <QtGui>
#include <QtWidgets>

#include "logwidget.h"

/**
 * \class LogWidget
 * \brief LogWidget is a simple widget that displays and filter log records sends by a SignalAppender of CuteLogger library.
 */

/**
 * @brief Constructor.
 * @param appender SignalAppender instance to listen.
 * @param settingsKey Optional key for storing widget geometry using QSettings. If set to empty string (default), the class will
 * not save/restore geometry of the widget.
 * @param parent Parent widget.
 */
LogWidget::LogWidget(SignalAppender *appender, const QString &settingsKey, QWidget *parent) : QWidget(parent) {
#ifndef LW_EMBEDDED
    Q_INIT_RESOURCE(logwidget);
#endif
    mAppender = nullptr;
    mFilterFlag = Logger::Info;
    mSettingsKey = settingsKey;
    QVBoxLayout *mainLayout = new QVBoxLayout(this);

    logBox = new QTextBrowser(this);
    toolBar = new QToolBar(this);

    QFont font = logBox->font();
    font.setFamily(QString::fromUtf8("Courier"));
    font.setFixedPitch(true);
    logBox->setFont(font);
    logBox->setLineWrapMode(QTextEdit::NoWrap);

    mToggleViewAction = new QAction(this);
    mToggleViewAction->setCheckable(true);
    mToggleViewAction->setText(windowTitle());
    QObject::connect(mToggleViewAction, &QAction::triggered, this, [=](bool checked) { setHidden(!checked); });
    QObject::connect(this, &LogWidget::visibilityChanged, mToggleViewAction, &QAction::setChecked);

    mSaveAction = toolBar->addAction(QIcon(":/logwidget/save.png"), QString(), this, SLOT(saveLog()));
    mClearAction = toolBar->addAction(QIcon(":/logwidget/clear.png"), QString(), this, SLOT(clearLog()));
    QWidget *spacer = new QWidget();
    spacer->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
    toolBar->addWidget(spacer);
    toolBar->setIconSize(QSize(16, 16));
    mShowErrorAction = toolBar->addAction(QIcon(":/logwidget/error.png"), QString());
    mShowErrorAction->setCheckable(true);
    QObject::connect(mShowErrorAction, &QAction::toggled, this, &LogWidget::toggleError_p);
    mShowWarningAction = toolBar->addAction(QIcon(":/logwidget/warn.png"), QString());
    mShowWarningAction->setCheckable(true);
    QObject::connect(mShowWarningAction, &QAction::toggled, this, &LogWidget::toggleWarning_p);
    mShowInfoAction = toolBar->addAction(QIcon(":/logwidget/info.png"), QString());
    mShowInfoAction->setCheckable(true);
    QObject::connect(mShowInfoAction, &QAction::toggled, this, &LogWidget::toggleInfo_p);
    mShowDebugAction = toolBar->addAction(QIcon(":/logwidget/debug.png"), QString());
    mShowDebugAction->setCheckable(true);
    QObject::connect(mShowDebugAction, &QAction::toggled, this, &LogWidget::toggleDebug_p);
    mShowErrorAction->setChecked(true);
    mShowWarningAction->setChecked(true);
    mShowInfoAction->setChecked(true);
    mShowDebugAction->setChecked(false);

    setAppender(appender);

    mainLayout->setSpacing(2);
    mainLayout->setContentsMargins(6, 6, 6, 6);
    mainLayout->addWidget(logBox);
    mainLayout->addWidget(toolBar);
    translateUI();
    loadUI();
}

/// @brief Destructor.
LogWidget::~LogWidget() {
    clearLog();
    setAppender(nullptr);
}

/// @brief Get the size hint of the widget
QSize LogWidget::sizeHint() const { return QSize(500, 800); }

/// @brief Get the toggleViewAction of the widget.
QAction *LogWidget::toggleViewAction() { return mToggleViewAction; }

/**
 * @brief Set SignalAppender.
 * @param appender SignalAppender instance to listen.
 * @details
 * Change the SignalAppender to listen to.
 *
 * Setting a nullptr for <i>appender</i> stops listening.
 */
void LogWidget::setAppender(SignalAppender *appender) {
    if (mAppender)
        QObject::disconnect(mAppender, &SignalAppender::logMessage, this, &LogWidget::append);

    mAppender = appender;
    if (mAppender)
        QObject::connect(mAppender, &SignalAppender::logMessage, this, &LogWidget::append);
}

/// @brief Get the current listened SignalAppender (may be null).
QPointer<SignalAppender> LogWidget::appender() {
    return mAppender;
}

/// @brief Show/Hide toolbar.
void LogWidget::setToolBarVisible(bool visible) { toolBar->setVisible(visible); }

/**
 * @brief Update log filter for error messages.
 * @param enabled True if error messages should be shown, false otherwise.
 */
void LogWidget::toggleError_p(bool enabled) {
    mFilterFlag = enabled ? mFilterFlag | Logger::Error : mFilterFlag - Logger::Error;
    refresh();
}
/**
 * @brief Update log filter for warning messages.
 * @param enabled True if warning messages should be shown, false otherwise.
 */
void LogWidget::toggleWarning_p(bool enabled) {
    mFilterFlag = enabled ? mFilterFlag | Logger::Warning : mFilterFlag - Logger::Warning;
    refresh();
}

/**
 * @brief Update log filter for information messages.
 * @param enabled True if information messages should be shown,false otherwise.
 */
void LogWidget::toggleInfo_p(bool enabled) {
    mFilterFlag = enabled ? mFilterFlag | Logger::Info : mFilterFlag - Logger::Info;
    refresh();
}

/**
 * @brief Update log filter for debug messages.
 * @param enabled @ti{true} if debug messages should be shown, @ti{false} otherwise.
 */
void LogWidget::toggleDebug_p(bool enabled) {
    mFilterFlag = enabled ? mFilterFlag | Logger::Debug : mFilterFlag - Logger::Debug;
    refresh();
}

/**
 * @brief Clear log.
 * @details The method clears both the internal list of messages and the TextBrowser.
 */
void LogWidget::clearLog() {
    logBox->clear();
    mLogList.clear();
}

/**
 * @brief Save current log in output file.
 * @return True on success, false otherwise (file cannot be created).
 */
void LogWidget::saveLog() {
    QString fileName = QFileDialog::getSaveFileName(this, QString());

    if (fileName.isEmpty())
        return;
    QFile file(fileName);
    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        // Log something :)
        LOG_ERROR(tr("Cannot write log file %1").arg(fileName));
        return;
    }
    QTextStream out(&file);
    for (const Log &m : qAsConst(mLogList))
        out << m.message << endl;
    file.close();
}

/// @brief Update internal TextBrowser to reflect changes in log filter.
void LogWidget::refresh() {
    QString log;
    QTextStream text(&log);
    for (const Log &l : qAsConst(mLogList))
        if (l.level & mFilterFlag)
            text << l.message << endl;
    logBox->setText(log);
    QScrollBar *sb = logBox->verticalScrollBar();
    sb->setValue(sb->maximum());
}

/** Write the log record in the logDockWidget.
 * @sa Logger::write()
 */
void LogWidget::append(const QString &message, int level) {
    mLogList.append(Log(Logger::LogLevel(level), message));
    refresh();
}

/// Translate widgets displayed text.
void LogWidget::translateUI() {
    mSaveAction->setToolTip(tr("Save log in file"));
    mClearAction->setToolTip(tr("Clear log"));
    mShowErrorAction->setToolTip(tr("Show/Hide errors"));
    mShowWarningAction->setToolTip(tr("Show/Hide warnings"));
    mShowInfoAction->setToolTip(tr("Show/Hide informations"));
    mShowDebugAction->setToolTip(tr("Show/Hide debug"));
}

/**
 * @brief Emit the visibilityChanged signal on hide.
 * @param event Associated event.
 * @internal
 * The method is part of the toggle view algorithm that mimics QDockWidget toggle view action.
 */
void LogWidget::hideEvent(QHideEvent *event) {
    emit visibilityChanged(false);
    saveUI();
    event->accept();
}

/// @brief Save UI layout on close event.
void LogWidget::closeEvent(QCloseEvent *event) {
    saveUI();
    event->accept();
}

/**
 * @brief Emit the visibilityChanged signal on show.
 * @param event Associated event.
 * @internal
 * The method is part of the toggle view algorithm that mimics QDockWidget toggle view action.
 */
void LogWidget::showEvent(QShowEvent *event) {
    loadUI();
    emit visibilityChanged(true);
    event->accept();
}

/// @brief Save UI layout.
void LogWidget::saveUI() {
    if (!mSettingsKey.isEmpty()) {
        QSettings settings;
        settings.setValue(QString("%1/geometry").arg(mSettingsKey), saveGeometry());
    }
}

/// @brief Load UI layout.
void LogWidget::loadUI() {
    if (!mSettingsKey.isEmpty()) {
        QSettings settings;
        restoreGeometry(settings.value(QString("%1/geometry").arg(mSettingsKey)).toByteArray());
    }
}
