// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024

#include "cpumemusage.hpp"

static QElapsedTimer timer;

#ifdef Q_OS_WIN

#include "Windows.h"

#include "Psapi.h"

static ULARGE_INTEGER lastCPU, lastSysCPU, lastUserCPU;
static DWORD numProcessors = 0;
static HANDLE self = 0;
static DWORDLONG totalMemory_ = 0;
static IO_COUNTERS lastIOcounters;

void initMonitoring(qint64 *totalMemory) {
    timer.start();
    MEMORYSTATUSEX memInfo;
    memInfo.dwLength = sizeof(MEMORYSTATUSEX);
    GlobalMemoryStatusEx(&memInfo);
    totalMemory_ = memInfo.ullTotalPhys;
    if (totalMemory)
        *totalMemory = totalMemory_;
    SYSTEM_INFO sysInfo;
    FILETIME ftime, fsys, fuser;

    GetSystemInfo(&sysInfo);
    numProcessors = sysInfo.dwNumberOfProcessors;

    GetSystemTimeAsFileTime(&ftime);
    memcpy(&lastCPU, &ftime, sizeof(FILETIME));

    self = GetCurrentProcess();
    GetProcessTimes(self, &ftime, &ftime, &fsys, &fuser);
    memcpy(&lastSysCPU, &fsys, sizeof(FILETIME));
    memcpy(&lastUserCPU, &fuser, sizeof(FILETIME));

    GetProcessIoCounters(self, &lastIOcounters);
}

void getStatistics(double *cpuUsage, int *memUsage, double *rio, double *wio, qint64 *memTotal) {
    static const qint64 million = 1024 * 1024;
    double elapsed = timer.restart() / 1000.0; // elapsed time since last call in sec
    // disk
    if (rio || wio) {
        IO_COUNTERS ioCounters;
        GetProcessIoCounters(self, &ioCounters);
        qint64 a = (1024.0 * elapsed);
        *rio = (ioCounters.ReadTransferCount - lastIOcounters.ReadTransferCount) / a;
        *wio = (ioCounters.WriteTransferCount - lastIOcounters.WriteTransferCount) / a;
        lastIOcounters = ioCounters;
    }
    // memory
    if (memUsage) {
        PROCESS_MEMORY_COUNTERS_EX pmc;
        ZeroMemory(&pmc, sizeof(PROCESS_MEMORY_COUNTERS_EX));
        GetProcessMemoryInfo(self, (PROCESS_MEMORY_COUNTERS *)&pmc, sizeof(pmc));
        *memUsage = static_cast<int>(pmc.WorkingSetSize / million);
    }
    if (memTotal)
        *memTotal = static_cast<int>(totalMemory_ / million);
    // cpu
    if (cpuUsage) {
        FILETIME ftime, fsys, fuser;
        ULARGE_INTEGER now, sys, user;
        double percent;
        GetSystemTimeAsFileTime(&ftime);
        memcpy(&now, &ftime, sizeof(FILETIME));
        GetProcessTimes(self, &ftime, &ftime, &fsys, &fuser);
        memcpy(&sys, &fsys, sizeof(FILETIME));
        memcpy(&user, &fuser, sizeof(FILETIME));
        percent = (sys.QuadPart - lastSysCPU.QuadPart) + (user.QuadPart - lastUserCPU.QuadPart);
        percent /= (now.QuadPart - lastCPU.QuadPart);
        percent /= numProcessors;
        lastCPU = now;
        lastUserCPU = user;
        lastSysCPU = sys;
        *cpuUsage = percent * 100;
    }
}

#else

#include "stdio.h"
#include "stdlib.h"
#include "string.h"
#include "sys/sysinfo.h"
#include "sys/times.h"
#include "sys/types.h"
#include "sys/vtimes.h"
#include <fstream>
#include <sstream>

static long long totalMemory_ = 0; // total mem in bytes
static clock_t lastCPU, lastSysCPU, lastUserCPU;
static int numProcessors;
static long long lastRIO;
static long long lastWIO;

void initMonitoring(qint64 *totalMemory) {
    // memory
    struct sysinfo memInfo;
    sysinfo(&memInfo);
    totalMemory_ = memInfo.totalram;
    // Multiply in next statement to avoid int overflow on right hand side...
    totalMemory_ *= memInfo.mem_unit;

    // processor
    FILE *file;
    struct tms timeSample;
    char line[128];

    lastCPU = times(&timeSample);
    lastSysCPU = timeSample.tms_stime;
    lastUserCPU = timeSample.tms_utime;

    file = fopen("/proc/cpuinfo", "r");
    numProcessors = 0;
    while (fgets(line, 128, file) != NULL) {
        if (strncmp(line, "processor", 9) == 0)
            numProcessors++;
    }
    fclose(file);

    if (totalMemory)
        *totalMemory = totalMemory_;
    lastRIO = 0L;
    lastWIO = 0L;
}

int parseLine(char *line) {
    // This assumes that a digit will be found and the line ends in " Kb".
    int i = strlen(line);
    const char *p = line;
    while (*p < '0' || *p > '9')
        p++;
    line[i - 3] = '\0';
    i = atoi(p);
    return i;
}

void getStatistics(double *cpuUsage, int *memUsage, double *rio, double *wio, qint64 *memTotal) {
    double elapsed = timer.restart() / 1000.0; // elapsed time since last call in sec

    // IO usage
    if (rio || wio) {
        std::ifstream in("/proc/self/io", std::ifstream::in);
        std::string tmp, token;
        long long crio = 0L, cwio = 0L;
        std::getline(in, tmp);
        {
            std::istringstream iss(tmp);
            std::getline(iss, token, ' '); // 1st token: trash
            std::getline(iss, token, ' '); // 2nd token: the value !
            crio = token.empty() ? 0L : std::stoll(token);
        }
        std::getline(in, tmp);
        {
            std::istringstream iss(tmp);
            std::getline(iss, token, ' '); // 1st token: trash
            std::getline(iss, token, ' '); // 2nd token: the value !
            cwio = token.empty() ? 0L : std::stoll(token);
        }
        in.close();
        *rio = (crio - lastRIO) / 1024.0 / elapsed;
        lastRIO = crio;
        *wio = (cwio - lastWIO) / 1024.0 / elapsed;
        lastWIO = cwio;
    }
    // mem usage
    if (memUsage) {
        FILE *file = fopen("/proc/self/status", "r");
        int result = 0;
        char line[128];
        while (fgets(line, 128, file) != NULL) {
            if (strncmp(line, "VmRSS:", 6) == 0) {
                result = parseLine(line);
                break;
            }
        }
        fclose(file);
        *memUsage = result / 1024;
    }
    if (memTotal)
        *memTotal = totalMemory_;
    // cpu usage
    if (cpuUsage) {
        *cpuUsage = 0.0;

        struct tms timeSample;
        clock_t now;
        double percent;

        now = times(&timeSample);
        if (now <= lastCPU || timeSample.tms_stime < lastSysCPU || timeSample.tms_utime < lastUserCPU) {
            // Overflow detection. Just skip this value.
            percent = -1.0;
        } else {
            percent = (timeSample.tms_stime - lastSysCPU) + (timeSample.tms_utime - lastUserCPU);
            percent /= (now - lastCPU);
            percent /= numProcessors;
            percent *= 100;
        }
        lastCPU = now;
        lastSysCPU = timeSample.tms_stime;
        lastUserCPU = timeSample.tms_utime;
        *cpuUsage = percent;
    }
}

#endif

/// @brief Get the total available memory in MB.
int totalMemory() { return static_cast<int>(totalMemory_ / 1024 / 1024); }

/**
 * @fn void initMonitoring(qint64 *totalMemory)
 * @brief Initialiaze the monitoring of resources.
 * @param totalMemory Optional total memory in bytes (output).
 * @details
 * The method that initialize the global variables for subsequent monitoring must be called before any calls of ::getStatistics
 * method.
 */

/**
 * @fn void getStatistics(double *cpuUsage, int *memUsage, double *rio, double *wio, qint64 *memTotal)
 * @brief Get CPU, RAM and IO statistics.
 * @param cpuUsage CPU percent usage.
 * @param memUsage RAM usage (in MB)
 * @param rio Read operations rate (KB/s)
 * @param wio Write operations rate (KB/s)
 * @param memTotal Total memory (in MB).
 * @details
 * All parameters are optional. The afferant operation is performed only if a non-null pointer is given.
 */
