// Copyright (c) URCA-CNRS-TOTALENERGIES - All Rights Reserved
// Authors: J. Burgalat
// 2016-2024
/**
 * @file AMDataset.cpp
 * @brief AMDataset class definitions file.
 * @author J. Burgalat
 * @date 2017-2018
 */
#include "amdataset.hpp"
#include <algorithm>
#include <iterator>
#include <numeric>

struct {
    bool operator()(ADataset *left, ADataset *right) { return left->startRef() < right->startRef(); }
} TimeComparer;

StringList getFiles(ADataset *set) {
    StringList out;
    if (!set)
        return out;
    for (int i = 0; i < set->blockCount(); i++) {
        std::string name = set->filePath(i);
        if (std::find(out.begin(), out.end(), name) == out.end())
            out.push_back(name);
    }
    return out;
}

/// @brief Constructor.
AMDataset::AMDataset(const QStringList &paths) {
    mName = "";
    addFiles(paths);
}

/// @brief Destructor.
AMDataset::~AMDataset() {
    for (ADataset *d : mFiles)
        delete d;
    mFiles.clear();
}

/**
 * @brief Add multiple files to the multi-dataset.
 * @param paths List of file paths to add.
 * @param sort True to sort data by time.
 * @param signalNeeded True to emit AMDataset::filesAdded.
 * @return True on success, false otherwise.
 */
bool AMDataset::addFiles(const QStringList &paths, bool sort, bool signalNeeded) {
    bool valid = false;
    for (QString path : paths) {
        if (addFile(path, false))
            valid = true;
    }
    if (sort)
        sortDataSets();
    if (valid && signalNeeded)
        emit filesAdded();
    return valid;
}

/**
 * @brief Add a single file to the dataset.
 * @param path Path of the file to add.
 * @param loadSpectras True to load spectras.
 * @param sort True to sort data by time.
 * @param signalNeeded True to emit AMDataset::filesAdded.
 * @return True on success, false otherwise.
 * @sa AMDataset::addFiles
 */
bool AMDataset::addFile(const QString &path, bool sort, bool signalNeeded) {
    ADataset *set = new ADataset(ISP::StringList{path.toStdString()});
    // invalid set -> do not add
    if (!set->valid())
        return false;
    if (mFiles.empty()) {
        // no dataset yet -> add and set AMDataset valid
        if (mName.isEmpty())
            mName = QFileInfo(path).baseName();
        mFiles.push_back(set);
        return true;
    } else if (alreadyIn(set)) {
        // already in -> do not add
        return false;
    } else if (set->descriptor().comparable(mFiles.front()->descriptor())) {
        // not in and match the file descriptor -> add
        mFiles.push_back(set);
        if (mName.isEmpty())
            mName = QFileInfo(path).baseName();
        if (sort)
            sortDataSets();
        if (signalNeeded)
            emit filesAdded();
        return true;
    }
    return false;
}

/**
 * @brief Add a dataset to the multi-dataset.
 * @param set Pointer to the dataset. The AMDataset takes ownership of the pointer.
 * @param sort True to sort the dataset.
 * @param signalNeeded True to emit AMDataset::filesAdded.
 * @return True on success, false otherwise.
 * @warning
 * If the method fails, the dataset is not added to the container. Then the pointer is not owned by the latter.
 * Such pointer should be deleted if it is not already part of the container. AMDataset::alreadyIn return
 * status should be used with the return of the current method to check whether or not the pointer has to be
 * deleted.
 */
bool AMDataset::addDataSet(ADataset *set, bool sort, bool signalNeeded) {
    if (!set->valid())
        return false;
    if (mFiles.empty()) {
        mFiles.push_back(set);
        if (mName.isEmpty())
            mName = QFileInfo(QString::fromStdString(set->filePath(0))).baseName();
        return true;
    } else if (alreadyIn(set)) {
        return false;
    } else if (set->descriptor().comparable(mFiles.front()->descriptor())) {
        if (sort) {
            QVector<ADataset *>::iterator it = sortedIterator(set);
            mFiles.insert(it, set);
            rebuildDiscontinuities();
        } else
            mFiles.push_back(set);
        if (mName.isEmpty())
            mName = QFileInfo(QString::fromStdString(set->filePath(0))).baseName();
        if (signalNeeded)
            emit filesAdded();
        return true;
    }
    return false;
}

/**
 * @brief Add a list of ADataset to this AMDataset.
 * @param sets List of ADataset to add.
 * @param sort Indicate whether or not this AMDataset should be sorted after insertion.
 * @return A vector with the datasets pointers that have not been added to the dataset. Such pointers are not owned by
 *the container and should be deleted if needed.
 *
 */
QVector<ADataset *> AMDataset::addDataSets(const QVector<ADataset *> &sets, bool sort, bool signalNeeded) {
    QVector<ADataset *> failed;
    for (QVector<ADataset *>::const_iterator it = sets.cbegin(); it != sets.cend(); it++) {
        if (!addDataSet(*it, false, false))
            failed.push_back(*it);
    }
    if (sort)
        sortDataSets();
    if (signalNeeded)
        emit filesAdded();
    return failed;
}

/// @brief Clear the multi-dataset.
void AMDataset::clear() {
    for (ADataset *d : mFiles)
        delete d;
    mFiles.clear();
}

/**
 * @brief Get a vector of block indexes (as double).
 * @param step Step Increment. Values are retrieved each step index.
 * @param noGap A control flag with true (default) to prevent the method of inserting ghost points between
 * discontinued internal datasets.
 * @return Vector of indexes.
 * @details
 * The output vector is not intended to perform subscript operations on vectors. It is only inteded to
 * have a raw time line. If noGap is set to false, the method inserts ghost points between discontinued
 * datasets.
 * Then the size oof the output vector can be greater than the actuall number of blocks in the multi-dataset.
 */
DoubleVector AMDataset::indexes(bool noGap, int step) const {
    DoubleVector out;
    if (mFiles.empty())
        return out;
    // get from 1st dataset
    for (int i = 0; i < mFiles.front()->blockCount(); i++)
        out.push_back(static_cast<double>(i));
    for (int i = 1, j = 0; i < mFiles.size(); i++, j++) {
        double iback = out.back();
        if (discontinued(j) && !noGap)
            out.push_back(0.5 + iback);
        iback += 1.0;
        for (int k = 0; k < mFiles[i]->blockCount(); k += step)
            out.push_back(iback + k);
    }
    return out;
}

/**
 * @brief Get the time line of this dataset.
 * @param step Step Increment. Values are retrieved each step index.
 * @param noGap A control flag with true (default) to prevent the method of inserting ghost points between
 * discontinued internal datasets.
 * @param ok A boolean control flag with true if time values are actually returned by the method or
 * false if only a vector of indexes is returned.
 * @return A vector of double with the time line as the number of seconds since Epoch.
 * @details
 * The method retrieves the internal time-line of each datasets and optionally insert ghost points
 * if noGap is set to false to reflect discontinuities in the dataset. The value of a ghost point
 * is computed as the mean between the last and first time of two datasets.
 *
 * If the first dataset does not store the parameter h_system a vector of indexes is returned and
 * ok is set to false (if any).
 *
 * If the dataset is not valid, an empty vector is returned.
 *
 * @todo
 * Check the consistency of the algorithm concerning the step argument.
 * The output vector should have the same size of the returned vector from
 * AMDataset::parameter(const AParam &parameter, bool noGap = true, int step = 1).
 */
DoubleVector AMDataset::timeLine(bool noGap, int step, bool *ok) const {
    DoubleVector out;
    size_t bt = blockCount(), bc = 0;
    out.reserve(bt);
    if (ok)
        *ok = true;
    if (!bt)
        return out;
    if (!mFiles[0]->hasParameter(AParam("h_system"))) {
        if (ok)
            *ok = false;
        for (int i = 0; i < mFiles.size(); i++) {
            for (int j = 0; j < mFiles[i]->blockCount(); j += step) {
                out.push_back(static_cast<double>(bc));
                bc++;
            }
            if (discontinued(i) && !noGap)
                out.push_back(static_cast<double>(bc) + 0.5);
        }
        return out;
    }
    for (int i = 0; i < mFiles.size(); i++) {
        const auto vec = mFiles[i]->times();
        int nb = std::distance(vec.begin(), vec.end());
        out.reserve(out.size() + nb);
        out.insert(out.end(), vec.begin(), vec.end());
        bc += nb;
        if (discontinued(i) && !noGap) {
            double t = (vec.back() + mFiles[i + 1]->times().front()) / 2.0;
            out.push_back(t);
        }
    }
    return out;
}

/// @brief Get the date time value at given block index.
/// @param index Block index.
double AMDataset::timeAt(size_t bIndex) const {
    long lidx = -1;
    long fidx = getFileIndex(bIndex, &lidx);
    if (fidx == -1 || lidx == -1)
        return qQNaN();
    return mFiles[fidx]->times().at(lidx);
}

/// @brief Get the average timestep of the files in seconds.
double AMDataset::timeStep() const {
    if (mFiles.empty())
        return qQNaN();
    double avgstep = 0;
    for (ADataset *s : mFiles) {
        avgstep += s->refStep(); // assume that time will always be the ref prm... who cares ?!
    }
    return avgstep / mFiles.size();
}

/// @brief Get the starting date time of the multi dataset.
double AMDataset::startTime() {
    if (mFiles.empty())
        return qQNaN();
    return mFiles.front()->startRef(); // assume that time will always be the ref prm... who cares ?!
}

/// @brief Get the ending date time of the multi dataset.
double AMDataset::endTime() const {
    if (mFiles.empty())
        return qQNaN();
    return mFiles.back()->endRef(); // assume that time will always be the ref prm... who cares ?!
}

/// @brief Get the list of internal datasets (const ref).
const QVector<ADataset *> &AMDataset::datasets() const { return mFiles; }

/// @brief Get the list of internal datasets.
QVector<ADataset *> &AMDataset::datasets() { return mFiles; }

/// @brief Get the ADataset at given global block index.
ADataset *AMDataset::dataset(int blockIndex) {
    int fileIndex = getFileIndex(blockIndex);
    if (fileIndex < 0 || fileIndex >= mFiles.size())
        return nullptr;
    return mFiles[fileIndex];
}

/// @brief Get the ADataset at given file index.
ADataset *AMDataset::datasetAt(long fileIndex) {
    if (fileIndex < 0 || fileIndex >= mFiles.size())
        return nullptr;
    return mFiles[fileIndex];
}

/// @brief Get a <b>copy</b> of the block at given index.
ABlock AMDataset::blockAt(int blockIndex) const {
    long relIndex;
    int fileIndex = getFileIndex(blockIndex, &relIndex);
    if (fileIndex < 0 || fileIndex >= mFiles.size())
        return ABlock();
    return mFiles[fileIndex]->blockAt(relIndex);
}

/// @brief Get a <b>copy</b> of all the blocks.
QVector<ABlock> AMDataset::blocks() const {
    QVector<ABlock> out;
    out.reserve(blockCount());
    for (ADataset *d : mFiles) {
        for (int i = 0; i < d->blockCount(); ++i)
            out.push_back(d->blockAt(i));
    }
    return out;
}

/// @brief Get the list of available categories.
QStringList AMDataset::categories() const {
    QStringList out;
    if (mFiles.empty())
        return out;
    for (const auto &s : mFiles.front()->categories())
        out.push_back(QString::fromStdString(s));
    return out;
}

/// @brief Get the list of parameters.
QVector<AParam> AMDataset::parameters() const {
    QVector<AParam> out;
    if (mFiles.empty())
        return out;
    const auto &prms = mFiles[0]->parameters();
    for(const auto & p : prms)
        out.push_back(p);
    return out;
}

/**
 * @brief Check whether or not the given parameter name is defined the dataset.
 * @param name Name of the parameter to search. The search is case-insensitive.
 * @return true if found, false otherwise.
 */
bool AMDataset::hasParameter(const AParam &parameter) const {
    return mFiles.empty() ? false : mFiles.front()->hasParameter(parameter);
}

/**
 * @brief Get a parameter value at given block index.
 * @param label Name of the parameter
 * @param bIndex Global block index.
 * @return Value of a NaN if either label does not matches any parameter or bIndex is out of range.
 */
double AMDataset::parameterAt(const AParam &parameter, long bIndex) const {
    long lidx = -1;
    long fidx = getFileIndex(bIndex, &lidx);
    if (fidx == -1 || lidx == -1)
        return qQNaN();
    return mFiles[fidx]->parameterAt(parameter, lidx);
}

/**
 * @brief Get the parameter values at the given block indexes.
 * @param parameter PArameter to process.
 * @param bIndexes Global block indexes.
 */
DoubleVector AMDataset::parameterAt(const AParam &parameter, const QVector<long> &bIndexes) const {
    typedef QPair<int, QVector<long>> APair;
    QVector<APair> pairs;
    int no = 0;
    for (long bi : bIndexes) {
        long lidx = -1;
        long fidx = getFileIndex(bi, &lidx);
        if (fidx == -1 || lidx == -1)
            continue;
        QVector<APair>::iterator it =
            std::find_if(pairs.begin(), pairs.end(), [=](const APair &p) { return p.first == fidx; });
        if (it == pairs.end()) {
            // add new pair
            pairs.append(qMakePair(fidx, QVector<long>{lidx}));
        } else {
            it->second.push_back(lidx);
        }
        no++;
    }
    DoubleVector out;
    out.reserve(no);
    for (const auto &p : pairs) {
        for (long i : p.second)
            out.push_back(mFiles[p.first]->parameterAt(parameter, (int)i));
    }
    return out;
}

/**
 * @brief Get all the values of the given parameter in the dataset.
 * @param parameter Parameter to process.
 * @param noGap A control flag with true (default) to prevent the method of inserting NaN value between
 * discontinued internal datasets.
 * @param step Step Increment. Values are retrieved each step index.
 * @return A vector of double with the values of the parameter or an empty vector if no parameter matches
 * in the dataset.
 */
DoubleVector AMDataset::parameter(const AParam &parameter, bool noGap, int step) const {
    DoubleVector out;

    if (parameter.name().compare("__index__") == 0) {
        return indexes(noGap, step);
    } else if (parameter.name().compare("__timeline__") == 0) {
        return timeLine(noGap, step);
    }
    if (!hasParameter(parameter))
        return out;
    int nf = mFiles.size();
    int nv = blockCount(noGap);
    out.reserve(nv);
    auto tmp = mFiles.front()->parameter(parameter);
    for (int i = 0; i < tmp.size(); i += step)
        out.push_back(tmp[i]);
    for (int i = 1, j = 0; i < nf; i++, j++) {
        if (!noGap && j < nf - 1 && discontinued(j))
            out.push_back(qQNaN());
        tmp = mFiles[i]->parameter(parameter);
        for (int i = 0; i < tmp.size(); i += step)
            out.push_back(tmp[i]);
    }
    return out;
}
/**
 * @brief Get the \a n firstvalues of the given parameter in the dataset.
 * @param n Maximum number of values to retrieve.
 * @param parameter Parameter to process.
 * @return A vector of double with the values of the parameter or an empty vector if no parameter matches.
 */
DoubleVector AMDataset::parameter(int n, const AParam &parameter) const {
    DoubleVector out;
    if (!hasParameter(parameter))
        return out;
    n = qMin(n, blockCount());
    out.reserve(n);
    int i = 0;
    bool getOut = false;
    for (auto *d : mFiles) {
        for (auto v : d->parameter(parameter)) {
            out.push_back(v);
            i++;
            if (i >= n) {
                getOut = true;
                break;
            }
        }
        if (getOut)
            break;
    }
    return out;
}

/// @brief Get the default size of spectra.
int AMDataset::spectraSize() const { return mFiles.empty() ? 0 : mFiles.front()->rampSize(); }

/**
 * @brief Get the spectra data at given block index and spectra index
 * @param bIndex Global block index in the dataset.
 * @param sIndex Index of the spectra in the block (refer to a molecule).
 * @return Vector of double with the data related to the spectra.
 */
DoubleVector AMDataset::spectraAt(int bIndex, int sIndex) {
    long lidx = -1;
    long fidx = getFileIndex(bIndex, &lidx);
    if (fidx == -1 || lidx == -1)
        return DoubleVector();
    return mFiles[fidx]->rampAt(lidx, sIndex);
}

/// @brief Get the labels of the spectras defined in the dataset (if any).
QStringList AMDataset::spectraLabels() const {
    QStringList out;
    if (mFiles.empty())
        return out;
    for (const auto &s : mFiles[0]->rampLabels())
        out.push_back(QString::fromStdString(s));
    return out;
}
/**
 * @brief Compute the spectra label for the given acquisition line identifier.
 * @param lineId Acquisition line identifier.
 * @return spectra label.
 * @details
 * The spectra label is generaly the list of molecules associated to the line separated by a comma.
 * If no molecule can be deduced, the method returns a generic name spectra_X where X is lineId.
 */
QString AMDataset::spectraLabel(int lineId) const {
    if (mFiles.empty())
        return QString();
    return QString::fromStdString(mFiles[0]->rampLabel(lineId));
}

/**
 * @brief Get the list of molecules associated to the given acquisition line.
 * @param lineId Acquisition line identifier.
 * @attention
 * The acquisition line identifier is not the spectra index. Line identifiers start from 1, by descriptor convention,
 * while spectra index starts from 0 (C++ convention).
 */
QStringList AMDataset::spectraMoleculesAt(int lineId, bool noUkn) const {
    QStringList out;
    if (mFiles.empty())
        return out;
    for (const auto &s : mFiles[0]->moleculesAtRamp(lineId, noUkn))
        out.push_back(QString::fromStdString(s));
    return out;
}

/**
 * @brief AMDataset::blockCount
 * @param noGap False to account for inserted ghost point. See AMDataset::parameter. Default to true.
 */
int AMDataset::blockCount(bool noGap) const {
    int num = 0;
    int nf = mFiles.size();
    if (!nf)
        return num;
    for (int i = 0, j = 0; i < nf; i++, j++) {
        num += mFiles[i]->blockCount();
        if (!noGap && j < nf - 1 && discontinued(j))
            num++;
    }
    return num;
}

/// @brief Get the number of actual parameters in the dataset.
int AMDataset::parameterCount() const {
    if (mFiles.empty())
        return 0;
    else
        return mFiles.front()->parameterCount();
}

/**
 * @brief Get the expected number of spectras per block.
 * @details
 * The expected number of spectras is computed from the Descriptor used to read the dataset.
 * Some blocks of the dataset may have no spectras.
 * @sa ABlock::lineCount
 */
int AMDataset::lineCount() const { return mFiles.empty() ? 0 : mFiles.front()->rampCount(); }

/// @brief Get the number of internal datasets.
size_t AMDataset::fileCount() const { return mFiles.size(); }

/// @brief Check whether or not the dataset is valid.
bool AMDataset::isValid() const { return !mFiles.empty(); }

/// @brief Get the list of file paths.
QStringList AMDataset::filePaths() {
    QStringList out;
    if (mFiles.size()) {
        for (QVector<ADataset *>::iterator it = mFiles.begin(); it != mFiles.end(); it++) {
            StringList tmp = getFiles((*it));
            for (const auto &t : tmp)
                out.push_back(QString::fromStdString(t));
        }
    }
    return out;
}

/// @brief Get the file descriptor
ABlock AMDataset::descriptor() const {
    return (mFiles.empty()) ? ABlock() : mFiles.front()->descriptor();
}

/// @brief Get the instrument number.
int AMDataset::instrument() const { return descriptor().instrument(); }

/// @brief Get the frimaware version.
int AMDataset::version() const { return descriptor().version(); }

/**
 * @brief Compute a valid block index from given index.
 * @param blockIndex block index.
 * @return valid block index in the dataset.
 * @details
 * The method shifts the given block index from ghost point added to AMDataset::parameter() vector in order to display
 * continuities
 * in plots.
 * @warning
 * The method is only intended to get actual block index of a dataset from the index of a data points in a graph in the
 * case of discontinued blocks.
 * @internal
 * The method replaces AMDataset's validBlockIndex method from older versions.
 */
size_t AMDataset::actualBlockIndex(size_t discontinuedIndex) const {
    int out = -1;
    if (discontinuedIndex < 0) {
        return out;
    }
    size_t count = 0;
    size_t offset = 0;
    for (int i = 0; i < mFiles.size(); i++) {
        count += mFiles[i]->blockCount();
        if (i < mFiles.size() - 1)
            offset += discontinued(i) ? 1 : 0;
        if (discontinuedIndex < count + offset)
            break;
    }
    return discontinuedIndex - offset - 1;
}

/**
 * @brief Convert the given block index from the AMDataset to a graph data index.
 * @param blockIndex Block index in this dataset.
 * @return Corresponding graph data index.
 * @warning
 * The method is only intended to get actual block index of a graph from the index of dataset in the case of
 * discontinued blocks.
 * @internal
 * The method replaces AMDataset's toGraphBlockIndex method from older versions.
 */
size_t AMDataset::discontinuedBlockIndex(size_t actualIndex) const {
    size_t count = 0;
    size_t offset = 0;
    for (int i = 0; i < mFiles.size(); i++) {
        count += mFiles[i]->blockCount();
        if (i < mFiles.size() - 1 && actualIndex > count)
            offset += discontinued(i) ? 1 : 0;
        else if (actualIndex + offset < count)
            break;
    }
    return actualIndex + offset;
}

/**
 * @brief Get the index of a dataset based on a global block index.
 * @param blockIndex Index of global block to seek.
 * @param relative A pointer to a int with (if valid) the value of the local block index.
 * @return Index of the afferent file.
 */
long AMDataset::getFileIndex(size_t blockIndex, long *relative) const {
    long out = -1;
    if (blockIndex < 0) {
        if (relative)
            *relative = out;
        return out;
    }
    size_t count = 0;
    for (int i = 0; i < mFiles.size(); i++) {
        count += mFiles[i]->blockCount();
        if (blockIndex < count) {
            if (relative) {
                *relative = static_cast<long>((i == 0) ? blockIndex : blockIndex - count + mFiles[i]->blockCount());
            }
            return static_cast<long>(i);
        }
    }
    return out;
}

/// @brief Sort and check for discontinued datasets.
void AMDataset::sortDataSets() {
    std::sort(mFiles.begin(), mFiles.end(), TimeComparer);
    mDiscontinued.clear();
    if (mFiles.size() > 1)
        for (int i = 1; i < mFiles.size(); i++) {
            ADataset *d1 = mFiles[i - 1];
            ADataset *d2 = mFiles[i];
            double dt = d1->refStep(); // assume time is the ref param ... again !
            double dt2 = (d2->startRef() - d1->endRef());
            mDiscontinued.push_back(dt2 > dt * 3.0);
        }
}

/**
 * @brief Rebuild the internal discontinuities vector.
 * @internal
 * Two dataset are continuous if the time delta betwen the end of the former and the beginning of the latter is less
 * than three time the average time step of the former.
 */
void AMDataset::rebuildDiscontinuities() {
    mDiscontinued.clear();
    if (mFiles.size() > 1)
        for (int i = 1; i < mFiles.size(); i++) {
            ADataset *d1 = mFiles[i - 1];
            ADataset *d2 = mFiles[i];
            double dt = d1->refStep();
            double dt2 = (d2->startRef() - d1->endRef()); // assume time is the ref param ... again !
            mDiscontinued.push_back(dt2 > dt * 3.0);
        }
}

/// @brief Check whether or not the file index is related to a discontinued dataset.
bool AMDataset::discontinued(int fileIndex) const {
    // out of range: not discontinued
    if (fileIndex < 0 || fileIndex >= mDiscontinued.size())
        return false;
    return mDiscontinued.at(fileIndex);
}

/// @brief Check whether or not a Dataset is already in the multi-dataset (by file path).
bool AMDataset::alreadyIn(ADataset *set) {
    QStringList paths = filePaths();
    for (auto &&p : getFiles(set)) {
        if (paths.contains(QString::fromStdString(p)))
            return true;
    }
    return false;
}

/// @brief Set a new name for the dataset.
void AMDataset::setName(const QString &name) { mName = name; }

/// @brief  Get the name of the dataset.
const QString &AMDataset::name() const { return mName; }

/**
 * @brief Computre and return the friendly name of the dataset.
 * @details
 * The method attempts to format the name of the datasety to get a more human-readable name.
 * If the name of the dataset (which is by default set as the first filename in the list of dataset) matches the
 * following pattern: <b>\d{12}_[a-zA-Z0-9]+.*</b>, then the method parses the string to compute the date and name of
 * the pattern and return a more friendly name. Otherwise, AMDataset::name is returned.
 */
QString AMDataset::friendlyName() const {
    static const QRegularExpression prefixPattern("(?<date>\\d{12})_(?<name>[a-zA-Z0-9]+)_\\d+_.*",
        QRegularExpression::OptimizeOnFirstUsageOption | QRegularExpression::CaseInsensitiveOption);
    QString name = mName;
    QRegularExpressionMatch match = prefixPattern.match(name);
    if (match.hasMatch()) {
        QDateTime datetime = QDateTime::fromString(match.captured("date"), "yyyyMMddHHmm");
        name = QString("%1 (%2)").arg(match.captured("name"), datetime.toString("dd/MM/yyyy, HH:mm"));
    }
    return name;
}

/// @brief Get an iterator pointing to the position of insertion of the given dataset.
QVector<ADataset *>::iterator AMDataset::sortedIterator(ADataset *set) {
    if (mFiles.empty())
        return mFiles.begin();
    return std::lower_bound(mFiles.begin(), mFiles.end(), set, [](ADataset *lhs, ADataset *rhs) {
        return (lhs->startRef() < rhs->startRef());
    });
}
